#!/usr/bin/env python

# Copyright (c) 2000-2004 Matthew Dixon Cowles <matt@mondoinfo.com>
#
# This program is free software; you may redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA



# Standard modules
import base64
import binascii # Just to get at the exception
import bisect
import calendar
import copy
import cPickle
import cStringIO
import datetime
import difflib
import email
import email.Charset
import email.Errors
import email.Generator
import email.Header
import email.Message
import email.Parser
import email.Utils
#import email
import errno
import mimetypes
import operator
import os
import popen2
import poplib
try:
  import pwd # Unix-specific
except ImportError:
  pass
import Queue
import quopri
import re
import sgmllib
# Docco seems to imply that having signal "available" (imported?) means
# that interrupts go to main thread rather than a random one. Maybe
# we want that.
import signal
import smtplib
import socket
import stat
import struct
import sys
import tempfile
import thread
import threading
import time
import tkFileDialog
import tkFont
from Tkinter import *
import types
import uu


# Non-standard modules
try:
  import metakit # For SWDB http://www.equi4.com/metakit/
except ImportError:
  pass
import Pmw # http://pmw.sourceforge.net/
## timeoutsocket is optional, http://www.timo-tasi.org/python/timeoutsocket.py
#try:
#  import timeoutsocket
#except ImportError:
#  pass


#
# General constants
#
# Not that they can't be changed at runtime. They just
# aren't meant to be.
#

kVersion="0.5.7"

kNormal,kWarning,kError=range(3)

kAddressExportDefaultName="swdb-exported.txt"
kBackslat="\\"
kBeginUURE=re.compile(r"^begin [0-7][0-7][0-7] (\S+)")
kButtonDotsPerChar=7
kButtonHeight=27
kButtonPaddingX=26
kCharsNeedingQuotingInMailAddress='"'+"'"+"\\"+"()<>@,;:.[]"
# How long the thread that runs checkRegularly should sleep. Controls
# how often it checks to see if it has been asked to quit or if it's
# time to check for new mail.
kCheckThreadSleepSecs=2
kCRLF="\r\n"
kDayAbbrs=("Mon","Tue","Wed","Thu","Fri","Sat","Sun")
# If we don't normally re-wrap a "Dear XXX," line if it's the first
# line of mail that's being composed, how do we identify a line like
# that?
kDearLineEndsWith=",:-;!"
kDearLineMaxLen=30
kDisplayableMainTypes=(None,"text","message")
kDoubleQuote='"'
kEighthBitSetRE=re.compile("[\177-\377]")
kEncodeQpriIfLineLongerThan=200
kEnoughMailForFancyMailboxOpening=30
kFileLineSep=os.linesep # Contantize this for speed, global lookup faster than module data lookup
kGeneralAfterDelayMS=100
kInAngleBracketsRE=re.compile("(<.*?>)") # Not even close to proper parsing
kIsThisTextBytesToRead=128
kIsThisTextMaxPortionNonASCII=0.1
kPaneHeightOverhead=5
kLastLogLineHeight=2
kLastLogLineWidth=50
kLogLinesToKeep=500
kMailboxContentsWidth=400
kMailboxHeaders=("N","Date","From","To","Sc","Subject")
kMainWinColWidths=(35,20) # Chars
kOutboxHeaders=("R","Date","From","To","Sc","Subject")
kMailboxOverviewLinesToShow=15
kMailboxStatsTextWidth=20
kMailOverviewLinesToShow=20
#kMailQuoteCharsRE=re.compile("^[>: ]+")
kMailQuoteCharsRE=re.compile("^[>: ]*[>:]")
kMailViewerTextWidth=80
kMaxDoubleClickDelayMS=500
kMaxFileCreationTries=100
kMaxViewerStringSearchSuccesses=100
kMaxUndos=50
kMIMEPreambleText="This is a multi-part message in MIME format."+os.linesep
kMonthAbbrs=("","Jan","Feb","Mar","Apr","May","Jun","Jul","Aug", \
             "Sep","Oct","Nov","Dec")
kMouseEntersActiveAddressBG="gray85"
kMyName="Swallow"
kMyReceivedRE=re.compile("^from .* by .* \(%s\); (.*)" % kMyName)
kNeverFetchBiggerThan=20*1024*1024
kNonASCIITextRE=re.compile("[\000-\010\016-\037\177-\377]")
# xbm data for null cursor from Steve Lidie of Lehigh University, archived
# at The FAQchest. Yes, this is the actual image file format.
kNullCursorData="""
  #define t_cur_width 1
  #define t_cur_height 1
  #define t_cur_x_hot 0
  #define t_cur_y_hot 0
  static unsigned char t_cur_bits[] = {  0x00};
  """
kOKMailboxChars="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_"
##kOverviewColumnDefaultWidths=(14,52,105,105,21,250)
kOverviewColumnDefaultWidths=(2,7,15,15,3,80) # In chars
kPasswordDialogSleepTime=0.1
kPasteTakesEffectDelayMS=25
kPathSep=os.sep # Contantize this for speed, global lookup faster than module data lookup
kPrefsPath="~/.swallowprefs.py"
kQuotes="'"+'"'
kReWrapIndentDepth=4
kScrollbarWidth=16
# How many columns to use for the radiobuttons that control searching
# in particular mailboxes.
kSearchMailMailboxCols=5
###kSearchMailStatusMessageWidth=60
kSearchMailBoxGridMaxHeight=200
kSearchMailMatchesBG="gray90"
kSearchMailSpacerSize=3
kSearchMailWindowWidth=660
kSecsIn36Hours=int(60*60*24*1.5)
kSecsInDay=60*60*24
kSecsInWeek=60*60*24*7
kSendingProgressSizeBigEnoughToReport=50*1024
##kSendingProgressReportEach=10*1024 # If this is less than a line length, you'll get odd results
##kShowFetchProgressEach=1024*5 # bytes
kShowFetchProgressForMailBiggerThan=1024*50 # bytes
kShowOpenOverviewProgressEvery=10 # messages processed
kShowProgressEvery=0.5 # seconds
kSocketTimeout=30
kStatePickleName="state.pickle"
kStatusCheckOffColor="gray40"
kStatusCheckOnColor="black"
kStatusLogBlinkColor="white"
kStatusLogBlinkMinDelay=100
kStatusLogErrorColor="red"
kStatusLogNormalColor="black"
kStatusLogNormalDelay=200 # ms
kStatusLogWarningColor="yellow"
kSubjJunkRE=re.compile("^re[\[\]0-9 ]* *: *") # Could be more sophisticated
# Could add message-id, references, in-reply-to if we wanted to speed up threading
#kSummaryHeaders=("subject","from","to","cc","bcc")
kSummaryHeaders=("subject","from","to","cc","bcc","references","message-id","in-reply-to")
kSummaryNotTheseMailboxes=("inbox","outbox","trash")
kSummaryVersion=2
kswdbDateFormat="%d %b %y"
kswdbDateTimeFormat="%d %b %y, %H:%M:%S"
kswdbDBViewName="addressbook"
kswdbDefaultColWidths=(10, 20, 20, 12, 16, 16, 40) # Overview default widths in chars
kswdbFieldLabels=("Nickname","Name","Email","Date created","Last heard from","Last mailed to","Comments")
kswdbFieldNames=("nick","name","email","dateCreated","dateLastHeardFrom","dateLastMailedTo","comments")
kswdbMkStruct="addressbook[email:S,name:S,nick:S,dateCreated:I,dateLastHeardFrom:I,dateLastMailedTo:I,comments:s]"
kswdbNoDateVal=0
kswdbROEntryBackgroundColor="gray90"
kswdbROFieldNames=("dateCreated","dateLastHeardFrom","dateLastMailedTo")
kswdbSearchDelayAfterKeystrokeMS=200 # See comment below where used
kswdbWindowStartSize={"height": 400, "width": 450}
kswdbWindowTitle="SWDB"
kMessageThreaderDebug=1
kTimeZoneNameInParensRE=re.compile("\(([A-Za-z0-9]{2,5})\)")
kTkDefaultBG="#d9d9d9" # Mostly used to turn off blinky active backgrounds
kTkHighlightBG="#ececec"
kUnableToParseMessage="(Unable to parse)"+os.linesep
kWhitespaceRunRE=re.compile("\s+")
kWinPrefsFile="swallowprefs.py" # Not path


#
# General utility routines
#

def nullCB(*positionArgsIgnored,**keywordArgsIgnored):
  return None


def nullEventCB(*positionArgsIgnored,**keywordArgsIgnored):
  return "break" # Signal Tkinter to do no more processing of keystroke


def pluralize(qty,word,pluralSuffix="s"):
  if qty<>1:
    return word+pluralSuffix
  else:
    return word


def qtyToReadable(val):
  scaleSuffixes=("","K","M","G","T","P")
  scale=0
  while val>=1024 and scale<len(scaleSuffixes)-1:
    val=val/1024.0
    scale+=1
  if val==int(val):
    decimals=0
  elif 10*val==int(10*val):
    decimals=1
  else:
    decimals=2
  return "%.*f%s" % (decimals,val,scaleSuffixes[scale])


def truncateWithEllipsis(str,maxLen):
  if len(str)<=maxLen:
    return str
  else:
    return str[:maxLen-3]+"..."


def stripQuotes(str):
  if len(str)<2:
    return str
  while str[0]==str[-1] and str[0] in kQuotes:
    str=str[1:-1]
  return str


def escapeDoubleQuotes(str):
  str=re.sub(r"\\",r"\\\\",str)
  str=re.sub('"',r'\\"',str)
  return str


# XXX configurable? do we like this one?
# Assumed to return a string that's always the same length
def formatDateForOverview(rfc822Date):
  try:
    timeTuple=email.Utils.parsedate(rfc822Date)
  except (OverflowError,TypeError,IndexError):
    return "baddate"
  if timeTuple==None:
    timeTuple=email.Utils.parsedate("Dummyday, "+rfc822Date)
  if timeTuple==None:
    return "baddate"
  try:
    return time.strftime("%d%b%y",timeTuple)
  except ValueError:
    return "baddate"


# os.path.join works fine here despite the odd (to me)
# behavior of returning "/b" for os.path.join("a","/b")
# but it's a hotspot in the code and this version runs
# a little more than twice as fast.
#
# .startswith() and .endswith() are prettier than [0]
# and [-1] but significantly slower

def myPathJoin(a,b):
  firstEndsSlant=(a[-1]==kPathSep)
  secondBeginsSlant=(b[0]==kPathSep)
  if firstEndsSlant and secondBeginsSlant:
    return a+b[1:]
  elif firstEndsSlant or secondBeginsSlant:
    return a+b
  else:
    return a+kPathSep+b


# Like above. Doesn't work just like os.path.split
# but it's about three times as fast and works for
# me.
def myPathSplit(p):
  i=p.rfind(kPathSep)+1
  return p[:i-1],p[i:]


def lastTwoPathElements(path):
  most,fn=myPathSplit(path)
  most,subdir=myPathSplit(most)
  return myPathJoin(subdir,fn)


def lowerIfPossible(x):
  try:
    return x.lower()
  except AttributeError:
    return x

def stripIfPossible(x):
  try:
    return x.strip()
  except AttributeError:
    return x


# A list comprehension would be nice here but is slower.
# And it turns out that using try/except and doing the
# return in the except is slower by a factor of four.
def removeAny(myList,item):
  while myList.count(item)>0:
    myList.remove(item)
  return myList


def stringFindAll(long,short):
    where=long.find(short)
    if where==-1:
      return None
    foundSpots=[]
    while where<>-1:
      foundSpots.append(where)
      where=long.find(short,where+1)
    return foundSpots


def clearSelection(event):
  event.widget.selection_clear()
  return None


def focusNext(event):
  nextWidget=event.widget.tk_focusNext()
  nextWidget.focus_set()
  return "break"


def focusPrev(event):
  prevWidget=event.widget.tk_focusPrev()
  prevWidget.focus_set()
  return "break"


def cleanupCharsForHeader(s):
  if s==None:
    return None
  s=s.replace("\t"," ")
  s=s.replace("\r"," ")
  s=s.replace("\n"," ")
  s=s.replace("\v"," ")
  s=kEighthBitSetRE.sub("?",s)
  return s


def standardMailboxNamesSortFunc(listOfNames):
  listOfNames.remove("inbox")
  listOfNames.remove("outbox")
  listOfNames.remove("trash")
  listOfNames.sort()
  listOfNames.insert(0,"trash")
  listOfNames.insert(0,"outbox")
  listOfNames.insert(0,"inbox")
  return listOfNames


# Very slightly modified from Alex Martelli's news post <9cpm4202cv1@news1.newsguy.com>
# of May 2, 2001, Subject: Stripping HTML tags from a string
# Thanks, Alex

class Cleaner(sgmllib.SGMLParser):
  entitydefs={"nbsp": " "} # I'll break if I want to

  def __init__(self):
    sgmllib.SGMLParser.__init__(self)
    self.result = []
  def do_p(self, *junk):
    self.result.append('\n')
  def do_br(self, *junk):
    self.result.append('\n')
  def handle_data(self, data):
    self.result.append(data)
  def cleaned_text(self):
    return ''.join(self.result)

def stripHTML(text):
  c=Cleaner()
  try:
    c.feed(text)
  except sgmllib.SGMLParseError:
    gLogger.add("Unable to parse HTML, can't remove tags")
    return text
  else:
    t=c.cleaned_text()
    return t
    # No longer necessary, can get returns in encoded
    # text parts, so they're stripped regardless.
    # Strip returns too
    #return t.replace("\r","")


# This assumes that the user will always enter text in the
# latin-1 character set. As far as I can tell, that's the
# case with Tkinter. If it's not and this should be made
# more general, someone will have to educate me.

# This is more complex than it needs to be because I don't
# want to encode actual addresses. That is, if a header is
# in the form
#
# Example E. Example <eee@example.com>
#
# but the human name part has non-ASCII characters in it,
# we encode only the human name. It would be perfectly legal
# to encode the address too but we don't. If something
# doesn't decode the the header, I'd prefer that the user
# be able to read it. The user could conceivably put a
# non-ASCII character in the address proper but that's the
# user's fault. Or so I claim for now.

def latin1HeaderToEncodedHeader(headerText,headerName):
  asciiCS=email.Charset.Charset("us-ascii")
  latin1CS=email.Charset.Charset("iso8859-1")

  h=email.Header.Header("",asciiCS,76,headerName,"\t")
  if headerName.lower() in ("from","to","cc","bcc","reply-to"):
    al=email.Utils.getaddresses([headerText])
    for ind in range(len(al)):
      addrTuple=al[ind]
      if ind>0:
        h.append(", ")
      humanName,emailAddr=addrTuple
      if humanName<>"":
        humanName=stripQuotes(humanName)
        lastChar=""
        count=0
        needsQuotes=0
        while count<len(humanName):
          thisChar=humanName[count]
          if count>0:
            lastChar=humanName[count-1]
          if thisChar==kDoubleQuote and lastChar<>kBackslat:
            humanName=humanName[:count]+kBackslat+humanName[count:]
            needsQuotes=1
            count+=1
            continue
          if thisChar in kCharsNeedingQuotingInMailAddress:
            needsQuotes=1
          count+=1
        if needsQuotes:
          humanName=kDoubleQuote+humanName+kDoubleQuote
        if kEighthBitSetRE.search(humanName):
          h.append(humanName,latin1CS)
          h.append(" <%s>" % emailAddr)
        else:
          h.append("%s <%s>" % (humanName,emailAddr))
      else:
        # No human name
        h.append(emailAddr)

  else:
    # Not address
    if kEighthBitSetRE.search(headerText):
      h.append(headerText,latin1CS)
    else:
      h.append(headerText)

  encodedVersion=h.encode()
  # Don't know why we get leading whitespace
  encodedVersion=encodedVersion.lstrip()
  return encodedVersion


def canonicalizeTextAddr(textAddr,useSWDBName=0):
  al=email.Utils.getaddresses([textAddr])
  canonList=canonicalizeAddrList(al,useSWDBName=useSWDBName)
  return ", ".join(canonList)


# Expects [(name,addr), (name,addr), ...]
def canonicalizeAddrList(addrList,useSWDBName=0):
  stringVersions=[]
  for humanName,emailAddr in addrList:

    if useSWDBName and gPrefs.get("swdbFile")<>None:
      addressbookName=gAddressBook.addrToName(emailAddr)
      if addressbookName<>None:
        humanName=addressbookName

    if humanName=="":
      if emailAddr<>None:
        stringVersions.append(emailAddr)
    else:
      humanName=stripQuotes(humanName)
      lastChar=""
      count=0
      needsQuotes=0
      while count<len(humanName):
        thisChar=humanName[count]
        if count>0:
          lastChar=humanName[count-1]
        if thisChar==kDoubleQuote and lastChar<>kBackslat:
          humanName=humanName[:count]+kBackslat+humanName[count:]
          needsQuotes=1
          count+=1
          continue
        if thisChar in kCharsNeedingQuotingInMailAddress:
          needsQuotes=1
        count+=1
      if needsQuotes:
        humanName=kDoubleQuote+humanName+kDoubleQuote
      stringVersions.append("%s <%s>" % (humanName,emailAddr))
  return stringVersions


def lengthOfLongestStr(myList):
  return max([len(s) for s in myList])


def isThisFileText(openFile):
  openFile.seek(0)
  b=openFile.read(kIsThisTextBytesToRead)
  nonASCII=kNonASCIITextRE.findall(b)
  return len(b)*kIsThisTextMaxPortionNonASCII>=len(nonASCII)


def makeMessageID():
  return "<%s.%s@%s>" % (str(time.time()),str(os.getpid()),gPrefs.get("fqdn"))


def getSubdirFromPath(p):
  return p.split(kPathSep)[-2]


def getMailboxFromPath(p):
  md=p.split(kPathSep)[-3]

  if md=="Maildir":
    return "inbox"
  else:
    return md


def moveToCur(oldPath):
  oldSubdirPath,oldFN=myPathSplit(oldPath)
  maildirPath=myPathSplit(oldSubdirPath)[0]
  newPath=myPathJoin(maildirPath,"cur")
  newPath=myPathJoin(newPath,oldFN)
  if not newPath.endswith(":2,S"):
    newPath=newPath+":2,S" # Seen
  assert not os.path.exists(newPath),"can't move from %s to %s, destination already exists" % \
    (oldPath,newPath)
  os.link(oldPath,newPath)
  os.unlink(oldPath)
  return newPath


def moveToNew(oldPath):
  oldSubdirPath,oldFN=myPathSplit(oldPath)
  maildirPath=myPathSplit(oldSubdirPath)[0]
  newPath=myPathJoin(maildirPath,"new")
  newPath=myPathJoin(newPath,oldFN)
  assert not os.path.exists(newPath),"can't move from %s to %s, destination already exists" % \
    (oldPath,newPath)
  os.link(oldPath,newPath)
  os.unlink(oldPath)
  return newPath


# Does this handle all cases correctly?
def splitRespectingQuotes(headerStr):
  addrs=[]
  charIndex=0
  inQuote=0
  inComment=0
  lastWasBackslat=0
  thisOneBegins=0
  while charIndex<len(headerStr):
    thisChar=headerStr[charIndex]

    if thisChar==kBackslat:
      lastWasBackslat=1
      charIndex+=1
      continue

    if thisChar=="(" and not inQuote and not lastWasBackslat:
      inComment=1
    elif thisChar==")" and not inQuote and not lastWasBackslat:
      inComment=0
    elif thisChar=='"' and not inComment and not lastWasBackslat:
      inQuote=not inQuote
    elif thisChar=="," and not inComment and not inQuote and not lastWasBackslat:
      addrs.append(headerStr[thisOneBegins:charIndex].strip())
      thisOneBegins=charIndex+1

    lastWasBackslat=0
    charIndex+=1

  if thisOneBegins<len(headerStr):
    addrs.append(headerStr[thisOneBegins:].strip())
  return addrs


def linesFromListOfStrs(inList,maxWidth,separator):
  outList=[""]
  outIndex=0
  outRanges=[]
  for inIndex in range(len(inList)):
    needSeparator=inIndex<len(inList)-1
    addWidth=len(inList[inIndex])
    if needSeparator:
      addWidth+=len(separator)
    if addWidth<=maxWidth:
      outStrLen=len(inList[inIndex])
      addStr=inList[inIndex]+needSeparator*separator
    else:
      outStrLen=maxWidth-needSeparator*len(separator)
      addStr=truncateWithEllipsis(inList[inIndex],maxWidth-needSeparator*len(separator))+needSeparator*separator

    if len(outList[outIndex])+len(addStr)>maxWidth:
      outIndex+=1
      outList.append("")

    outRanges.append(bunch(
      lineNo=outIndex,
      start=len(outList[outIndex]),
      end=len(outList[outIndex])+outStrLen,
      text=inList[inIndex]))
    outList[outIndex]+=addStr
  return outList,outRanges


def hardWrap(myText,maxLength,useFileLineSep=0):
  if useFileLineSep:
    myLineSep=kFileLineSep
  else:
    myLineSep="\n" # Tkinter widgets use newline regardless of OS

  likeToSplitOn=" \t-.,:;/_+=@<>"

  outputText=""

  st=cStringIO.StringIO(myText)
  inputLine=st.readline()

  while inputLine<>"":
    if len(inputLine)<=maxLength or (inputLine.endswith(myLineSep) and
      len(inputLine)<=maxLength+len(myLineSep)):
      outputText+=inputLine
      inputLine=st.readline()
      continue

    while len(inputLine)>maxLength:
      lastOnThisLine=-1
      for char in likeToSplitOn:
        whereThisChar=inputLine[:maxLength].rfind(char)
        if whereThisChar<>-1:
          lastOnThisLine=whereThisChar+1
          break

      if lastOnThisLine==-1:
        lastOnThisLine=maxLength

      outputLine=inputLine[:lastOnThisLine].strip()
      inputLine=inputLine[lastOnThisLine:].strip()
      outputText=outputText+outputLine+myLineSep

    if inputLine<>"":
      outputText=outputText+inputLine+myLineSep

    inputLine=st.readline()

  if outputText<>"" and outputText[-1]<>myLineSep:
    outputText=outputText+myLineSep
  return outputText


# This isn't much in the way of internationalization but I don't
# have data to test anything more sophisticated. Contributions
# are welcome!
#
# At least with the fonts that I have installed, Tkinter's Text
# widget seems to display Latin-1 by default. If there's a way to
# change that, I haven't figured it out.

def encodedHeaderToLatin1(str):
  if str.find("=?")==-1:
    return str

  strs=[]
  try:
    textTuples=email.Header.decode_header(str)
  except ValueError:
    return str

  for textTuple in textTuples:
    if textTuple[1]==None:
      strs.append(textTuple[0])
    else:

      try:
        unicodeVersion=textTuple[0].decode(textTuple[1])
      except (LookupError,UnicodeError,ValueError): # Lots of possibilities
        return str

      try:
        latin1Version=unicodeVersion.encode("latin-1")
      except UnicodeError:
        return str

      strs.append(latin1Version)
  return " ".join(strs)


def addrTextToDisplayText(addrText):
  outList=[]
  addrList=email.Utils.getaddresses([addrText])
  for addrTuple in addrList:
    name=addrTuple[0]
    name=name.strip()
    address=addrTuple[1]
    if address==None:
      outList.append("")
    else:
      nick=gAddressBook.addrToNick(addrTuple[1])
      if nick<>None: # Put people we know at the front
        outList.insert(0,nick)
      elif name<>"": # And people we don't at the end
        if name.find(",")<>-1:
          name='"'+name+'"'
        else:
          name=stripQuotes(name)
        outList.append(name)
      else:
        outList.append(address)
  # If we're in there, move to very front
  if outList.count("me")>0:
    outList.remove("me")
    outList=["me"]+outList
  return ", ".join(outList)


def charWidthsToDotWidths(charWidthsSeq):
  columnWidths=[]
  f=tkFont.Font(font=gPrefs.get("fixedWidthFontILike"))
  oneCharWidth=f.measure("x")
  for charWidth in charWidthsSeq:
    columnWidths.append(oneCharWidth*charWidth)
  return columnWidths


# Reasonably self-contained routines that implement some of
# the functions in the menus

def emptyTrash():
  gLogger.add("Emptying trash")
  ##m=maildir(myPathJoin(gPrefs.get("userMailboxDirParent"),"trash"))
  ##gSwitchboard.delete("trash",m)
  m=gAllMailboxes.getMaildir("trash")
  for mailObj in m:
    if not (gSwitchboard.getEditorByPath(mailObj.getPath()) or \
           gSwitchboard.getViewerByPath(mailObj.getPath())):
      mailObj.unlink()
    else:
      gLogger.add("Mail not deleted from trash because it's open")
  gSwitchboard.redisplayFromScratch("trash")
  gSwitchboard.counts("trash")
  gLogger.add("Trash emptying complete")
  return None


#~ def reloadPrefs():
  #~ global gMousePointerHider
  #~ gPrefs.load()
  #~ if gPrefs.get("hideMousePointerWhileTyping"):
    #~ gMousePointerHider=mousePointerHider()
  #~ #if not sys.modules.has_key("swdb") and gPrefs.get("swdbFile"):
  #~ #  try:
  #~ #    import swdb
  #~ #  except ImportError:
  #~ #    sys.path.append(os.path.dirname(sys.argv[0]))
  #~ #    import swdb
  #~ #  gAddressBook=swdb.addressBook(gPrefs,gLogger,gSwitchboard)
  #~ #else:
  #~ #  gAddressBook=addressBook()

  #~ gAddressBook.load()

  #~ gSwitchboard.redoLocationItems()
  #~ gLogger.add("Preferences file reloaded")
  #~ return None
  
# Helper classes for localRelDate(). The first implements
# the local timezone and the second the timezone the mail
# is from. First one swiped from Python docs, changed a
# trifle
class localTimezone(datetime.tzinfo):
  def __init__(self):
    self.stdOffset = datetime.timedelta(seconds = -time.timezone)
    # time.daylight indicates presence of DST, not that it's currently DST
    if time.daylight:
      self.dstOffset = datetime.timedelta(seconds = -time.altzone)
    else:
      self.dstOffset = self.stdOffset

    self.dstDiff = self.dstOffset - self.stdOffset
    return None
    
  def utcoffset(self, dt):
    if self.isdst(dt):
      return self.dstOffset
    else:
      return self.stdOffset

  def dst(self, dt):
    if self.isdst(dt):
      return self.dstDiff
    else:
      return datetime.timedelta(0)

  def tzname(self, dt):
    return time.tzname[self.isdst(dt)]

  def isdst(self, dt):
    tt = (dt.year, dt.month, dt.day,
        dt.hour, dt.minute, dt.second,
        dt.weekday(), 0, -1)
    stamp = time.mktime(tt)
    tt = time.localtime(stamp)
    return tt.tm_isdst > 0


class mailFixedTZ(datetime.tzinfo):

  def __init__(self,offsetMinsEast):
    self.offsetMinsEast=offsetMinsEast

  def utcoffset(self,dtIgnored):
    return datetime.timedelta(minutes=self.offsetMinsEast)


class newMailbox:
  def __init__(self):
    myFont=gPrefs.get("fontILike")
    self.newMailboxDialog=Pmw.Dialog(buttons=("OK","Cancel"),
      command=self.newMailboxDialogCB,
      defaultbutton=0,
      title="New mailbox")
    stuffParent=self.newMailboxDialog.interior()
    l=Label(stuffParent,text="New mailbox",font=myFont)
    l.pack(side=LEFT,padx=5,pady=5)
    border=Frame(stuffParent,bg="black",borderwidth=1)
    self.mailboxNameEntry=Entry(border,
      font=myFont,
      width=30,
      relief=FLAT,
      highlightthickness=0,
      bg="white")
    self.mailboxNameEntry.pack(side=LEFT)
    border.pack(side=LEFT,padx=5,pady=5)
    self.newMailboxDialog.component("buttonbox").component("Cancel").configure(
      font=myFont,activebackground=kTkDefaultBG)
    self.newMailboxDialog.component("buttonbox").component("OK").configure(
      font=myFont,activebackground=kTkDefaultBG)
    self.mailboxNameEntry.focus_set()
    return None

  def newMailboxDialogCB(self,dialogResult):
    if dialogResult=="OK":
      newName=self.mailboxNameEntry.get()
      badName=0
      for char in newName:
        if char not in kOKMailboxChars:
          badName=1
          break
      if badName:
        gLogger.add('Name "%s" has funny characters in it, I won\'t create it.' % newName)
      elif gAllMailboxes.isMaildir(newName):
        gLogger.add('There already is a mailbox "%s", can\'t create it.' % newName)
      else:
        gAllMailboxes.newMaildir(newName) # newMaildir() does the broadcast
        gLogger.add("New mailbox %s created." % newName)
    self.newMailboxDialog.destroy()
    self.newMailboxDialog=None
    return None


def quit(evenIfEditorsOpen=0):
  gLogger.add("Quitting %s" % kMyName)
  if not evenIfEditorsOpen:
    if len(gSwitchboard.getEditors())>0:
      gLogger.add("Didn't quit because one or more mail editors are open")
      return None
  gTkRoot.after_cancel(gSwitchboard.getOverview().checkQueueAfterID)
  if gChecker.getCheckingRegularly():
    gLogger.add("Checker thread running, asking it to terminate")
    gChecker.setCheckingRegularly(0)
  gAddressBook.cleanUpForQuit()
  saveState()
  if gPrefs.get("hideMousePointerWhileTyping"):
    gMousePointerHider.cleanUpForQuit()
  gSwitchboard.closeAll()
  wanna=gPrefs.get("emptyTrashOnQuit")
  if wanna:
    emptyTrash()
  for mailbox,daysOld in gPrefs.get("autoTrash"):
    moveToTrashIfOlderThan(mailbox,daysOld)
  gPrefs.get("cleanupForQuit")()
  gotIt=checker.checkRegularlyLock.acquire(0)
  if not gotIt:
    gLogger.add("Checker thread running, waiting for it to terminate")
    checker.checkRegularlyLock.acquire(1) # Block waiting
  gLogger.add("%s %s ends" % (kMyName,kVersion))
  sys.exit(0)


def moveToTrashIfOlderThan(mailbox,daysOld):
  fromMaildirObj=gAllMailboxes.getMaildir(mailbox)
  trashMaildirObj=gAllMailboxes.getMaildir("trash")
  olderThan=int(time.time())-(daysOld*kSecsInDay)
  for mailObj in fromMaildirObj.getMailObjs():
    if mailObj.getDateAsSecs()<olderThan:
      result=trashMaildirObj.copyToThisMaildir(mailObj)
      if result<>None:
        os.unlink(mailObj.path)
  return None


# Some of the time we do our own base64 decoding (but not encoding)
# here. It's not that Python's isn't good but rather than I want to
# relax the rules slightly. Since this routine is a little less than
# one twentieth the speed of the C implementation in Python's library,
# it's called as a second try if that module raises an exception.

# The reason for relaxing the rules some is that more and more software
# running on mailgates adds a disclaimer, a virus-check assurance, or
# some other sort of footer to outgoing mail. If a message is
# single-part and base64-encoded, doing that correctly would require
# changing the message's MIME structure to multipart and adding the
# footer as a new text part.  Alas, even some otherwise-excellent
# software doesn't do that.  Instead, lots of software just appends
# the footer to the message.

# Python's binascii module (and the base64 module that uses it, and
# the email module that uses the base64 module) complains when
# asked to decode a message like that since the decoder runs past the
# end of the stuff that's actually encoded and tries to decode the
# plain text footer. That's quite correct according to my reading of
# RFC 2045 but it's not a behavior I want for two reasons: First,
# failing to decode a base64 part leaves the user with gibberish. If
# it were a quoted-printable part at least the user coule probably
# make some sense of it. Second, the way to deal with the problem is
# obvious from even a glance at the message.

# So this routine skips any blank lines that come before a
# base64-encoded part, does the decoding, and quits at whichever comes
# first of end-of-file, a blank line, and the end of the base64 data.
# It's not always possible to determine the end of base64 data but it
# sometimes is: Base64 data is always encoded as groups of 4 bytes,
# but input data doesn't always fill out the last 4 bytes. For that
# reason "="s are used to pad the output to a multiple of 4.
# Therefore, any "=" in the data indicate the end. Stopping at an "="
# is sanctioned by RFC 2045. Stopping at a blank line is questionable
# but it's what I want to do.


def myBase64Decode(inFile,outFile):

  # First, try to use the speedier standard decoder
  try:
    base64.decode(inFile,outFile)
  except binascii.Error:
    gLogger.add("Base64 decoding error encountered, trying alternate decoder")
    inFile.seek(0)
    outFile.seek(0)
  else:
    return None

  valDict= \
    {
    "A": 0,  "B": 1,  "C": 2,  "D": 3,  "E": 4,  "F": 5,  "G": 6,  "H": 7,
    "I": 8,  "J": 9,  "K": 10, "L": 11, "M": 12, "N": 13, "O": 14, "P": 15,
    "Q": 16, "R": 17, "S": 18, "T": 19, "U": 20, "V": 21, "W": 22, "X": 23,
    "Y": 24, "Z": 25, "a": 26, "b": 27, "c": 28, "d": 29, "e": 30, "f": 31,
    "g": 32, "h": 33, "i": 34, "j": 35, "k": 36, "l": 37, "m": 38, "n": 39,
    "o": 40, "p": 41, "q": 42, "r": 43, "s": 44, "t": 45, "u": 46, "v": 47,
    "w": 48, "x": 49, "y": 50, "z": 51, "0": 52, "1": 53, "2": 54, "3": 55,
    "4": 56, "5": 57, "6": 58, "7": 59, "8": 60, "9": 61, "+": 62, "/": 63,
    "=": None,
    }
  seenDecodableData=0
  buf=""
  while 1:
    if len(buf)<4:
      l=inFile.readline()
      if l=="":
        break
      l=l.strip()
      l=l.rstrip("\0")
      if l=="":
        if not seenDecodableData:
          continue
        else:
          break
      if len(l)>76:
        raise ValueError("Base64-encoded line too long: is %i chars, should be <=76" % \
          len(l))
      buf=buf+l
    seenDecodableData=1
    inQuad=buf[:4]
    buf=buf[4:]
    inQuad=list(inQuad)
    outTriple=[None]*3
    for count in range(4):
      try:
        inQuad[count]=valDict[inQuad[count]]
      except KeyError:
        raise ValueError('Base64-encoded part contains illegal character "%s" ' \
          "should be in [A-Za-z+/=]" % inQuad[count])
      except IndexError: # very rare but seen in the wild
        raise ValueError("Wrong number of characters in base-64 encoded part")

    if inQuad[0]==None:
      raise ValueError('Base64 part has "=" pad character as the first character of a set of 4, '
        "that's not allowed")
    elif inQuad[1]==None:
      raise ValueError('Base64 part has "=" pad character as the second character of a set of 4, '
        "that's not allowed")
    elif inQuad[2]==None:
      outTriple[0]=inQuad[0] << 2 | inQuad[1] >> 6
      outBytes=struct.pack("B",outTriple[0])
      outFile.write(outBytes)
      break
    elif inQuad[3]==None:
      outTriple[0]=inQuad[0] << 2 | inQuad[1] >> 6
      outTriple[1]=(inQuad[1] << 4 | inQuad[2] >> 2) & 255
      outBytes=struct.pack("BB",outTriple[0],outTriple[1])
      outFile.write(outBytes)
      break
    outTriple[0]=inQuad[0] << 2 | inQuad[1] >> 4
    outTriple[1]=(inQuad[1] << 4 | inQuad[2] >> 2) & 255
    outTriple[2]=(inQuad[2] << 6 | inQuad[3]) & 255
    outBytes=struct.pack("BBB",outTriple[0],outTriple[1],outTriple[2])
    outFile.write(outBytes)

  if len(buf)>0:
    #print "Remains",repr(buf)
    raise ValueError("Length of base64-encoded part is not a multiple of 4")

  while 1:
    l=inFile.readline()
    if l=="":
      break
    outFile.write(l)

  return None


def decodePayload(emailMsg):
  payload=emailMsg.get_payload()
  # Put returned message headers in a text part
  if payload==None: ### ?
    return emailMsg.as_string()
  assert isinstance(emailMsg,email.Message.Message)
  assert type(payload)==types.StringType, "decodePayload called with message that has " \
    "payload type %s" % type(payload)
  cte=emailMsg.get("content-transfer-encoding","").lower()
  if cte in ("7bit","8bit","binary"):
    # No encoding, do nothing
    return emailMsg.get_payload()
  elif cte=="base64":
    inFile=cStringIO.StringIO(payload)
    outFile=cStringIO.StringIO()
    try:
      myBase64Decode(inFile,outFile)
    except ValueError,detail: # Even my lax decoder barfs
      gLogger.add("Base64 decoding failed: %s" % detail)
      return emailMsg.get_payload()
    retVal=outFile.getvalue()
    inFile.close()
    outFile.close()
    return retVal
  elif cte=="quoted-printable":
    inFile=cStringIO.StringIO(payload)
    outFile=cStringIO.StringIO()
    quopri.decode(inFile,outFile)
    retVal=outFile.getvalue()
    inFile.close()
    outFile.close()
    return retVal
  elif cte=="":
    # Not specified, do nothing
    return emailMsg.get_payload()
  else:
    gLogger.add('Unrecognized content-transfer-encoding "%s"' % cte)
    return emailMsg.get_payload()
  return None


# Should these next three be here?
# Parse a file containing a mail message that may have uuencoded
# hunks strewn through it and return a structure as though we had
# parsed a proper MIME message.
def getUUParts(msg):
  # Skip past headers
  while 1:
    l=msg.readline()
    if l=="" or l==kFileLineSep:
      break

  partsInfo=[]
  ordinaryText=cStringIO.StringIO()
  inUUPart=0
  for l in msg.readlines():
    partBeginsHere=kBeginUURE.search(l)
    # Normal case
    if not inUUPart and not partBeginsHere:
      ordinaryText.write(l)
      continue
    # Another line of a uuencoded part
    if inUUPart and l.rstrip()<>"end":
      thisUUPart.write(l)
      continue
    # The end of a uuencoded part
    if inUUPart and l.rstrip()=="end":
      thisUUPart.write(l)
      thisUUPart.seek(0)
      decodedVersion=cStringIO.StringIO()
      try:
        uu.decode(thisUUPart,decodedVersion)
      except (binascii.Error,uu.Error):
        partsInfo.append(bunch(
          part=kUnableToParseMessage,
          decodedPart=kUnableToParseMessage,
          name=uuFileName,
          mainType="text",
          type="text/plain",
          ))
      else: # Successful decoding
        if isThisFileText(decodedVersion):
          thisType="text"
        else:
          thisType="application"
        partsInfo.append(bunch(
          decodedPart=decodedVersion.getvalue(),
          name=uuFileName,
          mainType=thisType,
          ))
      thisUUPart.close()
      inUUPart=0
      continue
    # The beginning of a uuencoded part
    if not inUUPart and partBeginsHere:
      inUUPart=1
      thisUUPart=cStringIO.StringIO()
      thisUUPart.write(l)
      uuFileName=partBeginsHere.group(1)
      continue
    assert 1==0,"Unhandled condition in uudecoding"
  textPart=ordinaryText.getvalue()
  partsInfo.insert(0,bunch(
    decodedPart=textPart,
    mainType="text",
    name="Part-1",
    ))

  for part in partsInfo:
    if part.mainType=="text":
      textNeedsQuopri=0
      textAsFile=cStringIO.StringIO(part.decodedPart)
      while 1:
        l=textAsFile.readline()
        if l=="":
          break
        if len(l)>kEncodeQpriIfLineLongerThan:
          textNeedsQuopri=1
          break
        if kEighthBitSetRE.search(l):
          textNeedsQuopri=1
          break
      if textNeedsQuopri:
        textAsFile.seek(0)
        outFile=cStringIO.StringIO()
        quopri.encode(textAsFile,outFile,1) # 1 For quote tabs. No reason.
        part.part=outFile.getvalue()
        part.content_transfer_encoding="quoted-printable"
        part.content_type="text/plain"
        textAsFile.close()
        outFile.close()
      else: # No need to encode
        part.part=part.decodedPart
        part.content_transfer_encoding="7bit"
        part.content_type="text/plain"
    else: # Not text
      inFile=cStringIO.StringIO(part.decodedPart)
      outFile=cStringIO.StringIO()
      base64.encode(inFile,outFile)
      part.part=outFile.getvalue()
      part.content_transfer_encoding="base64"
      part.content_type="application/octet-stream"
      inFile.close()
      outFile.close()
  return partsInfo


# This deliberately flattens any nested MIME structure.
def getMimeParts(msg):
  partsInfo=[]
  getMimePartsRecurse(msg,partsInfo)
  return partsInfo

def getMimePartsRecurse(msg,partsInfo):
  # message/rfc822 parts are represented by the email module
  # as Message objects with the payloads of the original
  # message. If we ignore the MIME structure, we wouldn't decode
  # any encoded parts but if we only decode the contents, we'd
  # miss the headers of the original message. The orignal
  # message's headers may contain useful information such as the
  # address or name of the original sender. So we hack up an
  # extra MIME part that doesn't exist in the original which
  # holds the headers of the message/rfc822 part. There's even a
  # standard MIME type that's appropriate.

  # XXX Do we ever want to tidy these headers the way we do with a
  # regular mesage's headers?
  if msg.is_multipart() and msg.get_content_type()=="message/rfc822":
    m=msg.get_payload()[0]
    headersList=[ k+": "+m[k] for k in m.keys() ]
    headersText="\n".join(headersList)+"\n"
    partsInfo.append(bunch(
      part=headersText,
      decodedPart=headersText,
      content_type="message/rfc822-headers",
      mainType="message",
      content_description=msg.get_param("content-description"),
      name=None,
      charset=None,
      content_transfer_encoding=None,
      ))

  if msg.is_multipart():
    for subMsg in msg.get_payload():
      getMimePartsRecurse(subMsg,partsInfo)
  else:
    decoded=decodePayload(msg)
    if msg.get_content_maintype() in kDisplayableMainTypes and not decoded.endswith(kFileLineSep):
      decoded+=kFileLineSep
    # Some broken spamware adds newlines so we do a couple of strip()s.
    # Don't know if it's worth caring about content-description
    
    n=msg.get_param("name")
    if n==None:
      n=msg.get_filename()
      
    partsInfo.append(bunch(
      part=msg.get_payload(),
      decodedPart=decoded,
      content_type=stripIfPossible(msg.get_content_type()),
      mainType=stripIfPossible(msg.get_content_maintype()),
      content_description=msg.get_param("content-description"),
      name=n, #msg.get_param("name"),
      charset=msg.get_param("charset"),
      content_transfer_encoding=lowerIfPossible(msg.get("content-transfer-encoding")),
      ))
  return None


def formatRFC822Date(timeSeq):
  # Can't use the strings from the time module since they
  # vary by locale and RFC2822 requires English.

  #dayNames=("Mon","Tue","Wed","Thu","Fri","Sat","Sun")
  #monthNames=("","Jan","Feb","Mar","Apr","May","Jun","Jul","Aug", \
  #            "Sep","Oct","Nov","Dec")


  timeStr="%s, %s %s %s " % (
    kDayAbbrs[timeSeq[6]],
    timeSeq[2],
    kMonthAbbrs[timeSeq[1]],
    timeSeq[0])

  timeStr+=time.strftime("%H:%M:%S ",timeSeq[:9])

  zoneSecs=timeSeq[9]
  zoneAbsSecs=abs(zoneSecs)
  zoneMins=int(zoneAbsSecs/60)
  zoneHours=int(zoneMins/60) # Actually don't need the int(), already integer division
  zoneLeftoverMins=zoneMins-zoneHours*60

  if zoneSecs>0:
    sign="-"
  else:
    sign="+"

  timeStr+="%s%02i%02i" % (
    sign,
    zoneHours,
    zoneLeftoverMins)

  if timeSeq[10]<>"":
    timeStr+=" ("+timeSeq[10]+")"
  return timeStr


def getDateForTidyHeaders(dateStr):
  timeTuple=email.Utils.parsedate_tz(dateStr)
  if timeTuple==None:
    timeTuple=email.Utils.parsedate_tz("Dummyday, "+dateStr)
  
  if timeTuple==None or timeTuple[9]==None:
    return None
    
  tzMatch=kTimeZoneNameInParensRE.search(dateStr)
  if tzMatch==None:
    tz=None
  else:
    tz=tzMatch.group(1)
  
  d=canonicalizeRFC822Date(timeTuple,tz)
  r=localRelDate(timeTuple)
  if r==None:
    r="can't figure out what that would be here"
  else:
    r="%s here" % r
    
  return "%s [%s]" % (d,r)


##def localRelDate(theirTimeSeq):
##  myTimeSeq=list(time.localtime(time.time()))
##  
##  if myTimeSeq[8]==0:
##    myZoneSecs=time.timezone
##  else:
##    myZoneSecs=time.altzone
##  
##  #theirTimeSeq=email.Utils.parsedate_tz(dateStr)
##  if theirTimeSeq[9]==None:
##    return None
##
##  theirZoneSecs=-theirTimeSeq[9]
##
##  diffSecs=myZoneSecs-theirZoneSecs
##
##  if diffSecs==0:
##    return "same"
##
##  theirSecs=time.mktime(theirTimeSeq[:9])
##  theirTimeOurZoneSeq=list(time.localtime(theirSecs-diffSecs))
##  
##  if theirTimeOurZoneSeq[8]==0:
##    zoneSecs=time.timezone
##  else:
##    zoneSecs=time.altzone
##   
##  timeStr=kDayAbbrs[theirTimeOurZoneSeq[6]]+", "
##
##  timeStr+="%02i:%02i" % tuple(theirTimeOurZoneSeq[3:5])
##  return timeStr


def localRelDate(theirTimeTuple):

  theirTZ=mailFixedTZ(theirTimeTuple[9]/60)

  theirDatetime=datetime.datetime(year=theirTimeTuple[0],
    month=theirTimeTuple[1],
    day=theirTimeTuple[2],
    hour=theirTimeTuple[3],
    minute=theirTimeTuple[4],
    second=theirTimeTuple[5],
    tzinfo=theirTZ)
  
  # The time may not be valid. For example 2:01 AM on the day that
  # daylight saving time begins doesn't exist.
  try:
    myDatetime=theirDatetime.astimezone(localTimezone())
  except OverflowError:
    return None
    
  if myDatetime.utcoffset()==theirDatetime.utcoffset():
    return "same"
  else:
    timeStr=kDayAbbrs[myDatetime.weekday()]+", "
    timeStr+="%02i:%02i" % (myDatetime.hour,myDatetime.minute)
    return timeStr
    

def canonicalizeRFC822Date(t,tz):
  tList=list(t)
  tList[9]=-tList[9]
  try:
    tList[6]=calendar.weekday(tList[0],tList[1],tList[2])
  except ValueError:
    return None
##  tzMatch=kTimeZoneNameInParensRE.search(dateStr)
##  if tzMatch==None:
##    tList.append("")
##  else:
##    tList.append(tzMatch.group(1))
  if tz==None:
    tList.append("")
  else:
    tList.append(tz)
  return formatRFC822Date(tList)


def getRFC822Date():
  timeSeq=list(time.localtime(time.time()))
  if timeSeq[8]<>0:
    zoneSecs=time.altzone
  else:
    zoneSecs=time.timezone
  zoneName=time.tzname[timeSeq[8]]
  timeSeq+=[zoneSecs,zoneName]
  return formatRFC822Date(timeSeq)


def checkSwallowDir():
  swallowDir=gPrefs.get("swallowDir")
  if os.path.isdir(swallowDir):
    return None

  if not os.path.exists(swallowDir):
    most,tail=myPathSplit(swallowDir)
    if os.path.isdir(most):
      gLogger.add("Creating Swallow's direcotry %s" % swallowDir)
      os.umask(0077) # octal
      os.mkdir(swallowDir)
      return None

    if os.path.exists(most):
      print "The directory above your Swallow directory %s" % swallowDir
      print "exists but isn't a directory. I don't want to delete it."
      print "Can you sort that out for me?"
      sys.exit(0)
    else:
      print "The directory above your Swallow direcotry %s" % swallowDir
      print "doesn't exist. I'll be glad to create the directory itself"
      print "but I want to create only one level. Would you creae"
      print "%s for me if it's where you want me to put it?" % most
      sys.exit(0)



def checkTempDir():
  tempDir=gPrefs.get("tempDir")
  if os.path.exists(tempDir):

    if os.path.isdir(tempDir):
      # Temp dir looks fine, clean up old files

      aWeekAgo=int(time.time())-kSecsInWeek # Make this a constant?
      filenames=os.listdir(tempDir)
      for filename in filenames:
        fullPath=myPathJoin(tempDir,filename)
        if os.path.isfile(fullPath):
          mtime=os.path.getmtime(fullPath)
          if mtime<=aWeekAgo:
            gLogger.add("Unlinking old file %s, it's in the temp " \
              "directory and hasn't been changed in a week." % fullPath)
            os.unlink(fullPath)
        else:
          gLogger.add("Warning:")
          gLogger.add("%s is in the temp directory" % fullPath)
          gLogger.add("and it's not a file. That's very strange.")
          gLogger.add("I'm not going to touch it, would you please")
          gLogger.add("deal with it for me?")
      return None
    else:
      print "Temp directory %s exists but isn't a directory." % tempDir
      print "I need it but won't delete what's there."
      print "You're going to have to fix it for me."
      sys.exit(1)

  most,tail=myPathSplit(tempDir)
  if not os.path.exists(most):
    print "The directory above your temp dir doesn't exist. I'll be glad to"
    print "create the last bit but I really only want to create one level."
    print "Would you please create %s for me." % most
    sys.exit(1)

  if not os.path.isdir(most):
    print "The directory above your temp dir exists. But it's not actually"
    print "a directory. Would you please sort out what's going on for me."
    print "The file in question is %s" % most
    sys.exit(1)

  gLogger.add("Creating temp file directory %s" % tempDir)
  os.umask(0077) # octal
  os.mkdir(tempDir)
  return None

## rot-13 encoder included in Python 2.2, use it when we stop
## supporting 2.0
#def rot13(inStr):
#  myList=list(inStr)
#  for count in range(len(myList)):
#    myChar=myList[count]
#    if "a" <= myChar <= "m" or "A" <= myChar <= "M":
#      myChar=chr(ord(myChar)+13)
#    elif "n" <= myChar <= "z" or "N" <= myChar <= "Z":
#      myChar=chr(ord(myChar)-13)
#    myList[count]=myChar
#  outStr="".join(myList)
#  return outStr


def strPadOrTrim(str,len):
  return (str+len * " ")[:len]



#
# General classes
#

class nullClass:
  pass


# From http://www.activestate.com/ASPN/Python/Cookbook/Recipe/52308
# Thanks, Alex.
# __repr__(), __getitem__(), modification of __init__ to take dicts my own

class bunch:
  def __init__(self,*posArgs,**kwds):
    for arg in posArgs:
      self.__dict__.update(arg)
    self.__dict__.update(kwds)

  def __getitem__(self,key):
    return getattr(self,key)

  def __repr__(self):
    return self.__class__.__name__+": "+str(self.__dict__)


class multiValDict:
  def __init__(self):
    self.data={}
    return None

  def __setitem__(self,key,val):
    if type(val)<>types.ListType:
      val=[val]
    if self.data.has_key(key):
      self.data[key]+=val
    else:
      self.data[key]=val
    return None

  def __repr__(self):
    return "%s instance: %s" % (self.__class__.__name__,self.data)

  def __getitem__(self,key):
    return self.data[key]

  def keys(self):
    return self.data.keys()


class sendmail: # Perhaps should call this postfix <wink>.
  def __init__(self):
    self.conn=None
    
    self.changedTimeout=False
    self.oldTimeout=None
    return None

  def setup(self):
    if gPrefs.get("smtpTLS"):
      # Needed to work around a bug in the socket module
      self.changedTimeout=True
      self.oldTimeout=socket.getdefaulttimeout()
      socket.setdefaulttimeout(None)
      
    server=gPrefs.get("smtpServer")
    port=gPrefs.get("smtpPort")
    if port==25:
      gLogger.add("Connecting to SMTP server %s" % server)
    else:
      gLogger.add("Connecting to SMTP server %s, port %i" % (server,port))

    try:
      conn=smtplib.SMTP(server,port)

    except socket.error,detail:
      if detail==errno.ETIMEDOUT:
        gLogger.add("Connection to SMTP server %s timed out" % server,kError)
        return None
      else:
        gLogger.add("Unable to connect to SMTP server %s, socket reports: %s" % \
          (server,str(detail)),kError)
        return None
    except socket.timeout:
      if port==25:
        gLogger.add("Timeout connecting to SMTP server %s" % server,kError)
      else:
        gLogger.add("Timeout connecting to SMTP server %s, port %i" % (server,port),kError)
      return None

    self.conn=conn
    replyTuple=self.conn.helo(gPrefs.get("fqdn"))
    if str(replyTuple[0])[:1]<>"2":
      gLogger.add("Got error reply to my HELO: %i %s" % replyTuple,kError)
      # May as well quit now, we're not going to send any mail through
      # this connection
      self.conn.quit()
      return None
      
    if gPrefs.get("smtpTLS"):
      # Needed to work around a bug in the socket module
      self.changedTimeout=True
      self.oldTimeout=socket.getdefaulttimeout()
      socket.setdefaulttimeout(None)
      
      gLogger.add("Starting SMTP TLS")
      replyTuple=self.conn.starttls()
      if str(replyTuple[0])[:1]<>"2":
        gLogger.add("Got error reply to my STARTTLS: %i %s" % replyTuple,kError)
        self.conn.quit()
        return None
      # Must re-hello
      gLogger.add("Saying hello again")
      replyTuple=self.conn.ehlo()
      if str(replyTuple[0])[:1]<>"2":
        gLogger.add("Got error reply to my EHLO after STARTTLS: %i %s" % replyTuple,kError)
        self.conn.quit()
        return None
    
    if gPrefs.get("smtpAuth"):
      gLogger.add("Doing SMTP login")
      user=gPrefs.get("smtpUser")
      pw=gPrefs.get("smtpPass")
      pw=pw.encode("rot-13")
      replyTuple=self.conn.login(user,pw)
      if str(replyTuple[0])[:1]<>"2":
        gLogger.add("Got error reply to my LOGIN: %i %s" % replyTuple,kError)
        self.conn.quit()
        return None
      
    return 1 # Indicate success

  def sendOne(self,mailObj,progressCB=None):
    ## # If we have one, we need the other
    ## #assert (progressCB==None) == (progressReportEach==None)
    if gPrefs.get("sanityCheckOutboundMail"):
      s=outboundMailSanityChecker()
      mailPath=mailObj.path
      gLogger.add("Checking %s for sanity" % mailPath)
      s.check(mailPath)
      gLogger.add("Nothing insane found")

    # RFC 2822 points out that some SMTP servers copy data from all
    # RCPT TO commands into headers. Therefore, to preserve privacy
    # it suggests that it may be desirable to send to any BCC recipients
    # seaperately and each as a separate transaction. We do that.

    regularRecipients=mailObj.getToAndCCRecipients()
    bccRecipients=mailObj.getBCCRecipients()

    assert len(regularRecipients)+len(bccRecipients)>0,"In sendmail.sendOne() but no recipents"

    if len(regularRecipients)>0:
      replyTuple=self.conn.mail(gPrefs.get("myMailAddress"))
      if str(replyTuple[0])[:1]<>"2":
        gLogger.add("Got error reply to my MAIL FROM: %i %s" % replyTuple,kError)
        gLogger.add("Mail not sent")
        return None

      #gLogger.add('Sending message "%s" to <%s>' % (mailObj.get("subject",""),", ".join(regularRecipients)))
      lm='Sending message "%s" to ' % mailObj.get("subject","")
      lm+=", ".join([ "<%s>" % s for s in regularRecipients ])
      gLogger.add(lm)
      #gLogger.add('Sending message "%s" to <%s>' % (mailObj.get("subject",""),", ".join(regularRecipients)))

      for recipient in regularRecipients:
        replyTuple=self.conn.rcpt(recipient)
        if str(replyTuple[0])[:1]<>"2":
          gLogger.add("Got error reply to my RCPT TO: %i %s" % replyTuple,kError)
          gLogger.add("Mail not sent")
          return None

      replyTuple=self.conn.docmd("data")
      if str(replyTuple[0])[:1]<>"3":
        gLogger.add("Got error reply to my command DATA %i %s" % replyTuple,kError)
        gLogger.add("Mail not sent")
        return None
      self.doData(mailObj.path,progressCB)
      replyTuple=self.conn.getreply()

      if str(replyTuple[0])[:1]<>"2":
        gLogger.add("Got error reply to end of mail message: %i %s" % replyTuple,kError)
        gLogger.add("Mail not sent")
        return None
      gLogger.add("Message accepted, remote server said: %i %s" % replyTuple)

    for recipient in bccRecipients:
      gLogger.add('Sending message "%s" to <%s>' % (mailObj.get("subject",""),recipient))

      #gLogger.add("Sending message to %s" % recipient)
      replyTuple=self.conn.mail(gPrefs.get("myMailAddress"))
      if str(replyTuple[0])[:1]<>"2":
        gLogger.add("Got error reply to my MAIL FROM: %i %s" % replyTuple,kError)
        gLogger.add("Mail not sent")
        return None
      replyTuple=self.conn.rcpt(recipient)
      if str(replyTuple[0])[:1]<>"2":
        gLogger.add("Got error reply to my RCPT TO: %i %s" % replyTuple,kError)
        gLogger.add("Mail not sent")
        return None
      replyTuple=self.conn.docmd("data")
      if str(replyTuple[0])[:1]<>"3":
        gLogger.add("Got error reply to my command DATA %i %s" % replyTuple,kError)
        gLogger.add("Mail not sent")
        return None
      self.doData(mailObj.path,progressCB)
      replyTuple=self.conn.getreply()

      if str(replyTuple[0])[:1]<>"2":
        gLogger.add("Got error reply to end of mail message: %i %s" % replyTuple,kError)
        gLogger.add("Mail not sent")
        return None
      gLogger.add("Message accepted, remote server said: %i %s" % replyTuple)

    return 1 # Indicate success

  def doData(self,path,progressCB):
    linesSent=0
    bytesSent=0
    bytesSentToCompareWithFileSize=0.0 # Float because we want float division
    fileSize=os.stat(path).st_size

    f=open(path)
    if f==None:
      gLogger.add("Error sending mail") # More detail logged by getFileNameForSending()
      return None

    nextProgressAt=time.time()+kShowProgressEvery
    inHeaders=1
    lastWasBCC=0
    doingProgress=0
    while 1:
      line=f.readline()
      if line=="":
        break
      if line==kFileLineSep:
        inHeaders=0

      if time.time()>=nextProgressAt:
        portionSent=(bytesSentToCompareWithFileSize+len(line))/fileSize
        if progressCB<>None:
          progressCB((bytesSentToCompareWithFileSize+len(line))/fileSize)
        gLogger.add("%i%% sent" % (portionSent*100))
        doingProgress=1
        nextProgressAt+=kShowProgressEvery

      ## # Last two lines of if statement see that callback is called
      ## # if this line puts us past a multiple of the number of bytes
      ## # that represents the frequency with which to call the callback.
      ## if progressCB<>None and \
      ##   bytesSentToCompareWithFileSize % kSendingProgressReportEach >= \
      ##   (bytesSentToCompareWithFileSize+len(line)) % progressReportEach:
      ##   progressCB((bytesSentToCompareWithFileSize+len(line))/fileSize)
      bytesSentToCompareWithFileSize+=len(line)

      if inHeaders:
        if line[:5].lower()=="bcc: ":
          lastWasBCC=1
          continue
        elif lastWasBCC and line[0].isspace():
          continue
        else:
          lastWasBCC=0
      else: # Not in headers
        if line[0]==".":
          line="."+line

      if line.endswith(kFileLineSep):
        line=line[:-len(kFileLineSep)]
      line+=kCRLF
      self.conn.send(line)
      bytesSent+=len(line)
      linesSent+=1
    self.conn.send(kCRLF+"."+kCRLF)
    f.close()
    if doingProgress:
      gLogger.add("100% sent")
    gLogger.add("Sent %i lines, %s bytes" % (linesSent,qtyToReadable(bytesSent)))
    return None

  def tearDown(self):
    if self.changedTimeout:
      socket.setdefaulttimeout(self.oldTimeout)
    
    if self.conn<>None:
      try:
        self.conn.quit()
      except smtplib.SMTPServerDisconnected:
        pass
      self.conn=None
    gLogger.add("SMTP conversation complete")
    return None


def sendReady():
  gLogger.add("Sending mail from outbox")
  dirObj=gAllMailboxes.getMaildir("outbox")
  nInOutbox=dirObj.howmanyFiles()
  if nInOutbox==0:
    gLogger.add("Nothing sent, no mail in outbox")
    return None

  gLogger.add("%i %s in outbox" %
    (nInOutbox,pluralize(nInOutbox,"message")))


  mailToSend=[]
  for mailObj in dirObj:
    if mailObj.sendable():
      result=gSwitchboard.getEditorByPath(mailObj.getPath())
      if result==None:
        mailToSend.append(mailObj)

  nSendable=len(mailToSend)

  if nSendable==0:
    sendableStr="None"
  elif nSendable==nInOutbox:
    sendableStr="All"
  else:
    sendableStr=str(nSendable)

  gLogger.add("%s sendable" % sendableStr)

  if nSendable==0:
    return None

  s=sendmail()
  if not s.setup():
    return None
  for mailObj in mailToSend:
    success=s.sendOne(mailObj)
    if success:
      mailObj.fileBecauseSent()
      gSwitchboard.delete("outbox",[mailObj.path])
      mailObj.unlink()
      gSwitchboard.counts("outbox")
  s.tearDown()
  if gPrefs.get("swdbFile")<>None:
    for mailObj in mailToSend:
      addrs=[mailObj.get("to",""), mailObj.get("cc",""),
             mailObj.get("bcc","")]
      removeAny(addrs,"")
      gAddressBook.snarf(addrs,direction="out")
  return None

# After a couple of embarrassing errors, it seems sensible
# to sanity-check outbound mail before sending. That's not
# as good as an automated test suite but that's not really
# practical for a Tkinter app. The class is meant to be
# self-contained and as much as possible written in a
# different style from the rest of Swallow so that an error
# in one place is unlikely to be duplicated in the other.
class outboundMailSanityChecker:
  maxHeaderLineLen=500
  maxHeaderTextLen=10240
  minBodyLen=4
  maxBodyLen=1024 * 1024 * 20
  headerTextREs= \
    [
    ("Subject",re.compile(r"^Subject: .+",re.MULTILINE)),
    ("From",re.compile(r"^From: .+",re.MULTILINE)),
    ("To/CC/BCC",re.compile(r"^(To|C[Cc]|B[Cc][Cc]): .+",re.MULTILINE)),
    ("Message-ID",re.compile(r"^Message-[Ii][Dd]: .+",re.MULTILINE)),
    ("Date",re.compile(r"^Date: .+",re.MULTILINE)),
    ]
  headerLineRE=re.compile(r"^([A-Z][-A-Za-z]{1,30}:\s+).+$")

  def check(self,path):
    headers=""
    logicalLines=[]
    f=open(path)
    while 1:
      l=f.readline()
      if l=="" or l==os.linesep:
        break
      headers+=l
      if l[0].isspace():
        assert len(logicalLines)>0,"First header starts with whitespace"
        logicalLines[-1]=logicalLines[-1].rstrip()+" "+l.lstrip()
      else:
        logicalLines.append(l)
      assert len(l)<=outboundMailSanityChecker.maxHeaderLineLen,'Header too long: "%s..."' % l[:30]
      #assert outboundMailSanityChecker.headerLineRE.search(l),'Header "%s" doesn\'t look like a header' % l[:-1]

    body=f.read()
    for re in outboundMailSanityChecker.headerTextREs:
      assert re[1].search(headers),"Header %s insane in %s" % (re[0],os.path.basename(path))

    for logicalLine in logicalLines:
      assert outboundMailSanityChecker.headerLineRE.search(logicalLine),'Header "%s" doesn\'t look like a header' % logicalLine[:-1]

    assert len(body)>outboundMailSanityChecker.minBodyLen,"Body too short in %s" % os.path.basename(path)
    assert len(body)<=outboundMailSanityChecker.maxBodyLen,"Body too long in %s" % os.path.basename(path)
    assert len(headers)<=outboundMailSanityChecker.maxHeaderTextLen,"Headers too long in %s" % os.path.basename(path)
    assert body.endswith(os.linesep),"Body doesn't end with line end in %s" % os.path.basename(path)
    f.seek(0)
    p=email.Parser.Parser()
    try:
      m=p.parse(f)
    except email.Errors.BoundaryError,details:
      raise AssertionError,details
    f.close()
    addrs=email.Utils.getaddresses(
      [m.get("to",""),
       m.get("cc",""),
       m.get("bcc",""),
       ])
    addrs=[ a[1] for a in addrs]
    while len(addrs)>1:
      t=addrs[0]
      addrs=addrs[1:]
      assert t not in addrs,'Mail would be sent to addrsss "%s" more than once' % t
    return None


class mousePointerHider:

  def __init__(self):
    self.cursorHidden=0
    self.nullCursorFileName=tempfile.mktemp()
    os.umask(0177) # octal
    f=open(self.nullCursorFileName,"w")
    f.write(kNullCursorData)
    f.close()
    return None

  def hideCursor(self,event):
    if not self.cursorHidden:
      event.widget.configure(cursor="@%s white" % self.nullCursorFileName)
      self.cursorHidden=1
    if hasattr(event.widget,"extraKeyCB"):
      event.widget.extraKeyCB(event)
    return ""

  def showCursor(self,event,ignoreHidden=0):
    if self.cursorHidden or ignoreHidden:
      event.widget.configure(cursor="xterm")
      self.cursorHidden=0
    if hasattr(event.widget,"extraMotionCB"):
      event.widget.extraMotionCB(event)
    return ""

  def makeWidgetHideMouseOnKey(self,widget):
    widget.bind("<Key>",self.hideCursor,"+")
    widget.bind("<Motion>",self.showCursor,"+")
    # Don't know why next line is necessary but it is
    widget.bind("<Enter>",self.showCursorEvenIfWeThinkHidden,"+")
    return None

  def showCursorEvenIfWeThinkHidden(self,event):
    self.showCursor(event,ignoreHidden=1)
    return None

  def explicitShowMouse(self,widget):
    widget.configure(cursor="xterm")
    self.cursorHidden=0
    return None

  def cleanUpForQuit(self):
    os.unlink(self.nullCursorFileName)
    return None


###XXX Should we use this again? Combine it with decoding headers?
# Sticking what really wants to be a function inside an object
# and sticking that in a global is ugly but I haven't thought
# of a less ugly way to do it.
#class makeDisplayableOnly:
#  def __init__(self):
#    self.translationTable="".join([self.displayableCharFromOrd(x) for x in range(256)])
#
#  def displayableCharFromOrd(self,val):
#    if val<>9 and (val<32 or val>176):
#      return "?"
#    else:
#      return chr(val)
#
#  def __call__(self,str):
#    return str.translate(self.translationTable)


class raiseWinCBO:
  def __init__(self,win):
    self.win=win
    return None

  def __call__(self):
    self.win.root.lift()
    return None


class genericCBO:
  def __init__(self,func,*positionArgs,**kwArgs):
    self.func=func
    self.positionArgs=positionArgs
    self.kwArgs=kwArgs
    return None

  # Ignore args called with so that it can be called by an
  # event binding which passes the event as an argument.
  def __call__(self,*morePositionArgs,**moreKwArgs):
    pa=self.positionArgs+morePositionArgs
    ka=copy.copy(self.kwArgs)
    ka.update(moreKwArgs) # New ones override, no reason
    return self.func(*pa,**ka)

# To make a Pmw.PromptDialog synchronous. Don't know if this is the best way to
# do it.
class getPasswordDialog:
  def dialogCB(self,buttonClicked):
    if buttonClicked<>"OK":
      self.retPW=None
      self.dialogDone=1
    else:
      self.retPW=self.passwordEntry.get()
      self.dialogDone=1
    self.dialog.destroy()
    return None

  def __init__(self):
    self.retPW=None
    self.dialogDone=0
    myFont=gPrefs.get("fontILike")
    self.dialog=Pmw.Dialog(buttons=("OK","Cancel"),
      command=self.dialogCB,
      defaultbutton=0,
      title="POP password")
    stuffParent=self.dialog.interior()
    l=Label(stuffParent,text="POP password",font=myFont)
    l.pack(side=LEFT,padx=5,pady=5)
    border=Frame(stuffParent,bg="black",borderwidth=1)
    self.passwordEntry=Entry(border,
      font=myFont,
      width=30,
      relief=FLAT,
      highlightthickness=0,
      show="*",
      bg="white")
    self.passwordEntry.pack(side=LEFT)
    border.pack(side=LEFT,padx=5,pady=5)

    self.dialog.component("buttonbox").component("Cancel").configure(
      font=myFont,activebackground=kTkDefaultBG)
    self.dialog.component("buttonbox").component("OK").configure(
      font=myFont,activebackground=kTkDefaultBG)
    self.passwordEntry.focus_force()
    return None

  def waitForIt(self):
    while not self.dialogDone:
      gTkRoot.update()
      time.sleep(kPasswordDialogSleepTime)
    return self.retPW


# Wrap buttons in a class so that we can control their size exactly

class myButton:
  def __init__(self,parent,text,cb,widthInChars=None,widthInDots=None):
    if widthInDots==None:
      if widthInChars==None:
        widthInChars=len(text)
      widthInDots=(widthInChars*kButtonDotsPerChar)+kButtonPaddingX
    self.frame=Frame(parent,height=kButtonHeight,width=widthInDots)
    self.frame.pack_propagate(0)
    self.button=Button(self.frame,font=gPrefs.get("fontILike"),command=cb,text=text,takefocus=0, \
      activebackground=kTkDefaultBG)
    self.button.pack(fill=BOTH,expand=1)
    return None

  def pack(self,*positionArgsTuple,**kwArgsDict):
    self.frame.pack(*positionArgsTuple,**kwArgsDict)
    return None

  # Send size configuration to the frame, everything else to the button
  # Expects widthInChars or widthInDots, not plain width
  def configure(self,*positionArgsTuple,**kwArgsDict):
    if kwArgsDict.has_key("widthInDots"):
      self.frame.configure(width=kwArgsDict["widthInDots"])
      del kwArgsDict["widthInDots"]
    elif kwArgsDict.has_key("widthInChars"):
      widthInDots=(kwArgsDict["widthInChars"]*kButtonDotsPerChar)+kButtonPaddingX
      self.frame.configure(width=widthInDots)
      del kwArgsDict["widthInChars"]
    if kwArgsDict.has_key("height"):
      self.frame.configure(height=kwArgsDict["height"])
      del kwArgsDict["height"]
    self.button.configure(*positionArgsTuple,**kwArgsDict)
    return None


#
# Text widget with scrollbars and undo
#

# First, a helper

class undoStack:
  # Might be smart to limit the size of the text in addition to the number of items.
  # Could add new item to last one when it made sense.

  def __init__(self):
    self.undoData=[]
    self.redoData=[]
    return None

  def push(self,item):
    self.redoData=[]
    self.undoData.append(item)
    if len(self.undoData)>kMaxUndos:
      self.undoData=self.undoData[1:]
    return None

  def popUndo(self):
    item=self.undoData.pop()
    self.redoData.append(item)
    return item

  def popRedo(self):
    item=self.redoData.pop()
    self.undoData.append(item)
    return item

  def undoIsEmpty(self):
    return len(self.undoData)==0

  def redoIsEmpty(self):
    return len(self.redoData)==0


# Much of this code inspired by:
# http://www.arts.cuhk.edu.hk/humftp/Philosophy/Papers/undo.txt
# My thanks to the author.

class scrolledTextWithUndo(Pmw.ScrolledText):
  def __init__(self,parent,**kwArgs):
    # Pmw magic, the "()" indiciates that we don't add any options, I think
    self.defineoptions(kwArgs,())
    Pmw.ScrolledText.__init__(self,parent)
    self.textComponent=self.component("text") # For speedup
    self.textComponent.bind("<Control-z>",self.undoCB)
    self.textComponent.bind("<Control-y>",self.redoCB)
    self.textComponent.bind("<Key>",self.keyCB)
    self.textComponent.bind("<Button-2>",self.beforePasteCB)
    self.textComponent.bind("<ButtonRelease-2>",self.afterPasteCB)
    self.undoStack=undoStack()
    self.initialiseoptions(scrolledTextWithUndo) # More Pmw magic
    return None

  def keyCB(self,event):
    if event.char<>"" and len(event.char)==1:
      ascVal=ord(event.char)
    else:
      ascVal=None

    ins=event.widget.index(INSERT) # Cursor position

    if event.char<>"" and (ascVal>=32 or ascVal==9): # Ordinary character
      self.undoStack.push(bunch(type="insert",where=ins,text=event.char,cursorAt=ins))
      return ""

    if event.keysym=="BackSpace" or ascVal==8: # Backspace and control-h
      # If there's a selection, it will all be deleted so
      # remember the whole thing, otherwise remember the
      # character on the left
      r=self.textComponent.tag_ranges(SEL)
      if len(r)==0:
        t=self.textComponent.get(event.widget.index(INSERT+" - 1 chars"))
      else:
        t=self.textComponent.get(r[0],r[1])
      self.undoStack.push(bunch(type="delete",where=ins,text=t,cursorAt=ins))
      return ""

    if ascVal==15: # Control-o, insert newline to right
      self.undoStack.push(bunch(type="insert",where=ins,text="\n",cursorAt=ins))
      return ""

    atEndOfText=self.textComponent.compare(ins,"==",END+" - 1 chars")

    if ascVal==4 and not atEndOfText: # Control-d, kills to right
      t=self.textComponent.get(ins)
      self.undoStack.push(bunch(type="delete",where=ins,text=t,cursorAt=ins))
      return ""

    lineNo=int(float(ins))
    atEndOfLine=self.textComponent.compare(ins,"==","%i.end" % lineNo)

    # Control-k, kills to eol or if at eol, kills newline
    if ascVal==11 and not atEndOfText:
      if atEndOfLine:
        t="\n"
      else:
        t=self.textComponent.get(ins,"%i.end" % lineNo)
      self.undoStack.push(bunch(type="delete",where=ins,text=t,cursorAt=ins))
      return ""

    # Control-t, transpose, if not at eol transposes left and right, if at
    # eol, transposes left and one before that
    # Could handle this as a delete plus an insert but that seems to be
    # more work
    if ascVal==20:
      if atEndOfLine:
        t=self.textComponent.get(ins+" - 1 chars")
        self.undoStack.push(bunch(type="move",where=ins,to=ins+" - 2 chars",text=t, cursorAt=ins))
      else:
        t=self.textComponent.get(ins)
        self.undoStack.push(bunch(type="move",where=ins,to=ins+" - 1 chars",text=t,cursorAt=ins))
      return ""
    return "" # Continue processing keystroke
  
  # If we're going to paste something that contains returns, replce
  # them with newlines.
  def beforePasteCB(self,eventIgnored):
    try:
      t=self.textComponent.selection_get()
    except TclError:
      return ""
      
    if "\r" in t:
      t=t.replace("\r","\n")
      self.textComponent.selection_own()
      self.textComponent.selection_handle(clipboardCBO(t))
    return ""

  def afterPasteCB(self,event):
    # What text position was the click over?
    clickIndex=self.textComponent.index("@"+repr(event.x)+","+repr(event.y))
    # We send seletion_get() to the text component because we need to send to
    # to some Tkinter widet and the text component is handy. selection_get()
    # gets whatever's on the X clipboard, not just from the widget it's sent
    # to.

    # After the paste, if the insertion point isn't visible, scroll
    # so that it is. I think after_idle() ought to work here but it
    # doesn't.
    gTkRoot.after(50,self.textComponent.see,(INSERT,))

    # If there isn't a selection, an exception is raised
    try:
      s=self.textComponent.selection_get()
    except TclError: # no selection
      return None
    ins=event.widget.index(INSERT) # Cursor position
    # Don't know why + 1 chars is necessary
    self.undoStack.push(bunch(type="insert",where=clickIndex+"+ 1 chars",text=s,cursorAt=ins))
    return ""

  def undoCB(self,eventIgnored):
    self.undo()
    return None

  def redoCB(self,eventIgnored):
    self.redo()
    return None

  def undo(self):
    if self.undoStack.undoIsEmpty():
      return None
    inGroup=0
    while 1:
      item=self.undoStack.popUndo()
      if item.type=="insert":
        self.textComponent.see(item.where)
        self.textComponent.delete(item.where,item.where+" + %i chars" % len(item.text))
      elif item.type=="delete":
        self.textComponent.see(item.where)
        self.textComponent.insert(item.where,item.text)
      elif item.type=="move":
        self.textComponent.see(item.where)
        t=self.textComponent.get(item.to)
        self.textComponent.delete(item.to,item.to+" + %i chars" % len(t))
        self.textComponent.insert(item.where,t)
      elif item.type=="group":
        inGroup=not inGroup
      else:
        gLogger.add("Unrecognized type: %s" % item.type,kError)

      if hasattr(item,"cursorAt"):
        self.textComponent.mark_set(INSERT,item.cursorAt)

      if not inGroup:
        break
    return None

  def redo(self):
    if self.undoStack.redoIsEmpty():
      return None
    inGroup=0
    while 1:
      item=self.undoStack.popRedo()
      if item.type=="insert":
        self.textComponent.see(item.where)
        self.textComponent.insert(item.where,item.text)
      elif item.type=="delete":
        self.textComponent.see(item.where)
        self.textComponent.delete(item.where,item.where+" + %i chars" % len(item.text))
      elif item.type=="move":
        self.textComponent.see(item.where)
        t=self.textComponent.get(item.where)
        if t=="\n": # Tkinter's inconsistent line-end handling requires a special-case
          t=self.textComponent.get(item.where+" - 1 chars")
          self.textComponent.delete(item.where+" - 1 chars",item.where+" + %i chars" % (len(t)-1))
          self.textComponent.insert(item.to+" + 1 chars",t)
        else:
          self.textComponent.delete(item.where,item.where+" + %i chars" % len(t))
          self.textComponent.insert(item.to,t)
      elif item.type=="group":
        inGroup=not inGroup
      else:
        gLogger.add("Unrecognized type: %s" % item.type,kError)

      if not inGroup:
        break
    self.textComponent.mark_set(INSERT,item.cursorAt)
    return None

  def markUndoGroup(self):
    self.undoStack.push(bunch(type="group"))

  # Meant to be called from outside
  def deleteText(self,rangeStart,rangeEnd):
    t=self.textComponent.get(rangeStart,rangeEnd)
    self.textComponent.delete(rangeStart,rangeEnd)
    self.undoStack.push(bunch(type="delete",where=rangeStart,text=t, \
      cursorAt=self.textComponent.index(INSERT)))
    return None

  # Meant to be called from outside
  def insertText(self,where,text):
    self.undoStack.push(bunch(type="insert",where=where,text=text, \
      cursorAt=self.textComponent.index(INSERT)))
    self.textComponent.insert(where,text)
    return None


# Our own window for tracebacks from Tkinter

class callbackErrorHandler:
  def __init__(self):
    self.shown=0
    self.root=None
    self.queuedMsgs=[]
    return None

  def show(self):
    self.root=Toplevel()
    self.root.protocol("WM_DELETE_WINDOW", self.closeWinCB)
    self.root.title("Error in callback")
    self.createWidgets()
    self.root.bind("<Key>",self.keyCB)
    self.shown=1
    return None

  def newMsg(self,msg):
    gLogger.add("Traceback from Pmw")
    for line in "\n".split(msg):
      if line<>"":
        gLogger.add(line)
    gLogger.add("End of traceback from Pmw")
    if not self.shown:
      self.show()
      self.textArea.component("text").configure(state=NORMAL)
      self.textArea.settext(msg)
      self.textArea.component("text").configure(state=DISABLED)
      self.pendingLabel.configure(text="0 more")
    else:
      self.queuedMsgs.append(msg)
      self.pendingLabel.configure(text="%i more" % len(self.queuedMsgs))
    return None

  # Pmw wants a write() method to send tracebacks to.
  def write(self,msg):
    self.newMsg(msg)
    return None

  def keyCB(self,event):
    if event.char=="w":
      self.closeWinCB()
    elif event.keysym=="Up":
      self.textArea.yview("scroll","-1","units")
    elif event.keysym=="Down":
      self.textArea.yview("scroll","1","units")
    elif event.keysym=="Left":
      self.textArea.xview("scroll","-1","units")
    elif event.keysym=="Right":
      self.textArea.xview("scroll","1","units")
    elif event.char==" ":
      self.textArea.yview("scroll","1","pages")
    elif event.char=="u":
      self.textArea.yview("scroll","-1","pages")
    elif event.char=="-":
      self.textArea.yview("scroll","-1","pages")
    elif event.char=="w":
      self.closeWinCB()
    # No further processing of this event
    return "break"

  def closeWinCB(self):
    self.root.destroy()
    self.root=None
    self.shown=0
    self.queuedMsgs=[]
    return None

  def createWidgets(self):
    t=Text(self.root,font=gPrefs.get("boldFontILike"),relief=FLAT,height=4)
    t.insert(END,"I'm sorry. An error has occured.\n")
    t.insert(END,"The traceback is below. It may be useful in finding the problem.\n")
    t.insert(END,"The same information is in Swallow's log.\n")
    t.insert(END,"It may be possible to continue using Swallow.")
    t.configure(state=DISABLED)
    t.pack(side=TOP,fill=X,expand=1)
    textArea=Pmw.ScrolledText(self.root,
      hscrollmode="static",
      vscrollmode="static",
      text_wrap="none",
      text_bg="white",
      text_font=gPrefs.get("fontILike"),
      horizscrollbar_activebackground=kTkDefaultBG,
      vertscrollbar_activebackground=kTkDefaultBG,
      vertscrollbar_borderwidth=1,
      vertscrollbar_width=12,
      horizscrollbar_borderwidth=1,
      horizscrollbar_width=12,
      )
    textArea.pack(side=TOP,fill=BOTH,expand=1)
    self.textArea=textArea
    f=Frame(self.root)
    spacer=Frame(f,width=5)
    spacer.pack(side=LEFT)
    nextButton=myButton(f,"Next error",self.nextButtonCB)
    nextButton.pack(side=LEFT)
    spacer=Frame(f,width=5)
    spacer.pack(side=LEFT)
    pendingLabel=Label(f,font=gPrefs.get("fontILike"))
    pendingLabel.pack(side=LEFT)
    self.pendingLabel=pendingLabel
    f.pack(side=TOP,fill=X,expand=1)
    return None

  def nextButtonCB(self):
    if len(self.queuedMsgs)==0:
      return None

    self.textArea.component("text").configure(state=NORMAL)
    self.textArea.settext(self.queuedMsgs[0])
    self.textArea.component("text").configure(state=DISABLED)
    self.queuedMsgs=self.queuedMsgs[1:]
    if len(self.queuedMsgs)==0:
      self.pendingLabel.configure(text="0 more")
    else:
      self.pendingLabel.configure(text="%i more" % len(self.queuedMsgs))
    return None


# Used by menus in mailbox and viewer windows

class moveMailToOtherMailbox:
  def __init__(self,boxName,moveOrCopy,refToParent):
    assert moveOrCopy in ("move","copy")
    self.boxName=boxName
    self.refToParent=refToParent
    self.moveOrCopy=moveOrCopy
    return None

  def __call__(self):
    path=self.refToParent.getPathForMove()
    if path==None:
      gLogger.add("Can't move, nothing selected")
      return None

    boxAlreadyIn=getMailboxFromPath(path)
    if self.boxName==boxAlreadyIn:
      gLogger.add("Can't move a message to the mailbox it's in")
      return None # XXX error message?

    if self.moveOrCopy=="move" and gSwitchboard.getEditorByPath(path)<>None:
      gLogger.add("Can't move that message, it's open in an editor")
      return None

    myMailObj=mailObj(path)
    if self.boxName=="(Default per prefs)":
      filterFunc=gPrefs.get("getSaveFoldersInbound")
      newMBs=filterFunc(myMailObj.getHeaders())
    elif self.boxName=="(Alternate per prefs)":
      filterFunc=gPrefs.get("getSaveFoldersInboundAlt")
      newMBs=filterFunc(myMailObj.getHeaders())
    else:
      newMBs=[self.boxName]

    if type(newMBs)==types.StringType:
      newMBs=[newMBs]

    success=1
    for box in newMBs:

      if box==boxAlreadyIn:
        continue

      if not gAllMailboxes.isMaildir(box):
        gAllMailboxes.newMaildir(box)

      newMaildir=gAllMailboxes.getMaildir(box)
      result=newMaildir.copyToThisMaildir(myMailObj)

      if result==None:
        success=0
        gLogger.add("Error moving mail to %s" % box,kError)
      else:
        gSwitchboard.add(box,[result])

        if self.moveOrCopy=="move":
          gLogger.add("Mail moved to %s" % box)
        else:
          gLogger.add("Mail copied to %s" % box)

    if boxAlreadyIn not in newMBs and self.moveOrCopy=="move" and success:
      gSwitchboard.delete(boxAlreadyIn,[myMailObj.path])
      myMailObj.unlink()
      gSwitchboard.counts(boxAlreadyIn)
      gSwitchboard.counts(box)
    return None


class statusIndicator:
  def __init__(self,parent):
    self.parent=parent
    self.createWidgets()
    self.logBlinkAfterPending=None
    self.logBlinkLast=0
    #self.statusShown=gSwitchboard.getLogStatus()
    return None

  def createWidgets(self):
    myFont=gPrefs.get("fontILike")
    self.frame=Frame(self.parent.getParentWidget())
    border=Frame(self.frame,borderwidth=1,bg="black")
    self.checkLabel=Label(border,text="Checker",font=myFont)
    self.checkLabel.pack(side=LEFT)
    self.checkLabel.bind("<Button-1>",self.toggleCheckRegularly)
    logStatus=gSwitchboard.getLogStatus()
    if logStatus==kNormal:
      color=kStatusLogNormalColor
    elif logStatus==kWarning:
      color=kStatusLogWarningColor
    elif logStatus==kError:
      color=kStatusLogErrorColor
    else:
      assert 1==0, "Impossible log status: %s" % logStatus
    self.logLabel=Label(border,text="Log",font=myFont,fg=color)
    self.logLabel.pack(side=LEFT)
    self.logLabel.bind("<Button-1>",self.openLogWindow)
    spacer=Frame(self.frame,width=4,height=4)
    spacer.pack(side=RIGHT)
    border.pack(side=LEFT)
    return None

  def checkOn(self):
    self.checkLabel.configure(fg=kStatusCheckOnColor)
    return None

  def checkOff(self):
    self.checkLabel.configure(fg=kStatusCheckOffColor)
    return None

  def openLogWindow(self,eventIgnored):
    gSwitchboard.openLogWindow()
    return None

  def toggleCheckRegularly(self,eventIgnored):
    self.parent.toggleCheckRegularly()
    return None

  ### XXX clean this up
  def logBlink(self):
    status=gSwitchboard.getLogStatus()
    if status==kError:
      return None

    oldColor=self.logLabel.cget("fg").lower()
    if oldColor==kStatusLogErrorColor.lower():
      return None

    if time.time()-self.logBlinkLast<=kStatusLogBlinkMinDelay/1000.0:
      return None

    self.logBlinkLast=time.time()
    if oldColor==kStatusLogNormalColor.lower() or oldColor==kStatusLogWarningColor.lower():
      newColor=kStatusLogBlinkColor
      if self.logBlinkAfterPending<>None:
        gTkRoot.after_cancel(self.logBlinkAfterPending)
      self.logBlinkAfterPending=gTkRoot.after(kStatusLogNormalDelay,
        self.statusLogBackToNormal)
    elif oldColor==kStatusLogBlinkColor:
      newColor=kStatusLogNormalColor
    else:
      assert 1==0,"status has impossible color: %s" % oldColor
    self.logLabel.configure(fg=newColor)
    return None

  def statusLogBackToNormal(self):
    if self.logBlinkAfterPending<>None:
      gTkRoot.after_cancel(self.logBlinkAfterPending)
    self.logBlinkAfterPending=None
    # If log was being written as Swallow is quitting,
    # there may not be a widget by the time we're called
    # so ignore any error that would come from that.
    status=gSwitchboard.getLogStatus()
    if status==kWarning:
      newColor=kStatusLogWarningColor
    else:
      newColor=kStatusLogNormalColor

    try:
      self.logLabel.configure(fg=newColor)
    except TclError:
       pass
    gTkRoot.update_idletasks()
    return None

  def logErrorWarningSet(self):
    severity=gSwitchboard.getLogStatus()
    assert severity in (kError,kWarning), "Impossible severity %s" % severity
    if self.logBlinkAfterPending<>None:
      gTkRoot.after_cancel(self.logBlinkAfterPending)
      self.logBlinkAfterPending=None
    if severity==kWarning:
      color=kStatusLogWarningColor
    else:
      color=kStatusLogErrorColor
    self.logLabel.configure(fg=color)
    gTkRoot.update_idletasks()
    return None

  def logErrorReset(self):
    self.logLabel.configure(fg=kStatusLogNormalColor)
    return None

  def pack(self,*positionArgsTuple,**kwArgsDict):
    self.frame.pack(*positionArgsTuple,**kwArgsDict)
    return None


class mainMenubar:
  def __init__(self,parent):
    self.parent=parent
    self.locVar=StringVar()
    self.locVar.set(gPrefs.getLocation())
    self.checkRegularlyVar=IntVar()
    self.checkRegularlyVar.set(0)
    self.createWidgets()
    self.newItemsForWindowMenu()
    self.setCheckingState(gChecker.getCheckingRegularly())
    return None

  def createWidgets(self):
    self.myFrame=Frame(self.parent)
    self.myFrame.pack(side=LEFT,fill=X,expand=1)
    myFont=gPrefs.get("fontILike")
    mbar=Pmw.MenuBar(self.myFrame,padx=5,hotkeys=0)
    self.mbar=mbar
    mbar.pack(side=LEFT,fill=X)
    mbar.addmenu("File","",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("File","command",command=gSwitchboard.newMail,label="New mail",font=myFont)
    mbar.addmenuitem("File","command",command=emptyTrash,label="Empty trash",font=myFont)
    mbar.addmenuitem("File","command",command=self.newMailbox,label="New mailbox",font=myFont)
    #mbar.addmenuitem("File","command",command=self.reScan,label="Re-scan all",font=myFont)
    mbar.addmenuitem("File","separator")

    #mbar.addmenuitem("File","command",command=self.logError,label="Log error",font=myFont)
    #mbar.addmenuitem("File","command",command=self.logWarning,label="Log warning",font=myFont)
    #mbar.addmenuitem("File","separator")

    mbar.addmenuitem("File","command",command=quit,label="Quit",font=myFont)
    mbar.component("File-menu").configure(activebackground=kTkDefaultBG)##,activeborderwidth=1)

    mbar.addmenu("Get&Send","",font=myFont,activebackground=kTkDefaultBG)
    mbar.component("Get&Send-menu").configure(activebackground=kTkDefaultBG)
    mbar.addmenuitem("Get&Send","command",command=sendReady,label="Send ready",font=myFont)
    mbar.addmenuitem("Get&Send","separator")
    mbar.addmenuitem("Get&Send","command",command=self.checkOnce,label="Check now",font=myFont)
    mbar.addmenuitem("Get&Send","checkbutton",
      command=self.checkRegularlyToggled,
      variable=self.checkRegularlyVar,
      label="Check regularly",
      font=myFont,
      selectcolor="blue",
      )
    self.mbar.addmenuitem("Get&Send","separator")
    self.itemsAboveLocationsInMenu=5
    self.redoLocationItems()

    mbar.addmenu("Windows","",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","command",command=self.openAboutBox,label="About",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","command",command=self.openKeysWin,label="Key commands",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","separator")
    mbar.addmenuitem("Windows","command",command=self.openAddressBrowser,label="Addressbook",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","command",command=self.openSearchWin,label="Search mail",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","command",command=self.openLogWindow,label="Log",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","separator")
    # Next is trashed and replaced in newItemsForWindowMenu()
    self.mbar.addmenuitem("Windows","command",label="Open windows",font=gPrefs.get("boldFontILike"),state=DISABLED)
    self.standardWindowMenuLength=7
    self.statusIndicator=statusIndicator(self)
    self.statusIndicator.pack(side=RIGHT)
    return None

  def getParentWidget(self):
    return self.myFrame

  def beginAddingMenu(self):
    self.winButtonPackInfo=self.mbar.component("Windows-button").pack_info()
    return None

  def doneAddingMenu(self):
    self.mbar.component("Windows-button").pack(self.winButtonPackInfo)
    return None

  #def logError(self):
  #  gLogger.add("Test error",kError)
  #  return None

  def newMailbox(self):
    newMailbox()
    return None

  #def reScan(all):
  #  for mailboxName in gAllMailboxes.getNames():
  #    print mailboxName
  #    gSwitchboard.redisplayFromScratch(mailboxName,reReadDisk=1)
  #    gSwitchboard.counts(mailboxName)

  def locMenuCB(self):
    gPrefs.setLocation(self.locVar.get())
    gSwitchboard.setLocationMenus(self.locVar.get())
    return None

  def redoLocationItems(self):
    myFont=gPrefs.get("fontILike")
    # Bug workaround: deletemenuitems always deletes at least one item even
    # if the range is outside the items in the menu. So just tell it to delete
    # at least one (which we know is a separator) and add it back.
    self.mbar.deletemenuitems("Get&Send",self.itemsAboveLocationsInMenu-1,999)
    self.mbar.addmenuitem("Get&Send","separator")
    locations=gPrefs.getLocations()
    if len(locations)==1:
      self.mbar.addmenuitem("Get&Send","command",label="(Only one location)",font=myFont,state=DISABLED)
    else:
      for location in locations:
        self.mbar.addmenuitem("Get&Send","radiobutton",variable=self.locVar,
          label=location,font=myFont,selectcolor="blue",command=self.locMenuCB)
    return None

  def setCheckingState(self,val):
    self.checkRegularlyVar.set(val)
    if self.checkRegularlyVar.get():
      self.statusIndicator.checkOn()
    else:
      self.statusIndicator.checkOff()
    return None

  def toggleCheckRegularly(self):
    val=self.checkRegularlyVar.get()
    val=not val
    self.checkRegularlyVar.set(val)
    self.checkRegularlyToggled()
    return None

  ### Set the state then call this ### XXX
  def checkRegularlyToggled(self):
    # Tkinter toggles the state
    val=self.checkRegularlyVar.get()
    gChecker.setCheckingRegularly(val)
    gSwitchboard.checkRegularlyChanged(val)
    return None

  def checkOnce(self):
    gTkRoot.update() ## Helps??
    gChecker.checkOnce(invokedByThread=0)
    return None

  def newItemsForWindowMenu(self,items=None):
    myFont=gPrefs.get("fontILike")
    self.mbar.deletemenuitems("Windows",self.standardWindowMenuLength,999)
    ###self.mbar.addmenuitem("Window","separator")
    self.mbar.addmenuitem("Windows","command",
      label="Open windows",
      font=gPrefs.get("boldFontILike"),
      state=DISABLED,
      )
    self.mbar.addmenuitem("Windows","command",
      label="Mailboxes",
      font=myFont,
      command=self.raiseRootWin,
      activebackground=kTkDefaultBG,
      )
    if items<>None:
      for win in items:
        self.mbar.addmenuitem("Windows","command",
          label=win.root.title(),
          command=raiseWinCBO(win),
          font=myFont,
          activebackground=kTkDefaultBG,
          )
    return None

  def openAddressBrowser(self):
    gTkRoot.update()
    gSwitchboard.openAddressBrowser()
    return None

  def openAboutBox(self):
    gTkRoot.update()
    gSwitchboard.openAboutBox()
    return None

  def openSearchWin(self):
    gTkRoot.update()
    gSwitchboard.openSearchWin()
    return None

  def openLogWindow(self):
    gTkRoot.update()
    gSwitchboard.openLogWindow()
    return None

  def openKeysWin(self):
    gTkRoot.update()
    gSwitchboard.openKeysWin()
    return None

  def raiseRootWin(self):
    gSwitchboard.getOverview().root.lift()
    return None

  def pack(self,*positionArgsTuple,**kwArgsDict):
    self.mbar.pack(*positionArgsTuple,**kwArgsDict)
    return None


# Since we're likely to deliver more than one message per invocation,
# we need to add a "_" and a sequence number after the pid and maintain
# the sequence number state.
class newMaildirFileName:
  seq=0 # Class level data
  newMaildirFileNameLock=threading.Lock()

  def __call__(self):
    newMaildirFileName.newMaildirFileNameLock.acquire() # Will block
    name="%i.%i_%i.%s" % (int(time.time()),os.getpid(),self.seq, \
      gPrefs.get("fqdn"))
    newMaildirFileName.seq=newMaildirFileName.seq+1
    newMaildirFileName.newMaildirFileNameLock.release()
    return name


class allMailboxes:
  def __init__(self):
    self.readMailboxes()
    return None

  def readMailboxes(self):
    maildirPath=myPathJoin(gPrefs.get("maildirParent"),"Maildir")
    mailboxes={}
    mailboxes["inbox"]=maildirPath
    userMailboxDir=gPrefs.get("userMailboxDirParent")
    for item in os.listdir(userMailboxDir):
      if item=="inbox":
        raise ValueError('may not have user mailbox "inbox" in "%s"'  \
          "that's the name for your Maildir." % userMailboxDir)
      # No particular reason to exclude mailboxes beginning with ".", just sounds
      # like a good idea.
      fullPath=myPathJoin(userMailboxDir,item)
      if os.path.isdir(fullPath) and item[0]<>".":
        mailboxes[item]=fullPath
    if not mailboxes.has_key("outbox"):
      mailboxes["outbox"]=myPathJoin(userMailboxDir,"outbox")
    if not mailboxes.has_key("trash"):
      mailboxes["trash"]=myPathJoin(userMailboxDir,"trash")
    # For now we don't bother to save the maildir objects, just
    # create them temporarily so that the structure and permissions
    # get checked.
    #for dir in mailboxes.values():
    #  md=maildir(dir)
    self.mailboxes=mailboxes
    for boxName in mailboxes.keys():
      md=self.getMaildir(boxName)
      md.checkMaildir() # Check at startup
    return None

  def getMaildir(self,name,progressFunc=None):
    # If we're ever tempted to cache these, we'll need
    # an cache-invalidation protocol like the one used
    # for the overviews.
    if not self.mailboxes.has_key(name):
      self.newMaildir(name)
    path=self.mailboxes[name]
    #doSummary=0
    if name in kSummaryNotTheseMailboxes:
      doSummary=0
    else:
      doSummary=1
    return maildir(path,doSummary,progressFunc)

  def isMaildir(self,name):
    return self.mailboxes.has_key(name)

  def newMaildir(self,name):
    if self.mailboxes.has_key(name):
      raise ValueError("%s is already a mailbox name",name)
    newPath=myPathJoin(gPrefs.get("userMailboxDirParent"),name)
    self.mailboxes[name]=newPath
    os.chdir(gPrefs.get("userMailboxDirParent"))
    self.getMaildir(name) # automatically creates stuff
    gSwitchboard.informNewMailbox(name)
    return None

  def getNames(self):
    names=self.mailboxes.keys()
    names=gPrefs.get("mailboxNamesSortFunc")(names)
    return names

  def cleanupTmps(self):
    #for dir in self.mailboxes.values():
    #  m=maildir(dir)
    for boxName in self.mailboxes.keys():
      md=self.getMaildir(boxName)
      md.cleanupTmp()
    return None

  def __len__(self):
    return len(self.mailboxes.keys())


class mailObj:
  def __init__(self,path,summary=None,loadTheFileNow=None):
    self.path=path
    self.summary=summary

    if loadTheFileNow:
      f=open(path,"rb")
      t=f.read()
      f.close()
      self.file=cStringIO.StringIO(t)
      self.keepFile=1
    else:
      self.file=None
      self.keepFile=None

    self.headersLoaded=0
    self.score=None
    self.dateAsSecs=None
    self.receivedDateAsSecs=None
    self.overviewTextList=None
    return None

  def getFile(self):
    if self.file==None:
      self.file=open(self.path,"rb")
    else:
      self.file.seek(0)
    return None

  def doneWithFile(self):
    if self.keepFile==None:
      self.file.close()
      self.file=None
    return None

  def getText(self):
    try:
      return self.file.getvalue()
    except AttributeError:
      pass
    self.getFile()
    t=self.file.read()
    self.doneWithFile()
    return t

  def getBodyText(self):
    self.getFile()
    while 1:
      l=self.file.readline()
      if l==kFileLineSep or l=="":
        break
    t=self.file.read()
    self.doneWithFile()
    return t

  def getPath(self):
    return self.path

  def unlink(self):
    os.unlink(self.path)
    self.path=None
    return None

  def newPath(self,newPath):
    self.path=newPath

  def whichOfThreeDirs(self):
    return getSubdirFromPath(self.path)

  def sendable(self):
    return not(self.isDraft()) and self.hasAddresses()

  def isDraft(self):
    return self.whichOfThreeDirs()=="new"

  # xxx sum
  def hasAddresses(self):
    toHeader=self.get("to")
    if toHeader==None or toHeader=="undisclosed-recipients:;" and \
      self.get("cc")==None and self.get("bcc")==None:
      return 0
    else:
      return 1

  def moveToCur(self):
    oldPath=self.path
    self.path=moveToCur(oldPath)
    return None

  def moveToNew(self):
    oldPath=self.path
    self.path=moveToNew(oldPath)
    return None

  def fileBecauseSent(self):
    filterFunc=gPrefs.get("getSaveFoldersOutbound")

    if filterFunc==None:
      gLogger.add("Done sending, not saving sent mail")
      return None

    newMBs=filterFunc(self.getHeaders())
    if type(newMBs)==types.StringType:
      newMBs=[newMBs]

    gLogger.add("Done sending, saving message to %s" % ", ".join(newMBs))
    for box in newMBs:
      if box=="outbox":
        gLogger.add("Outbox in list of places to put sent mail -- doesn't make sense, skipped")
      newMaildir=gAllMailboxes.getMaildir(box)
      newPath=newMaildir.copyToThisMaildir(self)
      gSwitchboard.add(box,[newPath])
    return None

  def loadHeaders(self):
    if self.headersLoaded:
      return None

    self.getFile()
    p=email.Parser.HeaderParser()
    try:
      m=p.parse(self.file)
    except email.Errors.HeaderParseError:
      m=email.Message.Message() # Can't parse at all, use dummy message
    else:
      m.set_payload("") # Don't care about payloads here
    self.doneWithFile()
    self.headers=m
    self.headersLoaded=1
    return None

  def getHeaders(self):
    self.loadHeaders()
    return self.headers

  def getHeadersText(self):
    headers=""
    self.getFile()
    while 1:
      l=self.file.readline()
      if l=="" or l==kFileLineSep:
        break
      headers=headers+l
    self.doneWithFile()
    return headers

  # xxx sum
  # Allow a default. Caller may want to know if the header doesn't exist (default to return None
  # in that case) but call may also not care, could supply default of "".
  def get(self,h,default=None):

    if self.summary and self.summary["headers"].has_key(h):
      val=self.summary["headers"][h]
      if val==None and default<>None:
        return default
      else:
        return val
      #return self.summary["headers"].get(h,default)

    #print "disk for",h
    if not self.headersLoaded:
      self.loadHeaders()
    return self.headers.get(h.lower(),default)

  def __getitem__(self,h):
    return self.get(h)

  def getHeaderKeys(self):
    if not self.headersLoaded:
      self.loadHeaders()
    return self.headers.keys()

  # XXX not used
  def getSenderAddr(self):
    fromHeader=self.get("from")
    if fromHeader==None:
      return None
    return email.Utils.parseaddr(fromHeader)[1]

  def getRecipients(self):
    recipients=[]
    for hName in "to","cc","bcc":
      recipients+=email.Utils.getaddresses([self.get(hName,"")])
    for count in range(len(recipients)):
      recipients[count]=recipients[count][1]
    removeAny(recipients,"undisclosed-recipients:;")
    return recipients

  def getToAndCCRecipients(self):
    recipients=[]
    for hName in "to","cc":
      recipients+=email.Utils.getaddresses([self.get(hName,"")])
    for count in range(len(recipients)):
      recipients[count]=recipients[count][1]
    removeAny(recipients,"undisclosed-recipients:;")
    return recipients

  def getBCCRecipients(self):
    recipients=email.Utils.getaddresses([self.get("bcc","")])
    for count in range(len(recipients)):
      recipients[count]=recipients[count][1]
    return recipients

  def isNewText(self):
    if self.getMailboxName()=="outbox":
      if self.isDraft():
        return "D"
      elif not self.hasAddresses():
        return "X"
      else:
        return ""
    else:
      if self.whichOfThreeDirs()=="new" and self.getMailboxName()<>"outbox":
        return "N"
      else:
        return ""

  def getMailboxName(self):
    return getMailboxFromPath(self.path)

  def getReceivedDateAsSecs(self):
    if self.summary:
      return self.summary["receivedDateAsSecs"]

    if self.receivedDateAsSecs<>None:
      return self.receivedDateAsSecs

    receiveds=self.headers.get_all("received")
    if receiveds==None:
      self.receivedDateAsSecs=0
      return 0
    else:
      for r in receiveds:
        ans=kMyReceivedRE.search(r)
        if ans<>None:
          break
      else:
        self.receivedDateAsSecs=0
        return 0
      dateStr=ans.group(1)

    try:
      t=email.Utils.parsedate_tz(dateStr)
    except (ValueError,OverflowError,TypeError,AttributeError): # Lots of possiblities!
      self.receivedDateAsSecs=0
      return 0

    if t==None:
      # Some mailers don't include the day name and parsedate_tz() wants one
      t=email.Utils.parsedate_tz("Dummyday, "+dateStr)
    
    if t==None:
      self.receivedDateAsSecs=0
      return 0

    tzSecs=t[9]
    if tzSecs==None:
      tzSecs=0

    try:
      self.receivedDateAsSecs=time.mktime(t[:8]+(-1,))-tzSecs
    except (OverflowError,ValueError):
      self.receivedDateAsSecs=0
    return self.receivedDateAsSecs


  def getDateAsSecs(self):
    if self.summary:
      return self.summary["dateAsSecs"]

    if self.dateAsSecs<>None:
      return self.dateAsSecs

    if not self.headersLoaded:
      self.loadHeaders()

    dateStr=self.get("date")
    if dateStr==None:
      #receivedRE=re.compile("^from .* by .* \(%s\); (.*)" % kMyName)
      receiveds=self.headers.get_all("received")
      if receiveds==None:
        self.dateAsSecs=0
        return 0
      else:
        for r in receiveds:
          ans=kMyReceivedRE.search(r)
          if ans<>None:
            break
        else:
          self.dateAsSecs=0
          return 0
        dateStr=ans.group(1)
    try:
      t=email.Utils.parsedate_tz(dateStr)
    except (ValueError,OverflowError,TypeError,AttributeError): # Lots of possiblities!
      self.dateAsSecs=0
      return 0

    if t==None:
      t=email.Utils.parsedate_tz("Dummyday, "+dateStr)

    if t==None:
      self.dateAsSecs=0
      return 0

    tzSecs=t[9]
    if tzSecs==None:
      tzSecs=0

    try:
      self.dateAsSecs=time.mktime(t[:8]+(-1,))-tzSecs
    except (OverflowError,ValueError):
      self.dateAsSecs=0
    return self.dateAsSecs

  def __repr__(self):
    return "mailObject, file: "+self.path

  def getScore(self):
    if self.summary:
      return self.summary["score"]

    if self.score==None:
      s=gPrefs.get("calculateScore")(self.getHeaders())
      if s==None: # Default score calculator is nullCB and it returns None
        self.score=0
      else:
        self.score=s
    return self.score

  # XXX Don't know about deHTML here: stripping tags first
  # XXX seems often to foul up uuencoded parts, stripping
  # XXX them later fouls up line numbers. Doubt it's a big
  # XXX deal.
  def getTextHunksForUserUU(self,deHTML):
    commentaryLineNumbers=[]
    self.getFile()
    parts=getUUParts(self.file)
    self.doneWithFile()
    thisLineNo=0
    ordinaryTextAsFile=cStringIO.StringIO()
    decodedPartAsFile=cStringIO.StringIO(parts[0].decodedPart)
    while 1:
      l=decodedPartAsFile.readline()
      if l=="":
        break
      else:
        ordinaryTextAsFile.write(l.rstrip()+kFileLineSep)
        lastEndsSep=l.endswith(kFileLineSep)
        thisLineNo+=1
    if not lastEndsSep:
      ordinaryTextAsFile.write(kFileLineSep)
    for part in parts[1:]:
      thisLineNo+=1
      commentaryLineNumbers.append(bunch(lineNumber=thisLineNo,kind="decodeCommentary"))
      ordinaryTextAsFile.write("UUencoded part, name: %s" % part.name)
      if part.mainType=="text":
        partAsFile=cStringIO.StringIO(part.decodedPart)
        while 1:
          l=partAsFile.readline()
          if l=="":
            break
          else:
            lastEndsSep=l.endswith(kFileLineSep)
            thisLineNo+=1
            ordinaryTextAsFile.write(l.rstrip()+kFileLineSep)
      else: # Not text
        thisLineNo+=1
        ordinaryTextAsFile.write("(Not shown)"+kFileLineSep)
    textForViewing=ordinaryTextAsFile.getvalue()
    ordinaryTextAsFile.close()
    return commentaryLineNumbers,textForViewing,parts

  def getTextHunksForUserMIME(self,deHTML):
    self.getFile()
    p=email.Parser.Parser()
    try:
      msg=p.parse(self.file)
    except (email.Errors.BoundaryError,email.Errors.HeaderParseError):
      # Can't decode, return explanation and text
      self.file.seek(0)
      while 1:
        l=self.file.readline()
        if l==kFileLineSep or l=="":
          break
      textForViewing="Unable to parse MIME structure, showing as text\n"
      if deHTML:
        textForViewing+=stripHTML(self.file.read())
      else:
        textForViewing+=self.file.read()
      self.doneWithFile()
      commentaryLineNumbers=[bunch(lineNumber=1,kind="decodeCommentary")]
      partsInfo=[]
      return commentaryLineNumbers,textForViewing,partsInfo

    # Can decode
    self.doneWithFile()
    mimeList=getMimeParts(msg)
    commentaryLineNumbers=[]
    vqc=gPrefs.get("viewerQuoteColors")
    textForViewing=""
    lineNumber=1 # Count lines from 1 to match Tk. Bleh.
    for count in range(len(mimeList)):
      item=mimeList[count]
      if item.content_transfer_encoding not in (None,"7bit","8bit","binary") \
        or len(mimeList)>1 or item.mainType not in kDisplayableMainTypes:
        textForViewing+=self.makeCommentaryLine(item,count)
        commentaryLineNumbers.append(bunch(lineNumber=lineNumber,kind="decodeCommentary"))
        lineNumber+=1
      if item.name==None:
        item.name="Part-%i" % (count+1)
      if item.mainType in kDisplayableMainTypes:
        item.decodedPart=item.decodedPart.replace("\r","")
        item.decodedPart=item.decodedPart.replace("\x00","") # Weird but it can happen
        if deHTML:
          pf=cStringIO.StringIO(stripHTML(item.decodedPart))
        else:
          pf=cStringIO.StringIO(item.decodedPart)
        while 1:
          l=pf.readline()

          ##XXX
          if vqc<>None:  ###XXX###XXX
            result=kMailQuoteCharsRE.search(l)
            if result<>None:
              depth=len(result.group(0))-result.group(0).count(" ")
              depth=min(depth,len(vqc))
              commentaryLineNumbers.append(bunch(lineNumber=lineNumber,kind="quote%i" % depth))

          ##XXX
          if l=="":
            break
          if not l.endswith(kFileLineSep):
            l=l+kFileLineSep
          textForViewing+=l
          lineNumber+=1
      else:
        textForViewing+="(Not shown)\n"
        lineNumber+=1
    return commentaryLineNumbers,textForViewing,mimeList

  def makeCommentaryLine(self,mimeItem,partNo):
    commentary=[]
    if mimeItem.content_type<>None:
      commentary.append("type: %s" % mimeItem.content_type)
    if mimeItem.content_description<>None:
      commentary.append("description: %s" % mimeItem.content_description)
    if mimeItem.name<>None:
      commentary.append("name: %s" % mimeItem.name)
    if mimeItem.charset<>None:
      commentary.append("charset: %s" % mimeItem.charset)
    if mimeItem.decodedPart<>None and mimeItem.mainType<>None and mimeItem.mainType<>"text":
      commentary.append("%sB" % qtyToReadable(len(mimeItem.decodedPart)))

    if len(commentary)==0:
      commentaryLine="MIME part %i" % (partNo+1)
    else:
      commentaryLine="MIME part %i -- %s" % (partNo+1,", ".join(commentary))
    commentaryLine=commentaryLine+"\n"
    return commentaryLine

  def showInEditorNewReply(self,replyAddrChoice=None,deHTML=None):

    if self.whichOfThreeDirs()=="new":
      oldPath=self.path
      self.moveToCur()
      gSwitchboard.subdirChanged(oldPath,self.path)
      gSwitchboard.change(self.getMailboxName(),[self.path])

    useSWDBName=gPrefs.get("useSWDBNameOnReply")
    if replyAddrChoice=="to":
      h=self.get("to")
      if h==None:
        gLogger.add("Can't reply to the To address, there isn't one")
        return None
      replyAddr=canonicalizeTextAddr(h,useSWDBName=useSWDBName)
    elif replyAddrChoice=="all":
      al=[]
      for hName in ("from","to","cc","reply-to"):
        h=self.get(hName)
        if h<>None:
          al+=email.Utils.getaddresses([h])
      # Remove our own address
      al=[x for x in al if x[1]<>gPrefs.get("myMailAddress")]
      canonAl=canonicalizeAddrList(al,useSWDBName=useSWDBName)
      replyAddr=", ".join(canonAl)
    elif replyAddrChoice==None: # Normal reply
      h=self.get("reply-to")
      if h==None:
        h=self.get("from")
        if h==None:
          gLogger.add("Can't reply, no From or Reply-to address")
          return None
      replyAddr=canonicalizeTextAddr(h,useSWDBName=useSWDBName)
    else:
      assert 1==0,"showInEditorNewReply called with bad param: %s" % replyAddrChoice

    newMsg=email.Message.Message()
    newMsg["From"]="%s <%s>" % (gPrefs.get("myHumanName"),gPrefs.get("myMailAddress"))
    newMsg["To"]=replyAddr
    oldMessageID=cleanupCharsForHeader(self.get("message-id"))
    oldReferences=cleanupCharsForHeader(self.get("references"))
    oldInReplyTo=cleanupCharsForHeader(self.get("in-reply-to"))
    if oldMessageID<>None:
      oldMessageID=cleanupCharsForHeader(oldMessageID)
      newMsg["In-reply-to"]=oldMessageID
      if oldReferences<>None:
        newMsg["References"]=oldReferences+" "+oldMessageID
      elif oldInReplyTo<>None:
        newMsg["References"]=oldInReplyTo+" "+oldMessageID
      else:
        newMsg["References"]=oldMessageID
    oldSubj=cleanupCharsForHeader(self.get("subject"))
    if oldSubj==None or oldSubj.strip()=="":
      newSubj="Re: (no subject)"
    elif len(oldSubj)>=3 and oldSubj[:3].lower()=="re:":
      newSubj=oldSubj
    else:
      newSubj="Re: "+oldSubj
    newMsg["Subject"]=newSubj.strip()
    if gPrefs.get("replyTo")<>None:
      newMsg["Reply-to"]=gPrefs.get("replyTo")
    newMsg["Message-ID"]=makeMessageID()

    defaultReplyTextAsFile=cStringIO.StringIO()

    decodeImposs=0
    self.getFile()
    p=email.Parser.Parser()
    try:
      oldMsg=p.parse(self.file)
    except (email.Errors.BoundaryError,email.Errors.HeaderParseError):
      decodeImposs=1

    if decodeImposs:
      self.file.seek(0)
      # Skip past headers
      while 1:
        l=self.file.readline()
        if l==kFileLineSep or l=="":
          break
      while 1:
        l=self.file.readline()
        if l=="":
          break
        defaultReplyTextAsFile.write("> "+l)
        if not l.endswith(kFileLineSep):
          defaultReplyTextAsFile.write(kFileLineSep)
    else:
      mimeList=getMimeParts(oldMsg)
      for item in mimeList:
        if item.mainType in kDisplayableMainTypes:
          inFile=cStringIO.StringIO(item.decodedPart)
          while 1:
            l=inFile.readline()
            if l=="":
              break
            defaultReplyTextAsFile.write("> "+l)
          inFile.close()

    self.doneWithFile()
    if deHTML:
      newText=cStringIO.StringIO()
      newText.write(stripHTML(defaultReplyTextAsFile.getvalue()))
      defaultReplyTextAsFile=newText
    sigFilePath=gPrefs.get("sigFile")
    if sigFilePath<>None and sigFilePath<>"":
      defaultReplyTextAsFile.write(kFileLineSep)
      defaultReplyTextAsFile.write("-- "+kFileLineSep)
      f=open(sigFilePath)
      sig=f.read()
      defaultReplyTextAsFile.write(sig)
      if not sig.endswith(kFileLineSep):
        defaultReplyTextAsFile.write(kFileLineSep)
      f.close()

    newMsg.set_payload(defaultReplyTextAsFile.getvalue())

    if gPrefs.get("replyMunger"):
      gPrefs.get("replyMunger")(newMsg)

    outbox=gAllMailboxes.getMaildir("outbox")
    newPath=outbox.newFile()

    newOne=open(newPath,"wb")
    newOne.write(newMsg.as_string())
    defaultReplyTextAsFile.close()
    newOne.close()

    mailEditor(newPath,startInText=1)
    return None

  def getTidyHeaders(self,includeRanges=0):
    t=""
    allRanges=[]
    headerLines=0
    goodHeaders=gPrefs.get("nonBoringHeaders")
    headerNamePlusSpaceWidth=lengthOfLongestStr(goodHeaders)+1
    maxHeaderTextLen=kMailViewerTextWidth-headerNamePlusSpaceWidth
    for headerName in goodHeaders:
      headerName=headerName.capitalize()
      headerText=self.get(headerName)
      if headerText==None or headerText=="":
        continue
      headerText=encodedHeaderToLatin1(headerText) ###XXXXXXXXXXXX
      headerNamePadded=headerName.ljust(headerNamePlusSpaceWidth)
      headerText=re.sub(kWhitespaceRunRE," ",headerText)
      ###???headerText=gMakeDisplayableOnly(headerText)

      if headerName.lower() in ("from","to","cc","bcc","reply-to"):
        al=email.Utils.getaddresses([headerText])
        canonList=canonicalizeAddrList(al)
        lines,ranges=linesFromListOfStrs(canonList,maxHeaderTextLen,", ")
        for thisRange in ranges:
          thisRange.lineNo+=headerLines
          thisRange.start+=headerNamePlusSpaceWidth
          thisRange.end+=headerNamePlusSpaceWidth
        allRanges+=ranges
        for count in range(len(lines)):
          if count==0:
            t+=headerNamePadded
          else:
            t+="".ljust(headerNamePlusSpaceWidth)

          t+=lines[count]
          t+="\n"
          headerLines+=1
      elif headerName.lower()=="date": ### XXXX XXX
        #date=canonicalizeRFC822Date(headerText)
        date=getDateForTidyHeaders(headerText)
        if date==None: ### XXX XXX ### FIXME dateTuple==None:
          date=headerText
        thisLine=headerNamePadded+date
        thisLine=truncateWithEllipsis(thisLine,kMailViewerTextWidth)
        t+=thisLine
        t+="\n"
        headerLines+=1
      else:
        thisLine=headerNamePadded+headerText
        thisLine=truncateWithEllipsis(thisLine,kMailViewerTextWidth)
        t+=thisLine
        t+="\n"
        headerLines+=1
    if includeRanges:
      return t,allRanges
    else:
      return t

  def showInEditorNewForward(self,decodeStyle,deHTML=0):

    if self.whichOfThreeDirs()=="new":
      oldPath=self.path
      self.moveToCur()
      gSwitchboard.subdirChanged(oldPath,self.path)
      gSwitchboard.change(self.getMailboxName(),[self.path])

    newMsg=email.Message.Message()
    #headers={}
    newMsg["From"]="%s <%s>" % (gPrefs.get("myHumanName"),gPrefs.get("myMailAddress"))
    oldSubj=cleanupCharsForHeader(self.get("subject",""))
    if oldSubj=="":
      newMsg["subject"]="Fwd: (No subject)"
    elif oldSubj.lower().startswith("fwd:"): # String methods are fun!
      newMsg["subject"]=oldSubj.strip()
    else:
      newMsg["subject"]="Fwd: "+oldSubj.strip()
    newMsg["Message-id"]=makeMessageID()
    if gPrefs.get("replyTo")<>None:
      newMsg["reply-to"]=gPrefs.get("replyTo")

    defaultForwardTextAsFile=cStringIO.StringIO()
    decodeImposs=0

    self.getFile()
    if decodeStyle=="mime":
      p=email.Parser.Parser()
      try:
        oldMsg=p.parse(self.file)
      except email.Errors.BoundaryError:
        decodeImposs=1
    elif decodeStyle=="uudecode":
      try:
        uuParts=getUUParts(self.file)
      except (uu.Error,binascii.Error):
        decodeImposs=1

    if decodeImposs or decodeStyle=="nodecode":
      self.file.seek(0)
      if decodeImposs and decodeStyle<>"nodecode":
        # Skip past headers
        while 1:
          l=self.file.readline()
          if l==kFileLineSep or l=="":
            break
      defaultForwardTextAsFile.write("   --- Beginning of forwarded message ---"+kFileLineSep)
      if decodeImposs and decodeStyle<>"nodecode":
        defaultForwardTextAsFile.write(self.getTidyHeaders())
        defaultForwardTextAsFile.write(kFileLineSep)
      while 1:
        l=self.file.readline()
        if l=="":
          break
        defaultForwardTextAsFile.write(l)
      defaultForwardTextAsFile.write("   ---    End of forwarded message    ---"+kFileLineSep)

      self.doneWithFile()
      sigFilePath=gPrefs.get("sigFile")
      if sigFilePath<>None and sigFilePath<>"":
        defaultForwardTextAsFile.write(kFileLineSep)
        defaultForwardTextAsFile.write("-- "+kFileLineSep)
        f=open(sigFilePath)
        sig=f.read()
        defaultForwardTextAsFile.write(sig)
        if not sig.endswith(kFileLineSep):
          defaultForwardTextAsFile.write(kFileLineSep)
        f.close()

      newMsg.set_payload(defaultForwardTextAsFile.getvalue())
      defaultForwardTextAsFile.close()

      if gPrefs.get("forwardMunger"):
        gPrefs.get("forwardMunger")(newMsg)
      outbox=gAllMailboxes.getMaildir("outbox")
      newPath=outbox.newFile()
      newOne=open(newPath,"wb")
      newOne.write(newMsg.as_string())
      newOne.close()
      mailEditor(newPath,startInText=0)
      return None

    # Decodable
    if decodeStyle=="mime":
      encList=getMimeParts(oldMsg)
    elif decodeStyle=="uudecode":
      encList=uuParts
    else:
      assert 1==0,"Impossible value for decode style: %s" % decodeStyle

    defaultForwardTextAsFile.write("   --- Beginning of forwarded message ---"+kFileLineSep)
    defaultForwardTextAsFile.write(self.getTidyHeaders())
    defaultForwardTextAsFile.write(kFileLineSep)

    count=0
    while count<len(encList):
      if encList[count].mainType in kDisplayableMainTypes:
        if deHTML:
          defaultForwardTextAsFile.write(stripHTML(encList[count].decodedPart))
        else:
          defaultForwardTextAsFile.write(encList[count].decodedPart)
        del encList[count]
      else:
        count+=1
    defaultForwardTextAsFile.write("   ---    End of forwarded message    ---"+kFileLineSep)

    sigFilePath=gPrefs.get("sigFile")
    if sigFilePath<>None and sigFilePath<>"":
      defaultForwardTextAsFile.write(kFileLineSep)
      defaultForwardTextAsFile.write("-- "+kFileLineSep)
      f=open(sigFilePath)
      sig=f.read()
      defaultForwardTextAsFile.write(sig)
      if not sig.endswith(kFileLineSep):
        defaultForwardTextAsFile.write(kFileLineSep)
      f.close()

    if len(encList)==0:
      newMsg.set_payload(defaultForwardTextAsFile.getvalue())
      defaultForwardTextAsFile.close()
    else:
      del newMsg["Content-type"]
      newMsg["Content-type"]="multipart/mixed"
      newMsg.preamble=kMIMEPreambleText+kFileLineSep

      subMsg=email.Message.Message()
      subMsg.add_header("Content-type","text/plain",charset="ISO-8859-1") # Correct?
      subMsg.add_header("Content-transfer-encoding","7bit")
      subMsg.set_payload(defaultForwardTextAsFile.getvalue())
      defaultForwardTextAsFile.close()
      newMsg.attach(subMsg)
      for count in range(len(encList)):
        item=encList[count]
        subMsg=email.Message.Message()
        subMsg.set_payload(item.part)
        subMsg["Content-transfer-encoding"]=item.content_transfer_encoding
        if item.name==None:
          item.name="Part %i" % count
        subMsg.add_header("Content-type",item.content_type,name=item.name)
        newMsg.attach(subMsg)


    if gPrefs.get("forwardMunger"):
      gPrefs.get("forwardMunger")(newMsg)

    outbox=gAllMailboxes.getMaildir("outbox")
    newPath=outbox.newFile()
    newOne=open(newPath,"wb")
    g=email.Generator.Generator(newOne,mangle_from_=False)
    g.flatten(newMsg)
    #newOne.write(newMsg.as_string())
    newOne.close()
    mailEditor(newPath,startInText=0)
    return None

  def killOverviewCache(self):
    self.overviewTextList=None
    self.headersLoaded=0
    return None

  def getTextForOverview(self):
    # We don't cache isNewText, all else is assumed to remain constant
    if self.overviewTextList<>None:
      return [self.isNewText()]+self.overviewTextList

    myDate=formatDateForOverview(self.get("date",""))

    fromAddrText=self.get("from","")
    fromAddrText=kNonASCIITextRE.sub("?",fromAddrText)
    fromAddrText=encodedHeaderToLatin1(fromAddrText)
    fromAddrText=addrTextToDisplayText(fromAddrText)

    toHdr=self.get("to","")
    toHdr=kNonASCIITextRE.sub("?",toHdr)
    toHdr=encodedHeaderToLatin1(toHdr)
    toHdr=re.sub(kWhitespaceRunRE," ",toHdr)

    ccHdr=self.get("cc","")
    ccHdr=encodedHeaderToLatin1(ccHdr)
    ccHdr=re.sub(kWhitespaceRunRE," ",ccHdr)
    if toHdr=="" and ccHdr=="":
      toPlusCC=""
    elif toHdr=="":
      toPlusCC=ccHdr
    elif ccHdr=="":
      toPlusCC=toHdr
    else:
      toPlusCC=toHdr+", "+ccHdr

    toAddrText=addrTextToDisplayText(toPlusCC)

    subject=self.get("subject","")
    ###XXXXXXXX
    subject=kNonASCIITextRE.sub("?",subject)
    subject=encodedHeaderToLatin1(subject)

    # Shorten subject header by converting whitespace runs to once space.
    # I've actually run into subject headers that had newlines in them.
    #subject=gMakeDisplayableOnly(re.sub(kWhitespaceRunRE," ",subject))
    subject=re.sub(kWhitespaceRunRE," ",subject)

    score=str(self.getScore())

    self.overviewTextList=[myDate,fromAddrText,toAddrText,score,subject]
    return [self.isNewText()]+self.overviewTextList


class maildir:
  def __init__(self,dir,doSummary,progressDisplayFunc=None):
    #if dir=="":
    #  dir=os.path.expanduser(gPrefs.get("maildirParent"))
    self.dir=dir
    self.progressDisplayFunc=progressDisplayFunc
    self.mailObjs=[]
    self.mailObjsLoaded=0
    self.doSummary=doSummary
    if doSummary:
      self.summary={}
    else:
      self.summary=None
    self.checkMaildir()
    return None

  def getDir(self):
    return self.dir

  def getLastMod(self):
    # I don't think we should care about tmp
    last=max([ os.path.getmtime(myPathJoin(self.dir,subdir)) for subdir in ("cur","new") ])
    #return os.path.getmtime(self.dir)
    return last

  def getName(self):
    if self.dir==myPathJoin(gPrefs.get("maildirParent"),"Maildir"):
      return "inbox"
    else:
      return myPathSplit(self.dir)[1]

  def getPaths(self):
    paths=[]
    for subDir in ("new","cur"):
      for file in os.listdir(myPathJoin(self.dir,subDir)):
        fullPath=myPathJoin(myPathJoin(self.dir,subDir),file)
        paths.append(fullPath)
    return paths

  def loadMailObjsSummary(self):
    numFiles=self.howmanyFiles()
    wantToShowProgress=numFiles>=kEnoughMailForFancyMailboxOpening
    if self.progressDisplayFunc<>None and wantToShowProgress:
      self.progressDisplayFunc(0)

    picklePath=myPathJoin(self.dir,"summary.pickle")
    if not os.path.exists(picklePath):
      summaryFrom={}
    else:
      f=open(picklePath,"rb")
      try:
        summaryFrom=cPickle.load(f)
      except cPickle.UnpicklingError:
        f.close()
        gLogger.add("Summary %s couldn't be unpickled, discarding" % picklePath)
        summaryFrom={}
      else:
        f.close()
        if summaryFrom["version"]<>kSummaryVersion:
          gLogger.add("Old summary version found, discarding")
          summaryFrom={}

    summaryTo=self.getNewEmptySummary()
    myMailObjs=[]
    count=0.0
    nextProgressAt=time.time()+kShowProgressEvery
    #fileCounts=self.getFileCounts()
    #tot=fileCounts.new+fileCounts.cur


    for subdir in ("new","cur"):
      subdirPath=myPathJoin(self.dir,subdir)
      for file in os.listdir(subdirPath):
        # Icky? Now less so.
        count+=1
        if self.progressDisplayFunc<>None and wantToShowProgress and time.time()>=nextProgressAt:
          pctDone=round(count/numFiles*100)
          self.progressDisplayFunc(pctDone)
          nextProgressAt+=kShowProgressEvery

        fullPath=myPathJoin(subdirPath,file)
        lastTwo=myPathJoin(subdir,file)

        if lastTwo not in summaryFrom:
          thisSummary=self.makeOneSummary(fullPath) # obj too?
          summaryTo[lastTwo]=thisSummary
        else:
          # Are we concerned about this case?
          st=os.stat(fullPath)
          thisSummary=summaryFrom[lastTwo]
          if st.st_mtime==thisSummary["fileLastModified"]:
            summaryTo[lastTwo]=thisSummary
          elif st.st_mtime>thisSummary["fileLastModified"]:
            # File changed in place
            thisSummary=self.makeOneSummary(fullPath)
            summaryTo[lastTwo]=thisSummary
          else:
            gLogger.add("Time machine used, file %s modified earlier than summary" %
              fullPath)
            thisSummary=self.makeOneSummary(fullPath)
            summaryTo[lastTwo]=thisSummary

        myMailObjs.append(mailObj(fullPath,thisSummary))

    if self.progressDisplayFunc<>None and wantToShowProgress:
      self.progressDisplayFunc(100)
    os.umask(0077) # octal
    f=open(picklePath,"wb")
    cPickle.dump(summaryTo,f,True)
    f.close()
    self.summary=summaryTo

    self.mailObjs=myMailObjs
    self.mailObjsLoaded=1
    return None

  def loadMailObjsNoSummary(self):
    numFiles=self.howmanyFiles()
    wantToShowProgress=numFiles>=kEnoughMailForFancyMailboxOpening

    if self.progressDisplayFunc<>None and wantToShowProgress:
      self.progressDisplayFunc(0)

    count=0.0
    nextProgressAt=time.time()+kShowProgressEvery
    myMailObjs=[]

    for subDir in ("new","cur"):
      for file in os.listdir(myPathJoin(self.dir,subDir)):

        lastTwoElements=myPathJoin(subDir,file)
        fullPath=myPathJoin(self.dir,lastTwoElements)
        #if self.doSummary:
        #  myMailObjs.append(mailObj(fullPath,self.summary[lastTwoElements]))
        #else:
        #  myMailObjs.append(mailObj(fullPath))
        myMailObjs.append(mailObj(fullPath))

        count+=1
        if self.progressDisplayFunc<>None and wantToShowProgress and time.time()>=nextProgressAt:
          pctDone=round(count/numFiles*100)
          self.progressDisplayFunc(pctDone)
          nextProgressAt+=kShowProgressEvery

    self.mailObjs=myMailObjs
    self.mailObjsLoaded=1

    if self.progressDisplayFunc<>None and wantToShowProgress:
      self.progressDisplayFunc(100)

    return None

  def loadMailObjs(self):
    if self.doSummary:
      self.loadMailObjsSummary()
    else:
      self.loadMailObjsNoSummary()
    return None

  def getMailObjs(self):
    if not self.mailObjsLoaded:
      #if self.doSummary:
      #  self.loadMailObjsSummary(progressDisplayFunc)
      #else:
      #  self.loadMailObjsNoSummary()
      self.loadMailObjs()
    return self.mailObjs

  def howmanyFiles(self):
    fc=self.getFileCounts()
    return fc.new+fc.cur

  def getFileCounts(self):
    return bunch(new=len(os.listdir(myPathJoin(self.dir,"new"))),
                 cur=len(os.listdir(myPathJoin(self.dir,"cur"))))

  def __getitem__(self,index):
    if not self.mailObjsLoaded:
      self.loadMailObjs()
    return self.mailObjs[index]

  # Plenty of other ways to do this in Python but this is how the maildir
  # docs say to do it.
  def checkForFile(self,path):
    try:
      ignored=os.stat(path)
    except os.error,detail:
      if detail[0]<>errno.ENOENT:
        raise os.error,detail
      else:
        return 0
    else:
      return 1

  def newFile(self,flags=None):
    os.chdir(self.dir)
    os.umask(0177) # octal
    name=gNewMaildirFilename()
    if flags<>None:
      name=name+":2,"+flags
    path=myPathJoin("tmp",name)
    triesSoFar=0
    while self.checkForFile(path):
      # XXX better way to log?
      gLogger.add("weird: want to deliver to file %s but "
        "it already exists" % myPathJoin(os.getcwd(),path))
      assert triesSoFar<kMaxFileCreationTries, ("Tried %i times to create file %s "
          "but couldn't" % (triesSoFar,myPathJoin(os.getcwd(),path)))
      # A caught signal will interrupt our sleep. Make sure that
      # we have actually slept 2 seconds.
      zz=time.time()
      while time.time()-zz<2:
        time.sleep(2)
      name=gNewMaildirFilename()
      if flags<>None:
        name+=":2,"+flags
      path=myPathJoin("tmp",name)
      triesSoFar+=1
    fh=open(path,"wb")
    fh.close() # XXX Keep open if we're about to deliver to it?
    fullPath=myPathJoin(self.dir,path)
    return fullPath

  def getMailObj(self,i):
    if not self.mailObjsLoaded:
      self.loadMailObjs()
    return self.mailObjs[i]

  def checkMaildir(self):
    # We create missing directories (including all of them) but don't
    # try to fix other broken stuff

    dirs=[self.dir]
    for d in "tmp","new","cur":
      dirs.append(myPathJoin(self.dir,d))
    for d in dirs:
      try:
        s=os.stat(d)
      except os.error,detail:
        if detail[0]<>errno.ENOENT:
          raise os.error,detail
        else:
          gLogger.add("%s doesn't exist, creating" % d)
          os.umask(0077) # octal
          os.mkdir(d)
      else:
        mode=s[stat.ST_MODE]
        assert stat.S_ISDIR(mode), "%s is not a directory" % d

        #XXXX need to do something about win32 permissions
        if not gWin32:
          # Octal mode
          assert stat.S_IMODE(mode)==0700, ("%s has wrong permissions (%s)" %
            (d,oct(stat.S_IMODE(mode))))
    return None

  def cleanupTmp(self):
    p=myPathJoin(self.dir,"tmp")
    thirtySixHoursAgo=int(time.time())-kSecsIn36Hours
    assert os.path.isdir(p), "%s has no tmp subdirectory" % self.dir

    for fn in os.listdir(p):
      fnPath=myPathJoin(p,fn)
      if not os.path.isfile(fnPath):
        gLogger.add("Warning: %s is not a plain file" % fnPath)
        gLogger.add("and it should be since it's in a Maildir's tmp directory")
      aTime=os.path.getatime(fnPath)
      if aTime<=thirtySixHoursAgo:
        gLogger.add("Deleting %s," % fnPath)
        gLogger.add("it's in a Maildir's tmp directory and hasn't been accessed in 36 hours")
        os.unlink(fnPath)
    return None

  def copyToThisMaildir(self,oldMailObj):
    oldPath=oldMailObj.path
    if oldPath==None:
      flags="S"
    else:
      whereColonStuff=oldPath.find(":2,")
      if whereColonStuff<>-1 and whereColonStuff<len(oldPath)-3:
        flags=oldPath[whereColonStuff+3:]
      else:
        flags=None
    oldMailObj.getFile()
    newPath=self.deliver(oldMailObj.file,flags)
    oldMailObj.doneWithFile()
    if oldMailObj.path==None or getSubdirFromPath(oldMailObj.path)=="cur":
      newPath=moveToCur(newPath)
    return newPath

  def deliver(self,inFile,flags=None):
    outPath=self.newFile(flags)
    outFile=open(outPath,"wb")
    gulp=inFile.read()
    outFile.write(gulp)
    os.fsync(outFile.fileno())
    outFile.close()
    newPath=moveToNew(outPath)
    return newPath

  # This is nothing but a simpler and slower adaptation to Python
  # and Maildirs of Jamie Zawinski's mail summary files. They're
  # described at: http://www.jwz.org/doc/mailsum.html
  def makeOneSummary(self,path):
    # Get addresses in advance?
    msg=mailObj(path)
    sum={ "headers": {} }

    sum["score"]=msg.getScore()
    sum["dateAsSecs"]=msg.getDateAsSecs()
    sum["receivedDateAsSecs"]=msg.getReceivedDateAsSecs()
    sum["fileLastModified"]=os.stat(msg.path).st_mtime

    for headerName in kSummaryHeaders:
      val=msg.get(headerName)
      #if val<>None:
      sum["headers"][headerName]=val
    #fromHdr=msg.get("from","")
    #toHdr=msg.get("to","")
    #ccHdr=msg.get("cc","")
    #bccHdr=msg.get("bcc","")



    #sum["addresses"]["sender"]=email.Utils.parseaddr(fromHdr)[1]

    #sum["addresses"]["
    #recipientAddrs=email.Utils.getaddresses([toHdr,ccHdr,bccHdr])
    #recipientAddrs=[ x[1] for x in recipientAddrs ]
    #sum["recipientAddresses"]=recipientAddrs
    return sum

  def getNewEmptySummary(self):
    return copy.copy({"version": kSummaryVersion})

  # This speeds up opening a mailbox by about a factor of 2.
  # Can we improve that? Are pickles the best storage format?
  #
  # Stat-ing each file on open is at least a little annoying but
  # saves complexity. Common case for archives is that nothing
  # has changed. Could try to detect that case by stat-ing the
  # subdirectories and checking against cached last mod time.
  # But modifying a file in place doesn't update the directory's
  # last mod time. I don't know why you'd want to modify one of
  # the message files in place, but I think it's unreasonable to
  # interfere with user's ability to do that; one of the virtues
  # of maildirs is that the data is in plain files.

  def updateSummary(self):
    if not self.doSummary:
      return None

    picklePath=myPathJoin(self.dir,"summary.pickle")

    if not os.path.exists(picklePath):
      summaryFrom={}
    else:
      f=open(picklePath,"rb")
      try:
        summaryFrom=cPickle.load(f)
      except cPickle.UnpicklingError:
        f.close()
        gLogger.add("Summary %s couldn't be unpickled, discarding" % picklePath)
        summaryFrom={}
      else:
        f.close()
        if summaryFrom["version"]<>kSummaryVersion:
          gLogger.add("Old summary version found, discarding")
          summaryFrom={}

    summaryTo=self.getNewEmptySummary()

    diskPaths=self.getPaths()

    for diskPath in diskPaths:
      lastTwoElements=lastTwoPathElements(diskPath)
      if lastTwoElements not in summaryFrom:
        summaryTo[lastTwoElements]=self.makeOneSummary(diskPath)
      else:
        # Are we concerned about this case?
        st=os.stat(diskPath)
        thisSummary=summaryFrom[lastTwoElements]
        if st.st_mtime==thisSummary["fileLastModified"]:
          summaryTo[lastTwoElements]=summaryFrom[lastTwoElements]
        elif st.st_mtime>thisSummary["fileLastModified"]:
          summaryTo[lastTwoElements]=self.makeOneSummary(diskPath)
        else:
          gLogger.add("Time machine used, file %s modified earlier than summary" %
            diskPath)
          summaryTo[lastTwoElements]=self.makeOneSummary(diskPath)

    os.umask(0077) # octal
    f=open(picklePath,"wb")
    cPickle.dump(summaryTo,f,True)
    f.close()
    self.summary=summaryTo
    return None


#
# pop fetcher
#

class popMixin:

  def retrToFile(self,msgNum,openFile,msgSize,msgCount):
    self._putcmd("RETR %i" % msgNum)
    try:
      ignored=self._getresp()
    except poplib.error_proto,detail:
      gLogger.add("POP server said: %s" %detail,kError)
      return None # indicate failure
    except socket.timeout:
      gLogger.add("Timeout waiting for message to be retrieved",kError)
      return None # indicate failure

    myHostName=gPrefs.get("fqdn")
    theirHostName=gPrefs.get("popServer")
    inHeaders=1
    doneReceived=0
    bytesSoFar=0
    pctGot=0
    #lastProgressAt=0
    msgSize=float(msgSize) # So that we do floating-point division on it
    sizeStr=qtyToReadable(msgSize)
    nextProgressAt=time.time()+kShowProgressEvery
    # We trust the server to send us only stuff that makes sense
    # Across network, lines end CRLF, on disk they end LF only so we're receiving
    # more bytes that the POP server sees on disk and reports. Since we don't
    # want to go past 100%, reduce by 1 so the values are comparable.
    while 1:
      line,bytesInLine=self._getline()
      if line==".":
        break
      bytesInLine-=1
      if msgSize>kShowFetchProgressForMailBiggerThan:
        bytesSoFar+=bytesInLine
        #if bytesSoFar-lastProgressAt>=kShowFetchProgressEach:
        if time.time()>=nextProgressAt:
          #lastProgressAt=bytesSoFar
          pctGot=int(float(bytesSoFar)*100/msgSize)
          gLogger.add("Getting message %i of %i, %s bytes (%i%%)" % (msgNum,msgCount,sizeStr,pctGot))
          nextProgressAt+=kShowProgressEvery
      if line=="":
        inHeaders=0
      if (line.startswith("Received: ") or inHeaders==0) and doneReceived==0:
        doneReceived=1
        # Must match re in getDateAsSecs
        # XXX now in constants, put this there too
        openFile.write("Received: from %s by %s (%s); %s%s" % \
          (theirHostName,myHostName,kMyName,getRFC822Date(),kFileLineSep))
      openFile.write(line)
      openFile.write(kFileLineSep)
    os.fsync(openFile.fileno())
    openFile.close()
    # If we've been showing progress and haven't hit 100%,
    # show 100% for tidyness's sake.
    if msgSize>kShowFetchProgressForMailBiggerThan and pctGot<100:
      gLogger.add("Getting message %i of %i, %s bytes (%i%%)" % (msgNum,msgCount,sizeStr,100))
    return 1 # Success

class myPOP3(poplib.POP3,popMixin):
  pass
  
class myPOP3_SSL(poplib.POP3_SSL,popMixin):
  pass

#
# Checking for mail
#

# UI to display large mail not fetched

class unfetchedTableLine:
  def __init__(self,parent,unfetchedBunch):
    self.parent=parent
    self.unfetchedBunch=unfetchedBunch
    self.createWidgets()
    return None

  def createWidgets(self):
    myFont=gPrefs.get("fontILike")
    encFrame=Frame(self.parent,bg="white")

    self.fetchNextTimeVar=IntVar()

    # Jeez, it's a PITA to get these to look right. Worse still with grid().
    if self.unfetchedBunch.size>kNeverFetchBiggerThan:
      maxSize=qtyToReadable(kNeverFetchBiggerThan)
      l=Label(encFrame,text="Larger than %s, too big to fetch" % maxSize, \
        anchor=W,font=gPrefs.get("boldFontILike"),bg="white")
      l.pack(side=TOP,fill=X)
    else:
      c=Checkbutton(encFrame,text="Fetch this one next time",font=myFont, \
        bg="white",selectcolor="blue",activebackground="white", \
        highlightbackground="white",var=self.fetchNextTimeVar,anchor=W,padx=0,pady=0)
      c.pack(side=TOP,fill=X)

    f=Frame(encFrame,bg="white")
    sizeLabel=Label(f,text="Size",font=myFont,bg="white",borderwidth=0, \
      width=8,anchor=W,padx=2)
    sizeLabel.pack(side=LEFT)
    self.sizeDispay=Label(f,text=qtyToReadable(self.unfetchedBunch.size), \
      font=myFont,bg="white",borderwidth=0)
    self.sizeDispay.pack(side=LEFT)
    f.pack(side=TOP,fill=X)

    f=Frame(encFrame,bg="white")
    dateLabel=Label(f,text="Date",font=myFont,bg="white",borderwidth=0, \
      width=8,anchor=W,padx=2)
    dateLabel.pack(side=LEFT)
    self.dateDisplay=Label(f, \
      text=formatDateForOverview(self.unfetchedBunch.date), \
      font=myFont,bg="white",borderwidth=0)
    self.dateDisplay.pack(side=LEFT)
    f.pack(side=TOP,fill=X)

    f=Frame(encFrame,bg="white")
    fromLabel=Label(f,text="From",font=myFont,bg="white",borderwidth=0, \
      width=8,anchor=W,padx=2)
    fromLabel.pack(side=LEFT)
    self.fromDisplay=Label(f,text=canonicalizeTextAddr(self.unfetchedBunch.fromAddr),
      font=myFont,bg="white",borderwidth=0)
    self.fromDisplay.pack(side=LEFT)
    f.pack(side=TOP,fill=X)

    f=Frame(encFrame,bg="white")
    toLabel=Label(f,text="To",font=myFont,bg="white",borderwidth=0, \
      width=8,anchor=W,padx=2)
    toLabel.pack(side=LEFT)
    self.toLabel=Label(f,text=canonicalizeTextAddr(self.unfetchedBunch.toAddr),font=myFont, \
      bg="white",borderwidth=0)
    self.toLabel.pack(side=LEFT)
    f.pack(side=TOP,fill=X)

    f=Frame(encFrame,bg="white")
    subjectLabel=Label(f,text="Subject",font=myFont,bg="white", \
      borderwidth=0,width=8,anchor=W,padx=2)
    subjectLabel.pack(side=LEFT)
    self.subjectDisplay=Label(f,text=self.unfetchedBunch.subject, \
      font=myFont,bg="white",borderwidth=0)
    self.subjectDisplay.pack(side=LEFT)
    f.pack(side=TOP,fill=X)

    self.tkWidget=encFrame
    return None


class unfetchedMailNotifier:
  def __init__(self):
    self.summaryList=[]
    self.uidlsApproved=[]
    self.shown=0
    return None

  def show(self,summaryList):
    if self.shown==0:
      self.root=Toplevel()
      self.root.protocol("WM_DELETE_WINDOW",self.closeWinCB)
      self.root.title("Large mail not fetched")
      self.createWidgets()
      self.root.bind("<Key>",self.keyCB)
      self.shown=1
      self.addContents(summaryList)
    else:
      self.newContents(summaryList)
      # Prolly too annoying self.root.lift()
    return None

  def getUidlsApproved(self):
    return self.uidlsApproved

  # XXX Prolly evil of me to use methods with leading underscores.
  def keyCB(self,event):
    if event.char=="w":
      self.closeWinCB()
    elif event.keysym=="Up":
      self.contentsArea._yview("scroll","-1","units")
    elif event.keysym=="Down":
      self.contentsArea._yview("scroll","1","units")
    elif event.keysym=="Left":
      self.contentsArea._xview("scroll","-1","units")
    elif event.keysym=="Right":
      self.contentsArea._xview("scroll","1","units")
    elif event.char==" ":
      self.contentsArea._yview("scroll","1","pages")
    elif event.char=="u":
      self.contentsArea._yview("scroll","-1","pages")
    elif event.char=="-":
      self.contentsArea._yview("scroll","-1","pages")

    # No further processing of this event
    return "break"

  def windowClosed(self):
    return not self.root.winfo_exists()

  def closeWinCB(self):
    self.closeMe()
    return None

  def createWidgets(self):
    self.contentsArea=Pmw.ScrolledFrame(self.root,
      vscrollmode="static",
      hscrollmode="static",
      vertscrollbar_activebackground=kTkDefaultBG,
      horizscrollbar_activebackground=kTkDefaultBG,
      frame_bg="white",
      clipper_bg="white",
      vertscrollbar_borderwidth=1,
      vertscrollbar_width=12,
      horizscrollbar_borderwidth=1,
      horizscrollbar_width=12,
      )
    self.contentsArea.pack(side=TOP,fill=X,expand=1)

    self.controlsFrame=Frame(self.root)
    spacer=Frame(self.controlsFrame,width=5)
    spacer.pack(side=LEFT)
    okButton=myButton(self.controlsFrame,"OK",self.OKCB,widthInChars=10)
    okButton.pack(side=LEFT)
    spacer=Frame(self.controlsFrame,width=5)
    spacer.pack(side=RIGHT)
    canButton=myButton(self.controlsFrame,"Cancel",self.canCB,widthInChars=10)
    canButton.pack(side=RIGHT)
    self.controlsFrame.pack(side=TOP,fill=X)
    return None

  def addContents(self,summaryList):
    self.summaryList=summaryList
    for count in range(len(self.summaryList)):
      item=self.summaryList[count]
      myLine=unfetchedTableLine(self.contentsArea.interior(),item)
      myLine.tkWidget.pack(side=TOP,fill=X)
      item.tkWidget=myLine
      spacer=Frame(self.contentsArea.interior(),height=2,width=3,bg="white")
      spacer.pack(side=TOP,fill=X)
      if count<len(self.summaryList)-1:
        spacer=Frame(self.contentsArea.interior(),height=3,width=3,bg="gray")
        spacer.pack(side=TOP,fill=X)
    return None

  def removeContents(self):
    for item in self.contentsArea.interior().slaves():
      item.pack_forget()
    return None

  def OKCB(self):
    self.uidlsApproved=[]
    for item in self.summaryList:
      if item.tkWidget.fetchNextTimeVar.get()==1:
        self.uidlsApproved.append(item.uidl)
    self.closeMe()
    return None

  def canCB(self):
    self.closeMe()
    return None

  def closeMe(self):
    self.root.destroy()
    self.shown=0
    return None

  def newContents(self,summaryList):
    oldUidls=[item.uidl for item in self.summaryList]
    newUidls=[item.uidl for item in summaryList]
    oldUidls.sort()
    newUidls.sort()
    if oldUidls==newUidls:
      return None
    self.removeContents()
    self.addContents(summaryList)
    return None


class keysWin:
  def __init__(self):
    self.root=Toplevel()
    self.createWidgets()
    self.root.title("Key commands")
    self.root.bind("<Key>",self.keyCB)
    return None

  def createWidgets(self):
    myText= \
"""In the overview of mailboxes (main window)
m            New mail
up arrow     Move selection up
down arrow   Move selection down
return       Open viewer for selected mailbox
space        Open viewer for selected mailbox
click        Select mailbox
double-click Open viewer for selected mailbox
q            Quit program unless mail editors are open
Q            Quit program even if mail editors are open


In a mailbox viewer
up arrow     Move selection up
down arrow   Move selection down
left arrow   Scroll left
right arrow  Scroll right
return       Open selected mail
space        Open selected mail
click        Select mail
double-click Open selected mail
s            Save selected mail to mailbox(es) returned by getSaveFoldersInbound()
               in prefs, as long as the mail has been read
S            Save even if mail hasn't been read
a            Save selected mail to mailbox(es) returned by getSaveFoldersInboundtAlt()
               in prefs, as long as the mail has been read
A            A:a :: S:s
d            Move selected mail to trash folder if read
D            Trash even if mail hasn't been read
f            Forward selected mail, decoding text parts and enclosing binary
               enclosures
F            Forward selected mail without doing any MIME decoding
r            Reply to selected mail
R            (note capital) Reply to all
l            Reply to selected mail's To address (good for mailing
               list replies)
m            New mail
w            Close window
t            Raise toplevel window


Viewing mail
f            Forward current mail, decoding text parts and enclosing binary
               enclosures
F            Forward current mail without doing any MIME decoding
r            Reply
R            (note capital) Reply to all
l            Reply to mail's To address (good for mailing list replies)
h            Toggle display of all headers
w            Close window
i            Close window
S            (note capital) Save mail to file
P            (note capital) Print mail (see prefs)
m            New mail
left arrow   Previous mail
right arrow  Next mail
up arrow     Scroll up one line
down arrow   Scroll down one line
u            Scroll up one page
-            Scroll up one page
space        Scroll down one page
/            Search text
|            Pipe mail to external program specified in prefs
t            Raise toplevel window
b            Raise mailbox window


Editing outbound mail
control-l    Scroll to put insertion point near center of the window
tab          Go to next field (there is no way to put a tab in mail)
shift-tab    Go to previous field
control-z    Undo
control-y    Redo
control-up   Scroll up
control-down Scroll down
(The standard emacs-like bindings from the Tkinter Text widget also work.)


Other windows
up arrow     Scroll up one line
down arrow   Scroll down one line
space        Scroll down one page
u            Scroll up one page
-            Scroll up one page
w            Close window
"""
    self.textArea=Pmw.ScrolledText(self.root,
      text_bg="white",
      text_font=gPrefs.get("fontILike"),
      text_selectborderwidth=0,text_highlightbackground="white",
      text_selectbackground="white",
      text_cursor="", # Use normal cursor
      text_relief=FLAT,
      vertscrollbar_activebackground=kTkDefaultBG,
      vertscrollbar_borderwidth=1,
      vertscrollbar_width=12,
      horizscrollbar_borderwidth=1,
      horizscrollbar_width=12,
      )
    self.textArea.pack(fill=BOTH,expand=1)
    self.textArea.settext(myText)
    self.textArea.component("text").configure(state=DISABLED)
    return None

  def keyCB(self,event):
    if event.char=="w":
      self.close()
    elif event.keysym=="Up":
      self.textArea.yview("scroll","-1","units")
    elif event.keysym=="Down":
      self.textArea.yview("scroll","+1","units")
    elif event.char==" ":
      self.textArea.yview("scroll","+1","pages")
    elif event.char=="u":
      self.textArea.yview("scroll","-1","pages")
    elif event.char=="-":
      self.textArea.yview("scroll","-1","pages")
    # No further processing of this event
    return "break"

  def close(self):
    self.root.destroy()
    return None

class aboutBox:
  def __init__(self):
    aboutText1="%s, version %s" % (kMyName,kVersion)
    aboutText2="Provided with ABSOLUTELY NO WARRANTY,\nDistributable under the GPL.\n" \
      "See the documentation for details."
    self.root=Toplevel()
    self.root.resizable(0,0)
    self.root.title("About %s" % kMyName)
    l=Label(self.root,text=aboutText1,font=gPrefs.get("boldFontILike"),bg="white",padx=5)
    l.pack(side=TOP,fill=X,expand=1)
    l=Label(self.root,text=aboutText2,font=gPrefs.get("fontILike"),padx=5)
    l.pack(side=TOP,fill=X,expand=1)
    self.root.bind("<Key>",self.keyCB)
    return None

  def keyCB(self,event):
    if event.char=="w":
      self.close()
    return "break"

  def close(self):
    self.root.destroy()
    return None


# Routes messages among windows

class switchboard:
  def __init__(self):
    self.viewers=[]
    self.editors=[]
    self.mailboxBrowsers=[]
    #self.editorWithFocus=None
    self.mailboxOverview=None
    self.aboutBoxWin=None
    self.keysWin=None
    self.searchWin=None
    self.logStatus=kNormal
    self.messagesToCheckerThread=Queue.Queue()
    self.messagesFromCheckerThread=Queue.Queue()
    self.mailCheckLock=threading.Lock() ###XXX
    return None

  # Here rather than global since swdb wants to be able to call it
  def newMail(self,addr=None):
    m=email.Message.Message()
    m["From"]="%s <%s>" % (gPrefs.get("myHumanName"),gPrefs.get("myMailAddress"))
    m["Message-ID"]=makeMessageID()
    if addr<>None:
      m["To"]=addr
    if gPrefs.get("replyTo")<>None:
      m["Reply-to"]=gPrefs.get("replyTo")

    body=cStringIO.StringIO()

    sigFilePath=gPrefs.get("sigFile")
    if sigFilePath<>None and sigFilePath<>"":
      body.write(kFileLineSep * 3)
      body.write("-- "+kFileLineSep)
      sigFile=open(sigFilePath)
      t=sigFile.read()
      sigFile.close()
      body.write(t)

    m.set_payload(body.getvalue())
    body.close()

    if gPrefs.get("newMailMunger"):
      gPrefs.get("newMailMunger")(m)

    outbox=gAllMailboxes.getMaildir("outbox")
    newPath=outbox.newFile()
    newFile=open(newPath,"wb")
    newFile.write(m.as_string())
    newFile.close()
    mailEditor(newPath)
    return None

  def setOverview(self,win):
    self.mailboxOverview=win

  def getOverview(self):
    return self.mailboxOverview

  def addViewer(self,win):
    self.viewers.append(win)
    self.redoWindowMenus()
    return None

  def addEditor(self,win):
    self.editors.append(win)
    self.redoWindowMenus()
    return None

  # This is ugly but critical. The problem that we're avoiding is a
  # situation in which the user opens a mailbox browser just as the
  # delivery thread is saving new messages to the mailbox. That's
  # what the mailCheckLock is for. But the checking thread sends its
  # results back to the main thread via a queue. Since the queue isn't
  # processed immediately, it's possible for it to queue a notification
  # that there's new mail in a mailbox and have that notification sit
  # in the queue for a while. It's possible for the checker thread to
  # save the new mail just before the user opens a mailbox browser
  # (which sees the mail). Then when the queue is processed, the mailbox
  # browser is informed of the new mail that it has already seen. In
  # order to avoid that, we (potentailly) delay adding the browser to
  # the list of objects that want to be informed of new mail. We queue
  # the addition to that list in the same queue that the checker thread
  # uses to inform objects of new mail so that they're guaranteed to
  # happen in the right order.
  def addMailboxBrowser(self,win):
    self.messagesFromCheckerThread.put(
      bunch(func=self.realAddMailboxBrowser,args=(win,)))
    return None

  def realAddMailboxBrowser(self,win):
    self.mailboxBrowsers.append(win)
    self.redoWindowMenus()
    return None

  def removeViewer(self,win):
    self.remove(self.viewers,win)
    self.redoWindowMenus()
    return None

  def removeEditor(self,win):
    self.remove(self.editors,win)
    self.redoWindowMenus()
    return None

  def removeMailboxBrowser(self,win):
    self.remove(self.mailboxBrowsers,win)
    self.redoWindowMenus()
    return None

  def remove(self,list,win):
    try:
      list.remove(win)
    except ValueError:
      gLogger.add("Window not in list of open windows. That's either a bug now or one earlier.")
    return None

  def getBrowserByBox(self,boxName):
    for browser in self.mailboxBrowsers:
      if browser.mailboxName==boxName:
        return browser
    return None

  def getViewersByBox(self,boxName):
    retVal=[]
    for viewer in self.getViewers():
      if getMailboxFromPath(viewer.path)==boxName:
        retVal.append(viewer)
    return retVal

  def add(self,boxName,stuff):
    browser=self.getBrowserByBox(boxName)
    if browser<>None:
      browser.add(stuff)
    self.counts(boxName)
    return None

  def change(self,boxName,stuff):
    browser=self.getBrowserByBox(boxName)
    if browser<>None:
      browser.change(stuff)
    self.counts(boxName)
    return None

  def delete(self,boxName,stuff):
    browser=self.getBrowserByBox(boxName)
    if browser<>None:
      browser.delete(stuff)
    self.counts(boxName)
    return None

  def redisplayFromScratch(self,boxName):
    browser=self.getBrowserByBox(boxName)
    if browser<>None:
      browser.redisplayFromScratch()
    return None

  def counts(self,boxName):
    browser=self.getBrowserByBox(boxName)
    if browser<>None:
      browser.setCountLabel()
    self.getOverview().redoCounts(boxName)
    for viewer in self.getViewersByBox(boxName):
      viewer.setXOfYLabel()
    return None

  def redoWindowMenus(self):
    # Addressbook is always there, no need to list it agin
    wins=self.mailboxBrowsers+self.viewers+self.editors
    self.getOverview().mbar.newItemsForWindowMenu(wins)
    for win in wins:
      if hasattr(win,"mbar") and hasattr(win.mbar,"newItemsForWindowMenu"):
        win.mbar.newItemsForWindowMenu(wins)
    if gAddressBook.getState()<>"hidden":
      gAddressBook.mbar.newItemsForWindowMenu(wins)
    return None

  def informNewMailbox(self,boxName):
    for winObj in self.getViewers()+self.getMailboxBrowsers():
      winObj.updateMailboxNamesInMenu()
    self.getOverview().newMailboxCreated(boxName)
    return None

  def subdirChanged(self,oldPath,newPath):
    boxName=getMailboxFromPath(oldPath)
    browser=self.getBrowserByBox(boxName)
    if browser<>None:
      browser.subdirChanged(oldPath,newPath)
    return None

  def getByPath(self,path):
    retVal=self.getEditorByPath(path)
    if retVal<>None:
      return retVal
    else:
      return self.getViewerByPath(path)

  def getEditorByPath(self,path):
    for winObj in self.editors:
      if winObj.getPath()==path:
        return winObj
    return None

  def getViewerByPath(self,path):
    for winObj in self.viewers:
      if winObj.getPath()==path:
        return winObj
    return None

  def getEditors(self):
    return self.editors

  def getViewers(self):
    return self.viewers

  def getMailboxBrowsers(self):
    return self.mailboxBrowsers

  def closeAll(self):
    for winObj in self.getEditors()+self.getViewers()+self.getMailboxBrowsers():
      winObj.closeMe()
    return None

  def openAboutBox(self):
    if self.aboutBoxWin<>None and self.aboutBoxWin.root.winfo_exists() and self.aboutBoxWin.root.winfo_ismapped():
      self.aboutBoxWin.root.lift()
    else:
      self.aboutBoxWin=aboutBox()
    return None

  def openKeysWin(self):
    if self.keysWin<>None and self.keysWin.root.winfo_exists() and self.keysWin.root.winfo_ismapped():
      self.keysWin.root.lift()
      return None
    self.keysWin=keysWin()
    return None

  def openLogWindow(self):
    gLogger.show()
    return None

  def openAddressBrowser(self):
    if gAddressBook.getState()=="hidden":
      gAddressBook.show()
    else:
      gAddressBook.lift()
    return None

  def editInAddressbook(self,address):
    self.openAddressBrowser()
    gAddressBook.showInEditor(address)
    return None

  def openSearchWin(self):
    if self.searchWin==None:
      self.searchWin=searchWindow()
      return None
    try:
      ignored=self.searchWin.root.state()
    except TclError:
      self.searchWin=searchWindow()
      return None
    self.searchWin.root.lift()
    return None

  def redoLocationItems(self):
    self.getOverview().mbar.redoLocationItems()
    b=self.getBrowserByBox("outbox")
    if b<>None:
      b.mbar.redoLocationItems()
    return None

  def setLocationMenus(self,val):
    self.getOverview().mbar.locVar.set(val)
    b=self.getBrowserByBox("inbox")
    if b<>None:
      b.mbar.locVar.set(val)
    return None

  def checkRegularlyChanged(self,val):
    o=self.getOverview()
    o.mbar.setCheckingState(val)
    winObjs=self.getMailboxBrowsers()+self.getViewers()
    for winObj in winObjs:
      winObj.mbar.setCheckingState(val)
    return None

  def logBlink(self):
    if self.getOverview()<>None:
      self.getOverview().mbar.statusIndicator.logBlink()
    for winObj in self.getMailboxBrowsers():
      try: # Possile for window to close at just the wrong moment
        winObj.mbar.statusIndicator.logBlink()
      except TclError:
        pass
    return None

  def logErrorWarningSet(self,severity):
    self.logStatus=severity
    if self.getOverview()<>None:
      self.getOverview().mbar.statusIndicator.logErrorWarningSet()
    for winObj in self.getMailboxBrowsers():
      winObj.mbar.statusIndicator.logErrorWarningSet()
    return None

  def logErrorReset(self):
    self.logStatus=kNormal
    if self.getOverview()<>None:
      self.getOverview().mbar.statusIndicator.logErrorReset()
    for winObj in self.getMailboxBrowsers():
      winObj.mbar.statusIndicator.logErrorReset()
    return None

  def getLogStatus(self):
    return self.logStatus



class checker:
  checkRegularlyLock=threading.Lock()

  def __init__(self):
    self.checkerThread=None
    self.messagesToMainThread=None
    self.invokedByThread=0
    return None

  # For scan action
  def currentInboxAlreadyDisplayed(self):
    self.maildirNewSubdirFilenames=[]
    inbox=gAllMailboxes.getMaildir("inbox")
    for mail in inbox.getMailObjs():
      if mail.whichOfThreeDirs()=="new":
        self.maildirNewSubdirFilenames.append(mail.getPath())
    return None

  def scan(self):
    gLogger.add("Scanning inbox for new mail")
    pathsNow=[]
    newMailSinceLast=[]
    maxScore=0
    inbox=gAllMailboxes.getMaildir("inbox")
    for mailItem in inbox.getMailObjs():
      if mailItem.whichOfThreeDirs()=="new":
        p=mailItem.getPath()
        pathsNow.append(p)
        if not p in self.maildirNewSubdirFilenames:
          newMailSinceLast.append(mailItem)

          if gPrefs.get("swdbFile")<>None:
            addrs=[mailItem.get("to",""),mailItem.get("from",""), \
                   mailItem.get("cc",""),mailItem.get("bcc",""), \
                   mailItem.get("reply-to")]
            removeAny(addrs,"")
            gAddressBook.snarf(addrs,direction="in")
          maxScore=max(maxScore,mailItem.getScore())
    self.maildirNewSubdirFilenames=pathsNow

    numNew=len(newMailSinceLast)
    if numNew==0:
      self.logForFetch("No new mail in inbox")
      return None

    self.logForFetch("%i new %s in inbox" % (numNew,pluralize(numNew,"message")))
    gSwitchboard.add("inbox",newMailSinceLast)
    if maxScore>=gPrefs.get("noisyScore"):
      gPrefs.get("doHighScoreThing")()
    return None

  def setCheckingRegularly(self,val):
    if val:
      self.startCheckingRegularly()
    else:
      self.stopCheckingRegularly()
    return None

  def startCheckingRegularly(self):
    gotIt=checker.checkRegularlyLock.acquire(0)
    if not gotIt:
      gLogger.add("Checker thread failed to aquire checkRegularlyLock so won't do anything, " \
        "another check must be in progress -- shouldn't happen",kError)
    else:
      gLogger.add("Starting checker thread")
      self.checkerThread=threading.Thread(target=self.checkThreadRoutine)
      self.checkerThread.start()
    return None

  def stopCheckingRegularly(self):
    gLogger.add("Asking checker thread to stop")
    gSwitchboard.messagesToCheckerThread.put("stop")
    return None

  def getCheckingRegularly(self):
    return self.checkerThread<>None

  def checkThreadRoutine(self):
    gSwitchboard.messagesFromCheckerThread.put(bunch(func=gLogger.add,args=("Checker thread started",)))

    lastChecked=0
    delayTime=gPrefs.get("checkDelay")
    while 1:
      try:
        item=gSwitchboard.messagesToCheckerThread.get(0)
        if item=="stop":
          gSwitchboard.messagesFromCheckerThread.put(bunch(func=gLogger.add,args=("Checker thread sees a request to quit, exiting",)))
          checker.checkRegularlyLock.release()
          gChecker.checkerThread=None
          thread.exit()
        else:
          assert 1==0,"Unknown request to checker thread: %s" % item
      except Queue.Empty:
        pass
      if lastChecked+delayTime<=time.time():
        gSwitchboard.messagesFromCheckerThread.put(bunch(func=gLogger.add,args=("Checker thread checking for new mail",)))

        self.checkOnce(invokedByThread=1)
        lastChecked=int(time.time())
      time.sleep(kCheckThreadSleepSecs)

  def checkOnce(self,invokedByThread):
    if not gSwitchboard.mailCheckLock.acquire(0):
      gSwitchboard.messagesFromCheckerThread.put(bunch(func=gLogger.add,args=("checkOnce failed to aquire mail check lock so won't do anything",),kwArgs={"severity": kNormal}))
      return None
    try:
      self.invokedByThread=invokedByThread
      checkAction=gPrefs.get("mailCheckAction")
      if checkAction=="pop":
        self.doFetch()
      elif checkAction=="scan":
        self.scan()
      else:
        assert 1==0,'preference mailCheckAction neither "pop" nor "scan"'
    finally:
      gSwitchboard.mailCheckLock.release()
    return None

  def logForFetch(self,str,severity=kNormal):
    # Could log directly if we weren't invoked by the checker thread
    # but that adds complexity and going through the queue even when
    # unnecessary doesn't hurt anything.
    if self.invokedByThread:
      gSwitchboard.messagesFromCheckerThread.put(bunch(func=gLogger.add,args=(str,),kwArgs={"severity": severity}))
    else:
      gLogger.add(str,severity)
    return None

  def doFetch(self):
    self.logForFetch("Beginning fetch")
    retVal=None

    # Outer block to catch connection refused errors and log them
    try:

      retVal=self.doRealFetchingWork()
      ### XXX move instantiation out of doRealFetchingWork, combine with inbound filer
      self.logForFetch("POP conversation complete")

    # Catch connection refused or immediately closed
    except (socket.error,poplib.error_proto),detail:
      if (sys.exc_type==socket.error and detail[0]==111) or \
         (sys.exc_type==poplib.error_proto and str(detail)=="-ERR EOF"):
        self.logForFetch("Connection refused by POP server",kError)
        return None
      else:
        # If we caught something we don't know how to handle, throw it back
        raise

    if retVal<>None:
      gLogger.add("Processing retrieved messages")
      newPaths=multiValDict()

      filer=gPrefs.get("inboundFiler")
      scoreFunc=gPrefs.get("calculateScore")
      parseFunc=gPrefs.get("messageParseFunction")

      maxScore=0
      fromHeaders=[]


      nMessages=len(retVal.newPaths)
      for ind in range(nMessages):
        gLogger.add("Processing message %i of %i" % (ind+1,nMessages))
        path=retVal.newPaths[ind]
        f=open(path)
        try:
          msg=parseFunc(f)
        except (email.Errors.BoundaryError,email.Errors.HeaderParseError):
          f.close()
          newPaths["inbox"]=path
          gLogger.add("Can't parse %s so it can't be scored or auto-filed" % path)
          continue

        f.close()

        if scoreFunc<>None:
          maxScore=max(maxScore,scoreFunc(msg))

        fromHeaders.append(msg.get("from",""))

        if filer==None:
          newPaths["inbox"]=path
        else:
          mbs=filer(msg)
          if mbs==None:
            newPaths["inbox"]=path
            continue

          if type(mbs)==types.StringType:
            mbs=[mbs]

          for mb in mbs:
            if mb<>"inbox":
              newMaildir=gAllMailboxes.getMaildir(mb)
              myMailObj=mailObj(path)
              newPaths[mb]=newMaildir.copyToThisMaildir(myMailObj)
              gLogger.add("Auto-filed message %i of %i to mailbox %s" % (ind+1,len(retVal.newPaths),mb))
          if "inbox" in mbs:
            newPaths["inbox"]=path
            gLogger.add("Message %i of %i remains in inbox" % (ind+1,len(retVal.newPaths)))
          else:
            os.unlink(path)

      ##for k in newPaths.keys():
      ##  print k,newPaths[k]
      # Doesn't matter if we're not in the thread. It will get done anyhow.
      for k in newPaths.keys():
        gSwitchboard.messagesFromCheckerThread.put(
          bunch(func=gSwitchboard.add,
          args=(k,newPaths[k])))

      if maxScore>=gPrefs.get("noisyScore"):
        gPrefs.get("doHighScoreThing")()

      if len(retVal.unfetchedMailSummary)>0:
        gUnfetchedMailNotifier.show(retVal.unfetchedMailSummary)

      if gPrefs.get("swdbFile")<>None:
        self.logForFetch("Snarfing addresses")
        gSwitchboard.messagesFromCheckerThread.put(
          bunch(func=gAddressBook.snarf,
                args=(fromHeaders,),
                kwArgs={"direction": "in"}))
    self.logForFetch("Fetch complete")
    return None

  def doRealFetchingWork(self):
    pw=gPrefs.get("popPass")
    if pw==None or pw=="":
      passDialog=getPasswordDialog()
      pw=passDialog.waitForIt()
      if pw==None:
        self.logForFetch("Fetch cancelled")
        return None
    else:
      #pw=rot13(gPrefs.get("popPass"))
      pw=gPrefs.get("popPass").encode("rot-13")

    self.logForFetch("Connecting to POP server %s" % gPrefs.get("popServer")) ##XXX

    if gPrefs.get("popSSL"):
      c=myPOP3_SSL
    else:
      c=myPOP3
      
    try:
      popObj=c(gPrefs.get("popServer"),gPrefs.get("popPort"))
    except socket.error,detail:
      if detail==errno.ETIMEDOUT:
        self.logForFetch("Connection to POP server %s timed out" % gPrefs.get("popServer"),kError) ##XXX
        return None
      else:
        self.logForFetch("Unable to connect to POP server %s, socket reports: %s" % \
          (gPrefs.get("popServer"),str(detail)),kError) ###XXX
        return None
    except socket.timeout:
      self.logForFetch("Connection to POP server %s timed out" % gPrefs.get("popServer"),kError)##!!,kError) ##XXX
      return None

    if gPrefs.get("popAuth")=="apop":
      self.logForFetch("Doing APOP authentication")
      try:
        popObj.apop(gPrefs.get("popUser"),pw)
      except poplib.error_proto, detail:
        if str(detail)[:4]<>"-ERR":
          raise
        self.logForFetch("APOP authentication not accepted -- server said: %s" % detail,kError)
        popObj.quit()
        return None
    else:
      self.logForFetch("Doing user/pass authentication")
      try:
        # I doubt that a POP server would reject a username but may as well put
        # this line here as above the try.
        popObj.user(gPrefs.get("popUser"))
        popObj.pass_(pw)
      except poplib.error_proto, detail:
        if str(detail)[:4]<>"-ERR":
          raise
        self.logForFetch("POP password not accepted -- server said: %s" % detail,kError)
        popObj.quit()
        return None
      except socket.timeout:
        self.logForFetch("Timeout doing POP authentication",kError)
        popObj.quit()
        return None

    msgCount,totSize=popObj.stat()

    if msgCount==0:
      self.logForFetch("No mail waiting")
      popObj.quit()
      return None

    self.logForFetch("%i %s, %s bytes waiting" % (msgCount,pluralize(msgCount,"message"),
      qtyToReadable(totSize)))

    #maxScore=0
    maxBytes=gPrefs.get("askBeforeFetchingMailThisBig")
    if maxBytes==0 or maxBytes>kNeverFetchBiggerThan:
      maxBytes=kNeverFetchBiggerThan
    inbox=gAllMailboxes.getMaildir("inbox")
    newPaths=[]
    unfetchedMailSummary=[]
    #addressHeaders=[]

    for count in range(msgCount):
      msgNum=count+1
      try:
        resp=popObj.list(msgNum)
      except socket.timeout:
        gLogger.add("Timeout asking for message",kError)
        popObj.quit()
        return None

      bytes=int(resp.split()[2])
      if bytes>maxBytes:
        self.logForFetch("Message %i of %i is %s bytes (%s), that's larger than %s bytes (%s), " \
          "the maximum to fetch automatically" % \
          (msgNum,msgCount,bytes,qtyToReadable(bytes),maxBytes,qtyToReadable(maxBytes)))
        # May return "fetch" to fetch, None, to continue (punt), or
        # dict to add to summary for later display
        result=self.dealWithBigOne(popObj,bytes,msgNum)
        if result==None:
          continue
        if result<>"fetch":
          unfetchedMailSummary.append(result)
          continue

      self.logForFetch("Beginning fetch of message %i of %i, %s bytes" %
        (msgNum,msgCount,qtyToReadable(bytes)))
      tmpPath=inbox.newFile()
      tmpFile=open(tmpPath,"wb")
      status=popObj.retrToFile(msgNum,tmpFile,bytes,msgCount)
      if status==None:
        gLogger.add("Unable to retrieve",kError)
        tmpFile.close()
        os.unlink(tmpPath)
      else:
        tmpFile.close()
        newPath=moveToNew(tmpPath)

        newPaths.append(newPath)
        if gPrefs.get("popDele"):
          popObj.dele(msgNum)

        # XXX Not the right place to do this???
        #m=mailObj(newPath)
        #if gPrefs.get("swdbFile")<>None:
        #  addressHeaders+=[m.get("to",""),m.get("from",""),
        #                   m.get("cc",""),m.get("bcc",""),
        #                   m.get("reply-to","")]
        #maxScore=max(maxScore,m.getScore())
        # XXX

    popObj.quit()
    self.logForFetch("Done with fetch")
    return bunch(newPaths=newPaths,unfetchedMailSummary=unfetchedMailSummary)
                 #maxScore=maxScore,
                 #unfetchedMailSummary=unfetchedMailSummary
                 #addressHeaders=addressHeaders)

  def dealWithBigOne(self,popObj,bytes,msgNum):
    # Could be confused by switching servers after fetch
    # of large mail approved. Going to assume that's not
    # a problem unless it proves to be one in practice.
    try:
      uidlLine=popObj.uidl(msgNum)
    except poplib.error_proto, detail:
      if str(detail)[:4]<>"-ERR":
        raise
      # A lie. Could show it, just couldn't figure out which ones to
      # fetch later. Fix if that's a problem in practice.
      #gLogger.add("UIDL command not accepted -- server said: %s, can't show " \
      #  "summary of mail too big to fetch" % detail,kError)
      self.logForFetch("UIDL command not accepted -- server said: %s, can't show " \
        "summary of mail too big to fetch" % detail)##!!!,kError)
      return None # Continue but don't fetch

    uidl=uidlLine.split()[2] # Should be safe, UIDLs can't contain spaces
    if uidl in gUnfetchedMailNotifier.getUidlsApproved():
      self.logForFetch("Large mail has been approved for fetching")
      return "fetch"

    self.logForFetch("Message noted but not fetched")
    try:
      popResp=popObj.top(msgNum,0) # Get headers only
    except poplib.error_proto, detail:
      if str(detail)[:4]<>"-ERR":
        raise
      self.logForFetch("TOP command not accepted -- server said: %s, can't show " \
        "summary of mail to big to fetch" % detail,kError)
      return None

    lineList=popResp[1]
    s=cStringIO.StringIO("\n".join(lineList)) # StringIO seems to want newlines
    p=email.Parser.HeaderParser()
    m=p.parse(s)
    return bunch(size=bytes,fromAddr=m.get("from",""),toAddr=m.get("to",""), \
      subject=m.get("subject",""),date=m.get("date",""),uidl=uidl)


class logger:
  def __init__(self):
    self.threadWriteLock=threading.Lock()
    self.data=[]
    self.shown=0
    self.stayAtBottom=1
    self.lastLogLineWidget=None
    self.useFile=None
    self.logPath=None
    return None

  def readyToUseFile(self):
    self.useFile=1
    logPath=gPrefs.get("logFilePath")
    if logPath==None:
      self.logPath=None
    else:
      self.logPath=logPath+time.strftime("-%Y%m%d-%H%M%S",time.localtime(time.time()))
      if len(self.data)>0:
        # Catch up
        self.threadWriteLock.acquire()
        f=open(self.logPath,"a")
        for l in self.data:
          f.write(l.text+kFileLineSep)
        self.threadWriteLock.release()
        f.close()
    self.add("Writing log to file %s" % self.logPath)
    return None

  def add(self,str,severity=kNormal):
    assert severity in (kNormal,kError,kWarning), "Impossible logging severity %s" % severity
    self.threadWriteLock.acquire()
    oldStatus=gSwitchboard.getLogStatus()
    if severity<>kNormal:
      gSwitchboard.logErrorWarningSet(severity)
    else:
      gSwitchboard.logBlink()
    str=str.rstrip()
    ts=time.strftime("%a, %Y-%m-%d %H:%M:%S",time.localtime(time.time()))
    if severity==kError:
      l="%s ERROR: %s" % (ts,str)
    elif severity==kWarning:
      l="%s Warning: %s" % (ts,str)
    else:
      l="%s %s" % (ts,str)
    if not self.useFile:
      if len(self.data)>=kLogLinesToKeep:
        del self.data[:(len(self.data)-kLogLinesToKeep)+1]
      self.data.append(bunch(text=l,severity=severity))
      self.threadWriteLock.release()
      return None

    if self.shown:
      self.textArea.component("text").configure(state=NORMAL)
    if len(self.data)>=kLogLinesToKeep:
      del self.data[:(len(self.data)-kLogLinesToKeep)+1]
      if self.shown:
        self.textArea.delete("1.0","2.0")
    if self.lastLogLineWidget<>None and oldStatus<>kError:
      self.setLastLogLineText(l)
      if severity==kWarning:
        self.lastLogLineWidget.configure(bg=kStatusLogWarningColor)
      elif severity==kError:
        self.lastLogLineWidget.configure(bg=kStatusLogErrorColor)
    self.data.append(bunch(text=l,severity=severity))
    if self.shown:
      self.textArea.insert(END,l+"\n")
      if self.stayAtBottom:
        self.textArea.yview(kLogLinesToKeep+1)
      self.textArea.component("text").configure(state=DISABLED)
    self.threadWriteLock.release()
    gTkRoot.update_idletasks()
    if self.logPath<>None:
      f=open(self.logPath,"a")
      f.write(l+kFileLineSep)
      f.close()
    return None

  def setLastLogLineText(self,line):
    self.lastLogLineWidget.configure(state=NORMAL)
    self.lastLogLineWidget.delete("1.0",END)
    self.lastLogLineWidget.insert(END,line)
    self.lastLogLineWidget.configure(state=DISABLED)
    return None

  def setLastLogLineWidget(self,widget):
    self.lastLogLineWidget=widget
    self.setLastLogLineText(self.data[-1].text)
    return None

  def show(self):
    if gSwitchboard.getLogStatus()<>kNormal:
      self.lastLogLineWidget.configure(bg="white")
      gSwitchboard.logErrorReset()
    if self.shown:
      self.root.lift()
    else:
      self.root=Toplevel()
      self.root.title("Log %s" % os.path.basename(self.logPath))
      self.createWidgets()
      self.textArea.yview(kLogLinesToKeep+1)
      self.shown=1
      self.root.protocol("WM_DELETE_WINDOW",self.close)
    return None

  def stayAtBottomCB(self):
    self.stayAtBottom=self.stayAtBottomTK.get()

  def createWidgets(self):
    topFrame=Frame(self.root)
    markButton=myButton(topFrame,"Mark",self.markButtonCB)
    markButton.pack(side=LEFT)
    self.stayAtBottomTK=IntVar()
    self.stayAtBottomTK.set(self.stayAtBottom)
    bottomCheckbox=Checkbutton(topFrame,text="Stay at bottom",font=gPrefs.get("fontILike"), \
      selectcolor="blue", variable=self.stayAtBottomTK,command=self.stayAtBottomCB)
    bottomCheckbox["activebackground"]=kTkDefaultBG
    bottomCheckbox.pack(side=LEFT)

    topFrame.pack(side=TOP,fill=X)

    textArea=Pmw.ScrolledText(self.root,
      vscrollmode="static",
      hscrollmode="static",
      text_bg="white",
      text_font=gPrefs.get("fixedWidthFontILike"),
      text_wrap="none", # Sigh, "none" not None
      vertscrollbar_activebackground=kTkDefaultBG,
      horizscrollbar_activebackground=kTkDefaultBG,
      vertscrollbar_borderwidth=1,
      vertscrollbar_width=12,
      horizscrollbar_borderwidth=1,
      horizscrollbar_width=12,
      )
    textArea.tag_configure("warning",background=kStatusLogWarningColor)
    textArea.tag_configure("error",background=kStatusLogErrorColor)
    textArea.pack(side=TOP,fill=X)
    for item in self.data:
      textArea.insert(END,item.text)
      textArea.insert(END,"\n")
      if item.severity==kWarning:
        textArea.tag_add("warning","end-2 lines","end-1 line")
      elif item.severity==kError:
        textArea.tag_add("error","end-2 lines","end-1 line")
    textArea.component("text").configure(state=DISABLED)
    self.textArea=textArea
    self.root.bind("<Key>",self.keyCB)
    return None

  def keyCB(self,event):
    if event.char=="w":
      self.close()
    elif event.keysym=="Up":
      self.textArea.yview("scroll",-1,"units")
    elif event.keysym=="Down":
      self.textArea.yview("scroll",+1,"units")
    elif event.char==" ":
      self.textArea.yview("scroll",+1,"pages")
    elif event.char=="u":
      self.textArea.yview("scroll",-1,"pages")
    elif event.char=="-":
      self.textArea.yview("scroll",-1,"pages")
    elif event.keysym=="Right":
      self.textArea.xview("scroll",1,"units")
    elif event.keysym=="Left":
      self.textArea.xview("scroll",-1,"units")
    # No further processing of this event
    return "break"

  def markButtonCB(self):
    self.add("Mark")
    return None

  def close(self):
    self.stayAtBottom=self.stayAtBottomTK.get()
    self.root.destroy()
    self.shown=0
    return None


# mail search window

class searchWindow:

  def __init__(self):
    self.root=Toplevel()
    self.root.title("Search saved mail")
    self.stillSearching=0
    self.stopSearch=0
    self.allMatches=[]
    self.lastChildClicked=None
    self.lastClickTime=0
    ###self.highlightedChild=None ### XXXX should be here?
    self.searchWords=StringVar()
    self.searchInWhat=StringVar()
    self.searchInWhat.set("both")
    self.statusMsgVar=StringVar()
    self.createWidgets()
    self.root.bind("<Key-Up>",self.scrollUpCB)
    self.root.bind("<Key-Down>",self.scrollDownCB)
    self.root.protocol("WM_DELETE_WINDOW",self.closeWinCB)
    self.showStatusMsg("No search yet")
    return None

  def createWidgets(self):
    myFont=gPrefs.get("fontILike")

    f=Frame(self.root,width=kSearchMailWindowWidth)
    f.pack_propagate(0)
    f.pack(side=TOP)

    panes=Pmw.PanedWidget(self.root,orient="vertical")
    topPane=panes.add("top")
    bottomPane=panes.add("bottom")

    spacer=Frame(topPane,width=1,height=1)
    spacer.pack(side=TOP)

    f=Frame(topPane,width=kSearchMailWindowWidth)
    f.pack_propagate(0)
    f.pack(side=TOP)
    f=Frame(topPane)
    spacer=Frame(f,width=kSearchMailSpacerSize,height=kSearchMailSpacerSize)
    spacer.pack(side=LEFT)
    searchButton=myButton(f,"Search",self.searchCB,widthInChars=8)
    searchButton.pack(side=LEFT)
    stopSearchButton=myButton(f,"Stop search",self.stopSearchCB,widthInChars=8)
    stopSearchButton.pack(side=LEFT)
    clearButton=myButton(f,"Clear",self.clearCB,widthInChars=8)
    clearButton.pack(side=LEFT)
    spacer=Frame(f,height=kSearchMailSpacerSize,width=kSearchMailSpacerSize)
    spacer.pack(side=LEFT)
    statusMsgLabel=Label(f,textvariable=self.statusMsgVar,
      anchor=NW,font=myFont)
    statusMsgLabel.pack(side=LEFT,fill=X)
    f.pack(side=TOP,fill=X)

    spacer=Frame(topPane,width=kSearchMailSpacerSize,height=kSearchMailSpacerSize)
    spacer.pack(side=TOP)

    f=Frame(topPane)
    spacer=Frame(f,width=3,height=3)
    spacer.pack(side=LEFT)
    spacer=Frame(f,width=3,height=3)
    spacer.pack(side=RIGHT)
    l=Label(f,text="Search for",width=13,anchor=W,font=myFont)
    l.pack(side=LEFT)
    wordsEntryBorder=Frame(f,bg="black",borderwidth=1)
    wordsEntry=Entry(wordsEntryBorder,
      textvariable=self.searchWords,
      font=myFont,
      bg="white",
      highlightthickness=0,
      relief=FLAT)###GROOVE)##,relief=FLAT)
    if gPrefs.get("hideMousePointerWhileTyping"):
      gMousePointerHider.makeWidgetHideMouseOnKey(wordsEntry)
    wordsEntry.bind("<Key-Return>",self.returnHitCB)
    wordsEntry.pack(side=LEFT,fill=X,expand=1)
    wordsEntryBorder.pack(side=LEFT,fill=X,expand=1)
    f.pack(side=TOP,fill=X)

    spacer=Frame(topPane,width=3,height=3)
    spacer.pack(side=TOP)

    f=Frame(topPane)
    spacer=Frame(f,width=3,height=3)
    spacer.pack(side=LEFT)
    l=Label(f,text="Search mails'",width=13,anchor=W,font=myFont)
    l.pack(side=LEFT)
    headersRadio=Radiobutton(f,text="headers",
      font=myFont,
      value="headers",
      variable=self.searchInWhat,
      anchor=W,
      activebackground=kTkDefaultBG,
      selectcolor="blue",
      takefocus=0)
    headersRadio.pack(side=LEFT)
    bodyRadio=Radiobutton(f,text="body",font=myFont,
      value="body",variable=self.searchInWhat,anchor=W,
      activebackground=kTkDefaultBG,selectcolor="blue",
      takefocus=0)
    bodyRadio.grid(sticky=W,row=2,column=2)
    bodyRadio.pack(side=LEFT)
    bothRadio=Radiobutton(f,text="both",font=myFont,
      value="both",variable=self.searchInWhat,anchor=W,
      activebackground=kTkDefaultBG,selectcolor="blue",
      takefocus=0)
    bothRadio.pack(side=LEFT)
    f.pack(side=TOP,fill=X)

    spacer=Frame(topPane,width=3,height=3)
    spacer.pack(side=TOP)

    subFrame=Frame(topPane)
    spacer=Frame(subFrame,width=3)
    spacer.pack(side=LEFT,fill=Y)
    f2=Frame(subFrame)
    l=Label(f2,text="Mailboxes",font=myFont,anchor=W)
    l.pack(side=TOP,fill=X)
    allBoxesOnButton=myButton(f2,"All on",self.allBoxesOnButtonCB,widthInChars=8)
    allBoxesOnButton.pack(side=TOP)
    allBoxesOffButton=myButton(f2,"All off",self.allBoxesOffButtonCB,widthInChars=8)
    allBoxesOffButton.pack(side=TOP)
    f2.pack(side=LEFT,fill=Y)
    spacer=Frame(subFrame,width=3)
    spacer.pack(side=LEFT,fill=Y)
    subFrame.pack(side=TOP,fill=X)
    spacer=Frame(subFrame,height=3,width=3)
    spacer.pack(side=BOTTOM)
    mailboxesFrame=Pmw.ScrolledFrame(subFrame,
      vertfraction=0.2,
      clipper_bg="white",
      clipper_borderwidth=0,
      frame_bg="white",
      horizscrollbar_activebackground=kTkDefaultBG,
      vertscrollbar_activebackground=kTkDefaultBG,
      borderframe_relief=GROOVE,
      vertscrollbar_borderwidth=1,
      vertscrollbar_width=12,
      horizscrollbar_borderwidth=1,
      horizscrollbar_width=12,
      )
    myRow=0
    myCol=0
    boxSearchDict={}
    for name in gAllMailboxes.getNames():
      if myCol==kSearchMailMailboxCols:
        myRow+=1
        myCol=0
      v=IntVar()
      v.set(1)
      c=Checkbutton(mailboxesFrame.interior(),
        text=name,anchor=W,
        bg="white",
        font=myFont,
        activebackground="white",
        selectcolor="blue",
        highlightbackground="white",
        variable=v,
        takefocus=0,
        )
      boxSearchDict[name]=v
      c.grid(sticky=W,row=myRow,column=myCol)
      myCol+=1
    mailboxesFrame.pack(side=LEFT,fill=BOTH,expand=1)
    spacer=Frame(subFrame,width=3,height=3)
    spacer.pack(side=LEFT)
    gTkRoot.update_idletasks()
    bottomOfLastCheckbox=c.winfo_y()+c.winfo_height()+subFrame.winfo_y()
    self.boxSearchDict=boxSearchDict
    self.mailOverview=Pmw.ScrolledFrame(bottomPane,
      horizflex="expand",
      borderframe_relief=GROOVE,
      horizscrollbar_activebackground=kTkDefaultBG,
      vertscrollbar_activebackground=kTkDefaultBG,
      vertscrollbar_borderwidth=1,
      vertscrollbar_width=12,
      vertscrollbar_takefocus=0,
      horizscrollbar_borderwidth=1,
      horizscrollbar_width=12,
      horizscrollbar_takefocus=0,
      )
    self.mailOverview.pack(side=TOP,fill=BOTH,expand=1)
    panes.pack(side=TOP,fill=BOTH,expand=1)
    panes.configurepane("top",size=min(bottomOfLastCheckbox+kPaneHeightOverhead+3,9999))###,kSearchMailBoxGridMaxHeight))
    self.root.title("Search mail")
    wordsEntry.focus_set()
    return None

  def showStatusMsg(self,msg):
    self.statusMsgVar.set(msg)
    gTkRoot.update_idletasks()
    return None

  def scrollUpCB(self,eventIgnored):
    self.mailOverview.yview("scroll","-1","units")
    return None

  def scrollDownCB(self,eventIgnored):
    self.mailOverview.yview("scroll","1","units")
    return None

  def searchCB(self):
    self.stopSearch=0
    self.stillSearching=1
    nFilesSearched=0
    nFilesMatched=0

    for viewItem in self.mailOverview.interior().children.values():
      viewItem.destroy()
    self.childWidgets=[]
    self.lastChildClicked=None
    self.lastClickTime=0

    searchText=self.searchWords.get()
    searchText=searchText.strip()
    if searchText=="":
      self.showStatusMsg("No text to search for")
      return None
    self.showStatusMsg("Searching... ")
    searchText=searchText.lower()
    wordList=searchText.split()
    searchInWhat=self.searchInWhat.get()

    # Sort mailboxes in reverse date-modified order
    boxes=[]
    for boxName in gAllMailboxes.getNames():
      if self.boxSearchDict[boxName].get()==1:
        maildirObj=gAllMailboxes.getMaildir(boxName)
        mtime=maildirObj.getLastMod()
        boxes.append((mtime,maildirObj))

    if len(boxes)==0:
      self.showStatusMsg("No mailboxes selected for searching")
      return None

    boxes.sort()
    boxes.reverse()
    boxes=[x[1] for x in boxes] # Undecorate

    firstMatch=1

    for maildirObj in boxes:
      if self.stopSearch:
        break

      for fileCount in range(maildirObj.howmanyFiles()):
        self.root.update() # Necessary for clicks to get through
        if self.stopSearch: # Outer loop also checks so we break out of both loops
          break

        mailObj=maildirObj.getMailObj(fileCount)
        wordsMatchedDict={}
        linesMatchedList=[]
        nFilesSearched+=1
        searchFile=open(mailObj.path,"rb")
        lines=0

        if searchInWhat=="body":
          while 1:
            l=searchFile.readline()
            if l==kFileLineSep or l=="":
              break

        while 1:
          thisLine=searchFile.readline()
          if thisLine=="" or (searchInWhat=="headers" and thisLine==kFileLineSep):
            break
          lines+=1
          if lines % 100==0:
            self.root.update()
          thisLineLower=thisLine.lower()

          # The real excitement
          # For now the only search option is that all words must appear
          # somewhere in a message. Add others if they'd be useful.
          matchRuns=[]
          for word in wordList:
            where=stringFindAll(thisLineLower,word)
            if where<>None:
              # Record fact that word was found
              wordsMatchedDict[word]=1
              # matchRuns contains tuples of (where,len)
              matchRuns+=[(x,len(word)) for x in where]
          if len(matchRuns)>0:
            linesMatchedList.append(bunch(line=thisLine,runs=matchRuns))
        searchFile.close()

        allFoundInThisFile=len(wordsMatchedDict.keys())==len(wordList)
        if allFoundInThisFile:
          nFilesMatched+=1
          thisOne=bunch(path=mailObj.path,matches=linesMatchedList)
          if not firstMatch:
            line=Frame(self.mailOverview.interior(),bg="black",height=1)
            line.pack(side=TOP,fill=X,expand=1)
          else:
            firstMatch=0
          viewObj=searchOverviewObj(self.mailOverview.interior(),self,thisOne)
          viewObj.pack(side=TOP,fill=X,expand=1)
          self.childWidgets.append(viewObj)
          self.allMatches.append(thisOne)

        if nFilesSearched % 10 ==0 or allFoundInThisFile:
          if nFilesSearched==0:
            self.showStatusMsg("Searching in %s, %i files searched" % (maildirObj.getName(),nFilesSearched))
          elif nFilesMatched==0:
            self.showStatusMsg("Searching in %s, %i files searched, none found so far" % (maildirObj.getName(),nFilesSearched))
          else:
            self.showStatusMsg("Searching in %s, %i files searched, %i found so far" % (maildirObj.getName(),nFilesSearched,nFilesMatched))

    if self.stopSearch:
      msg="Search aborted, "
    else:
      msg="Search complete, "
    msg+="searched %i, " % nFilesSearched
    if nFilesMatched>0:
      msg+="found %i" % nFilesMatched
    else:
      msg+="not found"
    self.showStatusMsg(msg)
    self.stillSearching=0
    return None

  def clearCB(self):
    self.showStatusMsg("No search yet")
    self.searchWords.set("")
    self.allMatches=[]
    self.lastChildClicked=None
    self.lastClickTime=0
    for viewItem in self.mailOverview.interior().children.values():
      viewItem.destroy()
    return None

  # Fun! User might click Done button to close us while search still in progress. That would
  # kill the window out from under the process that's doing the searching and cause an ugly
  # error. So when user clicks Done, request that the search be stopped. If it is stopped
  # or wasn't underway in the first place (doesn't matter which), kill the window. Otherwise
  # schedule the same routine for the next time Tkinter is idle to check again.
  def closeWinCB(self):
    self.stopSearch=1
    self.root.destroy()
    return None

  def stopSearchCB(self):
    self.stopSearch=1
    return None

  def returnHitCB(self,eventIgnored):
    self.searchCB()
    return None

  def allBoxesOnButtonCB(self):
    for var in self.boxSearchDict.values():
      var.set(1)
    return None

  def allBoxesOffButtonCB(self):
    for var in self.boxSearchDict.values():
      var.set(0)
    return None

  def childClicked(self,childClicked):
    # Bugger me if I know why binding a <Double-Button-1> doesn't work here.
    if self.lastChildClicked==childClicked and \
      (time.time()-self.lastClickTime)*1000<=kMaxDoubleClickDelayMS:
      self.childDoubleClicked(childClicked)
    else:
      self.lastChildClicked=childClicked
      self.lastClickTime=time.time()
      for childWidget in self.childWidgets:
        if childWidget==childClicked:
          childWidget.highlightMe()
        else:
          childWidget.unHighlightMe()
    return None

  def childDoubleClicked(self,childDoubleClicked):
    mailViewer(childDoubleClicked.matchItem.path,self,moveToCur=0)
    return None

  def howmanyItems(self):
    return len(self.allMatches)

  def getPathByIndex(self,index):
    return self.allMatches[index].path

  def getIndexByPath(self,path):
    for count in range(len(self.allMatches)):
      if self.allMatches[count].path==path:
        return count
    assert 1==0,"Path %s not found in list of matched messages" % path


class searchOverviewObj:
  def __init__(self,parentWidget,parentObject,matchItem):
    self.parentWidget=parentWidget
    self.parentObject=parentObject
    self.matchItem=matchItem
    self.lastWidgetClicked=None
    self.createWidgets()
    return None

  def createWidgets(self):
    self.font=gPrefs.get("fontILike")
    l=Label(gTkRoot,text=" ",font=self.font)
    self.rowHeight=l.winfo_reqheight()+1
    self.widthsInDots=charWidthsToDotWidths(kOverviewColumnDefaultWidths)
    self.outerFrame=Frame(self.parentWidget)
    m=mailObj(self.matchItem.path)
    myMailboxLine=gridRow(self)##XXXXXXXX#mailboxLine(self)
    myMailboxLine.newText(m.getTextForOverview())
    ##line.pack(side=TOP)
    fileLabel=Label(self.outerFrame,anchor=NW,font=self.font,bg="white",text="File "+self.matchItem.path)
    fileLabel.pack(side=TOP,fill=X,expand=1)
    fileLabel.bind("<Button-1>",self.clickCB)
    self.childWidgets=[myMailboxLine,fileLabel]
    for lineInfoItem in self.matchItem.matches:
      t=Text(self.outerFrame,
        bg=kSearchMailMatchesBG,
        height=1,
        font=self.font,
        relief=FLAT,
        borderwidth=0,
        highlightthickness=0,
        selectbackground="black",
        selectborderwidth=0,
        exportselection=0,
        cursor="",
        wrap="none",
        xscrollcommand=self.dontScroll,
        yscrollcommand=self.dontScroll,
        takefocus=0,
        )
      t.insert(END,lineInfoItem.line)
      t.configure(state=DISABLED)
      t.bind("<Button-1>",self.clickCB)
      t.bind("<B1-Motion>",self.drag)
      t.pack(side=TOP,fill=X,expand=1)
      for run in lineInfoItem.runs:
        t.tag_add("highlight",
          "1.%i" % (run[0]),
          "1.%i" % (run[0]+run[1]))
      t.tag_configure("highlight",foreground="blue")
      self.childWidgets.append(t)
    return None

  def getGridParentWidget(self):
    return self.outerFrame

  # Ordinary bindings call this one
  def clickCB(self,event):
    self.lastWidgetClicked=event.widget
    self.parentObject.childClicked(self)
    return None

  # Mailbox overview lines calls thisone
  def childClicked(self,widget):
    self.lastWidgetClicked=widget
    self.parentObject.childClicked(self)
    return None

  def highlightMe(self):
    self.childWidgets[0].setColors("white","black")
    for w in self.childWidgets[1:]:
      w.configure(fg="white",bg="black")
    return None

  def unHighlightMe(self):
    self.childWidgets[0].setColors("black","white")
    self.childWidgets[1].configure(fg="black",bg="white")
    for w in self.childWidgets[2:]:
      w.configure(fg="black",bg=kSearchMailMatchesBG)
    return None

  # Even if we specify that the handler for B1-Motion should
  # return "break", Tkinter will scroll the widget to enable
  # the user to do a drag-selection. So for one, if we ever
  # start to scroll, scroll back immediately and if anybody
  # is doing any dragging, remove the selection immeidately.
  # Still can look a little funny but it's good enough.
  def dontScroll(self,low,high):
    if float(low)>0:
      self.lastWidgetClicked.see("1.0")
    return None

  def drag(self,eventIgnored):
    self.lastWidgetClicked.tag_remove(SEL,"1.0",END)
    return "break"

  def pack(self,*positionArgsTuple,**kwArgsDict):
    self.outerFrame.pack(*positionArgsTuple,**kwArgsDict)
    return None


#
# address book
#

# Used in browser as a callback "function" to handle clipboard requests.
class clipboardCBO:
  def __init__(self,str):
    self.str=str
    return None

  def __call__(self,ignored1,ignored2):
    return self.str

class addressBook:

  def __init__(self):
    self.root=None
    self.load()
    return None

  def show(self):
    self.root=Toplevel()
    self.root.title("Address browser")
    self.createWidgets()
    self.openEditWindow=None
    self.filterStr=""
    self.refresh()
    return None

  def hide(self):
    self.root.destroy()
    self.root=None
    return None

  def getState(self):
    if self.root==None:
      return "hidden"
    else:
      return "shown"

  def snarf(self,listOfText,direction):
    # Not implemented in this addressbook, only
    # in more elaborate optional one.
    return None

  def editAddr(self,text):
    # Not implemented in this addressbook, only
    # in more elaborate optional one.
    return None

  def cleanUpForQuit(self):
    # Not implemented in this addressbook, only
    # in more elaborate optional one.
    return None

  def lift(self):
    self.root.lift()
    return None

  def getAsSortedList(self):
    k=self.nickToAddress.keys()
    k.sort()
    l=[]
    for item in k:
      l.append({"nickname": item, "address": canonicalizeAddrList(self.nickToAddress[item])[0]})
    return l

  def nickToAddr(self,nick):
    return self.nickToAddress.get(nick,None)

  def getAllNicks(self):
    # XXX better interface later
    return self.nickToAddress.keys()

  def addrToNick(self,addr):
    # XXX better interface later
    return self.addressToNick.get(addr.lower(),None)

  def load(self):
    nickToAddress={}
    addressToNick={}
    data=gPrefs.get("addressbookData")
    for item in data:
      nick=item[0]
      addrString=item[1]
      if nick=="":
        continue
      if nickToAddress.has_key(nick):
        gLogger.add("In mail address file %s, nickname %s appears more than once" % (nick),kError)
        continue
      addressList=email.Utils.getaddresses([addrString])
      nickToAddress[nick]=addressList
      if len(addressList)==1:
        addressToNick[addressList[0][1].lower()]=nick
    # Always add "me"
    nickToAddress["me"]=[(gPrefs.get("myHumanName"),gPrefs.get("myMailAddress"))]
    addressToNick[gPrefs.get("myMailAddress").lower()]="me"
    self.nickToAddress=nickToAddress
    self.addressToNick=addressToNick
    return None

  def getText(self):
    aliases=[]
    aliasList=self.getAsSortedList()
    lenOfLongestNick=lengthOfLongestStr(self.getAllNicks())
    for item in aliasList:
      aliases.append(item["nickname"].ljust(lenOfLongestNick+1)+ \
        item["address"])
    self.aliases=aliases
    return None

  def refresh(self):
    oldYScroll=self.textArea.yview()[0]
    self.getText()
    self.filterAddresses()
    self.showText()
    self.textArea.yview("moveto",oldYScroll)
    return None

  def filterAddresses(self):
    filteredText=[]
    lowerFilterStr=self.filterStr.lower()
    for item in self.aliases:
      if self.filterStr=="" or item.lower().find(lowerFilterStr)<>-1:
        filteredText.append(item)
    self.filteredText=filteredText
    self.statusLabel.configure(text=repr(len(filteredText))+"/"+ \
      repr(len(self.aliases)))
    return None

  def showText(self):
    self.textArea.delete(0,END)
    for item in self.filteredText:
      self.textArea.insert(END,item)
    return None

  def createWidgets(self):
    topFrame=Frame(self.root,width=100,height=16)

    spacerFrame=Frame(topFrame,width=2,height=10)
    spacerFrame.pack_propagate(0)
    spacerFrame.pack(side=LEFT)

    entryBox=Entry(topFrame,font=gPrefs.get("fontILike"),bg="white",relief=FLAT)
    entryBox.pack(side=LEFT)

    spacerFrame=Frame(topFrame,width=2,height=10)
    spacerFrame.pack_propagate(0)
    spacerFrame.pack(side=LEFT)

    allButton=myButton(topFrame,"All",self.allButtonCB)
    allButton.pack(side=LEFT)

    statusLabel=Label(topFrame,font=gPrefs.get("boldFontILike"),width=9,anchor=E,padx=10)
    statusLabel.pack(side=LEFT)

    doneButton=myButton(topFrame,"Done",self.doneButtonCB) # Delete button and this?
    doneButton.pack(side=RIGHT)

    topFrame.pack(side=TOP,fill=X)

    bottomFrame=Frame(self.root)
    textArea=Listbox(bottomFrame,selectmode="single")
    textArea["background"]="white"
    textArea["font"]=gPrefs.get("fixedWidthFontILike")
    scrollX=Scrollbar(bottomFrame,
      orient=HORIZONTAL,
      activebackground=kTkDefaultBG,
      borderwidth=1,
      width=12,
      )
    scrollY=Scrollbar(bottomFrame,
      orient=VERTICAL,
      activebackground=kTkDefaultBG,
      borderwidth=1,
      width=12,
      )

    textArea["xscrollcommand"]=scrollX.set
    textArea["yscrollcommand"]=scrollY.set
    textArea["height"]=10
    textArea["width"]=50
    scrollX["command"]=textArea.xview
    scrollY["command"]=textArea.yview

    scrollX.pack(side=BOTTOM,fill=X)
    scrollY.pack(side=RIGHT,fill=Y)
    textArea.pack(side=LEFT,fill=BOTH,expand=1)

    bottomFrame.pack(side=TOP,expand=1,fill=BOTH)

    textArea.bind("<ButtonRelease-1>",self.selectCB)
    entryBox.bind("<KeyRelease>",self.entryKeyCB)
    entryBox.focus_set()

    self.textArea=textArea
    self.entryBox=entryBox
    self.statusLabel=statusLabel

    return None

  # We expect that if the user pastes the address, they don't want the nickname so
  # we have to fiddle the X selection. We do that by handing Tk a callback "function"
  # that will return the value we want on the X clibpoard. Bloody annoying.
  def selectCB(self,eventIgnored):
    myLine=self.textArea.get(self.textArea.curselection())
    myRE=re.compile(" +(.*)")
    found=myRE.search(myLine)
    if found:
      myAddr=found.group(1)
    else:
      myAddr=myLine

    myCBO=clipboardCBO(myAddr)
    self.textArea.selection_handle(myCBO)

    # We don't mind any default behavior (I don't think there is any on a mouse-up)
    # so we don't return "break"
    return ""

  def allButtonCB(self):
    self.entryBox.delete("0","end")
    self.filterStr=""
    self.filterAddresses()
    self.showText()
    return None

  def entryKeyCB(self,eventIgnored):
    newText=self.entryBox.get()
    self.filterStr=newText
    self.filterAddresses()
    self.showText()
    return None

  def doneButtonCB(self):
    self.root.destroy()
    self.root=None
    return None


#
# swdb
#


class swdbMenubar:
  def __init__(self,parent):
    self.parent=parent
    self.showOnlyWithNicksVar=IntVar()
    self.showOnlyWithNicksVar.set(1)
    self.createWidgets()
    self.newItemsForWindowMenu()
    return None

  def createWidgets(self):
    myFrame=Frame(self.parent.getMenuWidgetParent())
    myFrame.pack(side=LEFT,fill=X,expand=1)
    myFont=gPrefs.get("fontILike")
    mbar=Pmw.MenuBar(myFrame,padx=5,hotkeys=0)
    self.mbar=mbar
    mbar.pack(side=LEFT,fill=X)

    mbar.addmenu("Addressbook","",font=myFont,activebackground=kTkDefaultBG)

    mbar.addmenuitem("Addressbook","command",
      label="New",
      command=self.newAddressCB,
      font=myFont,
      activebackground=kTkDefaultBG)

    mbar.addmenuitem("Addressbook","separator")

    mbar.addmenuitem("Addressbook","checkbutton",
      command=self.nicksChangedCB,
      label="Show only with nicks",
      variable=self.showOnlyWithNicksVar,
      selectcolor="blue",
      font=myFont)

    mbar.addmenuitem("Addressbook","separator")

    mbar.addmenuitem("Addressbook","command",
      label="Import",
      command=self.importCB,
      font=myFont,
      activebackground=kTkDefaultBG)

    mbar.addmenuitem("Addressbook","command",
      label="Export",
      command=self.exportCB,
      font=myFont,
      activebackground=kTkDefaultBG)

    mbar.addmenuitem("Addressbook","separator")

    mbar.addmenuitem("Addressbook","command",
      label="Delete uninteresting",
      command=self.delUninterestingCB,
      font=myFont,
      activebackground=kTkDefaultBG)

    mbar.addmenu("Windows","",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","command",command=gSwitchboard.openAboutBox,label="About",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","command",command=gSwitchboard.openKeysWin,label="Key commands",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","separator")
    mbar.addmenuitem("Windows","command",command=gSwitchboard.openAddressBrowser,label="Addressbook",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","command",command=gSwitchboard.openSearchWin,label="Search mail",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","command",command=gSwitchboard.openLogWindow,label="Log",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","separator")
    # Next is trashed and replaced in newItemsForWindowMenu()
    self.mbar.addmenuitem("Windows","command",label="Open windows",font=gPrefs.get("boldFontILike"),state=DISABLED)
    self.standardWindowMenuLength=7
    return None

  def newItemsForWindowMenu(self,items=None):
    myFont=gPrefs.get("fontILike")
    self.mbar.deletemenuitems("Windows",self.standardWindowMenuLength,999)
    self.mbar.addmenuitem("Windows","command",
      label="Open windows",
      font=gPrefs.get("boldFontILike"),
      state=DISABLED,
      )
    self.mbar.addmenuitem("Windows","command",
      label="Mailboxes",
      font=myFont,
      command=self.raiseRootWin,
      activebackground=kTkDefaultBG,
      )
    if items<>None:
      for win in items:
        self.mbar.addmenuitem("Windows","command",
          label=win.root.title(),
          command=raiseWinCBO(win),
          font=myFont,
          activebackground=kTkDefaultBG,
          )
    return None

  def newAddressCB(self):
    self.parent.newRecord()
    return None

  def nicksChangedCB(self):
    self.parent.nicksCheckBoxChanged(self.showOnlyWithNicksVar.get())
    return None

  def delUninterestingCB(self):
    self.parent.delUninteresting()
    return None

  def importCB(self):
    self.parent.importAddresses()
    return None

  def exportCB(self):
    self.parent.exportAddresses()
    return None

  def raiseRootWin(self):
    gSwitchboard.getOverview().root.lift()
    return None

  def pack(self,*positionArgsTuple,**kwArgsDict):
    self.mbar.pack(*positionArgsTuple,**kwArgsDict)
    return None


class swdbAddressBook:
  def __init__(self):
    self.currState="table"
    self.editItem=None
    self.editIndex=None
    self.indirectView=None
    self.filterStr=""
    self.root=None
    self.searchDelayAlarmID=None
    self.showOnlyWithNicks=1
    filename=gPrefs.get("swdbFile")+".mk4"

    #if not os.path.exists(filename) or os.stat(filename).st_size]==0:
    if not os.path.exists(filename) or os.stat(filename).st_size==0:
      os.umask(0077) # Octal
      dbFile=metakit.storage(filename,1)
      ignored=dbFile.getas(kswdbMkStruct)
      dbFile.commit()

      oldFilePath=gPrefs.get("swdbFile")
      if os.path.exists(oldFilePath):
        self.convertDatabase()
    # If this is a new version and the structure has changed
    # update it. MetaKit is cool!
    else:
      dbFile=metakit.storage(filename,1)
      if dbFile.description("addressbook")<>kswdbMkStruct:
        dummy=dbFile.getas(kswdbMkStruct)
        dbFile.commit()

    self.dbFile=dbFile
    dbFile.autocommit()
    self.dbView=self.dbFile.view(kswdbDBViewName)
    self.mapFile=metakit.storage() # In memory, probably sufficient
    self.mapView=self.mapFile.getas("hash[_H:I,_R:I]") # Structure speecified by docco
    #self.hashView=self.dbView.hash(self.mapView,1)
    # Sometimes invalidated, re-created by the routine that needs them
    #self.mapView=None
    self.hashView=None
    self.filterDB()
    return None

  def show(self):
    if self.getState()=="hidden":
      self.createWidgets()
      self.root.title(kswdbWindowTitle)
      self.searchEntry.delete(0,END)
      self.searchEntry.insert(0,self.filterStr)
      self.filterDB()
      self.setCountsLabel()
      self.tableContainer.redisplay()
    else:
      gTkRoot.lift()
    return None

  def hide(self):
    self.root.destroy()
    self.root=None
    return None

  def getState(self):
    if self.root==None:
      return "hidden"
    else:
      return self.currState

  def nickToAddr(self,nick):
    assert nick<>""
    # Always return a sequence even though we don't yet have a way to specify
    # a list of addresses.
    if nick=="me":
      return (gPrefs.get("myHumanName"),gPrefs.get("myMailAddress"))
    startAt=0
    while 1:
      ind=self.dbView.find(nick=nick,start=startAt)
      if ind==-1:
        return None
      elif self.dbView[ind].email<>"":
        return (self.dbView[ind].name,self.dbView[ind].email)
      else:
        startAt=ind+1
    return None

  # Uses the hash view to speed up lookup
  def addrToNickOrName(self,addr,whichToReturn):
    assert whichToReturn in ("nick","name")
    addr=addr.lower()

    if addr==gPrefs.get("myMailAddress").lower():
      if whichToReturn=="nick":
        return "me"
      elif whichToReturn=="name":
        return gPrefs.get("myHumanName")
      else:
        assert 1==0,"Impossible value for whichToReturn: %s" % whichToReturn

    if self.hashView==None:
      self.hashView=self.dbView.hash(self.mapView,1)

    startAt=0

    while 1:
      index=self.hashView.find(email=addr,start=startAt)
      if index==-1:
        return None

      if whichToReturn=="nick" and self.dbView[index].nick<>"":
        return self.dbView[index].nick
      elif whichToReturn=="name" and self.dbView[index].name<>"":
        return self.dbView[index].name

      startAt=index+1
    return None

  def addrToNick(self,addr):
    return self.addrToNickOrName(addr,"nick")

    #~ if addr.lower()==gPrefs.get("myMailAddress").lower():
      #~ return "me"
    #~ if self.hashView==None:
      #~ self.hashView=self.dbView.hash(self.mapView,1)
    #~ startAt=0
    #~ while 1:
      #~ index=self.hashView.find(email=addr,start=startAt)
      #~ if index==-1:
        #~ return None
      #~ elif self.dbView[index].nick<>"":
        #~ return self.dbView[index].nick
      #~ startAt=index+1
    #~ return None

  def addrToName(self,addr):
    return self.addrToNickOrName(addr,"name")

    #~ assert addr<>""
    #~ if addr.lower()==gPrefs.get("myMailAddress").lower():
      #~ return gPrefs.get("myHumanName")
    #~ if self.hashView==None:
      #~ self.hashView=self.dbView.hash(self.mapView,1)
    #~ startAt=0
    #~ while 1:
      #~ index=self.hashView.find(email=addr,start=startAt)
      #~ if index==-1:
        #~ return None
      #~ elif self.dbView[index].name<>"":
        #~ return self.dbView[index].name
      #~ startAt=index+1
    #~ return None

  def lift(self):
    if self.root<>None:
      self.root.lift()
    return None

  def newState(self,newState):
    if newState==self.currState:
      return None
    if self.currState==None:
      pass
    elif self.currState=="edit":
      self.editContainer.pack_forget()
    elif self.currState=="table":
      self.tableAndControlsFrame.pack_forget()
      # Doesn't work either, foo
      #self.root.wait_window(self.tableAndControlsFrame)

      # Occasionally the table won't disappear and the edit widgets will
      # just be added to the table in the window. Don't know why. Perhaps
      # letting Tkinter have a moment to think after the pack_forget()
      # will help.
      # OK, update_idletasks() didn't help. Let's see if update() will.
      #self.root.update_idletasks()
      #self.root.update()
    else:
      assert 1==0, "Error: current state %s not recognized" % self.currState

    if newState=="edit":
      self.editContainer.pack(side=TOP,fill=BOTH,expand=1)
      self.editWidgets["nick"].focus_set()
    elif newState=="table":
      self.tableAndControlsFrame.pack(side=TOP,fill=BOTH,expand=1)
      self.tableContainer.redisplay()
      self.searchEntry.focus_set()
    else:
      assert 1==0, "Error: new state %s not recognized" % newState

    self.currState=newState
    return None

  def createWidgets(self):
    self.root=Toplevel(height=kswdbWindowStartSize["height"],width=kswdbWindowStartSize["width"])
    self.editWidgets={}
    self.root.protocol("WM_DELETE_WINDOW",self.closeWinCB)
    self.root.pack_propagate(0)
    myFont=gPrefs.get("fontILike")
    hideMouse=gPrefs.get("hideMousePointerWhileTyping")
    self.tableAndControlsFrame=Frame(self.root)
    self.menuFrame=Frame(self.tableAndControlsFrame)
    self.mbar=swdbMenubar(self)
    self.mbar.pack(side=LEFT)
    self.menuFrame.pack(side=TOP,fill=X)
    line=Frame(self.tableAndControlsFrame,bg="black",height=1)
    line.pack(side=TOP,fill=X)
    spacer=Frame(self.tableAndControlsFrame,height=4,width=1)
    spacer.pack(side=TOP)
    controlsFrame=Frame(self.tableAndControlsFrame)
    spacer=Frame(controlsFrame,width=8,height=1)
    spacer.pack(side=LEFT)
    l=Label(controlsFrame,font=myFont,text="Search for:")
    l.pack(side=LEFT)
    entryBorder=Frame(controlsFrame,bg="black",borderwidth=1)
    self.searchEntry=Entry(entryBorder,
      font=gPrefs.get("fontILike"),
      bg="white",relief=FLAT,
      highlightthickness=0,
      width=30)
    if hideMouse:
      gMousePointerHider.makeWidgetHideMouseOnKey(self.searchEntry)
    self.searchEntry.pack(side=LEFT)
    entryBorder.pack(side=LEFT)
    spacer=Frame(controlsFrame,width=8,height=1)
    spacer.pack(side=RIGHT)
    self.countLabel=Label(controlsFrame,font=gPrefs.get("fontILike"),anchor=E)
    self.countLabel.pack(side=RIGHT,fill=X)
    controlsFrame.pack(side=TOP,fill=X)
    spacer=Frame(self.tableAndControlsFrame,height=4,width=1)
    spacer.pack(side=TOP)
    lineAbove=Frame(self.tableAndControlsFrame,bg="black",height=1)
    lineAbove.pack(side=TOP,fill=X)
    self.tableContainer=gridWithHeaders(self,
      tableLinesToShow=15,
      headerStrs=kswdbFieldLabels,
      defaultWidthsChars=kswdbDefaultColWidths,
      textGetCB=self.getTextList,
      contentsLenGetCB=self.getViewLen,
      doubleClickCB=self.doubleClickCB,
      singleClickCB=self.singleClickCB,
      click3CB=self.click3CB,
      font=myFont,
      )
    self.tableContainer.pack(side=TOP,fill=BOTH,expand=1)
    for l in self.tableContainer.getHeaderLabels():
      l.bind("<Button-3>",self.headerClick3CB)

    self.tableAndControlsFrame.pack(side=TOP,fill=BOTH,expand=1)
    self.editContainer=Frame(self.root)
    buttonsFrame=Frame(self.editContainer)
    saveButton=myButton(buttonsFrame,"Save",self.saveButtonCB,widthInChars=7)
    saveButton.pack(side=LEFT)
    abandonButton=myButton(buttonsFrame,"Abandon",self.abandonButtonCB,widthInChars=7)
    abandonButton.pack(side=RIGHT)
    spacerFrame=Frame(buttonsFrame,width=20)
    spacerFrame.pack(side=RIGHT)
    delButton=myButton(buttonsFrame,"Delete",self.delButtonCB,widthInChars=7)
    delButton.pack(side=RIGHT)
    buttonsFrame.pack(side=TOP,fill=X)
    fieldsFrame=Frame(self.editContainer)
    labelWidth=lengthOfLongestStr(kswdbFieldLabels)+1
    for count in range(len(kswdbFieldNames)-1):
      rowFrame=Frame(fieldsFrame)
      l=Label(rowFrame,
        text=kswdbFieldLabels[count],
        font=myFont,width=labelWidth,
        anchor=W,
        )
      l.pack(side=LEFT)
      line=Frame(rowFrame,width=1,bg="black")
      line.pack(side=LEFT,fill=Y)
      entryFrame=Frame(rowFrame)
      if count==0:
        line=Frame(entryFrame,height=1,bg="black")
        line.pack(side=TOP,fill=X,expand=1)
      w=Entry(entryFrame,
        font=myFont,
        highlightthickness=0,
        bg="white",
        width=20,
        relief=FLAT)
      w.bind("<FocusIn>",clearSelection)
      if hideMouse:
        gMousePointerHider.makeWidgetHideMouseOnKey(w)
      if kswdbFieldNames[count] in kswdbROFieldNames:
        w.configure(bg=kswdbROEntryBackgroundColor)
      w.pack(side=TOP,fill=X,expand=1)
      line=Frame(entryFrame,height=1,bg="black")
      line.pack(side=TOP,fill=X,expand=1)
      entryFrame.pack(side=LEFT,fill=X,expand=1)
      line=Frame(rowFrame,width=1,bg="black")
      line.pack(side=LEFT,fill=Y)
      self.editWidgets[kswdbFieldNames[count]]=w
      rowFrame.pack(side=TOP,fill=X,expand=1)
    rowFrame=Frame(fieldsFrame)
    subRowFrame=Frame(rowFrame)
    l=Label(subRowFrame,
      text=kswdbFieldLabels[-1],
      font=myFont,
      anchor=W,
      width=labelWidth)
    l.pack(side=TOP)
    subRowFrame.pack(side=LEFT,fill=Y)
    textFrame=Frame(rowFrame)
    line=Frame(textFrame,width=1,bg="black")
    line.pack(side=LEFT,fill=Y)
    w=Pmw.ScrolledText(textFrame,
      text_font=myFont,
      text_bg="white",
      text_height=15,
      text_width=50,
      text_relief=FLAT,
      text_highlightthickness=0,
      text_wrap="word",
      horizscrollbar_activebackground=kTkDefaultBG,
      vertscrollbar_activebackground=kTkDefaultBG,
      vertscrollbar_borderwidth=1,
      vertscrollbar_width=12,
      horizscrollbar_borderwidth=1,
      horizscrollbar_width=12,
      vertscrollbar_takefocus=0,
      vscrollmode="static",
      )
    textArea=w.component("text")
    textArea.bind("<Key-Tab>",focusNext)
    textArea.bind("Control-i>",focusNext)
    textArea.bind("<Shift-Tab>",focusPrev)
    w.pack(side=LEFT)
    line=Frame(textFrame,width=1,bg="black")
    line.pack(side=LEFT,fill=Y)
    textFrame.pack(side=TOP)
    rowFrame.pack(side=TOP)
    if hideMouse:
      gMousePointerHider.makeWidgetHideMouseOnKey(textArea)
    self.editWidgets["comments"]=w

    fieldsFrame.pack(side=TOP)
    self.searchEntry.bind("<KeyRelease>",self.searchKeyCB)
    self.searchEntry.focus_force()
    self.root.bind("<Key-Up>",self.upAndDownArrowCB)
    self.root.bind("<Key-Down>",self.upAndDownArrowCB)
    self.root.bind("<Control-Up>",self.controlArrowCB)
    self.root.bind("<Control-Down>",self.controlArrowCB)
    return None

  def getMenuWidgetParent(self):
    return self.menuFrame

  def upAndDownArrowCB(self,event):
    if event.keysym=="Up":
      self.tableContainer.moveSelection(-1)
    elif event.keysym=="Down":
      self.tableContainer.moveSelection(1)
    else:
      assert 1==0,"Key %s received by upAndDownArrowCB but it doesn't know how to deal with that character" % event.keysym
    return "break"

  def controlArrowCB(self,event):
    if event.keysym=="Up":
      self.tableContainer.scrollLines(-1)
    elif event.keysym=="Down":
      self.tableContainer.scrollLines(1)
    else:
      assert 1==0,"Key %s received by controlArrowCB but it doesn't know how to deal with that character" % event.keysym
    return "break"

  def getWidgetParent(self):
    return self.tableAndControlsFrame

  def makeDateFromInt(self,intDateAsStr):
    unixIntDate=int(intDateAsStr)
    if unixIntDate==kswdbNoDateVal:
      return ""
    timeTuple=time.localtime(unixIntDate)
    return time.strftime(kswdbDateFormat,timeTuple)

  def makeDateAndTimeFromInt(self,intDateAsStr):
    unixIntDate=int(intDateAsStr)
    if unixIntDate==kswdbNoDateVal:
      return ""
    timeTuple=time.localtime(unixIntDate)
    return time.strftime(kswdbDateTimeFormat,timeTuple)

  def getTextList(self,index):
    r=self.dbView[self.indirectView[index].index]
    vals=[ getattr(r,fn) for fn in kswdbFieldNames ]
    vals[3]=self.makeDateFromInt(vals[3])
    vals[4]=self.makeDateFromInt(vals[4])
    vals[5]=self.makeDateFromInt(vals[5])
    vals[6]=vals[6].replace("\n"," ") # Remove newlines from comments
    return vals

  def getViewLen(self):
    return len(self.indirectView)

  def doubleClickCB(self,index):
    self.editRecord(index)
    return None

  # XXX ??? User must click to get address on clibpoard.
  # XXX ??? Should we set it to the highlighted value always?
  def singleClickCB(self,index):
    self.setClipboardToDataInRow(index)
    return None

  def setClipboardToDataInRow(self,rowIndex):
    row=self.dbView[self.indirectView[rowIndex].index]
    addr=canonicalizeAddrList([(row.name,row.email)])[0]
    gTkRoot.selection_own()
    gTkRoot.selection_handle(clipboardCBO(addr))
    return None

  def sortCB(self,colNo):
    self.mapView=self.mapView[0:0] # Empty the map view
    self.hashView=None
    self.dbView=self.dbView.sort(kswdbFieldNames[colNo])
    self.filterDB()
    self.setCountsLabel()
    self.tableContainer.redisplay()
    return None

  def filterDBFunc(self,withNicksOnly,filterStr,row):
    if withNicksOnly==1 and row.nick=="":
      return 0
    if filterStr<>"" and (
      row.nick.lower().find(filterStr)==-1 and
      row.email.lower().find(filterStr)==-1 and
      row.name.lower().find(filterStr)==-1 and
      row.comments.lower().find(filterStr)==-1
      ):
      return 0
    return 1

  def filterDB(self):
    withNicksOnly=self.showOnlyWithNicks
    filterStr=self.filterStr
    filterCBO=genericCBO(self.filterDBFunc,withNicksOnly,filterStr)
    self.indirectView=self.dbView.filter(filterCBO)
    return None

  def setCountsLabel(self):
    self.countLabel.configure(text="%i shown of %i" %
      (len(self.indirectView),len(self.dbView)))
    return None

  def newMailToCB(self,addr):
    gSwitchboard.newMail(addr)
    return None

  def copyAddr(self,addr):
    myCBO=clipboardCBO(addr)
    gTkRoot.selection_own()
    gTkRoot.selection_handle(myCBO)
    return None

  def click3CB(self,colIndex,rowIndex,event):
    row=self.dbView[self.indirectView[rowIndex].index]
    addr=canonicalizeAddrList([(row.name,row.email)])[0]
    myFont=gPrefs.get("fontILike")
    m=Menu(gTkRoot,tearoff=0,activebackground=kTkDefaultBG)

    m.add_command(label="Sort on field: %s" % self.tableContainer.headerStrs[colIndex],
      command=genericCBO(self.sortCB,colIndex),font=myFont)

    m.add_command(label="New mail to: %s" % addr,
      command=genericCBO(self.newMailToCB,addr),font=myFont)

    m.add_command(label="Copy to clipboard: %s" % addr,
      command=genericCBO(self.copyAddr,addr),font=myFont)

    m.post(event.x_root,event.y_root)
    gTkRoot.after(250,m.grab_set)
    return "break"

  def headerClick3CB(self,event):
    labels=[ l.interior() for l in self.tableContainer.getHeaderLabels() ]
    colIndex=labels.index(event.widget)

    myFont=gPrefs.get("fontILike")
    m=Menu(gTkRoot,tearoff=0,activebackground=kTkDefaultBG)

    m.add_command(label="Sort on field: %s" % self.tableContainer.headerStrs[colIndex],
      command=genericCBO(self.sortCB,colIndex),font=myFont)
    m.post(event.x_root,event.y_root)
    # Delay to ignore the first mouse-up so that menu stays visible when
    # user releases mouse click that opened it. There's prolly a better
    # way.
    gTkRoot.after(250,m.grab_set)
    return "break"

  def newRecord(self):
    self.editRecord(None)
    return None

  def cleanUpForQuit(self):
    self.dbFile.commit()
    self.mapView=None
    self.hashView=None
    self.dbView=None
    self.dbFile=None
    return None

  def closeWinCB(self):
    self.root.destroy()
    self.currState=None
    self.root=None
    return None

  # User may type several chars right after one another and search is slow enough that
  # it would be annoying to user to search on each as typed, especially if the first
  # was common. So delay actual searching until there's a pause.
  def searchKeyCB(self,keyIgnored):
    if self.searchDelayAlarmID<>None:
      gTkRoot.after_cancel(self.searchDelayAlarmID)
    self.searchDelayAlarmID=gTkRoot.after(kswdbSearchDelayAfterKeystrokeMS,self.doSearch)
    return None

  def doSearch(self):
    self.root.configure(cursor="watch")
    self.filterStr=self.searchEntry.get().lower()
    self.filterDB()
    self.setCountsLabel()
    # If we've shotened the list, see that we're not off the bottom.
    self.tableContainer.redisplay()
    self.root.configure(cursor="")
    return None

  def nicksCheckBoxChanged(self,val):
    self.showOnlyWithNicks=val
    self.root.configure(cursor="watch")
    # Redraw the button immediately.
    gTkRoot.update() # update_idletasks() doesn't work, don't know why
    self.filterDB()
    self.setCountsLabel()
    self.tableContainer.rowAtTop=0
    self.tableContainer.redisplay()
    self.root.configure(cursor="")
    return None

  # Not necessarily best to send list of headers
  def snarf(self,listOfHeaders,direction):
    removeAny(listOfHeaders,"")
    if self.hashView==None:
      self.hashView=self.dbView.hash(self.mapView,1)
    for header in listOfHeaders:
      al=email.Utils.getaddresses([header])
      for name,addr in al:
        if addr==None:
          continue
        name=stripQuotes(name)
        name=name.strip()
        addr=addr.lower() # Not technically correct but unlikely to cause a problem in practice
        ind=self.hashView.find(email=addr)
        if ind==-1:
          newItem={"email": addr,
                   "name": name,
                   "dateCreated": int(time.time()),
                   "dateLastMailedTo": 0,
                   "dateLastHeardFrom": 0}
          if direction=="in":
            newItem["dateLastHeardFrom"]=int(time.time())
          elif direction=="out":
            newItem["dateLastMailedTo"]=int(time.time())
          else:
            assert 1==0,"Impossible value for direction: %s" % direction
          self.hashView.append(newItem)
        else:
          row=self.hashView[ind]
          if row.name=="":
            row.name=name
          if direction=="in":
            row.dateLastHeardFrom=int(time.time())
          elif direction=="out":
            row.dateLastMailedTo=int(time.time())
          else:
            assert 1==0,"Impossible value for direction: %s" % direction
    return None

  def showInEditor(self,textAddress):
    addrTuple=email.Utils.parseaddr(textAddress)
    assert addrTuple[1]<>None
    ind=self.dbView.find(email=addrTuple[1])
    assert ind<>-1
    self.editRecord(ind,indirect=0) ###XXX
    return None

  def editRecord(self,index,indirect=1):
    self.editIndex=index
    if self.editIndex==None:
      item={"nick": "",
            "name": "",
            "email": "",
            "dateLastHeardFrom": 0,
            "dateLastMailedTo": 0,
            "dateCreated": int(time.time()),
            "comments": ""}
      self.editItem=item
    else:
      if indirect:
        r=self.dbView[self.indirectView[index].index]
      else:
        r=self.dbView[index]
      item={"nick": r.nick,
            "name": r.name,
            "email": r.email,
            "dateLastHeardFrom": r.dateLastHeardFrom,
            "dateLastMailedTo": r.dateLastMailedTo,
            "dateCreated": r.dateCreated,
            "comments": r.comments}
      self.editItem=r

    self.newState("edit")
    self.editWidgets["nick"].delete(0,END)
    self.editWidgets["nick"].insert(0,item["nick"])

    self.editWidgets["name"].delete(0,END)
    self.editWidgets["name"].insert(0,item["name"])

    self.editWidgets["email"].delete(0,END)
    self.editWidgets["email"].insert(0,item["email"])

    self.editWidgets["dateLastHeardFrom"].configure(state=NORMAL)
    self.editWidgets["dateLastHeardFrom"].delete(0,END)
    self.editWidgets["dateLastHeardFrom"].insert(0,self.makeDateAndTimeFromInt(item["dateLastHeardFrom"]))
    self.editWidgets["dateLastHeardFrom"].configure(state=DISABLED)

    self.editWidgets["dateLastMailedTo"].configure(state=NORMAL)
    self.editWidgets["dateLastMailedTo"].delete(0,END)
    self.editWidgets["dateLastMailedTo"].insert(0,self.makeDateAndTimeFromInt(item["dateLastMailedTo"]))
    self.editWidgets["dateLastMailedTo"].configure(state=DISABLED)

    self.editWidgets["dateCreated"].configure(state=NORMAL)
    self.editWidgets["dateCreated"].delete(0,END)
    self.editWidgets["dateCreated"].insert(0,self.makeDateAndTimeFromInt(item["dateCreated"]))
    self.editWidgets["dateCreated"].configure(state=DISABLED)

    self.editWidgets["comments"].clear()
    self.editWidgets["comments"].settext(item["comments"])
    return None

  def saveButtonCB(self):
    self.mapView=self.mapView[0:0] # Empty the map view
    self.hashView=None

    # It's not technically correct to smash the address to lowercase
    # but it's unlikely to cause a problem in practice and it means
    # that an address-to-nick or address-to-name lookup done by key
    # (and therefore fast) will match regardless of case.

    if self.editIndex==None:
      newItem=self.editItem
      newItem["nick"]=self.editWidgets["nick"].get().strip()
      newItem["name"]=self.editWidgets["name"].get().strip()
      newItem["email"]=self.editWidgets["email"].get().strip().lower()
      newItem["comments"]=self.editWidgets["comments"].get("1.0",END).strip()
      self.dbView.append(newItem)
    else:
      r=self.editItem
      r.name=self.editWidgets["name"].get().strip()
      r.nick=self.editWidgets["nick"].get().strip()
      r.email=self.editWidgets["email"].get().strip().lower()
      r.comments=self.editWidgets["comments"].get("1.0",END).strip()
    self.filterDB()
    self.setCountsLabel()
    self.dbFile.commit()
    self.newState("table")
    return None

  def abandonButtonCB(self):
    self.newState("table")
    return None

  def delButtonCB(self):
    if self.editIndex<>None:

      self.mapView=self.mapView[0:0] # Empty the map view
      self.hashView=None

      self.dbView.delete(self.indirectView[self.editIndex].index)

      self.dbFile.commit()
      self.filterDB()
    self.newState("table")
    self.tableContainer.setHighlightIndex(None)
    self.tableContainer.redisplay()
    return None

  def delUninteresting(self):
    # XXX Should we compress the file?

    # XXX Prolly need to do some mutual exlusion here and elsewhere
    # XXX since fetch thread can write to the file
    self.root.configure(cursor="watch")
    self.root.update()
    self.dbFile.commit()
    filename=gPrefs.get("swdbFile")+".mk4"
    backupFilename=filename+".bak"
    os.umask(0177) # octal
    gLogger.add("Backing up swdb file %s before deleting uninteresting records" % filename)
    f=open(backupFilename,"wb")
    self.dbFile.save(f)
    f.close()
    gLogger.add("File successfully backed up to %s" % backupFilename)
    interestingFunc=gPrefs.get("swdbShouldKeepFunc")
    count=0
    all=0
    kept=0
    deleted=0
    while count<len(self.dbView):
      all+=1
      r=self.dbView[count]
      if interestingFunc(r):
        kept+=1
        count+=1
      else:
        self.dbView.delete(count)
        deleted+=1
    self.dbFile.commit()
    self.filterDB()
    self.setCountsLabel()
    gLogger.add("%i records previously, %i deleted, %i kept" % (all,deleted,kept))
    self.root.configure(cursor="")
    return None

  def importAddresses(self):
    global gLastSaveOrOpenDir
    path=tkFileDialog.askopenfilename(initialdir=gLastSaveOrOpenDir)
    if path==None or path=="":
      return None
    gLastSaveOrOpenDir=myPathSplit(path)[0]
    gLogger.add("Checking data in import file %s before importing" % path)
    numFields=len(kswdbFieldLabels)
    f=open(path,"rb")
    error=0
    while 1:
      l=f.readline()
      if l=="":
        break
      l=l[:-len(kFileLineSep)]
      fields=l.split("\t")
      if len(fields)>numFields:
        gLogger.add("Import found an error in the file %s, import not done" % path)
        gLogger.add('Line "%s" has %i %s but I want only %i' %
          (l,len(fields),pluralize(len(fields),"field"),numFields),kError)
        error=1
        break
      if len(fields)<numFields:
        gLogger.add("Import found an error in the file %s, import not done" % path)
        gLogger.add('Line "%s" has %i %s but I need %i' %
          (l,len(fields),pluralize(len(fields),"field"),numFields),kError)
        error=1
        break
      for count in range(3,6):
        v=stripQuotes(fields[count])
        try:
          ignored=float(v)
        except ValueError:
          gLogger.add("Import found an error in the file %s, import not done" % path)
          gLogger.add('In line "%s", value "%s" (%s field) should be numeric but isn\'t' %
            (l,v,kswdbFieldLabels[count]),kError)
          error=1
          break

    if error:
      d=Pmw.MessageDialog(title="Error",
        message_font=gPrefs.get("fontILike"),
        dialogchildsite_background="white",
        message_background="white",
        message_text="There is an error in the file to be imported. Details are in the log.")
      d.component("buttonbox").component("OK").configure(activebackground=kTkDefaultBG)
      return None


    self.mapView=self.mapView[0:0] # Empty the map view
    self.hashView=None

    added=0
    f.seek(0)
    while 1:
      l=f.readline()
      if l=="":
        break
      l=l[:-len(kFileLineSep)]
      fields=l.split("\t")
      d={}
      d["name"]=stripQuotes(fields[0])
      d["email"]=stripQuotes(fields[1])
      d["nick"]=stripQuotes(fields[2])
      d["dateCreated"]=int(float(stripQuotes(fields[3]))) # int(float()) to allow a decimal point
      d["dateLastHeardFrom"]=int(float(stripQuotes(fields[4])))
      d["dateLastMailedTo"]=int(float(stripQuotes(fields[5])))
      d["comments"]=stripQuotes(fields[6])
      self.dbView.append(d)
      added+=1

    self.dbFile.commit()
    f.close()
    gLogger.add("%i records added to SWDB" % added)
    return None

  def exportAddresses(self):
    global gLastSaveOrOpenDir
    toFP=tkFileDialog.asksaveasfile(initialdir=gLastSaveOrOpenDir,initialfile=kAddressExportDefaultName)
    if toFP==None:
      return None

    toFP.write(("%s\t%s\t%s\t%s\t%s\t%s\t%s"+kFileLineSep) % kswdbFieldLabels)
    for row in self.dbView:
      comments=row.comments
      if comments.find("\t")<>-1 or comments.find("\r")<>-1 or comments.find("\n")<>-1:
        comments=kDoubleQuote+escapeDoubleQuotes(comments)+kDoubleQuote
      toFP.write(("%s\t%s\t%s\t%s\t%s\t%s\t%s"+kFileLineSep) %
        (row.nick,
        row.name,
        row.email,
        row.dateCreated,
        row.dateLastHeardFrom,
        row.dateLastMailedTo,
        comments))
    toFP.close()
    gLogger.add("SWDB addressbook exported to %s" % toFP.name)
    gLastSaveOrOpenDir=myPathSplit(toFP.name)[0]
    return None

  def convertDatabase(self):
    convertableVersionStrs=["Version: 1","Version: 2","Version: 3"]

    oldFilePath=gPrefs.get("swdbFile")
    newFilePath=oldFilePath+".mk4"

    oldFile=open(oldFilePath,"rb")
    versionStr=oldFile.readline().rstrip()
    oldFile.close()

    assert versionStr in convertableVersionStrs

    backupFilePath=oldFilePath+".old-version"
    os.rename(oldFilePath,backupFilePath)

    gLogger.add("File %s backed up as %s" % (oldFilePath,backupFilePath))

    rf=open(backupFilePath,"rb")
    ignoreVersionLine=rf.readline()

    dbFile=metakit.storage(newFilePath,1)
    dbView=dbFile.view(kswdbDBViewName)

    count=0

    while 1:
      inL=rf.readline()
      if inL=="":
        break

      count+=1

      inL=inL[:-1]
      fields=inL.split("\t")
      if len(fields)<5:
        fields=fields+([""] * (5-len(fields)))
      if len(fields)==5:
        fields.insert(4,"0")
      if len(fields)==6:
        fields.insert(5,"0")
      d={}
      d["name"]=fields[0]
      d["email"]=fields[1]
      d["nick"]=fields[2]
      d["dateCreated"]=int(fields[3])
      d["dateLastHeardFrom"]=int(fields[4])
      d["dateLastMailedTo"]=int(fields[5])
      d["comments"]=fields[6]
      dbView.append(d)

    dbFile.commit()
    rf.close()
    gLogger.add("File %s created with %i records" % (newFilePath,count))
    return None


#
# mail viewer
#

class savePart:
  def __init__(self,partInfo):
    self.partInfo=partInfo
    return None

  def __call__(self):
    global gLastSaveOrOpenDir
    toFP=tkFileDialog.asksaveasfile(initialdir=gLastSaveOrOpenDir,initialfile=self.partInfo.name)
    if toFP<>None:
      toFP.write(self.partInfo.decodedPart)
      toFP.close()
      gLastSaveOrOpenDir=myPathSplit(toFP.name)[0]
    return None


class mailViewer:
  def __init__(self,path,refToCreator,moveToCur=1):
    self.root=Toplevel()
    self.root.protocol("WM_DELETE_WINDOW",self.closeBoxCB)
    self.root.bind("<Key>",self.keyCB)
    self.refToCreator=refToCreator
    self.moveToCur=moveToCur
    self.path=None
    self.showAllHeaders=0
    self.searchStr=""
    self.numMbNames=0
    self.charsToSearchRoutine=0
    #self.headersWidgetPackData=None
    self.partsInfo=[]
    self.interestingHeadersOnly=IntVar()
    self.decodeStyle=StringVar()
    self.decodeStyle.set("mime") # Also "nodecode" and "uudecode"
    self.deHTML=IntVar()
    self.createWidgets()
    self.loadNewMail(path)
    gSwitchboard.addViewer(self)
    return None

  def createWidgets(self):
    myFont=gPrefs.get("fontILike")
    topFrame=Frame(self.root)
    self.mbar=mainMenubar(topFrame)
    self.mbar.beginAddingMenu()

    self.mbar.mbar.addmenu("Viewer","",font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.component("Viewer-menu").configure(activebackground=kTkDefaultBG)

    self.mbar.mbar.addmenuitem("Viewer","checkbutton",command=self.headersMenuCB,label="Interesting headers only",
      variable=self.interestingHeadersOnly,selectcolor="blue",font=myFont)
    self.mbar.mbar.addmenuitem("Viewer","checkbutton",command=self.deHtmlMenuCB,label="Strip HTML tags",
      variable=self.deHTML,selectcolor="blue",font=myFont)

    self.mbar.mbar.addmenuitem("Viewer","separator")

    self.mbar.mbar.addmenuitem("Viewer","radiobutton",command=self.decodeMenuCB,label="MIME decode",
      variable=self.decodeStyle,value="mime",selectcolor="blue",font=myFont)
    self.mbar.mbar.addmenuitem("Viewer","radiobutton",command=self.decodeMenuCB,label="UUdecode",
      variable=self.decodeStyle,value="uudecode",selectcolor="blue",font=myFont)
    self.mbar.mbar.addmenuitem("Viewer","radiobutton",command=self.decodeMenuCB,label="No decode",
      variable=self.decodeStyle,value="nodecode",selectcolor="blue",font=myFont)

    self.mbar.mbar.addmenuitem("Viewer","separator")

    self.mbar.mbar.addcascademenu("Viewer","Reply",font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addmenuitem("Reply","command",label="Reply",command=self.replyMenuCB,font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addmenuitem("Reply","command",label="Reply to all",command=self.replyAllMenuCB,font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addmenuitem("Reply","command",label="Reply to To",command=self.replyListMenuCB,font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addcascademenu("Viewer","Forward",font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addmenuitem("Forward","command",label="Forward",command=self.mimeForwardCB,font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addmenuitem("Forward","command",label="Forward verbatim",command=self.verbatimForwardCB,font=myFont,activebackground=kTkDefaultBG)

    self.mbar.mbar.addmenuitem("Viewer","separator")

    self.mbar.mbar.addcascademenu("Viewer","Move to",font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addcascademenu("Viewer","Copy to",font=myFont,activebackground=kTkDefaultBG)
    self.updateMailboxNamesInMenu()
    self.mbar.mbar.addmenuitem("Viewer","command",command=self.saveMenuCB,label="Save message to file",font=myFont)
    self.mbar.mbar.addcascademenu("Viewer","Save parts",font=myFont,activebackground=kTkDefaultBG)


    self.mbar.doneAddingMenu()
    self.mbar.pack(side=LEFT)

    nextButton=myButton(topFrame,"Next",self.nextButtonCB,widthInChars=6)
    nextButton.pack(side=RIGHT)

    prevButton=myButton(topFrame,"Prev",self.prevButtonCB,widthInChars=6)
    prevButton.pack(side=RIGHT)

    spacerFrame=Frame(topFrame,width=2,height=2)
    spacerFrame.pack(side=RIGHT)

    # This produces a blink when the window is opened. It seems that Tkinter
    # draws the frame (all black) and then draws the label on top. Not obvious
    # how to avoid. Tried doing the color later but that doesn't seem to help.
    xOfYLabelBorder=Frame(topFrame,borderwidth=1,bg="black")
    xOfYLabel=Label(xOfYLabelBorder,width=9,font=gPrefs.get("fontILike"))
    xOfYLabel.pack(side=LEFT)
    xOfYLabelBorder.pack(side=RIGHT)
    self.xOfYLabel=xOfYLabel

    topFrame.pack(side=TOP,fill=X)

    # midFrame is only packed when needed
    self.midFrame=Frame(self.root)
    spacerFrame=Frame(self.midFrame,width=2,height=2)
    spacerFrame.pack(side=LEFT)
    self.searchLabelBorder=Frame(self.midFrame,borderwidth=1)
    self.searchLabelBorder.pack(side=LEFT)
    self.searchLabel=Label(self.searchLabelBorder,padx=2,font=gPrefs.get("fontILike"),anchor=W)
    self.searchLabel.pack(side=LEFT)

    #headersWidget=Text(self.root,
    #  width=kMailViewerTextWidth,
    #  bg="white",
    #  font=gPrefs.get("fixedWidthFontILike"),
    #  borderwidth=1,
    #  height=len(gPrefs.get("nonBoringHeaders")),
    #  )
    #headersWidget.pack(side=TOP,fill=X,expand=1)

    #self.headersWidget=headersWidget

    textWidget=Pmw.ScrolledText(self.root,
      vscrollmode="static",
      hscrollmode="none",
      text_wrap="word",
      text_width=kMailViewerTextWidth,
      vertscrollbar_activebackground=kTkDefaultBG,
      text_bg="white",
      text_font=gPrefs.get("fixedWidthFontILike"),
      text_borderwidth=1,
      text_height=30,
      vertscrollbar_borderwidth=1,
      vertscrollbar_width=12,
      horizscrollbar_borderwidth=1,
      horizscrollbar_width=12,
      )
    textWidget.pack(side=TOP,fill=BOTH,expand=1)
    self.textWidget=textWidget

    #d=headersWidget.pack_info()
    #d["before"]=textWidget
    #self.headersWidgetPackData=d

    return None

  def loadNewMail(self,path):
    self.path=path
    self.interestingHeadersOnly.set(1)
    self.decodeStyle.set("mime")
    self.deHTML.set(0)
    self.mailObj=mailObj(path,loadTheFileNow=1)
    if self.mailObj.whichOfThreeDirs()=="new" and self.moveToCur:
      self.mailObj.moveToCur()
      gSwitchboard.subdirChanged(path,self.mailObj.path)
      gSwitchboard.change(self.mailObj.getMailboxName(),[self.mailObj.path])
    fromText=self.mailObj.get("from","")
    fromText=encodedHeaderToLatin1(fromText)
    fromText=addrTextToDisplayText(fromText)
    self.root.title("From %s (%s)" % (fromText,os.path.basename(path)))
    self.getText()
    self.setXOfYLabel()
    self.resetSearch()
    return None

  def getPath(self):
    return self.mailObj.path

  # For mail moving object, needs to match mailbox window method
  def getPathForMove(self):
    return self.getPath()

  def updateMailboxNamesInMenu(self):
    self.mbar.mbar.deletemenuitems("Move to",0,999)
    self.mbar.mbar.deletemenuitems("Copy to",0,999)
    saveNames=["(Default per prefs)"]
    if gPrefs.get("getSaveFoldersInboundAlt"):
      saveNames+=["(Alternate per prefs)"]
    saveNames+=gAllMailboxes.getNames()
    myFont=gPrefs.get("fontILike")
    for name in saveNames:
      self.mbar.mbar.addmenuitem("Move to","command",
        label=name,command=moveMailToOtherMailbox(name,"move",self),
          font=myFont,activebackground=kTkDefaultBG)
      self.mbar.mbar.addmenuitem("Copy to","command",
        label=name,command=moveMailToOtherMailbox(name,"copy",self),
          font=myFont,activebackground=kTkDefaultBG)
    return None

  def setXOfYLabel(self):
    num=self.refToCreator.howmanyItems()
    myIndex=self.refToCreator.getIndexByPath(self.mailObj.path)
    if myIndex==None:
      t=""
    else:
      myIndex=str(myIndex+1)
      t=("%s/%i" % (myIndex,num)).rjust(9)
    self.xOfYLabel.configure(text=t)
    return None

  def decodeMenuCB(self):
    # Menu does the toggle by itself. Sigh.
    self.getText()
    return None

  def deHtmlMenuCB(self):
    self.getText()
    return None

  def headersMenuCB(self):
    self.getText()
    return None

  def headersKeyCB(self):
    self.interestingHeadersOnly.set(not self.interestingHeadersOnly.get())
    self.getText()
    return None

  def keyCB(self,event):
    if self.charsToSearchRoutine and event.char<>"":
      self.anotherSearchChar(event.char)
    elif event.char=="f":
      self.mimeForwardCB()
    elif event.char=="F":
      self.verbatimForwardCB()
    elif event.char=="r":
      self.doReply()
    elif event.char=="R":
      self.doReply(replyAddr="all")
    elif event.char=="l":
      self.doReply(replyAddr="to")
    elif event.char=="m":
      gSwitchboard.newMail()
    elif event.char=="h":
      self.headersKeyCB()
    elif event.char=="w": # like Mac
      self.closeMe()
    elif event.char=="i": # like mutt
      self.closeMe()
    elif event.char=="S":
      self.saveToFile()
    elif event.keysym=="Left":
      self.prevMail()
    elif event.keysym=="Right":
      self.nextMail()
    elif event.keysym=="Down":
      self.scrollDownOneLine()
    elif event.keysym=="Up":
      self.scrollUpOneLine()
    elif event.char==" ":
      self.scrollDownOnePage()
    elif event.char=="u":
      self.scrollUpOnePage()
    elif event.char=="-":
      self.scrollUpOnePage()
    elif event.char=="P":
      self.printCB()
    elif event.char=="/":
      self.startSearchMode()
    elif event.char=="|":
      self.pipeCB()
    elif event.char=="b":
      self.raiseParentWindow()
    elif event.char=="t":
      self.raiseRootWindow()
    return "break"

  def raiseParentWindow(self):
    mailbox=getMailboxFromPath(self.mailObj.getPath())
    win=gSwitchboard.getBrowserByBox(mailbox)
    if win<>None:
      win.lift()
    return None

  def raiseRootWindow(self):
    gTkRoot.lift()
    return None

  def anotherSearchChar(self,char):
    if char=="\r":
      if self.searchStr=="":
        self.resetSearch()
      else:
        self.charsToSearchRoutine=None
    else:
      if ord(char)==8:
        self.searchStr=self.searchStr[:-1]
        self.searchLabel.configure(text="Search for: %s" % self.searchStr)
      else:
        self.searchStr=self.searchStr+char
        self.searchLabel.configure(text="Search for: %s" % self.searchStr)
      self.textWidget.tag_delete("search")
      if len(self.searchStr)>0:
        nFound=0
        commentaryRanges=self.textWidget.tag_ranges("commentary")
        self.textWidget.tag_config("search",foreground=gPrefs.get("textSearchColors")[0], \
          background=gPrefs.get("textSearchColors")[1])
        searchStart="1.0"
        while 1:
          where=self.textWidget.search(self.searchStr,searchStart,nocase=1,stopindex=END)
          if where=="" or nFound>kMaxViewerStringSearchSuccesses:
            break
          for count in range(0,len(commentaryRanges),2):
            if self.textWidget.compare(where,">=",commentaryRanges[count]) and \
               self.textWidget.compare(where,"<=",commentaryRanges[count+1]):
                 break
          else:
            self.textWidget.tag_add("search",where,"%s + %i chars" % (where,len(self.searchStr)))
            nFound+=1
          searchStart=where+"+ 1 chars"

        if nFound<=kMaxViewerStringSearchSuccesses:
          self.searchLabel.configure(text="Search for: %s (%i)" % (self.searchStr,nFound))
        else:
          self.searchLabel.configure(text="Search for: %s (first %i highlighted)" % \
            (self.searchStr,kMaxViewerStringSearchSuccesses))
    return None

  def startSearchMode(self):
    self.midFrame.pack(side=TOP,fill=X,before=self.textWidget)
    self.textWidget.tag_delete("search")
    self.searchLabel.configure(text="Search for: ")
    self.searchLabelBorder.configure(bg="black")
    self.searchStr=""
    self.charsToSearchRoutine=1
    return None

  def resetSearch(self):
    self.charsToSearchRoutine=None
    self.searchLabel.configure(text="")
    self.searchLabelBorder.configure(bg=kTkDefaultBG)
    self.midFrame.pack_forget()
    return None

  def saveMenuCB(self):
    self.saveToFile()
    return None

  def saveToFile(self):
    global gLastSaveOrOpenDir
    newFP=tkFileDialog.asksaveasfile(initialdir=gLastSaveOrOpenDir)
    if newFP==None:
      return None
    newFP.write(self.mailObj.getText())
    newFP.close()
    gLastSaveOrOpenDir=myPathSplit(newFP.name)[0]
    return None

  def printCB(self):
    cmd=gPrefs.get("printCommand")
    gLogger.add("Sending mail viewer contents to %s" % cmd)
    t=self.textWidget.get()
    # XXX removed for testing log analyzer
    # t=hardWrap(t,gPrefs.get("outboundMailWrapLen"),useFileLineSep=1)
    p=os.popen(cmd,"w")
    p.write(t)
    p.close()
    gLogger.add("Printing done")
    return None

  def pipeCB(self):
    cmd=gPrefs.get("pipeCommand")
    if cmd==None:
      return None
    gLogger.add("Sending mail viewer contents to %s" % cmd)
    # I use popen2 because if I use popen, this code waits
    # at the pipe close for the child to exit and that isn't
    # what I want. The read pipe isn't used.
    readPipe,writePipe=popen2.popen2(cmd)
    writePipe.write(self.mailObj.getText())
    writePipe.close()
    readPipe.close()
    gLogger.add("Pipe done")
    return None

  def scrollDownOnePage(self):
    self.textWidget.yview("scroll","+1","pages")
    return None

  def scrollUpOnePage(self):
    self.textWidget.yview("scroll","-1","pages")
    return None

  def scrollDownOneLine(self):
    self.textWidget.yview("scroll","+1","units")
    return None

  def scrollUpOneLine(self):
    self.textWidget.yview("scroll","-1","units")
    return None

  def replyMenuCB(self):
    self.doReply()
    return None

  def replyAllMenuCB(self):
    self.doReply(replyAddr="all")
    return None

  def replyListMenuCB(self):
    self.doReply(replyAddr="to")
    return None

  def doReply(self,replyAddr=None):
    self.mailObj.showInEditorNewReply(replyAddr,deHTML=self.deHTML.get())
    return None

  def verbatimForwardCB(self):
    self.mailObj.showInEditorNewForward(decodeStyle="nodecode")
    return None

  def mimeForwardCB(self):
    self.mailObj.showInEditorNewForward(decodeStyle=self.decodeStyle.get(),deHTML=self.deHTML.get())
    return None

  def nextButtonCB(self):
    self.nextMail()
    return None

  def nextMail(self):
    ind=self.refToCreator.getIndexByPath(self.mailObj.path)
    if ind<>None:
      ind+=1
      if ind<=self.refToCreator.howmanyItems()-1:
        # Note that we don't want to re-use that mail object:
        # it deliberately doesn't load the text and we do.
        self.loadNewMail(self.refToCreator.getPathByIndex(ind))
    return None

  def prevButtonCB(self):
    self.prevMail()
    return None

  def prevMail(self):
    ind=self.refToCreator.getIndexByPath(self.mailObj.path)
    if ind<>None:
      ind-=1
      if ind>=0:
        self.loadNewMail(self.refToCreator.getPathByIndex(ind))
    return None

  def getText(self):
    textWidgetContents=self.textWidget.component("text")
    textWidgetContents.configure(state=NORMAL)
    self.textWidget.clear()

    #self.headersWidget.configure(state=NORMAL) ### XXX
    #self.headersWidget.delete("1.0",END) ### XXXX

    # Headers
    adjustCommentaryLineNumbers=0
    if not self.interestingHeadersOnly.get():

      # It seems that there's no easy way to find if the widget is packed
      #try:
      #  self.textWidget.pack_forget()
      #except TclError:
      #  pass

      headers=self.mailObj.getHeadersText()
      self.textWidget.insert(END,headers)
      adjustCommentaryLineNumbers+=headers.count(kFileLineSep)
      #self.headersWidget.configure(height=0) ### XXX
    else:
      # Tidy headers
      tidyHeaders,ranges=self.mailObj.getTidyHeaders(includeRanges=1)
      #try:
      #  self.headersWidget.pack(self.headersWidgetPackData)
      #except TclError:
      #  pass # Must already be packed
      #self.headersWidget.configure(state=NORMAL)
      #self.headersWidget.delete("1.0",END)
      self.textWidget.insert(END,tidyHeaders)
      #self.headersWidget.insert(END,tidyHeaders)
      #self.headersWidget.configure(state=DISABLED)###,height=tidyHeaders.count(kFileLineSep))
      if gPrefs.get("swdbFile")<>None:
        for count in range(len(ranges)):
          thisRange=ranges[count]
          name="address%i" % count
          tkStart="%i.%i" % (thisRange.lineNo+1,thisRange.start)
          tkEnd="%i.%i" % (thisRange.lineNo+1,thisRange.end)
          self.textWidget.tag_add(name,tkStart,tkEnd)
          self.textWidget.tag_bind(name,"<Button-3>",
            genericCBO(self.mouse3Address,thisRange.text))
          self.textWidget.tag_bind(name,"<Enter>",genericCBO(self.mouseEntersAddress,name))
          self.textWidget.tag_bind(name,"<Leave>",genericCBO(self.mouseLeavesAddress,name))
      adjustCommentaryLineNumbers+=tidyHeaders.count(kFileLineSep)

    # End of headers, blank line
    self.textWidget.insert(END,"\n")
    adjustCommentaryLineNumbers+=1

    # Body
    decodeStyle=self.decodeStyle.get()
    if decodeStyle=="nodecode":
      commentaryLineNumbers=[]
      textForViewer=self.mailObj.getBodyText()
      if self.deHTML.get():
        textForViewer=stripHTML(textForViewer)
    elif decodeStyle=="mime":
      commentaryLineNumbers,textForViewer,partsInfo= \
        self.mailObj.getTextHunksForUserMIME(deHTML=self.deHTML.get())
    elif decodeStyle=="uudecode":
      commentaryLineNumbers,textForViewer,partsInfo= \
        self.mailObj.getTextHunksForUserUU(deHTML=self.deHTML.get())
    else:
      assert 1==0,"Impossible value for decode style: %s" % decodeStyle

    self.textWidget.insert(END,textForViewer)

    self.textWidget.tag_configure("decodeCommentary",foreground=gPrefs.get("decodeCommentaryFG"),
      background=gPrefs.get("decodeCommentaryBG"))

    vqc=gPrefs.get("viewerQuoteColors")
    if vqc<>None:
      for count in range(len(vqc)):
        self.textWidget.tag_configure("quote%i" % (count+1),foreground=vqc[count])

    for item in commentaryLineNumbers:
      adjLineNo=item.lineNumber+adjustCommentaryLineNumbers
      self.textWidget.tag_add(item.kind,"%i.0" % adjLineNo,"%i.end" % adjLineNo)

    textWidgetContents.configure(state=DISABLED)

    self.mbar.mbar.deletemenuitems("Save parts",0,999) # XXX figure out how many?
    if decodeStyle=="nodecode" or len(partsInfo)==0:
      self.mbar.mbar.addmenuitem("Save parts","command", \
        command=nullCB,label="(No parts)",font=gPrefs.get("fontILike"),activebackground=kTkDefaultBG)
    else:
      for count in range(len(partsInfo)):
        item=partsInfo[count]
        if hasattr(item,"content_type"):
          l="Part %i, %s, %s" % (count+1,item.content_type,item.name)
        else:
          l="Part %i, %s" % (count+1,item.name)
        if item.part==kUnableToParseMessage:
          self.mbar.mbar.addmenuitem("Save parts","command",font=gPrefs.get("fontILike"), \
            label=l, command=savePart(item),activebackground=kTkDefaultBG,state=DISABLED)
        else:
          self.mbar.mbar.addmenuitem("Save parts","command",font=gPrefs.get("fontILike"), \
            label=l, command=savePart(item),activebackground=kTkDefaultBG)
    return None

  #~ def makeCommentaryLine(self,mimeItem,partNo):
    #~ commentary=[]
    #~ if mimeItem.content_type<>None:
      #~ commentary.append("type: %s" % mimeItem.content_type)
    #~ if mimeItem.content_description<>None:
      #~ commentary.append("description: %s" % mimeItem.content_description)
    #~ if mimeItem.name<>None:
      #~ commentary.append("name: %s" % mimeItem.name)
    #~ if mimeItem.charset<>None:
      #~ commentary.append("charset: %s" % mimeItem.charset)
    #~ if len(commentary)==0:
      #~ commentaryLine="MIME part %i" % partNo+1
    #~ else:
      #~ commentaryLine="MIME part %i -- %s" % (partNo+1,", ".join(commentary))
    #~ commentaryLine=commentaryLine+"\n"
    #~ return commentaryLine

  def mouseEntersAddress(self,tagName,eventIgnored):
    self.textWidget.tag_configure(tagName,background=kMouseEntersActiveAddressBG)
    return "break"

  def mouseLeavesAddress(self,tagName,eventIgnored):
    self.textWidget.tag_configure(tagName,background="white")
    return "break"

  def mouse3Address(self,textClickedOn,event):
    myFont=gPrefs.get("fontILike")
    myBoldFont=gPrefs.get("boldFontILike")
    m=Menu(gTkRoot,tearoff=0,activebackground=kTkDefaultBG)
    m.add_command(label=textClickedOn,font=myBoldFont,state=DISABLED)
    m.add_separator()
    m.add_command(label="Copy",font=myFont,command=genericCBO(self.putTextOnClipboard,textClickedOn))
    m.add_separator()
    m.add_command(label="Edit in addressbook",font=myFont,
      command=genericCBO(gSwitchboard.editInAddressbook,textClickedOn))
    m.post(event.x_root,event.y_root)
    # Delay to ignore the first mouse-up so that menu stays visible when
    # user releases mouse click that opened it. There's prolly a better
    # way.
    gTkRoot.after(250,m.grab_set)
    return None

  def putTextOnClipboard(self,text):
    # Doesn't work, don't know why
    #gTkRoot.clipboard_clear()
    #gTkRoot.clipboard_append(text)
    gTkRoot.selection_own()
    gTkRoot.selection_handle(clipboardCBO(text))
    return None

  def closeBoxCB(self):
    self.closeMe()
    return None

  def closeMe(self):
    gSwitchboard.removeViewer(self)
    self.root.destroy()
    return None


class enclosuresDialog:
  def __init__(self,mailEditor):
    self.mailEditor=mailEditor
    self.actions=[]
    self.dialog=Pmw.Dialog(buttons=("Add","Remove","Cancel","Done"),defaultbutton="Done", \
      command=self.buttonCB,title="Add/remove enclosures")
    for button in ("Add","Remove","Cancel","Done"):
      self.dialog.component("buttonbox").component(button).configure(activebackground=kTkDefaultBG)
    enclNames=[]
    if self.mailEditor.msg.is_multipart():
      for count in range(1,len(self.mailEditor.msg.get_payload())):
        name=self.mailEditor.msg.get_payload(count).get_param("name")
        size=qtyToReadable(len(self.mailEditor.msg.get_payload(count).get_payload(decode=1))) # 1 for decode
        enclNames.append("%s (%sB)" % (name,size))
    self.infoLabel=Label(self.dialog.interior(),font=gPrefs.get("fontILike"),anchor=W)
    self.infoLabel.pack(side=TOP,fill=X)
    self.listbox=Pmw.ScrolledListBox(self.dialog.interior(),
      items=enclNames,
      listbox_bg="white",
      listbox_font=gPrefs.get("fontILike"),
      vertscrollbar_activebackground=kTkDefaultBG,
      vertscrollbar_borderwidth=1,
      vertscrollbar_width=12,
      horizscrollbar_activebackground=kTkDefaultBG,
      horizscrollbar_borderwidth=1,
      horizscrollbar_width=12,

      )
    self.listbox.pack(side=TOP,expand=1,fill=BOTH)
    self.setMyInfoLabel()
    self.dialog.activate(geometry="centerscreenalways")
    return None

  def setMyInfoLabel(self):
    nEncls=self.listbox.size()
    if nEncls==0:
      self.infoLabel.configure(text="No enclosures")
    else:
      self.infoLabel.configure(text="%i %s" % (nEncls,pluralize(nEncls,"enclosure")))
    return None

  def buttonCB(self,button):
    global gLastSaveOrOpenDir
    if button=="Done":
      if self.mailEditor.msg.is_multipart():
        subMsgs=self.mailEditor.msg.get_payload()
      else:
        subMsgs=[None] # Placeholder for first part, is user's text
      for action in self.actions:
        if action.action=="del":
          del subMsgs[action.index+1] # Skip past user's text
        elif action.action=="add":
          inFile=open(action.path,"rb")
          outFile=cStringIO.StringIO()
          newSubMsg=email.Message.Message()
          cType=mimetypes.guess_type(action.path)[0]
          if cType==None:
            cType="application/octet-stream"
          newSubMsg.add_header("Content-type",cType,name=os.path.basename(action.path))
          partIsText=cType[:cType.find("/")]=="text"
          if partIsText:
            newSubMsg.add_header("Content-transfer-encoding","quoted-printable")
            quopri.encode(inFile,outFile,1)
          else:
            newSubMsg.add_header("Content-transfer-encoding","base64")
            base64.encode(inFile,outFile)
          newSubMsg.set_payload(outFile.getvalue())
          inFile.close()
          outFile.close()
          subMsgs.append(newSubMsg)
        else:
          assert 1==0,'Impossible action "%s" recorded in enclosures dialog' % action.action
      if len(subMsgs)==1:
        del self.mailEditor.msg["Content-type"]
        self.mailEditor.msg.set_payload(subMsgs[0])
      else:
        self.mailEditor.msg.set_payload(subMsgs)
      self.mailEditor.setEnclosuresMenu()
      self.dialog.destroy()
      return None
    elif button=="Cancel" or button==None:
      self.dialog.destroy()
      return None
    elif button=="Add":
      path=tkFileDialog.askopenfilename(initialdir=gLastSaveOrOpenDir)
      if path==None:
        return None
      if not os.path.isfile(path):
        return None # XXX error message
      gLastSaveOrOpenDir=myPathSplit(path)[0]
      self.actions.append(bunch(action="add",path=path))
      fSize=qtyToReadable(int(os.stat(path).st_size)) # If the size won't fit in an int, I don't want to send it
      oldVals=list(self.listbox.get())
      self.listbox.setlist(oldVals+["%s (%sB)" % (os.path.basename(path),fSize)])
      self.setMyInfoLabel()
    elif button=="Remove":
      cs=self.listbox.curselection()
      if len(cs)==0:
        return None # XXX error message
      items=list(self.listbox.get())
      index=int(cs[0])
      del items[index]
      self.actions.append(bunch(action="del",index=index))
      self.listbox.setlist(items)
      self.setMyInfoLabel()
    else:
      assert 1==0,'No such button "%s" in enclosures dialog' % button
    return None


# Spelling checker

class wordChooserWindow:
  def __init__(self,clickedWordRange,mWordItem,parent):
    self.parent=parent
    self.parent.addWordChooserWindow(self)
    self.clickedWordRange=clickedWordRange
    self.mWordItem=mWordItem
    self.root=Toplevel()
    self.root.protocol("WM_DELETE_WINDOW",self.closeWinCB)
    myFont=gPrefs.get("fontILike")
    self.root.title("Unrecognized word")
    topFrame=Frame(self.root,width=100,height=16)
    topFrame.label=Label(topFrame,text=" "+self.mWordItem.word+" ", \
      font=myFont)
    topFrame.label.pack(side=LEFT)

    if not self.mWordItem.word.islower():
      addLCButton=myButton(topFrame,"Add lc",
        self.addLCButtonCB,
        widthInChars=6)
      addLCButton.pack(side=RIGHT)
    addButton=myButton(topFrame,"Add",
      self.addButtonCB,
      widthInChars=6)
    addButton.pack(side=RIGHT)
    if mWordItem.suggestions<>None:
      replaceButton=myButton(topFrame,"Replace",
      self.replaceButtonCB,
      widthInChars=6)
      replaceButton.pack(side=RIGHT)

    topFrame.pack(side=TOP,fill=X)

    if mWordItem.suggestions<>None:
      self.root.bind("<Enter>",self.mouseEntersCB)
      self.root.bind("<Leave>",self.mouseLeavesCB)
      bottomFrame=Frame(self.root)
      list=Listbox(bottomFrame,
        selectmode="single",
        font=myFont,
        exportselection=0,
        bg="white")
      for str in mWordItem.suggestions:
        list.insert("end",str)
      list.pack(side=LEFT,fill=BOTH,expand=1)
      yScrollbar=Scrollbar(bottomFrame,
        activebackground=kTkDefaultBG,
        borderwidth=1,
        width=12,
        )
      yScrollbar.pack(side=RIGHT,fill=Y)
      bottomFrame.pack(side=TOP,fill=BOTH)

      list["yscrollcommand"]=yScrollbar.set
      yScrollbar["command"]=list.yview
      list.select_set(0) # This one's numbered from 0
      self.list=list
      self.savedSelection=[0]
    return None

  def closeButtonCB(self):
    self.closeWinCB()
    return None

  def closeWinCB(self):
    self.parent.removeWordChooserWindow(self)
    self.root.destroy()
    return None

  def addLCButtonCB(self):
    self.doRealAdd(self.mWordItem.word.lower())
    return None

  def addButtonCB(self):
    self.doRealAdd(self.mWordItem.word)
    return None

  def doRealAdd(self,word):
    # \n# required by aspell. Prolly won't hurt ispell.
    myCmd='echo "*%s\n#" | %s -a >/dev/null 2>&1' % (word,gPrefs.get("ispellLoc"))
    os.system(myCmd)
    self.root.destroy()
    self.parent.removeWordChooserWindow(self)
    return None

  def replaceButtonCB(self):
    if self.list.curselection()==[]:
      return None
    replacement=self.list.get(self.list.curselection())
    self.mWordItem.widget.delete(self.clickedWordRange[0],self.clickedWordRange[1])
    self.mWordItem.widget.insert(self.clickedWordRange[0],replacement)
    self.root.destroy()
    self.parent.removeWordChooserWindow(self)
    return None

  def mouseEntersCB(self,eventIgnored):
    if self.list.curselection==[]:
      self.list.select_set(self.savedSelection[0])
    return None

  def mouseLeavesCB(self,eventIgnored):
    self.savedSelection=self.list.curselection()
    return None


class editorMenubar:
  def __init__(self,parent):
    self.parent=parent
    self.createWidgets()
    self.newItemsForWindowMenu()
    return None

  def createWidgets(self):
    myFrame=Frame(self.parent.getMenuWidgetParent())
    myFrame.pack(side=LEFT,fill=X,expand=1)
    myFont=gPrefs.get("fontILike")
    mbar=Pmw.MenuBar(myFrame,padx=5,hotkeys=0)
    self.mbar=mbar
    mbar.pack(side=LEFT,fill=X)

    mbar.addmenu("Text","",font=myFont,activebackground=kTkDefaultBG)

    mbar.addmenuitem("Text","command",
      label="Re-wrap",
      command=self.rewrapCB,
      font=myFont,
      activebackground=kTkDefaultBG)

    mbar.addmenuitem("Text","command",
      label="Re-wrap indented",
      command=self.rewrapIndentedCB,
      font=myFont,
      activebackground=kTkDefaultBG)

    mbar.addmenuitem("Text","command",
      label="Quote",
      command=self.quoteCB,
      font=myFont,
      activebackground=kTkDefaultBG)

    mbar.addmenuitem("Text","separator")

    mbar.addmenuitem("Text","command",
      label="Check spelling",
      command=self.spellCB,
      font=myFont,
      activebackground=kTkDefaultBG)
    mbar.addmenuitem("Text","command",
      label="Remove spelling highlights",
      command=self.clearSpellHighlightsCB,
      font=myFont,
      activebackground=kTkDefaultBG)

    mbar.addmenuitem("Text","separator")

    mbar.addmenuitem("Text","command",
      label="Insert file as text",
      command=self.insertFileCB,
      font=myFont,
      activebackground=kTkDefaultBG)

    mbar.addmenuitem("Text","separator")

    mbar.addmenuitem("Text","command",
      label="Print",
      command=self.printCB,
      font=myFont,
      activebackground=kTkDefaultBG)


    mbar.addmenu("Enclosures","",font=myFont,activebackground=kTkDefaultBG)
    # This item will always be re-done with the number of enclosures
    mbar.addmenuitem("Enclosures","command",
      label="Add/remove (number of enclosures not set)",
      command=self.enclosuresCB,
      font=myFont,
      activebackground=kTkDefaultBG)

    mbar.addmenu("Windows","",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","command",command=gSwitchboard.openAboutBox,label="About",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","command",command=gSwitchboard.openKeysWin,label="Key commands",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","separator")
    mbar.addmenuitem("Windows","command",command=gSwitchboard.openAddressBrowser,label="Addressbook",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","command",command=gSwitchboard.openSearchWin,label="Search mail",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","command",command=gSwitchboard.openLogWindow,label="Log",font=myFont,activebackground=kTkDefaultBG)
    mbar.addmenuitem("Windows","separator")
    # Next is trashed and replaced in newItemsForWindowMenu()
    self.mbar.addmenuitem("Windows","command",label="Open windows",font=gPrefs.get("boldFontILike"),state=DISABLED)
    self.standardWindowMenuLength=7
    return None

  def spellCB(self):
    # Make Tkinter redraw, otherwise area that was covered
    # by the menu doesn't get redrawn until the dialog is
    # open and that's ugly.
    gTkRoot.update()
    self.parent.spellCB()
    return None

  def clearSpellHighlightsCB(self):
    gTkRoot.update() # As above
    self.parent.clearSpellHighlightsCB()
    return None

  def printCB(self):
    headers="From     %s <%s>\n" % (gPrefs.get("myHumanName"),gPrefs.get("myMailAddress"))

    toVal=self.parent.toEntry.get()
    if toVal<>"":
      headers+="To       %s\n" % toVal

    ccVal=self.parent.ccEntry.get()
    if ccVal<>"":
      headers+="CC       %s\n" % ccVal

    bccVal=self.parent.bccEntry.get()
    if bccVal<>"":
      headers+="BCC      %s\n" % bccVal

    replyToVal=self.parent.replyToEntry.get()
    if replyToVal<>"":
      headers+="Reply-to %s\n" % replyToVal

    headers+="Subject  %s\n\n" % self.parent.subjectText.get("1.0",END).rstrip()

    cmd=gPrefs.get("printCommand")
    gLogger.add("Sending mail editor contents to %s" % cmd)
    t=self.parent.textWidget.get()
    t=hardWrap(t,gPrefs.get("outboundMailWrapLen"),useFileLineSep=1)
    p=os.popen(cmd,"w")
    p.write(headers)
    p.write(t)
    p.close()
    gLogger.add("Printing done")
    return None

  def enclosuresCB(self):
    gTkRoot.update() # As above
    self.parent.enclosuresCB()
    return None

  def quoteCB(self):
    gTkRoot.update() # As above
    self.parent.quoteCB()
    return None

  def rewrapCB(self):
    gTkRoot.update() # As above
    self.parent.rewrapCB()
    return None

  def rewrapIndentedCB(self):
    gTkRoot.update() # As above
    self.parent.rewrapCB(indentDepth=kReWrapIndentDepth)
    return None

  def insertFileCB(self):
    gTkRoot.update()
    self.parent.insertFileCB()
    return None

  def setNumberOfEnclosures(self,num):
    self.mbar.deletemenuitems("Enclosures",1,999)
    if num==0:
      enclText="currently no enclosures"
    elif num==1:
      enclText="currently 1 enclosure"
    else:
      enclText="currently %i enclousres" % num

    self.mbar.addmenuitem("Enclosures","command",
      label="Add/remove (%s)" % enclText,
      command=self.enclosuresCB,
      font=gPrefs.get("fontILike"),
      activebackground=kTkDefaultBG)
    return None

  def redoLocationItems(self):
    myFont=gPrefs.get("fontILike")
    # Bug workaround: deletemenuitems always deletes at least one item even
    # if the range is outside the items in the menu. So just tell it to delete
    # at least one (which we know is a separator) and add it back.
    self.mbar.deletemenuitems("Get&Send",self.itemsAboveLocationsInMenu-1,999)
    self.mbar.addmenuitem("Get&Send","separator")
    locations=gPrefs.getLocations()
    if len(locations)==1:
      self.mbar.addmenuitem("Get&Send","command",label="(Only one location)",font=myFont,state=DISABLED)
    else:
      for location in locations:
        self.mbar.addmenuitem("Get&Send","radiobutton",variable=self.locVar, \
          label=location,font=myFont,selectcolor="blue",command=self.locMenuCB)
    return None

  def newItemsForWindowMenu(self,items=None):
    myFont=gPrefs.get("fontILike")
    self.mbar.deletemenuitems("Windows",self.standardWindowMenuLength,999)
    self.mbar.addmenuitem("Windows","command",
      label="Open windows",
      font=gPrefs.get("boldFontILike"),
      state=DISABLED,
      )
    self.mbar.addmenuitem("Windows","command",
      label="Mailboxes",
      font=myFont,
      command=self.raiseRootWin,
      activebackground=kTkDefaultBG,
      )
    if items<>None:
      for win in items:
        self.mbar.addmenuitem("Windows","command",
          label=win.root.title(),
          command=raiseWinCBO(win),
          font=myFont,
          activebackground=kTkDefaultBG,
          )
    return None

  def raiseRootWin(self):
    gSwitchboard.getOverview().root.lift()
    return None

  def pack(self,*positionArgsTuple,**kwArgsDict):
    self.mbar.pack(*positionArgsTuple,**kwArgsDict)
    return None


class mailEditor:

  def __init__(self,path=None,startInText=0,isNew=1):
    self.mWords=None
    self.isNew=isNew
    self.path=path
    self.root=Toplevel()
    self.openWordChooserWindows=[]
    self.root.protocol("WM_DELETE_WINDOW",self.closeWinCB)
    self.toVar=StringVar("")
    self.ccVar=StringVar("")
    self.bccVar=StringVar("")
    self.replyToVar=StringVar("")
    self.dontSendVar=IntVar()
    f=open(self.path)
    p=email.Parser.Parser()
    self.msg=p.parse(f)
    f.close()
    self.createWidgets(startInText) # Assigns to self.textArea
    self.loadWidgets()
    gSwitchboard.addEditor(self)
    return None

  def createWidgets(self,startInText):
    myFont=gPrefs.get("fontILike")
    hideMouse=gPrefs.get("hideMousePointerWhileTyping")
    upperFrame=Frame(self.root)
    self.mbarFrame=Frame(upperFrame) ###
    self.mbar=editorMenubar(self)
    self.mbar.pack(side=LEFT)
    self.mbarFrame.pack(side=TOP,fill=X)
    buttonFrame=Frame(upperFrame)
    saveButton=myButton(buttonFrame,"Save to outbox",self.saveCB,widthInChars=12)
    saveButton.pack(side=LEFT)
    self.sendNowButton=myButton(buttonFrame,"Send now",self.sendNowButtonCB,widthInChars=12)
    self.sendNowButton.pack(side=LEFT)
    abortButton=myButton(buttonFrame,"Abort",self.abortButtonCB,widthInChars=12)
    abortButton.pack(side=RIGHT)

    buttonFrame.pack(side=TOP,fill=X)

    toFrame=Frame(upperFrame)
    toLabel=Label(toFrame,font=myFont,text="To",width=8,anchor=W)
    toLabel.pack(side=LEFT)
    line=Frame(toFrame,width=1,bg="black")
    line.pack(side=LEFT,fill=Y)
    toEntryBorder=Frame(toFrame,borderwidth=0,bg="black")
    line=Frame(toEntryBorder,height=1,bg="black")
    line.pack(side=TOP,fill=X,expand=1)
    toEntry=Entry(toEntryBorder,width=20,textvariable=self.toVar,font=myFont, \
      highlightthickness=0,bg="white",relief=FLAT)
    if hideMouse:
      gMousePointerHider.makeWidgetHideMouseOnKey(toEntry)
    toEntry.pack(side=TOP,fill=X,expand=1)
    line=Frame(toEntryBorder,height=1,bg="black")
    line.pack(side=TOP,fill=X,expand=1)
    toEntry.bind("<FocusIn>",clearSelection)#self.insertionEntersCB)
    toEntry.bind("<FocusOut>",self.insertionLeavesCB)
    toEntry.bind("<Key-Down>",self.downArrowInEntryCB)
    toEntry.bind("<Key-Up>",self.upArrowInEntryCB)
    toEntry.bind("<ButtonRelease-2>",self.pasteAddressCB)
    toEntryBorder.pack(side=LEFT,fill=X,expand=1)
    line=Frame(toFrame,width=1,bg="black")
    line.pack(side=LEFT,fill=Y)
    spacerFrame=Frame(toFrame,width=5)
    spacerFrame.pack(side=LEFT)
    toFrame.pack(side=TOP,fill=X,expand=1)
    self.toEntry=toEntry

    ccFrame=Frame(upperFrame)
    ccLabel=Label(ccFrame,font=myFont,text="CC",width=8,anchor=W)
    ccLabel.pack(side=LEFT)
    line=Frame(ccFrame,width=1,bg="black")
    line.pack(side=LEFT,fill=Y)
    ccEntryBorder=Frame(ccFrame,borderwidth=0,bg="black")
    ccEntry=Entry(ccEntryBorder,width=20,textvariable=self.ccVar,font=myFont,
      highlightthickness=0,bg="white",relief=FLAT)
    if hideMouse:
      gMousePointerHider.makeWidgetHideMouseOnKey(ccEntry)
    ccEntry.bind("<FocusIn>",clearSelection)#self.insertionEntersCB)
    ccEntry.bind("<FocusOut>",self.insertionLeavesCB)
    ccEntry.bind("<Key-Down>",self.downArrowInEntryCB)
    ccEntry.bind("<Key-Up>",self.upArrowInEntryCB)
    ccEntry.bind("<ButtonRelease-2>",self.pasteAddressCB)
    ccEntry.pack(side=TOP,fill=X,expand=1)
    self.ccEntry=ccEntry
    line=Frame(ccEntryBorder,height=1,bg="black")
    line.pack(side=TOP,fill=X,expand=1)
    ccEntryBorder.pack(side=LEFT,fill=X,expand=1)
    line=Frame(ccFrame,width=1,bg="black")
    line.pack(side=LEFT,fill=Y)
    spacerFrame=Frame(ccFrame,width=5)
    spacerFrame.pack(side=LEFT)
    ccFrame.pack(side=TOP,fill=X,expand=1)

    bccFrame=Frame(upperFrame)
    bccLabel=Label(bccFrame,font=myFont,text="BCC",width=8,anchor=W)
    bccLabel.pack(side=LEFT)
    line=Frame(bccFrame,width=1,bg="black")
    line.pack(side=LEFT,fill=Y)
    bccEntryBorder=Frame(bccFrame,borderwidth=0,bg="black")
    bccEntry=Entry(bccEntryBorder,width=20,textvariable=self.bccVar,font=myFont, \
      highlightthickness=0,bg="white",relief=FLAT)
    if hideMouse:
      gMousePointerHider.makeWidgetHideMouseOnKey(bccEntry)
    bccEntry.bind("<FocusIn>",clearSelection)#self.insertionEntersCB)
    bccEntry.bind("<FocusOut>",self.insertionLeavesCB)
    bccEntry.bind("<Key-Down>",self.downArrowInEntryCB)
    bccEntry.bind("<Key-Up>",self.upArrowInEntryCB)
    bccEntry.bind("<ButtonRelease-2>",self.pasteAddressCB)
    # Ordinary shift-tab behavior would be fine. But we need to bind to the tab key
    # and unless we bind shift-tab, the tab binding gets called for the shift-tab.
    # Such are the wonders of Tk.
    #bccEntry.bind("<Shift-Tab>",self.shiftTabInBCCCB)
    bccEntry.pack(side=TOP,fill=X,expand=1)
    self.bccEntry=bccEntry
    line=Frame(bccEntryBorder,height=1,bg="black")
    line.pack(side=TOP,fill=X,expand=1)
    bccEntryBorder.pack(side=LEFT,fill=X,expand=1)
    line=Frame(bccFrame,width=1,bg="black")
    line.pack(side=LEFT,fill=Y)
    spacerFrame=Frame(bccFrame,width=5)
    spacerFrame.pack(side=LEFT)
    bccFrame.pack(side=TOP,fill=X,expand=1)

    replyToFrame=Frame(upperFrame)
    replyToLabel=Label(replyToFrame,font=myFont,text="Reply-to",width=8,anchor=W)
    replyToLabel.pack(side=LEFT)
    line=Frame(replyToFrame,width=1,bg="black")
    line.pack(side=LEFT,fill=Y)
    replyToEntryBorder=Frame(replyToFrame,borderwidth=0,bg="black")
    replyToEntry=Entry(replyToEntryBorder,width=20,textvariable=self.replyToVar,font=myFont,
      highlightthickness=0,bg="white",relief=FLAT)
    if hideMouse:
      gMousePointerHider.makeWidgetHideMouseOnKey(replyToEntry)
    replyToEntry.bind("<FocusIn>",clearSelection)#self.insertionEntersCB)
    replyToEntry.bind("<FocusOut>",self.insertionLeavesCB)
    replyToEntry.bind("<Key-Down>",self.downArrowInEntryCB)
    replyToEntry.bind("<Key-Up>",self.upArrowInEntryCB)
    replyToEntry.bind("<Key-Return>",nullEventCB)
    replyToEntry.bind("<ButtonRelease-2>",self.pasteAddressCB)
    replyToEntry.pack(side=TOP,fill=X,expand=1)
    line=Frame(replyToEntryBorder,height=1,bg="black")
    line.pack(side=TOP,fill=X,expand=1)
    replyToEntryBorder.pack(side=LEFT,fill=X,expand=1)
    line=Frame(replyToFrame,width=1,bg="black")
    line.pack(side=LEFT,fill=Y)
    spacerFrame=Frame(replyToFrame,width=5)
    spacerFrame.pack(side=LEFT)
    replyToFrame.pack(side=TOP,fill=X,expand=1)
    self.replyToEntry=replyToEntry

    subjectFrame=Frame(upperFrame)
    subjectLabel=Label(subjectFrame,font=myFont,text="Subject",width=8,anchor=W)
    subjectLabel.pack(side=LEFT)
    line=Frame(subjectFrame,width=1,bg="black")
    line.pack(side=LEFT,fill=Y)
    subjectTextBorder=Frame(subjectFrame,borderwidth=0,bg="black")
    subjectText=Text(subjectTextBorder,width=20,height=1,font=myFont, \
      highlightthickness=0,bg="white",relief=FLAT,insertwidth=2)
    if hideMouse:
      gMousePointerHider.makeWidgetHideMouseOnKey(subjectText)
    subjectText.bind("<Key-Down>",self.downArrowInEntryCB)
    subjectText.bind("<Key-Up>",self.upArrowInEntryCB)
    subjectText.bind("<Key-Tab>",self.tabInSubjectCB)
    subjectText.bind("<Control-i>",self.tabInSubjectCB)
    subjectText.bind("<Shift-Tab>",self.shiftTabInSubjectCB)
    subjectText.bind("<Key-Return>",nullEventCB)
    subjectText.bind("<ButtonRelease-2>",self.pasteInSubjectCB)
    subjectText.pack(side=TOP,fill=X,expand=1)
    line=Frame(subjectTextBorder,height=1,bg="black")
    line.pack(side=TOP,fill=X,expand=1)
    subjectTextBorder.pack(side=LEFT,fill=X,expand=1)
    line=Frame(subjectFrame,width=1,bg="black")
    line.pack(side=LEFT,fill=Y)
    spacerFrame=Frame(subjectFrame,width=5)
    spacerFrame.pack(side=LEFT)
    subjectFrame.pack(side=TOP,fill=X,expand=1)
    self.subjectText=subjectText

    spacer=Frame(upperFrame,height=3)
    spacer.pack(side=TOP)

    moreControlsFrame=Frame(upperFrame)

    spacer=Frame(moreControlsFrame,width=1,height=4)
    spacer.pack(side=LEFT)
    dscBorder=Frame(moreControlsFrame,
      borderwidth=1,
      bg="black")
    forceHeight=Frame(dscBorder,height=kButtonHeight-6,width=22*kButtonDotsPerChar)
    forceHeight.pack_propagate(0)
    forceHeight.pack(side=LEFT)
    self.dontSendCheckbox=Checkbutton(forceHeight,#dscBorder,
      text="Draft, don't send yet",
      selectcolor="blue",
      variable=self.dontSendVar,
      takefocus=0,
      font=myFont,
      width=22,
      activebackground=kTkDefaultBG)
    self.dontSendCheckbox.pack(side=LEFT)
    dscBorder.pack(side=LEFT)
    spacerFrame=Frame(moreControlsFrame,width=2,height=10)
    spacerFrame.pack_propagate(0)
    spacerFrame.pack(side=LEFT)

    spellLabelBorder=Frame(moreControlsFrame,borderwidth=1,bg="black")
    forceHeight=Frame(spellLabelBorder,height=kButtonHeight-6,width=22*kButtonDotsPerChar)
    forceHeight.pack_propagate(0)
    forceHeight.pack(side=LEFT)
    self.spellLabel=Label(forceHeight,
      font=myFont,
      width=22,
      text="No spelling check yet")
    self.spellLabel.pack(side=LEFT,fill=X,expand=1)
    spellLabelBorder.pack(side=LEFT)


    moreControlsFrame.pack(side=TOP,fill=X)
    upperFrame.pack(side=TOP,fill=X,expand=1)

    spacer=Frame(self.root,width=1,height=2)
    spacer.pack(side=TOP)

    textArea=scrolledTextWithUndo(self.root,
      text_wrap="word",
      vscrollmode="static",
      text_background="white",
      text_font=gPrefs.get("fixedWidthFontILike"),
      text_highlightthickness=0,
      text_width=gPrefs.get("outboundMailWrapLen"),
      vertscrollbar_activebackground=kTkDefaultBG,
      vertscrollbar_takefocus=0,
      vertscrollbar_borderwidth=1,
      vertscrollbar_width=12,
      horizscrollbar_borderwidth=1,
      horizscrollbar_width=12,
      )
    textArea.pack(side=TOP,fill=Y,expand=1)

    textWidget=textArea.component("text")
    if hideMouse:
      gMousePointerHider.makeWidgetHideMouseOnKey(textWidget)

    textWidget.bind("<Key-Tab>",self.tabInTextCB)
    textWidget.bind("<Control-i>",self.tabInTextCB)
    textWidget.bind("<Shift-Tab>",self.shiftTabInTextCB)
    textWidget.bind("<Key-Up>",self.upArrowInTextCB)
    textWidget.bind("<Key-Down>",self.downArrowInTextCB)
    textWidget.bind("<Control-l>",self.controlLInTextCB)
    textWidget.bind("<Control-Down>",self.controlDownInTextCB)
    textWidget.bind("<Control-Up>",self.controlUpInTextCB)

    self.textArea=textArea

    if startInText:
      textWidget.focus_force()
      textWidget.mark_set(INSERT,"1.0")
    else:
      toEntry.focus_set()
    return None

  def loadWidgets(self):
    toText=self.msg.get("to","")
    toText=kWhitespaceRunRE.sub(" ",toText)
    toText=encodedHeaderToLatin1(toText)
    if toText=="undisclosed-recipients:;":
      toText=""

    if self.isNew:
      titleText="New mail"
    else:
      titleText="To "+addrTextToDisplayText(toText)
    self.root.title("%s (%s)" % (titleText,os.path.basename(self.getPath())))

    self.toVar.set(toText)


    ccText=kWhitespaceRunRE.sub(" ",self.msg.get("cc",""))
    ccText=encodedHeaderToLatin1(ccText)
    self.ccVar.set(ccText)

    bccText=kWhitespaceRunRE.sub(" ",self.msg.get("bcc",""))
    bccText=encodedHeaderToLatin1(bccText)
    self.bccVar.set(bccText)

    replyToText=kWhitespaceRunRE.sub(" ",self.msg.get("reply-to",""))
    replyToText=encodedHeaderToLatin1(replyToText)
    self.replyToVar.set(replyToText)

    subjectText=kWhitespaceRunRE.sub(" ",self.msg.get("subject",""))
    subjectText=encodedHeaderToLatin1(subjectText)
    self.subjectText.insert(END,subjectText)

    # For some reason, the default for the insertion point is at the end.
    # Not what we want here.
    self.subjectText.mark_set(INSERT,"1.0")
    #self.setEnclosuresLabel()
    self.setEnclosuresMenu()
    if getSubdirFromPath(self.path)=="new":
      self.dontSendVar.set(1)
    if not self.msg.is_multipart():
      text=self.msg.get_payload(decode=1)
    else:
      text=self.msg.get_payload(0).get_payload()
    self.textArea.insert(END,text)
    self.textArea.mark_set("insert","1.0")
    return None

  def getMenuWidgetParent(self):
    return self.mbarFrame

  def getPath(self):
    return self.path

  def rightMouseOnHighlightedWordCB(self,event):
    if event.widget==self.textArea.component("text"):
      widgetClicked=self.textArea.component("text")
    elif event.widget==self.subjectText:
      widgetClicked=self.subjectText
    else:
      assert 1==0,"Unknown widget clicked on: %s" % widgetClicked

    clickIndex=widgetClicked.index("@%s,%s" % (event.x,event.y))
    allTagRanges=widgetClicked.tag_ranges("spell")+widgetClicked.tag_ranges("spellNoSugg")
    for count in range(0,len(allTagRanges),2):
      if widgetClicked.compare(allTagRanges[count],"<=",clickIndex) and \
        widgetClicked.compare(clickIndex,"<=",allTagRanges[count+1]):
        break
    else:
      assert 1==0,"Range not found"

    clickedWordRange=(allTagRanges[count],allTagRanges[count+1])
    clickedWord=widgetClicked.get(allTagRanges[count],allTagRanges[count+1])

    for item in self.mWords:
      if item.word==clickedWord and widgetClicked==item.widget:
        break
    else:
      assert 1==0,"word not found"
      return None

    for win in self.openWordChooserWindows:
      if win.clickedWordRange==clickedWordRange:
        win.root.lift()
        break
    else:
      wordChooserWindow(clickedWordRange,item,self)
    return None

  def enclosuresCB(self):
    enclosuresDialog(self)
    return None

  def setEnclosuresMenu(self):
    if not self.msg.is_multipart():
      self.mbar.setNumberOfEnclosures(0)
    else:
      n=len(self.msg.get_payload())-1 # First is user's text.
      self.mbar.setNumberOfEnclosures(n)
    return None

  def controlDownInTextCB(self,eventIgnored):
    self.textArea.yview("scroll",1,"units")
    return "break"

  def controlUpInTextCB(self,eventIgnored):
    self.textArea.yview("scroll",-1,"units")
    return "break"

  def controlLInTextCB(self,eventIgnored):
    charInfo=self.textArea.bbox("insert")
    if charInfo==None:
      self.textArea.see(self.textArea.index("insert"))
    else:
      # Jeez, that was harder than it ought to have been. Next time, try without
      # the whiskey.
      (x,y,charWidth,charHeight)=charInfo
      height=self.textArea.winfo_height()
      halfHeight=int(height/2)-charHeight # Actually already int
      linesToScroll=(y-halfHeight)/charHeight
      self.textArea.yview("scroll",linesToScroll,"units")
    return None

  def upArrowInTextCB(self,ignored):
    charInfo=self.textArea.bbox("insert")
    if charInfo==None: # Insert not on screen probably due to paste
      self.textArea.yview_pickplace(self.textArea.index("insert"))
      charInfo=self.textArea.bbox("insert")
    (x,y,charWidth,charHeight)=charInfo
    newY=y-charHeight
    if newY>0:
      self.textArea.mark_set("insert","@"+repr(x)+","+repr(newY))
    else:
      topStuff=self.textArea.bbox("1.0")
      # If we're at the top line and so can't scroll, go to the left margin
      if topStuff<>None and topStuff[1]==y:
        self.textArea.mark_set("insert","1.0")
      else:
        self.textArea.yview("scroll",-1,"units")
    return "break"

  def downArrowInTextCB(self,ignored):
    charInfo=self.textArea.bbox("insert")

    if charInfo==None: # Insert not on screen probably due to paste
      self.textArea.yview_pickplace(self.textArea.index("insert"))

      charInfo=self.textArea.bbox("insert")

    (x,y,charWidth,charHeight)=charInfo
    maxY=self.textArea.winfo_height()-charHeight
    newY=y+charHeight
    if newY<maxY:
      self.textArea.mark_set("insert","@"+repr(x)+","+repr(newY))
    else:
      # If wer're already on the last line, go to the end. I don't
      # know why the "- 1 chars" is required.
      bottomStuff=self.textArea.bbox("end - 1 chars")
      if bottomStuff<>None and bottomStuff[1]==y:
        self.textArea.mark_set("insert","end")
      else:
        self.textArea.yview("scroll",1,"units")
    return "break"

  def tabInSubjectCB(self,eventIgnored):
    self.textArea.component("text").focus_force()
    return "break"

  def shiftTabInSubjectCB(self,eventIgnored):
    self.replyToEntry.focus_force()
    return "break"

  def tabInTextCB(self,eventIgnored):
    self.toEntry.focus_force()
    return "break" # Charming way to indicate that Tkinter shouldn't process the event further

  def shiftTabInTextCB(self,eventIgnored):
    self.subjectText.focus_force()
    return "break" # Charming way to indicate that Tkinter shouldn't process the event further

  def downArrowInEntryCB(self,event):
    event.widget.selection_clear()
    if event.widget.widgetName=="text":
      event.widget.mark_set(INSERT,END)
    else:
      event.widget.icursor(END)
    return "break"

  def upArrowInEntryCB(self,event):
    event.widget.selection_clear()
    if event.widget.widgetName=="text":
      event.widget.mark_set(INSERT,"1.0")
    else:
      event.widget.icursor(0)
    return "break"

  def pasteInSubjectCB(self,eventIgnored):
    # Strange that the value pasted doesn't appear at once
    # even though we're called on mouse-up.
    gTkRoot.after(kPasteTakesEffectDelayMS,self.realPasteInSubjectCB)
    return None

  def realPasteInSubjectCB(self):
    subj=self.subjectText.get("1.0",END)
    subj=subj.rstrip()
    newSubj=cleanupCharsForHeader(subj)
    if subj<>newSubj:
      self.subjectText.delete("1.0",END)
      self.subjectText.insert(END,newSubj)
    return None

  def pasteAddressCB(self,event):
    try:
      pasteText=gTkRoot.selection_get()
    except TclError: # no selection
      return "break"
    if type(pasteText)==types.UnicodeType:
      pasteText=pasteText.encode("latin-1","replace")
    if pasteText.find("@")<>-1:
      addr=canonicalizeTextAddr(pasteText)
    else:
      addrTuple=gAddressBook.nickToAddr(pasteText)
      if addrTuple<>None:
        addr=", ".join(canonicalizeAddrList([addrTuple]))

    if pasteText.find("@")==-1:
      pasteText=pasteText.replace("\n"," ")
      pasteText=pasteText.replace("\r"," ")
      pasteText=pasteText.replace("\t"," ")
      relX=event.x-event.widget.winfo_x()
      f=tkFont.Font(font=gPrefs.get("fixedWidthFontILike"))
      # Don't know why the value returned by measure
      # needs to be adjusted but it does.
      oneCharWidth=float(f.measure("x")-1)
      charsIn=int(relX/oneCharWidth)
      prevText=event.widget.get()
      if prevText=="":
        event.widget.insert(END,pasteText)
      else:
        newText=prevText[:charsIn]+pasteText+prevText[charsIn:]
        event.widget.delete(0,END)
        event.widget.insert(END,newText)
        event.widget.icursor(charsIn)
        event.widget.focus_set()
      return "break"

    prevText=event.widget.get()
    if prevText=="":
      event.widget.insert(END,addr)
    else:
      strs=splitRespectingQuotes(prevText)
      relX=event.x-event.widget.winfo_x()
      f=tkFont.Font(font=gPrefs.get("fixedWidthFontILike"))
      # Don't know why the value returned by measure
      # needs to be adjusted but it does.
      oneCharWidth=float(f.measure("x")-1)
      charsIn=int(relX/oneCharWidth)
      soFar=0

      for count in range(len(strs)):
        lenThisOne=len(strs[count])
        if charsIn<=soFar+int(lenThisOne/2):
          whereInsert=count
          break
        soFar=soFar+lenThisOne
      else:
        whereInsert=len(strs)

      strs.insert(whereInsert,addr)
      event.widget.delete(0,END)
      event.widget.insert(END,", ".join(strs))
      event.widget.icursor(charsIn)
      event.widget.focus_set()
    return "break"

  def expandAliases(self,headerStr):
    strs=splitRespectingQuotes(headerStr)
    removeAny(strs,"")
    for count in range(len(strs)):
      if strs[count].find("@")<>-1:
        continue
      realAddr=gAddressBook.nickToAddr(strs[count])
      if realAddr==None:
        continue
      strs[count]=canonicalizeAddrList((realAddr,))[0]
    backTogether=", ".join(strs)
    return backTogether

  def insertionLeavesCB(self,event):
    oldVal=event.widget.get()
    newVal=self.expandAliases(oldVal)
    if newVal<>oldVal:
      event.widget.delete("0",END)
      event.widget.insert(END,newVal)
    return None

  def spellCB(self):
    whichIspell=gPrefs.get("ispellLoc")
    if whichIspell==None:
      gLogger.add("Location of ispell isn't set in your preferences " \
        "can't check spelling without it")
      return None
    self.spellLabel.configure(text="Checking...")

    # Don't know exactly why this is necessary but if we don't do it
    # the mouse pointer sometimes disappears when it's over these
    # widgets during a spelling check.
    if gPrefs.get("hideMousePointerWhileTyping"):
      #self.textArea.component("text").configure(cursor="xterm")
      gMousePointerHider.explicitShowMouse(self.textArea.component("text"))
      gMousePointerHider.explicitShowMouse(self.toEntry)
      gMousePointerHider.explicitShowMouse(self.ccEntry)
      gMousePointerHider.explicitShowMouse(self.bccEntry)
      gMousePointerHider.explicitShowMouse(self.subjectText)
    gTkRoot.update_idletasks()
    self.clearSpellHighlights()
    tempFileName=tempfile.mktemp()
    os.umask(0177) # octal
    spellCommand="%s -a -S > %s" % (gPrefs.get("ispellLoc"),tempFileName)
    ispellRegex1=re.compile("^. ([^ ]+) [0-9]+ ([0-9]+): (.*)$")
    ispellRegex2=re.compile("^. ([^ ]+) ([0-9]+)$")

    # First the subject line
    mWords=[]
    s=os.popen(spellCommand,"w")
    s.write("!%s" % kFileLineSep)
    s.write("^")
    l=self.subjectText.get("1.0",END)
    # For some reason (not described in the man page it seems) a $
    # causes ispell to stop checking.
    l=l.replace("$"," ")
    l=l.replace(":"," ")
    s.write(l)
    s.close()
    tfh=open(tempFileName,"r")
    copyrightIgnored=tfh.readline() # Throw away ispell's copyright message
    lineNo=1 # Count lines from 1 to match Tk. Bleh.
    thisLine=tfh.readline()
    while thisLine<>"":
      # Oh, and the only way of figuring out what line we're on is to
      # count the blank lines ispell emits. Bleh.
      if thisLine==kFileLineSep:
        thisLine=tfh.readline()
        continue
      whatSort=thisLine[0]
      if whatSort=="#":
        whereMatch=ispellRegex2.search(thisLine)
        badWord=whereMatch.group(1)
        charsIn=int(whereMatch.group(2))-1
        startIndex=repr(lineNo)+"."+repr(charsIn)
        endIndex=repr(lineNo)+"."+repr(charsIn+len(badWord))
        suggList=None
      else:
        whereMatch=ispellRegex1.search(thisLine)
        badWord=whereMatch.group(1)
        charsIn=int(whereMatch.group(2))-1
        suggStr=whereMatch.group(3)
        suggList=suggStr.split(", ")

      mWords.append(bunch(
        word=badWord,
        charsIn=charsIn,
        lineNo=lineNo,
        suggestions=suggList,
        widget=self.subjectText,
        ))
      thisLine=tfh.readline()
    tfh.close()
    os.unlink(tempFileName)

    # Then the main text
    myText=self.textArea.get("1.0","end")

    # If there's an 8-bit character, value comes back as Unicode.
    if type(myText)==types.UnicodeType:
      myText=myText.encode("latin-1","replace") # Right charset?

    st=cStringIO.StringIO(myText)
    s=os.popen(spellCommand,"w")

    # It seems that you can switch various of ispell's options on and off
    # by beginning lines with particular characters. It doesn't appear to
    # be possible to set those options any other way and it doesn't seem
    # to be possible to turn that behavior off. So we begin by setting
    # terse mode by sending a bang. Then we have to prefix each line with
    # a sharkfin so that ispell will check the line even if the line begins
    # with an option character (like dash) that can readily appear at
    # the beginning of a line in mail. We also have to fix up the char
    # position reported since the sharkfin is included in the count.
    # Bleah.

    s.write("!%s" % kFileLineSep)
    thisLine=st.readline()
    while thisLine<>"":
      s.write("^")
      thisLine=thisLine.replace("$"," ")
      s.write(thisLine)
      thisLine=st.readline()
    s.close()
    tfh=open(tempFileName,"r")
    copyrightIgnored=tfh.readline() # Throw away ispell's copyright message
    lineNo=1 # Count lines from 1 to match Tk. Bleh.

    thisLine=tfh.readline()
    while thisLine<>"":

      # Oh, and the only way of figuring out what line we're on is to
      # count the blank lines ispell emits. Bleh.
      if thisLine==kFileLineSep:
        lineNo+=1
        thisLine=tfh.readline()
        continue

      whatSort=thisLine[0]

      if whatSort=="#":
        whereMatch=ispellRegex2.search(thisLine)
        badWord=whereMatch.group(1)
        charsIn=int(whereMatch.group(2))-1
        startIndex=repr(lineNo)+"."+repr(charsIn)
        endIndex=repr(lineNo)+"."+repr(charsIn+len(badWord))
        suggList=None
      else:
        whereMatch=ispellRegex1.search(thisLine)
        badWord=whereMatch.group(1)
        charsIn=int(whereMatch.group(2))-1
        suggStr=whereMatch.group(3)
        suggList=suggStr.split(", ")

      mWords.append(bunch(
        word=badWord,
        charsIn=charsIn,
        lineNo=lineNo,
        suggestions=suggList,
        widget=self.textArea.component("text"),
        ))

      thisLine=tfh.readline()

    tfh.close()
    os.unlink(tempFileName)
    for item in mWords:
      startIndex=repr(item.lineNo)+"."+repr(item.charsIn)
      endIndex=repr(item.lineNo)+"."+repr(item.charsIn+len(item.word))
      if item.suggestions==None:
        item.widget.tag_add("spellNoSugg",startIndex,endIndex)
      else:
        item.widget.tag_add("spell",startIndex,endIndex)
    self.textArea.tag_configure("spellNoSugg",foreground=gPrefs.get("spellingCheckColors")[1])
    self.textArea.tag_configure("spell",foreground=gPrefs.get("spellingCheckColors")[0])
    self.textArea.tag_bind("spell","<Button-3>",self.rightMouseOnHighlightedWordCB)
    self.textArea.tag_bind("spellNoSugg","<Button-3>",self.rightMouseOnHighlightedWordCB)
    self.subjectText.tag_configure("spellNoSugg",foreground=gPrefs.get("spellingCheckColors")[1])
    self.subjectText.tag_configure("spell",foreground=gPrefs.get("spellingCheckColors")[0])
    self.subjectText.tag_bind("spell","<Button-3>",self.rightMouseOnHighlightedWordCB)
    self.subjectText.tag_bind("spellNoSugg","<Button-3>",self.rightMouseOnHighlightedWordCB)

    if gPrefs.get("twoSpacesOK"):
      spaceRunRE="   +"
    else:
      spaceRunRE="  +"

    for widget in (self.textArea,self.subjectText):
      startIndex="1.0"
      srLen=IntVar()
      
      while 1:
        spaceRunStart=widget.search(spaceRunRE,startIndex,stopindex=END,regexp=1,count=srLen)
        
        if spaceRunStart=="":
          break
        
        runLength=srLen.get()
        secondIndex="%s + %s chars" % (spaceRunStart,runLength)
        widget.tag_add("spaceRuns",spaceRunStart,secondIndex)
        startIndex=widget.index("%s + %i chars" % (spaceRunStart,runLength+1))

    self.textArea.tag_configure("spaceRuns",background=gPrefs.get("spellingCheckColors")[2])
    self.subjectText.tag_configure("spaceRuns",background=gPrefs.get("spellingCheckColors")[2])

    self.spellLabel.configure(text="%i unrecognized %s" % (len(mWords),pluralize(len(mWords),"word")))
    self.mWords=mWords
    return None

  def clearSpellHighlightsCB(self):
      self.clearSpellHighlights()
      self.spellLabel.configure(text="No spelling check yet")
      return None

  def clearSpellHighlights(self):
    self.textArea.tag_delete("spell")
    self.textArea.tag_delete("spaceRuns")
    self.textArea.tag_delete("spellNoSugg")
    self.subjectText.tag_delete("spell")
    self.subjectText.tag_delete("spaceRuns")
    self.subjectText.tag_delete("spellNoSugg")
    return None

  def getRangeToFiddle(self):
    selStart=""
    try:
      selStart=self.textArea.index("sel.first")
    except TclError:
      pass

    if selStart=="":
      selStart=self.textArea.index("insert")
      selEnd=selStart
      oneBeforeSelStart=selStart+"-1 chars"
      if self.textArea.get(selStart)=="\n" and self.textArea.get(oneBeforeSelStart)=="\n":
         return None,None
    else:
      # All "paras" in selection
      selEnd=self.textArea.index("sel.last")

    # Now find the line after a blank line
    thisLineNo=int(float(selStart))
    while 1:
      indexPrevLine=repr(thisLineNo)+".0-1 lines"
      firstCharPrevLine=self.textArea.get(indexPrevLine)
      if thisLineNo==1 or firstCharPrevLine=="\n":
        break
      else:
        thisLineNo-=1

    # Special case for "Dear X," line. If it looks like we want to fiddle with the
    # first line of the text but the user hasn't explictly selected it or clicked
    # on it, and it's short and it ends with punctuation, don't fiddle with it.
    if gPrefs.get("specialCaseRewrapDearLine") and \
      thisLineNo==1 and int(float(self.textArea.index(INSERT)))<>1 and \
      (len(self.textArea.tag_ranges(SEL))==0 or \
      int(float(self.textArea.tag_ranges(SEL)[0]))>1):
      firstLine=self.textArea.get("1.0","1.end")
      if len(firstLine)<=kDearLineMaxLen and firstLine[-1] in kDearLineEndsWith:
        thisLineNo=2

    startLineNo=thisLineNo

    thisLineNo=int(float(selEnd))
    endIndex=self.textArea.index(END)
    lastLine=int(float(endIndex))
    lastChar=self.textArea.get(endIndex+"-1 chars")
    if lastChar=="\n":
      lastLine-=1

    while 1:
      indexNextLine=repr(thisLineNo)+".0+1 lines"
      firstCharNextLine=self.textArea.get(indexNextLine)
      if thisLineNo==lastLine or firstCharNextLine=="\n":
        break
      else:
        thisLineNo+=1

    endLineNo=thisLineNo
    return startLineNo,endLineNo

  def quoteCB(self):
    startLineNo,endLineNo=self.getRangeToFiddle()
    if startLineNo==None:
      return None
    self.textArea.markUndoGroup()
    for count in range(startLineNo,endLineNo+1):
      l=self.textArea.get("%i.0" % count,"%i.end" % count)
      self.textArea.deleteText("%i.0" % count,"%i.end" % count)
      if l<>"" and (l[0]==">" or l[0]==":"):
        l=">"+l
      else:
        l="> "+l
      self.textArea.insertText("%i.0" % count,l)
    self.textArea.markUndoGroup()
    return None

# Called if window manager closes window
  def closeWinCB(self):
    for win in self.openWordChooserWindows:
      win.root.destroy()
    self.closeMe()
    return None

  def abortButtonCB(self):
    self.closeMe()
    return None

  def closeMe(self):
    for win in self.openWordChooserWindows:
      win.root.destroy()
    if self.path<>None and getSubdirFromPath(self.path)=="tmp":
      os.unlink(self.path)
    self.root.destroy()
    gSwitchboard.removeEditor(self)
    return None

  def saveCB(self):
    self.saveText()
    gSwitchboard.counts("outbox")
    self.root.destroy()
    gSwitchboard.removeEditor(self)
    return None

  def sendProgressUpdateCB(self,fracDone):
    self.sendNowButton.configure(text="Sending... %i%%" % round(fracDone*100))
    gTkRoot.update_idletasks()
    return None

  def sendNowButtonCB(self):
    if self.dontSendVar.get()==1:
      # If don't send is checked, blink it.
      self.dontSendCheckbox.configure(background=kTkHighlightBG)
      gLogger.add("Don't send is selected, so I won't.")
      gTkRoot.after(1000,self.dontSendCheckbox.configure(background=kTkDefaultBG))
      return None

    self.sendNowButton.configure(text="Sending...",bg=kTkHighlightBG,
      state=DISABLED,disabledforeground="black")

    try:
      gTkRoot.update_idletasks()
      self.saveText()

      doProgress=os.stat(self.path).st_size>=kSendingProgressSizeBigEnoughToReport

      myMailObj=mailObj(self.path)
      if not myMailObj.sendable():
        gLogger.add("Mail says it's not sendable so I won't send it.")
        self.sendNowButton.configure(text="Send now",bg=kTkDefaultBG,state=NORMAL)
        return None

      s=sendmail()
      success=s.setup()
      if success:
        if doProgress:
          success=s.sendOne(myMailObj,self.sendProgressUpdateCB)
        else:
          success=s.sendOne(myMailObj)
      s.tearDown()

    except:
      # Even if we're going to barf, don't leave the button wrong
      self.sendNowButton.configure(text="Send now",bg=kTkDefaultBG,state=NORMAL)
      raise

    if success:
      if doProgress:
        self.sendProgressUpdateCB(1) # Always say 100% done at end
      # Not necessarily best to send list of headers
      if gPrefs.get("swdbFile")<>None:
        addrs=[myMailObj.get("to",""), myMailObj.get("cc",""),
          myMailObj.get("bcc","")]
        removeAny(addrs,"")
        gAddressBook.snarf(addrs,direction="out")
      myMailObj.fileBecauseSent()
      gSwitchboard.delete("outbox",[self.path])
      os.unlink(self.path)
      gSwitchboard.counts("outbox")
      self.root.destroy()
      gSwitchboard.removeEditor(self)
    else:
      self.sendNowButton.configure(text="Send now",bg=kTkDefaultBG,state=NORMAL)
    return None

  def saveText(self):
    # XXX Date gets overwritten on each save. Means that it will be reasonably current on send
    # unless message has spent a long time in the outbox. Good idea? RFC822 only says that
    # the date header is the "creation date".
    del self.msg["To"]
    #to=canonicalizeTextAddr(self.toVar.get())
    to=latin1HeaderToEncodedHeader(self.toVar.get(),"to")
    if to<>"":
      self.msg["To"]=to

    del self.msg["CC"]
    #cc=canonicalizeTextAddr(self.ccVar.get())
    cc=latin1HeaderToEncodedHeader(self.ccVar.get(),"cc")
    if cc<>"":
      self.msg["CC"]=cc

    del self.msg["BCC"]
    #bcc=canonicalizeTextAddr(self.bccVar.get())
    bcc=latin1HeaderToEncodedHeader(self.bccVar.get(),"bcc")
    if bcc<>"":
      self.msg["BCC"]=bcc

    if to=="" and cc=="":
      self.msg["To"]="undisclosed-recipients:;"

    del self.msg["Reply-to"]
    #replyTo=canonicalizeTextAddr(self.replyToVar.get())
    replyTo=latin1HeaderToEncodedHeader(self.replyToVar.get(),"reply-to")
    if replyTo<>"":
      self.msg["Reply-to"]=replyTo

    del self.msg["Subject"]
    subject=self.subjectText.get("1.0",END).strip()
    subject=latin1HeaderToEncodedHeader(subject,"subject")
    if subject<>"":
      self.msg["Subject"]=subject

    del self.msg["Date"]
    self.msg["Date"]=getRFC822Date()

    bodyText=self.textArea.get("1.0",END)
    # If there's an 8-bit character, value comes back as Unicode.
    if type(bodyText)==types.UnicodeType:
      bodyText=bodyText.encode("latin-1","replace") # Right charset?
    bodyText=bodyText.rstrip()+kFileLineSep
    bodyText=hardWrap(bodyText,gPrefs.get("outboundMailWrapLen"),useFileLineSep=1)
    textNeedsQuopri=0
    userTextAsFile=cStringIO.StringIO(bodyText)
    while 1:
      l=userTextAsFile.readline()
      if l=="":
        break
      if len(l)>kEncodeQpriIfLineLongerThan:
        textNeedsQuopri=1
        break
      if kEighthBitSetRE.search(l):
        textNeedsQuopri=1
        break
    if textNeedsQuopri:
      userTextAsFile.seek(0)
      outFile=cStringIO.StringIO()
      quopri.encode(userTextAsFile,outFile,1) # 1 For quote tabs. No reason.
      bodyText=outFile.getvalue()
      userTextAsFile.close()
      outFile.close()

    if not self.msg.is_multipart():
      self.msg.set_payload(bodyText)
      del self.msg["Content-transfer-encoding"]
      del self.msg["Mime-version"]
      self.msg["Mime-version"]="1.0"
      if textNeedsQuopri:
        self.msg["Content-transfer-encoding"]="quoted-printable"
    else:
      self.msg.preamble=kMIMEPreambleText+kFileLineSep
      self.msg.epilogue=""
      subMessage=email.Message.Message()
      subMessage.add_header("Content-type","text/plain",charset="ISO-8859-1") # Correct?
      subMessage.add_header("Content-transfer-encoding","7bit")
      subMessage.set_payload(bodyText)
      allPayloads=self.msg.get_payload()
      allPayloads[0]=subMessage
      del self.msg["Mime-version"]
      self.msg["Mime-version"]="1.0"
      del self.msg["Content-type"]
      self.msg["Content-type"]="multipart/mixed"
      self.msg.set_payload(allPayloads)

    oldSubdirPath,oldFN=myPathSplit(self.path)
    maildirPath,subdir=myPathSplit(oldSubdirPath)
    if subdir=="tmp":
      f=open(self.path,"wb")
      g=email.Generator.Generator(f,mangle_from_=0)
      g.flatten(self.msg,unixfrom=0)
      # Would f.flush() be sufficient? No, flush() flushes user-space
      # buffers only, fsync() guarantees that kernel-space buffers are
      # also flushed.
      os.fsync(f.fileno())
      f.close()
    else:
      # Be paranoid when writing changes back. Not exactly
      # Maildir-standard but this is the outbox, nothing else
      # should write here.
      tmpVersion=myPathJoin(maildirPath,"tmp")
      tmpVersion=myPathJoin(tmpVersion,oldFN)
      f=open(tmpVersion,"wb")
      g=email.Generator.Generator(f,mangle_from_=0)
      g.flatten(self.msg,unixfrom=0)
      os.fsync(f.fileno())
      f.close()
      os.unlink(self.path)
      os.link(tmpVersion,self.path)
      os.unlink(tmpVersion)


    if self.dontSendVar.get()==1 or (self.toVar.get=="" and self.ccVar.get=="" and self.bccVar.get==""):
      wantToBe="new"
    else:
      wantToBe="cur"
    whereNow=getSubdirFromPath(self.path)
    if wantToBe<>whereNow:
      if wantToBe=="new":
        oldPath=self.path
        self.path=moveToNew(self.path)
      else:
        oldPath=self.path
        self.path=moveToCur(self.path)


    if self.isNew:
      gSwitchboard.add("outbox",[self.path])
    else:
      if wantToBe<>whereNow:
        gSwitchboard.subdirChanged(oldPath,self.path)
      gSwitchboard.change("outbox",[self.path])
    return None

  def rewrapCB(self,indentDepth=0):
    startLineNo,endLineNo=self.getRangeToFiddle()
    if startLineNo==None:
      return None

    whereWeWere=self.textArea.index("insert")

    self.textArea.markUndoGroup()
    # This is goofy: We need to make sure that the text ends with a \n or else
    # Tkinter counts lines wrong. To find out if it does, we ask for the
    # character two before the end. For some reason, the last character is always
    # reported as a control-d. In order to work around one kind of brain-damage
    # we need to work around another kind.
    if self.textArea.get(END+" - 2 chars")<>"\n":
      self.textArea.insertText(END,"\n")

    # We re-wrap individually units that are between blank lines or
    # which have different quote strings.

    stuff=[]

    thisLineNo=startLineNo
    thisLine=self.textArea.get(repr(thisLineNo)+".0",repr(thisLineNo)+".end")
    if type(thisLine)==types.UnicodeType:
        thisLine=thisLine.encode("latin-1","replace")

    m=kMailQuoteCharsRE.match(thisLine)
    if m==None:
      quoteChars=""
    else:
      quoteChars=thisLine[:m.end(0)]
      thisLine=thisLine[m.end(0):]

    currentText=thisLine.strip()
    lastQuoteChars=quoteChars

    thisLineNo+=1

    while thisLineNo<=endLineNo:
      thisLine=self.textArea.get(repr(thisLineNo)+".0",repr(thisLineNo)+".end")
      if type(thisLine)==types.UnicodeType:
        ### XXX some characters can't be encoded
        thisLine=thisLine.encode("latin-1","replace")

      m=kMailQuoteCharsRE.match(thisLine)
      if m==None:
        quoteChars=""
      else:
        quoteChars=thisLine[:m.end(0)]
        thisLine=thisLine[m.end(0):].lstrip()

      if quoteChars<>lastQuoteChars or thisLine=="":
        stuff.append({"quoteChars":lastQuoteChars.replace(" ",""), \
          "text":currentText})
        lastQuoteChars=quoteChars
        currentText=thisLine.strip()
      else:
        currentText=(currentText+" "+thisLine).strip()

      if thisLine=="":
        stuff.append({"quoteChars":"","text":"\n"})
        currentText=""
        lastQuoteChars=""

      thisLineNo+=1

    stuff.append({"quoteChars":lastQuoteChars.replace(" ",""), \
      "text":currentText})

    newStr=""
    ####XXXX ### XXXX clean up
    generalWrapLen=gPrefs.get("outboundMailWrapLen")
    generalWrapLen-=indentDepth

    for item in stuff:
      if item["quoteChars"]<>"":
        lenToWrap=generalWrapLen-(len(item["quoteChars"])+1) # space after quote chars
      else:
        lenToWrap=generalWrapLen

      #if item["quoteChars"]=="":
      #  lenToWrap=gPrefs.get("outboundMailWrapLen")
      #else:
      #  # We put a space after our quote chars
      #  lenToWrap=gPrefs.get("outboundMailWrapLen")-(len(item["quoteChars"])+1)
      wrappedText=hardWrap(item["text"],lenToWrap)
      st=cStringIO.StringIO(wrappedText)
      while 1:
        myLine=st.readline()
        if myLine=="":
          break
        newStr+=" "*indentDepth
        if item["quoteChars"]=="":
          newStr+=myLine
        else:
          newStr=newStr+item["quoteChars"]+" "+myLine

    self.textArea.deleteText(repr(startLineNo)+".0",repr(endLineNo)+".end+1 chars")
    self.textArea.insertText(repr(startLineNo)+".0",newStr)
    self.textArea.markUndoGroup()

    self.textArea.mark_set("insert",whereWeWere)
    return None

  def insertFileCB(self):
    global gLastSaveOrOpenDir
    path=tkFileDialog.askopenfilename(initialdir=gLastSaveOrOpenDir)
    if path==None:
      return None
    self.textArea.importfile(path,where=INSERT)
    gLastSaveOrOpenDir=myPathSplit(path)[0]
    return None

  def addWordChooserWindow(self,winObj):
    self.openWordChooserWindows.append(winObj)
    return None

  def removeWordChooserWindow(self,winObj):
    self.openWordChooserWindows.remove(winObj)
    return None


class state:
  def __init__(self):
    self.mailboxSortSettings={}
    return None

  def forgetAboutOldMailboxes(self):
    namesThatExist=gAllMailboxes.getNames()
    for k in self.mailboxSortSettings.keys():
      if k not in namesThatExist:
        del self.mailboxSortSettings[k]
    return None

def loadState():
  global gState
  path=myPathJoin(gPrefs.get("swallowDir"),kStatePickleName)
  if os.path.exists(path):
    stateStat=os.stat(path)
    mode=stateStat.st_mode
    if not stat.S_ISREG(mode):
      print 'My state file "%s" is not a regular file.' % path
      print "That's too weird for me. I don't want to run."
      sys.exit(0)

    if not pwd.getpwnam(os.environ["LOGNAME"])[2]==stateStat.st_uid:
      print 'My state file "%s" is not owned by you (user %s), it is owned by user %s.' % \
        (path,os.environ["LOGNAME"],pwd.getpwuid(stateStat.st_uid)[0])
      print "I don't think that can be right. I don't want to run."
      sys.exit(0)
    permissions=stat.S_IMODE(mode)
    if (permissions & 0020) or (permissions & 0002): # constants octal
      print 'My state file "%s" has' % path
      print "group or world write permissions (%s)." % oct(permissions)
      print "I don't think that's safe since stuff in there could cause me to do"
      print "things you don't want. Of course, you could be in your very own group"
      print "but I'm not smart enough to check that. Please fix the permissions."
      sys.exit(0)
    f=open(path,"rb")
    try:
      gState=cPickle.load(f)
    except cPickle.UnpicklingError:
      gLogger.add("State file %s couldn't be unpickled, ignoring it" % path)
      gState=state()
    f.close()
  else:
    gState=state()
  return None

def saveState():
  path=myPathJoin(gPrefs.get("swallowDir"),kStatePickleName)
  if os.path.exists(path):
    stateStat=os.stat(path)
    mode=stateStat.st_mode
    if not stat.S_ISREG(mode):
      gLogger.add('My state file "%s" is not a regular file.' % path)
      gLogger.add("That's too weird for me. I'm not going to save my state.")
      return None

    if not pwd.getpwnam(os.environ["LOGNAME"])[2]==stateStat.st_uid:
      gLogger.add('My state file "%s" is not owned by you (user %s), it is owned by user %s.' % \
        (path,os.environ["LOGNAME"],pwd.getpwuid(stateStat.st_uid)[0]))
      gLogger.add("I don't think that can be right. I'm not going to save to it.")
      return None

    permissions=stat.S_IMODE(mode)
    if (permissions & 0020) or (permissions & 0002): # constants octal
      gLogger.add('My state file "%s" has' % path)
      gLogger.add("group or world write permissions (%s)." % oct(permissions))
      gLogger.add("I don't think that's safe since so I'm not going to save to it.")
      gLogger.add("Of course, you could be in your very own group but I'm not")
      gLogger.add("smart enough to check that. Please fix the permissions.")
      return None

  os.umask(0077) # octal
  f=open(path,"wb")
  cPickle.dump(gState,f,True)
  f.close()
  return None


#
# prefs
#
# Used as global.
#

class location:
  pass

class prefs:

  allPrefsNames=(
    "addressbookData",
    "askBeforeFetchingMailThisBig",
    "autoStartCheckRegularly",
    "autoTrash",
    "boldFontILike",
    "calculateScore",
    "checkDelay",
    "cleanupForQuit",
    "decodeCommentaryFG",
    "decodeCommentaryBG",
    "defaultLocation",
    "defaultMailboxSort",
    "doHighScoreThing",
    "emptyTrashOnQuit",
    "fixedWidthFontILike",
    "fontILike",
    "forwardMunger",
    "fqdn",
    "getDefaultLocation",
    "getDefaultMailboxSort",
    "getSaveFoldersInbound",
    "getSaveFoldersInboundAlt",
    "getSaveFoldersOutbound",
    "hideMousePointerWhileTyping",
    "inboundFiler",
    "ispellLoc",
    "logFilePath",
    "mailboxNamesSortFunc",
    "mailboxSortOptions",
    "mailCheckAction",
    "maildirParent",
    "mailServer",
    "messageParseFunction",
    "mimeTypes",
    "myHumanName",
    "myMailAddress",
    "newMailMunger",
    "noisyScore",
    "nonBoringHeaders",
    "outboundMailWrapLen",
    "outboxColors",
    "pipeCommand",
    "popAuth",
    "popDele",
    "popPass",
    "popPort",
    "popServer",
    "popSSL",
    "popUser",
    "printCommand",
    "replyMunger",
    "replyTo",
    "sanityCheckOutboundMail",
    "sigFile",
    "smtpAuth",
    "smtpPort",
    "smtpServer",
    "smtpUser",
    "smtpPass",
    "smtpTLS",
    "spellingCheckColors",
    "specialCaseRewrapDearLine",
    "swallowDir",
    "swdbFile",
    "swdbShouldKeepFunc",
    "tempDir",
    "textSearchColors",
    "twoSpacesOK",
    "useSWDBNameOnReply",
    "userMailboxDirParent",
    "viewerQuoteColors",
    )

  def __init__(self):
    self.prefsPath=None
    self.globalPrefs=None
    self.currentLocation=None
    self.locationPrefs={}
    self.load()
    return None

  def load(self):
    self.getPath()
    self.checkFileAttributes()
    self.getDicts()
    self.applyDefaults()
    self.setInitialLocation()
    self.assembleSearchDict()
    # Could check the values before applying the defaults. As it is,
    # the user might be confused if defaults routine set something
    # wrong and user couldn't find that error in their file. But I
    # think it's better to catch the error.
    self.checkValues()
    return None

  def getPath(self):
    if len(sys.argv)==2:
      self.prefsPath=os.path.expanduser(sys.argv[1])
    if gWin32:
      # if no path specified, use the directory swallow.py is running from
      self.prefsPath=myPathJoin(os.path.dirname(sys.argv[0]),kWinPrefsFile)
    else: # Unix
      self.prefsPath=os.path.expanduser(kPrefsPath)
    return None

  def checkFileAttributes(self):
    if not os.path.exists(self.prefsPath):
      print 'My prefs file "%s" doesn\'t exist.' % self.prefsPath
      print "I can't run without it. Please see the documentation."
      sys.exit(0)
    prefsStat=os.stat(self.prefsPath)#.st_mode
    mode=prefsStat.st_mode
    if not stat.S_ISREG(mode):
      print 'My prefs file "%s" is not a regular file.' % self.prefsPath
      print "That's too weird for me. I don't want to run."
      sys.exit(0)

    if not gWin32: # Unix-specific checks
      if not pwd.getpwnam(os.environ["LOGNAME"])[2]==prefsStat.st_uid:
        print 'My prefs file "%s" is not owned by you (user %s), it is owned by user %s.' % \
          (self.prefsPath,os.environ["LOGNAME"],pwd.getpwuid(prefsStat.st_uid)[0])
        print "I don't think that can be right. I don't want to run."
        sys.exit(0)
      permissions=stat.S_IMODE(mode)
      if (permissions & 0020) or (permissions & 0002): # constants octal
        print 'My prefs file "%s" has' % self.prefsPath
        print "group or world write permissions (%s)." % oct(permissions)
        print "I don't think that's safe since stuff in there could cause me to do"
        print "things you don't want. Of course, you could be in your very own group"
        print "but I'm not smart enough to check that. Please fix the permissions."
        sys.exit(0)
    return None

  def getDicts(self):
    # Could use rexec module's r_execfile() and r_eval() here but
    # user's getDefaultLocation() could want to use the socket
    # module (to get the machine's IP, for example) and we've checked
    # that the prefs file has the right ownership and permissions so
    # it doesn't seem worth it.

    #d={"location": location, "gSwitchboard": gSwitchboard} # XXX test
    d={"location": location, "gLogger": gLogger}
    # Load file as Python code, writing stuff to new namespace
    execfile(self.prefsPath,d)
    locationPrefs={}
    for k in d.keys():
      if isinstance(d[k],location):
        locationPrefs[k]=d[k].__dict__
        del d[k]
    self.globalPrefs=d
    self.locationPrefs=locationPrefs
    return None

  def applyDefaults(self):
    # This is slightly obscure. First for all sets of preferences, set preferences
    # that depend on other perference settings.

    allPrefsDicts=[self.globalPrefs]+self.locationPrefs.values()
    for thisDict in allPrefsDicts:
      if thisDict.has_key("mailServer") and not thisDict.has_key("smtpServer"):
        thisDict["smtpServer"]=thisDict["mailServer"]
      if thisDict.has_key("mailServer") and not thisDict.has_key("popServer"):
        thisDict["popServer"]=thisDict["mailServer"]
      if thisDict.has_key("popSSL") and thisDict["popSSL"] and not thisDict.has_key("popPort"):
        thisDict["popPort"]=995

      # If userMailboxDirParent not abusolute, prepend home
      if thisDict.has_key("userMailboxDirParent") and not os.path.isabs(thisDict["userMailboxDirParent"]):
        thisDict["userMailboxDirParent"]=myPathJoin(os.path.expanduser("~"),thisDict["userMailboxDirParent"])

    # Then apply the real defaults. They're only applied to the global (i.e. not
    # location-dependant) prefs. Since prefs that are for the current location
    # are searched first, any settings that are for the current location will
    # override these defaults.

    p=self.globalPrefs

    p.setdefault("smtpPort",25)
    p.setdefault("smtpAuth",False)
    p.setdefault("smtpTLS",False)
    p.setdefault("smtpUser","")
    p.setdefault("smtpPass","")
    p.setdefault("popPort",110)
    p.setdefault("popDele",1)
    p.setdefault("popAuth","userpass")
    p.setdefault("popSSL",False)
    #XXXX need to do something about this
    if not gWin32:
      p.setdefault("popUser",os.environ["LOGNAME"])
    p.setdefault("popPass",None)
    p.setdefault("checkDelay",300)
    p.setdefault("outboxColors",("black","dodgerblue","red"))
    p.setdefault("maildirParent",os.path.expanduser("~"))
    p.setdefault("userMailboxDirParent",os.path.expanduser("~/Mail"))
    p.setdefault("emptyTrashOnQuit",0)
    p.setdefault("doHighScoreThing",nullCB)
    p.setdefault("noisyScore",99)
    p.setdefault("nonBoringHeaders",["from","to","cc","bcc","subject","date","reply-to"])
    p.setdefault("ispellLoc","/usr/bin/ispell")
    p.setdefault("fontILike","7x13")
    p.setdefault("fixedWidthFontILike","7x13")
    p.setdefault("boldFontILike","7x13bold")
    p.setdefault("autoTrash",[])
    #if p.has_key("calculateScore"):
    #  p.setdefault("sortOnScoreDefault",1)
    #else:
    #  p.setdefault("sortOnScoreDefault",0)
    p.setdefault("calculateScore",None)
    p.setdefault("twoSpacesOK",0)
    p.setdefault("replyTo",None)
    p.setdefault("printCommand","/bin/cat > /dev/null")
    p.setdefault("spellingCheckColors",("red","purple","yellow1"))
    p.setdefault("textSearchColors",("white","blue"))
    p.setdefault("sigFile",None)
    p.setdefault("pipeCommand",None)
    p.setdefault("addressbookData",[])
    p.setdefault("swdbFile",None)
    p.setdefault("swdbShouldKeepFunc", lambda x: 1)
    p.setdefault("askBeforeFetchingMailThisBig",10*1024*1024)
    p.setdefault("decodeCommentaryFG","white")
    p.setdefault("decodeCommentaryBG","black")
    p.setdefault("mailCheckAction","pop")
    p.setdefault("outboundMailWrapLen",70)
    p.setdefault("replyMunger",None)
    p.setdefault("newMailMunger",None)
    p.setdefault("forwardMunger",None)
    p.setdefault("inboundFiler",None)
    p.setdefault("fqdn",socket.getfqdn())
    p.setdefault("specialCaseRewrapDearLine",1)
    p.setdefault("sanityCheckOutboundMail",1)
    p.setdefault("hideMousePointerWhileTyping",0)
    p.setdefault("autoStartCheckRegularly",0)
    p.setdefault("viewerQuoteColors",None)
    p.setdefault("useSWDBNameOnReply",None)
    p.setdefault("getSaveFoldersInbound",lambda : ["saved"])
    p.setdefault("getSaveFoldersInboundAlt",None)
    p.setdefault("getSaveFoldersOutbound",lambda : ["sent"])
    p.setdefault("swallowDir",os.path.expanduser("~/.swallow"))
    p.setdefault("cleanupForQuit",nullCB)
    p.setdefault("tempDir",myPathJoin(p["swallowDir"],"temp"))
    p.setdefault("logFilePath",myPathJoin(p["tempDir"],"swallowlog"))
    p.setdefault("messageParseFunction",email.message_from_file)

    # The function getDefaultMailboxSort takes precedence over
    # the dict defaultMailboxSort. So we could do a setdefault()
    # with the dict here regardless. But then it wouldn't make
    # sense to complain below in checkValues() if they're both
    # set
    if p.has_key("getDefaultMailboxSort"):
      p.setdefault("defaultMailboxSort",None)
    else:
      if p.has_key("calculateScore"):
        sc=1
      else:
        sc=0
      p.setdefault("defaultMailboxSort",{"thread": 1, "score": sc, "date": "ascending", "whichDate": "sent"})
    p.setdefault("getDefaultMailboxSort",None)
    return None

  def setInitialLocation(self):
    if len(self.locationPrefs.keys())>0:
      if self.globalPrefs.has_key("defaultLocation"):
        loc=self.globalPrefs["defaultLocation"]
      elif self.globalPrefs.has_key("getDefaultLocation"):
        # XXX check that returned value makes sense XXX yes, do
        loc=self.globalPrefs["getDefaultLocation"]()
      else:
        loc=self.locationPrefs.keys()[0]
        gLogger.add("Caution:")
        gLogger.add("Current location not set by preferences file.")
        gLogger.add("Arbitrarily setting location to %s." % loc)
      self.currentLocation=loc
    return None

  def assembleSearchDict(self):
    mainDict=copy.copy(self.globalPrefs)
    # XXX make sure that currentLocation makes sense
    if self.currentLocation<>None:
      mainDict.update(self.locationPrefs[self.currentLocation])
    self.searchDict=mainDict
    return None

  def checkValues(self):
    err=0
    passInPrefs=0
    allPrefsDicts=[self.globalPrefs]+self.locationPrefs.values()
    for item in allPrefsDicts:
      if item.has_key("popPass") and item["popPass"]<>"" and item["popPass"]<>None:
        passInPrefs=1
      else:
        if item.has_key("autoStartCheckRegularly") and item["autoStartCheckRegularly"]:
          gLogger.add("You have autoStartCheckRegularly set but no password stored.",severity=kWarning)
          gLogger.add("That doesn't make much sense.",severity=kWarning)

    #XXXX need to do something about win32 permissions
    if not gWin32:
      permissions=stat.S_IMODE(os.stat(self.prefsPath).st_mode)
      if passInPrefs and (permissions & 0040) or (permissions & 0004): # constants octal
        err=1
        print "My prefs file %s has group or world read" % self.prefsPath
        print "permissions (%s) and you have passwords stored in it. You must" % oct(permissions)
        print "make it readable only by you before I'll be willing to run. Of"
        print "course you could be in your very own group but I'm not smart enough"
        print "to check that."

    if self.globalPrefs.has_key("defaultLocation") and self.globalPrefs.has_key("getDefaultLocation"):
      print "Your preferences file defines both defaultLocation and"
      print "getDefaultLocation. Only defaultLocation will be used."

    for loc in self.locationPrefs.keys():
      d=self.locationPrefs[loc]
      for pref in ("defaultLocation","getDefaultLocation"):
        if d.has_key(pref):
          print "The location %s in your preferences file defines" % loc
          print "has the value %s defined. That preference can only" % pref
          print "be set at the top level."

    val=self.get("myMailAddress")
    if val=="example@example.org" or val=="":
      err=1
      print "You must set your address in the preferences file."

    val=self.get("myHumanName")
    if val=="Swallow user":
      err=1
      print "You must set your name in the preferences file."

    val=self.get("checkDelay")
    if type(val)<>types.IntType or val<1:
      err=1
      print "Value for checkDelay (%s) in current location in %s" % (str(val),os.path.basename(self.prefsPath))
      print "doesn't make any sense."
    elif val<300 and self.get("mailCheckAction")=="pop":
      gLogger.add("Caution:")
      gLogger.add("checkDelay in your current location in %s is very small" % os.path.basename(self.prefsPath))
      gLogger.add("(%i seconds). Are you sure that your postmaster won't yell at you?" % val)

    val=self.get("userMailboxDirParent")
    if not os.path.isdir(val):
      most,tail=myPathSplit(val)
      if os.path.exists(val):
        err=1
        print "The path that you specified for your (saved-mail, etc.)"
        print "mailboxes (%s) exists but it's not a directory." % val
        print "I'm not going to try to guess what you want here."
      elif not os.path.exists(most):
        err=1
        print "The directory above the one you specified for your (saved-mail, etc.)"
        print "mailboxes doesn't exist. I'll be happy to create the last bit, but"
        print "would you please create %s for me?" % most
      elif not os.path.isdir(most):
        err=1
        print "The directory above the one you specified for your (saved-mail, etc.)"
        print "mailboxes (%s) exists but it's not a directory." % most
        print "Would you please sort that out for me?"
      else:
        gLogger.add("Creating %s" % val)
        os.umask(0077)
        os.mkdir(val)

    val=self.get("tempDir")
    home=os.path.expanduser("~")
    if not gWin32 and not os.path.commonprefix((val,home))==home:
      gLogger.add("Caution:")
      gLogger.add("The temp directory set in the prefs file is not in your home directory.")
      gLogger.add("In my opinion, that's generally an unsafe thing to do.")

    for addressTuple in self.get("addressbookData"):
      if addressTuple[0].find("@")<>-1:
        gLogger.add("Warning:")
        gLogger.add('At-sign in address nickname "%s"' % addressTuple[0],)
        gLogger.add("that nickname won't work")
      if addressTuple[0].find(",")<>-1:
        gLogger.add("Warning:")
        gLogger.add('Comma in address nickname "%s"' % addressTuple[0],)
        gLogger.add("that nickname won't work")
      if addressTuple[0].find(" ")<>-1:
        gLogger.add("Warning:")
        gLogger.add('Space in address nickname "%s"' % addressTuple[0],)
        gLogger.add("that nickname won't work")
      if addressTuple[0]=="me":
        gLogger.add("Warning:")
        gLogger.add('Nickname "me" in addressbook.',)
        gLogger.add('"me" is added automatically with your address.')
        gLogger.add("The one in your addressbook won't work.")

    if self.get("swdbFile") and hasattr(self.globalPrefs,"addressbookData"):
      gLogger.add("Caution:")
      gLogger.add("You have addressbook data in your prefs file and you also have swdb")
      gLogger.add("enabled. I will use only the data from swdb.")

    val=self.get("mailCheckAction")
    if val<>"pop" and val<>"scan":
      err=1
      print 'Preference mailCheckAction is "%s" but must be one of "pop" or "scan"'

    val=self.get("ispellLoc")
    if val<>None:
      if not os.path.exists(val):
        gLogger.add("Warning:")
        gLogger.add("Path specified in prefs for ispell (%s) doesn't exist" % val)
      elif not os.path.isfile(val):
        gLogger.add("Warning:")
        gLogger.add("Path specified in prefs for ispell (%s) exists but isn't a file" % val)
      elif not stat.S_IMODE(os.stat(val).st_mode) & 0001:
        gLogger.add("Warning:")
        gLogger.add("Path specified in prefs for ispell (%s) exists but isn't an executable program" % val)

    val=self.get("outboundMailWrapLen")
    if val<60 or val>80:
      gLogger.add("Caution:")
      gLogger.add("Value for outboundMailWrapLen (%i) is unusual. A value" % val)
      gLogger.add("between 60 and 80 is recommended")

    val1=self.get("getDefaultMailboxSort")
    val2=self.get("defaultMailboxSort")
    if val1<>None and val2<>None:
      gLogger.add("Your preferences defines both defaultMailboxSort and")
      gLogger.add("getDefaultMailboxSort in the same location. Only")
      gLogger.add("getDefaultMailboxSort will be used.")
      
    val1=self.get("smtpAuth")
    val2=self.get("smtpUser")
    val3=self.get("smtpPass")
    if val1:
      if val2=="":
        gLogger.add("SMTP authentication turned on but SMTP username not set.")
        err=True
      if val3=="":
        gLogger.add("SMTP authentication turned on but SMTP password not set.")
        err=True
    else:
      if val2<>"":
        gLogger.add("Caution: SMTP authentication turned off but SMTP username set.")
      if val3<>"":
        gLogger.add("Cuation: SMTP authentication turned off but SMTP password set.")

    # Not so many of them that a linear search is a big problem
    for dict in allPrefsDicts:
      for name in dict.keys():
        if name=="location" or name=="gLogger" or name=="__builtins__":
          continue

        val=dict[name]
        if isinstance(val,types.ModuleType):
          continue

        if name not in prefs.allPrefsNames:
          suggList=difflib.get_close_matches(name,prefs.allPrefsNames)

          # User may define utuility functions in prefs file and we don't
          # want to complain about them. So don't complain about a function
          # name unless it's close to something that should be there. Not
          # ideal but it's the best compromise I can think of.
          if isinstance(val,types.FunctionType) and len(suggList)==0:
            continue

          gLogger.add("Warning:")
          gLogger.add('Preference "%s" is set in your preferences file but there is' % name)
          gLogger.add("no preference of that name.")

          if len(suggList)>0:
            if len(suggList)==1:
              strSugg=suggList[0]
            elif len(suggList)==2:
              strSugg=" or ".join(suggList)
            else:
              strSugg=", ".join(suggList[:-1])+", or "+suggList[-1]
            gLogger.add('Did you perhaps mean "%s"?' % strSugg)


    if err:
      sys.exit(1)
    else:
      return None

  def get(self,prefName):
    return self.searchDict[prefName]

  def getLocations(self):
    locs=self.locationPrefs.keys()
    locs.sort()
    return locs

  def setLocation(self,newLocation):
    assert self.locationPrefs.has_key(newLocation)
    self.currentLocation=newLocation
    self.assembleSearchDict()
    # In case we're switching to a location that uses the scanner
    gChecker.currentInboxAlreadyDisplayed()
    return None

  def getLocation(self):
    return self.currentLocation


class gridRow:
  def __init__(self,parent):
    self.parent=parent
    self.labels=[]
    self.blank=None
    self.createWidgets()
    return None

  def createWidgets(self):
    self.outerFrame=Frame(self.parent.getGridParentWidget())
    # We're normally used in the overview of a mailbox in
    # which case, specs for width and height come from the
    # mailbox (width can change). But we're also used in
    # the overview of found items in the search window. In
    # that case, the parent doesn't supply that information.
    howmany=len(self.parent.widthsInDots)
    for count in range(howmany):
      l=labelWithSizeInDots(self.outerFrame,
        width=self.parent.widthsInDots[count],
        height=self.parent.rowHeight,
        anchor=W,
        bg="white",
        font=self.parent.font,
        )
      l.pack(side=LEFT)
      l.bind("<Button-1>",self.childClickedCB)
      l.bind("<Button-3>",self.childClicked3CB)
      self.labels.append(l)
      # Spacer but not after the last one
      if count<len(self.parent.widthsInDots)-1:
        spacer=Frame(self.outerFrame,width=1,height=self.parent.rowHeight)
        spacer.pack(side=LEFT)
        spacer.bind("<Button-1>",self.childClickedCB)
    self.outerFrame.pack(side=TOP,fill=X,expand=1) ###?????????
    return None

  def newWidth(self,index,width):
    self.labels[index].configure(width=width)
    return None

  def newText(self,textItems):
    if len(textItems)==0:
      self.blank=1
      for count in range(len(self.labels)):
        try:
          self.labels[count].configure(text="")
        except TclError:
          pass
    else:
      self.blank=0
      for count in range(len(self.labels)):
        try:
          self.labels[count].configure(text=textItems[count])
        except TclError:
          pass
    return None

  def childClickedCB(self,eventIngored):
    if not self.isBlank():
      self.parent.childClicked(self)
    return "break"

  def childClicked3CB(self,event):
    if not self.isBlank():
      colNum=[l.interior() for l in self.labels].index(event.widget)
      self.parent.childClick3(self,colNum,event)
    return "break"

  def isBlank(self):
    return self.blank

  def getLabels(self):
    return self.labels

  def highlight(self,onOff):
    for child in self.labels:
      if onOff:
        child.configure(fg="white",bg="black")
      else:
        child.configure(fg="black",bg="white")
    return None

  def setColors(self,foreground,background):
    for child in self.labels:
      child.configure(fg=foreground,bg=background)
    return None

  def destroy(self):
    self.outerFrame.destroy()
    return None

# Much of the code in this widget inspired by the
# Pmw.ScrolledFrame widget. Thanks to Greg McFarlane.

# Limitation here: If content rows are narrower than
# outer frame's initial size, they won't expand to fill
# it. Start with width same as or smaller than row width.
# Hard to avoid the limitaiton because sizes aren't
# calculated until the thing is drawn and fixing it after
# it's drawn looks funny. If it's expanded later, the
# frame will grow just fine as long as the last label
# widget is packed with fill=X, expand=1.

class gridWithHeaders:
  ## defaults here, rowHeight, widths, xjump
  def __init__(self,parent,
      tableLinesToShow=None,
      headerStrs=None,
      defaultWidthsChars=None,
      rowHeightDots=None,
      textGetCB=None,
      rowConfigureCB=None,
      contentsLenGetCB=None,
      doubleClickCB=None,
      singleClickCB=None,
      click3CB=None,
      xScrollJump=None,
      font=None,
      ):

    self.parent=parent
    self.tableLinesToShow=tableLinesToShow
    assert len(headerStrs)==len(defaultWidthsChars)
    self.headerStrs=headerStrs
    self.nColumns=len(self.headerStrs)
    self.font=font

    # calc widths if None, lineHeight, rowHeight, nColumns
    self.widthsInDots=self.charWidthsToDotWidths(defaultWidthsChars)

    if rowHeightDots==None:
      l=Label(self.parent.getWidgetParent(),text=" ",font=font)
      rowHeightDots=l.winfo_reqheight()+1
    self.rowHeight=rowHeightDots

    if xScrollJump==None:
      xScrollJump=15
    self.xScrollJumpDots=xScrollJump
    self.colMinWidth=5
    self.rowsAreNarrowerThanWidget=0 ###XXX
    #self.initScrollbarSetDelay=1000
    self.checkForWidthsSortedOutDelay=10
    self.maxDoubleClickDelayMS=500
    self.scrollAgainDelayMS=100
    self.textGetCB=textGetCB
    self.contentsLenGetCB=contentsLenGetCB
    self.doubleClickCB=doubleClickCB
    self.singleClickCB=singleClickCB
    self.childClick3CB=click3CB
    self.rowConfigureCB=rowConfigureCB

    self.headerLabels=[]
    self.headerDividers=[]
    self.rows=[]
    #self.headerDraggingInfo=None
    self.oldHeight=0
    self.oldWidth=0
    self.partRowVisible=0
    self.startX=0
    self.highlightedContentsIndex=0
    self.indexAtTop=0
    self.sbDragStartX=None
    self.sbDragStartY=None
    self.lastChildClicked=None
    self.draggingInfo=None
    self.lastClickTime=0
    self.mouseStillDown=0
    self.createWidgets()
    # Could send after to any widget
    self.outerFrame.after(self.checkForWidthsSortedOutDelay,self.setValuesWhenWidthsSortedOut)
    return None

  def createWidgets(self):
    self.outerFrame=Frame(self.parent.getWidgetParent())
    self.headersClipper=Frame(self.outerFrame,height=self.rowHeight-1)
    self.headersClipper.pack_propagate(0)
    self.headersFrame=Frame(self.headersClipper,height=self.rowHeight)
    for count in range(self.nColumns):
      l=labelWithSizeInDots(self.headersFrame,
        anchor=W,
        width=self.widthsInDots[count],
        height=self.rowHeight-1,
        text=self.headerStrs[count],
        font=self.font,
        )
      l.pack(side=LEFT)
      l.bind("<Button-1>",self.labelClickedCB)
      l.bind("<ButtonRelease-1>",self.labelUnClickedCB)
      l.bind("<B1-Motion>",self.labelMovedCB)
      l.bind("<Motion>",self.mouseInLabelCB)
      self.headerLabels.append(l)
      if count<len(self.widthsInDots)-1:
        f=Frame(self.headersFrame,width=1,bg="black",cursor="sb_h_double_arrow")
        f.pack(side=LEFT,fill=Y)
        f.bind("<Button-1>",self.dividerClickedCB)
        f.bind("<ButtonRelease-1>",self.dividerUnClickedCB)
        f.bind("<B1-Motion>",self.dividerMovedCB)
        self.headerDividers.append(f)
    self.headersFrame.place(x=0,y=0)
    self.headersClipper.pack(side=TOP,fill=X)
    lineBelow=Frame(self.outerFrame,bg="black",height=1)
    lineBelow.pack(side=TOP,fill=X)
    
    self.lowerFrame=Frame(self.outerFrame,height=self.tableLinesToShow*self.rowHeight+kScrollbarWidth)

    vsb=Scrollbar(self.lowerFrame,
      activebackground=kTkDefaultBG,
      width=kScrollbarWidth-2, # Subtract 2 for border
      borderwidth=1,
      )
    vsb.place(relheight=1,
      height=-kScrollbarWidth,
      width=kScrollbarWidth,
      relx=1,
      x=-kScrollbarWidth,
      )
    vsb.bind("<Button-1>",self.vScrollbarClickedCB)
    vsb.bind("<ButtonRelease-1>",self.vScrollbarUnClickedCB)
    vsb.bind("<Button1-Motion>",self.vScrollbarDraggedCB)
    self.vsb=vsb

    hsb=Scrollbar(self.outerFrame,
      orient=HORIZONTAL,
      activebackground=kTkDefaultBG,
      width=kScrollbarWidth-2, # Subtract 2 for border
      borderwidth=1,
      )
    hsb.place(height=kScrollbarWidth,
      relwidth=1,
      width=-kScrollbarWidth,
      rely=1,
      y=-kScrollbarWidth
      )
    hsb.bind("<Button-1>",self.hScrollbarClickedCB)
    hsb.bind("<ButtonRelease-1>",self.hScrollbarUnClickedCB)
    hsb.bind("<Button1-Motion>",self.hScrollbarDraggedCB)
    self.hsb=hsb
    
    self.tableClipper=Frame(self.lowerFrame,
      height=self.tableLinesToShow*self.rowHeight,
      bg="white",
      )
    self.tableClipper.bind("<Configure>",self.resizeContentsCB)
    self.tableFrame=Frame(self.tableClipper,height=self.tableLinesToShow*self.rowHeight)
    self.tableFrame.place(x=0,y=0)###,width=400)###XXXX
    self.tableClipper.place(x=0,y=0,
      relheight=1,
      relwidth=1,
      height=-kScrollbarWidth,
      width=-kScrollbarWidth)
    #self.dummy=Frame(self.tableClipper,height=300,width=10,bg="green")
    #self.dummy.pack()
    self.lowerFrame.pack(side=TOP,fill=BOTH,expand=1)
    self.msgLabel=Label(self.tableClipper,bg="white",font=self.font)
    return None

  def setValuesWhenWidthsSortedOut(self):
    if self.tableFrame.winfo_width()==1:
      self.tableFrame.after(self.checkForWidthsSortedOutDelay,self.setValuesWhenWidthsSortedOut)
    else:
      self.setHScrollbar()
      self.checkForNarrowRows()
      #self.checkForPartRowVisible()
    return None

  def getGridParentWidget(self):
    return self.tableFrame

  def getHeaderLabels(self):
    return self.headerLabels

  def resizeContentsCB(self,event):
    print "eh",event.height ### XXX odd!
    if event.height<>self.oldHeight:
      nRows=int(event.height/self.rowHeight)
      if nRows*self.rowHeight<event.height:
        self.partRowVisible=1
        nRows+=1
      else:
        self.partRowVisible=0

      if len(self.rows)>nRows:
        for count in range(nRows,len(self.rows)):
          self.rows[count].destroy()
        self.rows=self.rows[:nRows]
      elif len(self.rows)<nRows:
        for count in range(len(self.rows),nRows):
          m=gridRow(self)
          if self.rowConfigureCB<>None:
            self.rowConfigureCB(m)
          self.rows.append(m)
      self.oldHeight=event.height
    self.redisplay()
    if self.rowsAreNarrowerThanWidget and event.width<>self.oldWidth:
      self.tableFrame.place(width=self.tableClipper.winfo_width())
      self.oldWidth=event.width
      
    print "r",len(self.rows)
    return None

##  ## XXX
##  def checkForPartRowVisible(self):
##    print "x"
##    print "h", self.tableClipper.winfo_height()
##    print "q", self.tableClipper.winfo_reqheight()
##    print "r",self.rowHeight
##    print "l",len(self.rows)
##    if self.tableClipper.winfo_height() % self.rowHeight<>0:
##      print "y"
##      self.partRowVisible=1
##    else:
##      print "z"
##      self.partRowVisible=0
##    return None
##

  def checkForNarrowRows(self):
    self.rowsAreNarrowerThanWidget=self.tableFrame.winfo_width()<=self.tableClipper.winfo_width()
    if self.rowsAreNarrowerThanWidget:
      self.tableFrame.place(width=self.tableClipper.winfo_width())
    return None

  def getHighlightIndex(self):
    return self.highlightedContentsIndex

  def setHighlightIndex(self,ind):
    self.highlightedContentsIndex=ind
    return None

  def moveSelection(self,direction):
    assert abs(direction)==1

    if self.highlightedContentsIndex==None:
      return None

    if direction==-1 and self.highlightedContentsIndex==0:
      return None
    lenContents=self.contentsLenGetCB()
    if direction==1 and self.highlightedContentsIndex==lenContents-1:
      return None

    self.highlightedContentsIndex+=direction

    if self.highlightedContentsIndex<self.indexAtTop:
      self.indexAtTop-=1
      self.redisplay()
    elif self.highlightedContentsIndex>=self.indexAtTop+len(self.rows):
      self.indexAtTop+=1
      self.redisplay()
    else:
      self.redoHighlight()
    return None

  def redisplay(self):
    contentsLen=self.contentsLenGetCB()

    # Contents length may have shrunk. If so, fix scroll.
    maxIndexAtTop=contentsLen-len(self.rows)
    if self.partRowVisible:
      maxIndexAtTop+=1
    maxIndexAtTop=max(0,maxIndexAtTop)
    if self.indexAtTop>maxIndexAtTop:
      self.indexAtTop=maxIndexAtTop

    # If there aren't any items, don't highlight but
    # if we weren't highlighting and now there are items
    # highlight the first.
    if contentsLen==0:
      self.highlightedContentsIndex=None
    elif self.highlightedContentsIndex==None:
      self.highlightedContentsIndex=0
    # Don't let highlight fall off end of used rows
    elif self.highlightedContentsIndex>0 and self.highlightedContentsIndex>=self.contentsLenGetCB():
      self.highlightedContentsIndex=self.contentsLenGetCB()-1

    for count in range(len(self.rows)):
      contentsIndex=count+self.indexAtTop
      if contentsIndex>=contentsLen:
        text=[]
      else:
        text=self.textGetCB(contentsIndex)
      self.rows[count].newText(text)
      if count+self.indexAtTop==self.highlightedContentsIndex:
        self.rows[count].highlight(1)
      else:
        self.rows[count].highlight(0)
    self.setVScrollbar()
    self.setHScrollbar()
    return None

  def updateDisplayOf(self,index):
    rowIndex=index-self.indexAtTop
    if rowIndex>=0 and rowIndex<len(self.rows):
      self.rows[rowIndex].newText(self.textGetCB(index))
    return None

  def redoHighlight(self):
    for count in range(len(self.rows)):
      if count+self.indexAtTop==self.highlightedContentsIndex:
        self.rows[count].highlight(1)
      else:
        self.rows[count].highlight(0)
    return None

  def childClicked(self,child):
    childIndex=self.rows.index(child)
    if self.lastChildClicked==child and (time.time()-self.lastClickTime)*1000<=self.maxDoubleClickDelayMS:
      self.lastChildClicked=None
      self.lastClickTime=0
      self.doubleClickCB(childIndex+self.indexAtTop)
    else:
      self.lastChildClicked=child
      self.lastClickTime=time.time()
      self.highlightedContentsIndex=childIndex+self.indexAtTop
      self.redoHighlight()
    if self.singleClickCB<>None:
      self.singleClickCB(self.highlightedContentsIndex)
    return None

  def childClick3(self,child,colNo,event):
    if self.childClick3CB<>None:
      self.childClick3CB(colNo,self.rows.index(child)+self.indexAtTop,event)
    return None

  def dividerClickedCB(self,event):
    self.processDividerClicked(event.widget,event.x_root)
    return "break"

  def dividerUnClickedCB(self,eventIgnored):
    self.processDividerUnClicked()
    return "break"

  def processDividerClicked(self,widget,x):
    myIndex=self.headerDividers.index(widget)
    self.draggingInfo=bunch(index=myIndex,
      startX=x,
      startWidth=self.headerLabels[myIndex].winfo_width(),
      lastHeaderWidth=self.headerLabels[-1].winfo_width(),
      )
    return None

  def processDividerUnClicked(self):
    self.draggingInfo=None
    return None

  def dividerMovedCB(self,event):
    assert self.draggingInfo<>None,"Divider drag without click"
    self.processDividerDrag(event.x_root)
    return "break"

  def processDividerDrag(self,x):
    delta=int(x-self.draggingInfo.startX)
    #labelToResize=self.headerLabels[self.draggingInfo.index]
    newWidth=max(self.colMinWidth,self.draggingInfo.startWidth+delta)
    self.headerLabels[self.draggingInfo.index].configure(width=newWidth)
    if self.rowsAreNarrowerThanWidget:
      self.tableFrame.place(width=self.tableClipper.winfo_width())

    self.newColumnWidth(self.draggingInfo.index,newWidth)
    # Do update only occasionally?
    self.outerFrame.update_idletasks() # Could send to any widget
    return None

  def labelClickedCB(self,event):
    for labelIndex in range(len(self.headerLabels)):
      if self.headerLabels[labelIndex].interior()==event.widget:
        break
    else:
      assert 1==0,"Label not found"

    # If clicked close to a divider, treat it as clicking on the divider
    if event.x<=2 and labelIndex>0:
      self.processDividerClicked(self.headerDividers[labelIndex-1],event.x_root)
    elif event.widget.winfo_width()-event.x<=2 and labelIndex<len(self.headerLabels)-1:
      self.processDividerClicked(self.headerDividers[labelIndex],event.x_root)
    return "break"

  def labelUnClickedCB(self,eventIgnored):
    self.processDividerUnClicked()
    return "break"

  # Called repeatedly while mouse is in label.
  # The dividers are only one dot wide and therefore a nuisance to click on.
  # So we treat a click on a label that's close to a divider as though it
  # were on the divider. Here, we set the mouse pointer so that it
  # looks right for what the user is going to get.
  def mouseInLabelCB(self,event):
    for labelIndex in range(len(self.headerLabels)):
      if self.headerLabels[labelIndex].interior()==event.widget:
        break
    else:
      assert 1==0,"Label not found"

    if (event.x<=2 and labelIndex>0) or (event.widget.winfo_width()-event.x<=2 and labelIndex<len(self.headerLabels)-1):
      event.widget.configure(cursor="sb_h_double_arrow")
    else:
      event.widget.configure(cursor="")
    return None

  def labelMovedCB(self,event):
    if self.draggingInfo<>None:
      self.processDividerDrag(event.x_root)
    return "break"

  def newColumnWidth(self,index,width):
    self.widthsInDots[index]=width
    self.headerLabels[index].configure(width=width)
    for r in self.rows:
      r.newWidth(index,width)
    return None

  def vScrollbarClickedCB(self,event):
    self.mouseStillDown=1
    whatClicked=self.vsb.identify(event.x,event.y)
    self.mouseDownOn=whatClicked
    if whatClicked=="arrow1":
      self.scrollLines(-1)
      self.outerFrame.after(self.scrollAgainDelayMS,self.scrollAgain,"up-line")
    elif whatClicked=="arrow2":
      self.scrollLines(1)
      self.outerFrame.after(self.scrollAgainDelayMS,self.scrollAgain,"down-line")
    elif whatClicked=="trough1":
      self.scrollPagesV(-1)
      self.outerFrame.after(self.scrollAgainDelayMS,self.scrollAgain,"up-page")
    elif whatClicked=="trough2":
      self.scrollPagesV(1)
      self.outerFrame.after(self.scrollAgainDelayMS,self.scrollAgain,"down-page")
    elif whatClicked=="slider":
      self.sbDragStartY=event.y
    elif whatClicked=="":
      pass # Possible to click on the border and get "" for identify
    else:
      assert 1==0,'Scroll click on "%s", don\'t know what that is'% whatClicked
    return None

  def vScrollbarUnClickedCB(self,eventIgnored):
    self.mouseStillDown=0
    return None

  def vScrollbarDraggedCB(self,event):
    if self.mouseStillDown and self.mouseDownOn=="slider":
      contentsLen=self.contentsLenGetCB()
      frac=self.vsb.delta(0,event.y-self.sbDragStartY)
      scrollLines=int(frac*contentsLen)
      if scrollLines<>0:
        self.scrollLines(scrollLines)
        self.sbDragStartY=event.y
    return None

  def hScrollbarClickedCB(self,event):
    self.mouseStillDown=1
    whatClicked=self.hsb.identify(event.x,event.y)
    self.mouseDownOn=whatClicked
    if whatClicked=="arrow1":
      self.scrollXJump(-1)
      self.outerFrame.after(self.scrollAgainDelayMS,self.scrollAgain,"left-char")
    elif whatClicked=="arrow2":
      self.scrollXJump(1)
      self.outerFrame.after(self.scrollAgainDelayMS,self.scrollAgain,"right-char")
    elif whatClicked=="trough1":
      self.scrollPagesX(-1)
      self.outerFrame.after(self.scrollAgainDelayMS,self.scrollAgain,"left-page")
    elif whatClicked=="trough2":
      self.scrollPagesX(1)
      self.outerFrame.after(self.scrollAgainDelayMS,self.scrollAgain,"right-page")
    self.sbDragStartX=event.x
    return None

  def hScrollbarUnClickedCB(self,eventIgnored):
    self.mouseStillDown=0
    return None

  def hScrollbarDraggedCB(self,event):
    if self.mouseDownOn=="slider":
      maxScrollDots=-(self.tableFrame.winfo_width())
      scrollDots=self.hsb.delta(event.x-self.sbDragStartX,0)*maxScrollDots
      scrollDots=-int(round(scrollDots,0))
      if scrollDots<>0:
        self.scrollX(scrollDots)
        self.sbDragStartX=event.x
    return None

  def scrollAgain(self,direction):
    if self.mouseStillDown:
      if direction=="up-line":
        self.scrollLines(-1)
      elif direction=="down-line":
        self.scrollLines(1)
      elif direction=="up-page":
        self.scrollPagesV(-1)
      elif direction=="down-page":
        self.scrollPagesV(1)
      elif direction=="left-char":
        self.scrollXJump(-1)
      elif direction=="right-char":
        self.scrollXJump(1)
      elif direction=="left-page":
        self.scrollPagesX(-1)
      elif direction=="right-page":
        self.scrollPagesX(1)
      self.outerFrame.after(self.scrollAgainDelayMS,self.scrollAgain,direction)
    return None

  def setHScrollbar(self):
    contentsWidth=float(self.tableFrame.winfo_width())
    left=-self.startX/contentsWidth
    right=(-self.startX+self.tableClipper.winfo_width())/contentsWidth
    self.hsb.set(left,right)
    return None

  def setVScrollbar(self):
    contentsLen=self.contentsLenGetCB()
    if contentsLen==0:
      self.vsb.set(0.0,1.0)
    else:
      top=float(self.indexAtTop)/contentsLen
      bottom=min(1.0,float((self.indexAtTop+len(self.rows)))/contentsLen)
      self.vsb.set(top,bottom)
    return None

  def scrollXJump(self,direction):
    self.scrollX(direction*self.xScrollJumpDots)
    return None

  def scrollPagesX(self,direction):
    units=direction*(self.tableClipper.winfo_width()/2)
    self.scrollX(units)
    return None

  def scrollX(self,numDots):
    maxX=0
    minX=self.tableClipper.winfo_width()-self.tableFrame.winfo_width()
    wantX=self.startX-numDots
    getX=min(max(minX,wantX),maxX)
    self.startX=getX
    #self.parent.columnLabels.newXOrigin(getX)
    self.tableFrame.place(x=self.startX,rely=0,relheight=1)
    self.headersFrame.place(x=self.startX,y=0)
    self.setHScrollbar()
    return None

  def scrollLines(self,nLines):
    lenContents=self.contentsLenGetCB()
    wantIndexAtTop=self.indexAtTop+nLines
    maxIndexAtTop=lenContents-len(self.rows)
    if self.partRowVisible:
      maxIndexAtTop+=1
    getIndexAtTop=min(maxIndexAtTop,wantIndexAtTop)
    getIndexAtTop=max(0,getIndexAtTop)
    if getIndexAtTop==self.indexAtTop:
      return None
    self.indexAtTop=getIndexAtTop
    self.redisplay()
    return None

  def scrollPagesV(self,nPages):
    onePageLines=len(self.rows)-1
    self.scrollLines(nPages*onePageLines)
    return None

  def charWidthsToDotWidths(self,charWidthsSeq):
    columnWidths=[]
    f=tkFont.Font(font=self.font)
    oneCharWidth=f.measure("x")
    for charWidth in charWidthsSeq:
      columnWidths.append(oneCharWidth*charWidth)
    return columnWidths

  def showMessage(self,message):
    self.msgLabel.configure(text=message)
    self.msgLabel.place(x=0,y=0,relwidth=1,relheight=1)
    # If update_idletasks() is in when object is shown with
    # label already open, object is transparent when window
    # opens. Can safely leave out at least for now because
    # the only thing that uses the message also writes to the
    # log (not when opened but during long sort) and log has
    # its own update_idletasks()

    # Later: seems no longer to be true, putting back in.

    gTkRoot.update_idletasks()
    return None

  def hideMessage(self):
    self.msgLabel.place_forget()
    return None

  def pack(self,*positionArgsTuple,**kwArgsDict):
    self.outerFrame.pack(*positionArgsTuple,**kwArgsDict)
    return None


class simpleMailSorter:

  def __init__(self,mailObjs,sortOrderIndex,useScore,whichDate,messagesToMe=None,messagesFromMe=None):
    self.sortOrderIndex=sortOrderIndex
    self.useScore=useScore
    self.whichDate=whichDate
    self.messagesToMe=messagesToMe
    self.messagesFromMe=messagesFromMe
    self.pathIndex={}
    self.doProgress("Getting sort data... 0%")
    mailObjs=mailObjs[:]
    nObjs=len(mailObjs)
    nObjsFloat=float(nObjs)
    nextProgressAt=time.time()+kShowProgressEvery
    #self.weWantToShowMessage=nObjs>=kEnoughMailThreadSorter
    count=0 # Not set by next line if nObjs is 0
    for count in range(nObjs):
      if time.time()>=nextProgressAt:
        self.doProgress("Getting sort data... %i%%" % int(100*(count/nObjsFloat)))
        self.checkForMessagesToMe()
        nextProgressAt+=kShowProgressEvery

      if useScore:
        score=mailObjs[count].getScore()
        score=-score # Sort increasing
      else:
        score=0

      if self.whichDate=="sent":
        secs=mailObjs[count].getDateAsSecs()
      else:
        secs=mailObjs[count].getReceivedDateAsSecs()
      if sortOrderIndex==1:
        secs=-secs

      mailObjs[count]=((score,secs),mailObjs[count])

    self.doProgress("Getting sort data... %i%%" % 100)
    self.checkForMessagesToMe()

    self.doProgress("Sorting...")

    mailObjs.sort()
    self.mailObjs=mailObjs
    self.checkForMessagesToMe()
    self.doProgress("Indexing...")
    self.makeIndex()
    if self.messagesFromMe<>None:
      self.messagesFromMe.put(bunch(type="done",data=self))
    return None

  def doProgress(self,progressMessage):
    if self.messagesFromMe<>None:
      self.messagesFromMe.put(bunch(type="progress",
        message=progressMessage))
    return None

  def checkForMessagesToMe(self):
    if self.messagesToMe==None:
      return None
    try:
      item=self.messagesToMe.get(0)
    except Queue.Empty:
      return None

    if item=="stop":
      self.messagesFromMe.put(bunch(type="done",data=None))
      thread.exit()
    else:
      assert 1==0,"Unknown message %s" % item

  def makeIndex(self,fromHereOn=None):
    if fromHereOn==None:
      pathIndex={}
      fromHereOn=0
    else:
      pathIndex=self.pathIndex
    mailObjs=self.mailObjs
    for ind in range(fromHereOn,len(mailObjs)):
      pathIndex[mailObjs[ind][1].getPath()]=ind
    self.pathIndex=pathIndex
    return None

  def getIndexByPath(self,path):
    return self.pathIndex[path]

  def deleteByPath(self,path):
    ix=self.getIndexByPath(path)
    del self.mailObjs[ix]
    self.makeIndex(ix)
    return None

  def __getitem__(self,ind):
    return self.mailObjs[ind][1]

  def add(self,paths):
    earliestChanged=None
    for path in paths:
      obj=mailObj(path)

      if self.useScore:
        score=obj.getScore()
        score=-score
      else:
        score=0

      secs=obj.getDateAsSecs()
      if self.sortOrderIndex==1:
        secs=-secs

      item=((score,secs),obj)
      where=bisect.bisect_right(self.mailObjs,item)
      if earliestChanged==None or earliestChanged>where:
        earliestChanged=where
      self.mailObjs.insert(where,item)
    self.makeIndex(earliestChanged)
    return None

  def __len__(self):
    return len(self.mailObjs)

  def getAsList(self):
    return [ x[1] for x in self.mailObjs ]

  def newPath(self,oldPath,newPath):
    ix=self.getIndexByPath(oldPath)
    self.mailObjs[ix][1].newPath(newPath)
    self.pathIndex[newPath]=self.pathIndex[oldPath]
    del self.pathIndex[oldPath]
    return ix


class messageContainer:
  def __init__(self):
    self.messages=[]
    self.parent=None
    self.children=[]
    self.messageID=None
    self.subject=None
    self.strippedSubject=None
    self.score=None
    return None

  def __repr__(self):
    if self.parent==None:
      parentStr="no"
    else:
      parentStr="yes"

    if self.messageID==None:
      idStr="(none)"
    else:
      idStr=self.messageID

    return "%s, Message-ID: %s parent: %s children: %i messages: %i" % \
      (self.__class__.__name__,idStr,parentStr,len(self.children),len(self.messages))


class messageThreader:

  def tableCount(self,idTable):
    tot=0
    for container in idTable.values():
      tot+=len(container.messages)
    return tot


  def countMessages(self,rootSet):

    def recurseCountMessages(container):
      tot=len(container.messages)
      for child in container.children:
        tot+=recurseCountMessages(child)
      return tot

    tot=0
    for container in rootSet:
      tot+=recurseCountMessages(container)
    return tot


  def countExtraMessages(self,rootSet):

    def recurseCountExtras(container):
      tot=0
      if len(container.messages)>1:
        tot+=len(container.messages)-1
      for child in container.children:
        tot+=recurseCountExtras(child)
      return tot

    tot=0
    for container in rootSet:
      tot+=recurseCountExtras(container)
    return tot


  def countContainers(self,rootSet):

    def recurseCountContainers(container):
      tot=1
      for child in container.children:
        tot+=recurseCountContainers(child)
      return tot

    tot=0
    for container in rootSet:
      tot+=recurseCountContainers(container)
    return tot


  def countContainersWithoutMessages(self,rootSet):

    def recurseCountNoMessages(container):
      empties=0
      if len(container.messages)==0:
        empties+=1
        for child in container.children:
          empties+=recurseCountNoMessages(child)
      return empties

    tot=0
    for container in rootSet:
      tot+=recurseCountNoMessages(container)
    return tot


  def countEmpties(self,rootSet):

    def recurseCountEmpties(container):
      empties=0
      if len(container.messages)==0 and len(container.children)==0:
        empties+=1
        for child in container.children:
          empties+=recurseCountEmpties(child)
      return empties

    tot=0
    for container in rootSet:
      tot+=recurseCountEmpties(container)
    return tot


  def countContainersWithoutKeys(self,rootSet):

    def recurseCountNoKeys(container):
      noKeys=0
      if not hasattr(container,"dateKey") or \
        not hasattr(container,"score") or \
        not isinstance(container.score,int) or \
        not isinstance(container.dateKey,int):
        noKeys+=1
        for child in container.children:
          noKeys+=recurseCountNoKeys(child)
      return noKeys

    tot=0
    for container in rootSet:
      tot+=recurseCountNoKeys(container)
    return tot


  def countNoMessagesTopLevel(self,rootSet):
    empties=0
    for container in rootSet:
      if len(container.messages)==0:
        empties+=1
    return empties

  def countTopLevelWithParents(self,rootSet):
    n=0
    for container in rootSet:
      if container.parent<>None:
        n+=1
    return n

  def topLevelSubjectsUnique(self,rootSet):
    subjects=[]
    for container in rootSet:
      if container.strippedSubject<>"" and container.strippedSubject in subjects:
        return 0
      subjects.append(container.strippedSubject)
    return 1



  def chaseChildrenForID(self,container,msgID):
    while 1:
      if container.messageID==msgID:
        return 1
      if len(container.children)==0:
        return 0
      for child in container.children:
        result=self.chaseChildrenForID(child,msgID)
        if result:
          return 1
      return 0


  def chaseParentsForID(self,container,msgID):
    while 1:
      if container.messageID==msgID:
        return 1
      if container.parent==None:
        return 0

      container=container.parent


  # Swallow doesn't display messages in a tree because I don't
  # think that's necessary for a mail reader. This function
  # flattens the trees into a list. Some people would say that
  # it thereby undoes much of the good that the threading
  # algorithm does.

  def flattenTrees(self):
    if kMessageThreaderDebug:
      msgsInTreesAtStart=self.countMessages(self.rootSet)

    def recurseFlattenTree(container):
      l=container.messages[:]
      for child in container.children:
        l+=recurseFlattenTree(child)
      return l

    l=[]
    for container in self.rootSet:
      l+=recurseFlattenTree(container)

    if kMessageThreaderDebug:
      assert msgsInTreesAtStart==self.countMessages(self.rootSet)
      assert len(l)==self.countMessages(self.rootSet)

    self.flat=l
    return None

  # Could lose Fw:, Fwd:, etc
  def stripSubjJunk(self,subj):
    while 1:
      subj=subj.lower()
      where=kSubjJunkRE.search(subj)
      if where==None:
        return subj.strip()
      subj=subj[len(where.group(0)):]


  def addToIDTable(self,message):
    self.messagesSeen+=1
    idTable=self.idTable

    refsHeader=message.get("references","")
    msgIDHeader=message.get("message-id","")
    irtHeader=message.get("in-reply-to","")

    if idTable.has_key(msgIDHeader):
      thisMessageContainer=idTable[msgIDHeader]
      idTable[msgIDHeader].messages.append(message)
      if len(thisMessageContainer.messages)>1: # If duplicate message-id
        self.dupMessageIDsSeen+=1
        return None
    else:
      thisMessageContainer=messageContainer()
      thisMessageContainer.messageID=msgIDHeader
      self.containersAddedForMessages+=1
      thisMessageContainer.messages.append(message)
      idTable[msgIDHeader]=thisMessageContainer

    refs=refsHeader.split() # Naive but may be sufficient

    # Make sure that no reference is repeated. Weird
    # but it can happen.
    refsSet={}
    ind=0
    while ind<len(refs):
      if refs[ind] in refsSet:
        del refs[ind]
      else:
        refsSet[refs[ind]]=1
        ind+=1

    ans=kInAngleBracketsRE.search(irtHeader)
    if ans<>None:
      irtID=ans.group(1)
      # I understand that this can happen sometimes
      if irtID not in refsSet:
        refs.append(irtID)

    for count in range(len(refs)-1):
      thisRef=refs[count]
      if idTable.has_key(thisRef):
        thisContainer=idTable[thisRef]
      else:
        self.containersAddedForMessages+=1
        thisContainer=messageContainer()
        thisContainer.messageID=thisRef
        idTable[thisRef]=thisContainer

      nextRef=refs[count+1]
      if idTable.has_key(nextRef):
        nextContainer=idTable[nextRef]
      else:
        self.containersAddedForMessages+=1
        nextContainer=messageContainer()
        nextContainer.messageID=nextRef
        idTable[nextRef]=nextContainer

      # First clause a PITA to find need for
      if nextContainer.parent==None and \
        not nextContainer in thisContainer.children and \
        not self.chaseChildrenForID(nextContainer,thisContainer.messageID) and \
        not self.chaseParentsForID(thisContainer,nextContainer.messageID):
        thisContainer.children.append(nextContainer)
        nextContainer.parent=thisContainer

    if len(refs)>0:
      lastRef=refs[-1]
      if idTable.has_key(lastRef):
        lastContainer=idTable[lastRef]
      else:
        self.containersAddedForMessages+=1
        lastContainer=messageContainer()
        lastContainer.messageID=lastRef
        idTable[lastRef]=lastContainer

      # If we've already got a parent, don't believe it.
      if thisMessageContainer.parent<>None and thisMessageContainer.parent<>lastContainer:
        thisMessageContainer.parent.children.remove(thisMessageContainer)
        thisMessageContainer.parent=None

      if thisMessageContainer.parent==None and \
        not thisMessageContainer in lastContainer.children and \
        not self.chaseChildrenForID(thisMessageContainer,lastContainer.messageID) and \
        not self.chaseParentsForID(lastContainer,thisMessageContainer.messageID):
        lastContainer.children.append(thisMessageContainer)
        thisMessageContainer.parent=lastContainer

    self.idTable=idTable
    return None


  def makeIDTable(self):
    self.idTable={}
    #for msg in self.originalList:
    #  self.addToIDTable(msg)
    originalList=self.originalList
    nObjsFloat=float(len(originalList))
    nextProgressAt=time.time()+kShowProgressEvery
    for ind in range(len(originalList)):
      self.addToIDTable(originalList[ind])
      #if ind>0 and ind % kShowOpenOverviewProgressEvery==0:
      if time.time()>=nextProgressAt:
        self.doProgress("Threading, step 1... %i%%" % int(100*(ind/nObjsFloat)))
        self.checkForMessagesToMe()
        nextProgressAt+=kShowProgressEvery
    return None


  def findRootSet(self):
    rootSet=[]
    for container in self.idTable.values():
      if container.parent==None:
        rootSet.append(container)
    self.rootSet=rootSet
    return None


  def recursePrune(self,container):
    ind=0
    containersRemoved=0
    while ind<len(container.children):
      child=container.children[ind]
      containersRemoved+=self.recursePrune(child)

      if len(child.messages)==0 and len(child.children)==0:
        container.children.remove(child)
        containersRemoved+=1
      elif len(child.messages)==0 and len(child.children)>0:
        container.children+=child.children
        container.children.remove(child)
        containersRemoved+=1
      else:
        ind+=1
    return containersRemoved


  def pruneEmpties(self):
    for container in self.rootSet:
      self.containersRemoved+=self.recursePrune(container)

    ind=0
    while ind<len(self.rootSet):
      container=self.rootSet[ind]
      if len(container.messages)==0 and len(container.children)==0:
        del self.rootSet[ind]
        self.containersRemoved+=1
      elif len(container.messages)==0 and len(container.children)==1:
        self.rootSet[ind]=container.children[0]
        self.rootSet[ind].parent=None
        self.containersRemoved+=1
        ind+=1
      else:
        ind+=1

    return None


  def makeSubjectTable(self):
    subjectTable={}
    for thisContainer in self.rootSet:
      if len(thisContainer.messages)>0:
        subjHeader=thisContainer.messages[0].get("subject","")
      else:
        subjHeader=thisContainer.children[0].messages[0].get("subject","")
      strippedSubject=self.stripSubjJunk(subjHeader)

      thisContainer.subject=subjHeader
      thisContainer.strippedSubject=strippedSubject

      if strippedSubject=="":
        continue

      if not subjectTable.has_key(strippedSubject):
        subjectTable[strippedSubject]=thisContainer
      else:
        containerInTable=subjectTable[strippedSubject]
        if len(containerInTable.messages)>0 and len(thisContainer.messages)==0:
          subjectTable[strippedSubject]=thisContainer
        elif len(containerInTable.subject)<>len(containerInTable.strippedSubject) and \
          len(thisContainer.subject)==len(thisContainer.strippedSubject):
          subjectTable[strippedSubject]=thisContainer
    self.subjectTable=subjectTable
    return None


  def joinTreesBySubject(self):
    ind=0
    while ind<len(self.rootSet):
      thisContainer=self.rootSet[ind]
      strippedSubject=thisContainer.strippedSubject

      if not self.subjectTable.has_key(strippedSubject):
        assert strippedSubject==""
        ind+=1
        continue

      containerInTable=self.subjectTable[strippedSubject]

      if thisContainer==containerInTable:
        ind+=1
        continue

      if len(thisContainer.messages)==0 and len(containerInTable.messages)==0:
        containerInTable.children+=thisContainer.children
        self.containersRemoved+=1
        del self.rootSet[ind]
        continue

      if len(containerInTable.messages)==0:
        assert len(thisContainer.messages)>0
        containerInTable.children.append(thisContainer)
        del self.rootSet[ind]
        continue

      if len(thisContainer.strippedSubject)<>len(thisContainer.subject) and \
        len(containerInTable.strippedSubject)==len(containerInTable.subject):
        containerInTable.children.append(thisContainer)
        del self.rootSet[ind]
        continue

      newOne=messageContainer()
      self.containersAddedForOtherReasons+=1
      newOne.children.append(thisContainer)
      newOne.children.append(containerInTable)
      newOne.strippedSubject=strippedSubject
      del self.rootSet[ind]
      where=self.rootSet.index(containerInTable)
      self.rootSet[where]=newOne
      self.subjectTable[strippedSubject]=newOne

    return None

  def checkInvariants(self):
    if self.idTable<>None:
      assert self.containersAddedForMessages==len(self.idTable.keys())
      assert self.messagesSeen==self.tableCount(self.idTable)
    if self.rootSet<>None:
      assert self.dupMessageIDsSeen==self.countExtraMessages(self.rootSet)
      assert self.messagesSeen==self.countMessages(self.rootSet)
      assert self.countTopLevelWithParents(self.rootSet)==0
      assert self.containersAddedForMessages+self.containersAddedForOtherReasons \
        -self.containersRemoved==self.countContainers(self.rootSet)
      if self.havePruned:
        assert self.countContainersWithoutMessages(self.rootSet)==self.countNoMessagesTopLevel(self.rootSet)
        assert self.countEmpties(self.rootSet)==0
      if self.haveMergedSubjects:
        assert self.topLevelSubjectsUnique(self.rootSet)
    if self.haveSortedSiblings:
      assert self.countContainersWithoutKeys(self.rootSet)==0
    return None


  def jwzThread(self):
    self.doProgress("Threading, step 1...")

    self.makeIDTable()

    if kMessageThreaderDebug:
      self.checkInvariants()

    self.doProgress("Threading, step 2...")
    self.checkForMessagesToMe()

    self.findRootSet()
    self.idTable=None

    if kMessageThreaderDebug:
      self.checkInvariants()

    self.doProgress("Threading, step 3...")
    self.checkForMessagesToMe()

    self.pruneEmpties()

    if kMessageThreaderDebug:
      self.havePruned=1
      self.checkInvariants()

    self.doProgress("Threading, step 4...")
    self.checkForMessagesToMe()

    self.makeSubjectTable()
    self.joinTreesBySubject()

    if kMessageThreaderDebug:
      self.haveMergedSubjects=1
      self.checkInvariants()

    return None


  def sortSiblingsRecurse(self,node):
    totKidScore=0
    avgKidScore=0
    if len(node.children)>0:
      for ind in range(len(node.children)):
        kid=node.children[ind]
        self.sortSiblingsRecurse(kid)
        totKidScore+=kid.score
        node.children[ind]=(-kid.score,kid.dateKey,kid)
      node.children.sort()
      for count in range(len(node.children)):
        node.children[count]=node.children[count][2] # List comprehensions a little slower
      avgKidScore=int(totKidScore/len(node.children))

    if len(node.messages)>0:
      if self.useScore:
        node.score=int(node.messages[0].getScore()+avgKidScore)/2
      else:
        node.score=0
      if self.whichDate=="sent":
        secs=node.messages[0].getDateAsSecs()
      else:
        secs=node.messages[0].getReceivedDateAsSecs()
      if self.sortOrderIndex==1:
        secs=-secs
      node.dateKey=secs
    else:
      # Must be root-level node
      node.score=avgKidScore
      node.dateKey=node.children[0].dateKey
    return None


  def sortSiblings(self):
    for ind in range(len(self.rootSet)):
      node=self.rootSet[ind]
      self.sortSiblingsRecurse(node)
      self.rootSet[ind]=(-node.score,node.dateKey,node)
    self.rootSet.sort()
    for ind in range(len(self.rootSet)):
      self.rootSet[ind]=self.rootSet[ind][2]

    if kMessageThreaderDebug:
      self.siblingsSorted=1
      self.checkInvariants()
    return None


  # Since add() just appends to the original list of
  # messages we need to remove the message from there
  # in addition to removing it from the result list,
  # otherwise it would appear in the result list after
  # add() is next run.
  def deleteByPath(self,path):
    ind=self.getIndexByPath(path)
    del self.flat[ind]
    fi=self.flatIndex

    for k in fi.keys():
      if fi[k]>ind:
        fi[k]-=1

    return None


  def add(self,paths):
    self.idTable=None
    self.rootSet=None
    self.subjectTable=None
    self.havePruned=0
    self.haveMergedSubjects=0
    self.haveSortedSiblings=0
    self.messagesSeen=0
    self.containersAddedForMessages=0
    self.containersAddedForOtherReasons=0
    self.containersRemoved=0
    self.dupMessageIDsSeen=0

    new=[ mailObj(path) for path in paths ]
    # We overwrite the old list because if we added to the
    # original one, we'd need to delete from that list too
    # and that would require either a sequential search or
    # another index.
    self.originalList=self.flat+new

    self.jwzThread()
    self.sortSiblings()
    self.flattenTrees()
    self.makeFlatIndex()
    return None


  def getIndexByPath(self,path):
    return self.flatIndex[path]

  def newPath(self,oldPath,newPath):
    ix=self.getIndexByPath(oldPath)
    self.flat[ix].newPath(newPath)
    self.flatIndex[newPath]=self.flatIndex[oldPath]
    del self.flatIndex[oldPath]
    return ix

  def __getitem__(self,ind):
    return self.flat[ind]

  def __len__(self):
    return len(self.flat)

  def getAsList(self):
    return self.flat

  def makeFlatIndex(self):
    flatIndex={}
    flat=self.flat
    for ind in range(len(self.flat)):
      flatIndex[flat[ind].getPath()]=ind
    self.flatIndex=flatIndex
    return None

  def doProgress(self,progressMessage):
    if self.weWantToShowMessage: #### and self.messagesFromMe<>None: ###XXXXXX
      self.messagesFromMe.put(bunch(type="progress",
        message=progressMessage))
    return None

  def checkForMessagesToMe(self):
    if self.messagesToMe==None:
      return None
    #print "check"###XXX
    try:
      item=self.messagesToMe.get(0)
    except Queue.Empty:
      return None

    if item=="stop":
      self.messagesFromMe.put(bunch(type="done",data=None))
      thread.exit()
    else:
      assert 1==0,"Unknown message %s" % item

  def __init__(self,listOfMessages,sortOrderIndex,useScore,whichDate,
      messagesToMe=None,messagesFromMe=None):
    self.originalList=listOfMessages
    self.messagesToMe=messagesToMe
    self.messagesFromMe=messagesFromMe
    self.sortOrderIndex=sortOrderIndex
    self.useScore=useScore
    self.whichDate=whichDate
    self.idTable=None
    self.rootSet=None
    self.subjectTable=None
    self.flat=None
    self.flatIndex=None

    self.havePruned=0
    self.haveMergedSubjects=0
    self.haveSortedSiblings=0

    self.messagesSeen=0
    self.containersAddedForMessages=0
    self.containersAddedForOtherReasons=0
    self.containersRemoved=0
    self.dupMessageIDsSeen=0
    self.containersRemoved=0
    self.weWantToShowMessage=len(listOfMessages)>=kEnoughMailForFancyMailboxOpening
    self.jwzThread()

    self.checkForMessagesToMe()
    self.doProgress("Threading, step 5...")


    self.sortSiblings()

    self.checkForMessagesToMe()
    self.doProgress("Threading, step 6...")

    self.flattenTrees()
    self.checkForMessagesToMe()
    self.makeFlatIndex()

    self.doProgress("Threading done")
    if self.messagesFromMe<>None:
      self.messagesFromMe.put(bunch(type="done",data=self))
    return None



class labelWithSizeInDots:
  def __init__(self,parent,**kwArgsDict):
    frameArgsDict,labelArgsDict=self.splitArgsDict(kwArgsDict)
    self.frame=Frame(parent,**frameArgsDict)
    self.frame.pack_propagate(0)
    self.label=Label(self.frame,**labelArgsDict)
    self.label.pack(fill=BOTH)
    return None

  def configure(self,**kwArgsDict):
    frameArgsDict,labelArgsDict=self.splitArgsDict(kwArgsDict)
    if len(frameArgsDict)>0:
      try:
        self.frame.configure(**frameArgsDict)
      except TclError:
        pass
    if len(labelArgsDict)>0:
      try:
        self.label.configure(**labelArgsDict)
      except TclError:
        pass
    return None

  def splitArgsDict(self,dict):
    frameArgsDict={}
    if dict.has_key("height"):
      frameArgsDict["height"]=dict["height"]
      del dict["height"]
    if dict.has_key("width"):
      frameArgsDict["width"]=dict["width"]
      del dict["width"]
    return frameArgsDict,dict

  def winfo_width(self):
    return self.frame.winfo_width()

  def interior(self):
    return self.label

  def bind(self,event,cb):
    self.label.bind(event,cb)
    return None

  def pack(self,*positionArgsTuple,**kwArgsDict):
    self.frame.pack(*positionArgsTuple,**kwArgsDict)
    return None


class mailboxBrowser:
  def __init__(self,mailboxName):
    self.processingKeystrokeLock=threading.Lock()
    self.mailboxName=mailboxName
    self.root=Toplevel()
    self.root.title(mailboxName)
    self.root.protocol("WM_DELETE_WINDOW",self.closeWinCB)
    self.root.bind("<Control-Up>",self.controlArrowCB)
    self.root.bind("<Control-Down>",self.controlArrowCB)
    self.root.bind("<Key>",self.keyCB)
    # Grid widget gets <Configure> events as soon as it's built
    # so give it harmless things to eat for contents until
    # contents actually ready
    self.mailObjs=None
    self.maildirObj=None
    self.sortThreadCheckerAfter=None
    self.messagesToSorterThread=Queue.Queue()
    self.messagesFromSorterThread=Queue.Queue()
    self.sorterThread=None

    if gState.mailboxSortSettings.has_key(mailboxName):
      defaultSort=gState.mailboxSortSettings[mailboxName]
    else:
      defaultSortFunc=gPrefs.get("getDefaultMailboxSort")
      if defaultSortFunc==None:
        defaultSort=gPrefs.get("defaultMailboxSort")
      else:
        defaultSort=defaultSortFunc(mailboxName)
    if defaultSort["date"]=="ascending":
      val=0
    elif defaultSort["date"]=="descending":
      val=1
    else:
      assert 1==0,'Impossible value for defaultSort["date"]: %s' % defaultSort["date"]

    self.dateSortVar=IntVar()
    self.dateSortVar.set(val)

    self.whichDateVar=StringVar()
    self.whichDateVar.set(defaultSort["whichDate"])

    self.useScoreVar=IntVar()
    self.useScoreVar.set(bool(defaultSort["score"]))

    self.threadMessagesVar=IntVar()
    self.threadMessagesVar.set(bool(defaultSort["thread"]))

    self.highlightedChild=None
    l=Label(gTkRoot,text=" ",font=gPrefs.get("fontILike"))
    self.contentsItemHeight=l.winfo_reqheight()+1 # Dunno why +1
    self.columnWidths=charWidthsToDotWidths(kOverviewColumnDefaultWidths)
    self.createWidgets()

    # If we ever start delivering mail to other
    # mailboxes, we'll need to change this.
    # Well, we've started delivering to other mailboxes so grab
    # the lock regardless
    #if mailboxName=="inbox":
    gSwitchboard.mailCheckLock.acquire() # Will block

    self.redisplayFromScratch()
    #self.setCountLabel()
    gSwitchboard.addMailboxBrowser(self)
    gSwitchboard.mailCheckLock.release()
    #self.setCountLabel()
    return None

  def createWidgets(self):
    myFont=gPrefs.get("fontILike")
    controlsFrame=Frame(self.root,width=600,height=27)
    controlsFrame.pack_propagate(0)
    self.mbar=mainMenubar(controlsFrame)
    self.mbar.beginAddingMenu()
    self.mbar.mbar.addmenu("Mailbox","",font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.component("Mailbox-menu").configure(activebackground=kTkDefaultBG)

    self.mbar.mbar.addmenuitem("Mailbox","command",label="Sort",font=gPrefs.get("boldFontILike"),state=DISABLED)

    self.mbar.mbar.addmenuitem("Mailbox","radiobutton",
      value=0,
      label="Earlier first",
      font=myFont,
      activebackground=kTkDefaultBG,
      command=self.sortChanged,
      selectcolor="blue",
      variable=self.dateSortVar,
      )

    self.mbar.mbar.addmenuitem("Mailbox","radiobutton",
      value=1,
      label="Later first",
      font=myFont,
      activebackground=kTkDefaultBG,
      command=self.sortChanged,
      selectcolor="blue",
      variable=self.dateSortVar,
      )

    self.mbar.mbar.addmenuitem("Mailbox","separator")

    self.mbar.mbar.addmenuitem("Mailbox","radiobutton",
      value="sent",
      label="Use sent date",
      font=myFont,
      activebackground=kTkDefaultBG,
      command=self.sortChanged,
      selectcolor="blue",
      variable=self.whichDateVar,
      )

    self.mbar.mbar.addmenuitem("Mailbox","radiobutton",
      value="received",
      label="Use received date",
      font=myFont,
      activebackground=kTkDefaultBG,
      command=self.sortChanged,
      selectcolor="blue",
      variable=self.whichDateVar,
      )

    self.mbar.mbar.addmenuitem("Mailbox","separator")

    self.mbar.mbar.addmenuitem("Mailbox","checkbutton",
      label="Thread messages",
      font=myFont,
      activebackground=kTkDefaultBG,
      command=self.threadToggled,
      selectcolor="blue",
      variable=self.threadMessagesVar,
      )

    self.mbar.mbar.addmenuitem("Mailbox","checkbutton",
      label="Use score",
      font=myFont,
      activebackground=kTkDefaultBG,
      command=self.scoreToggled,
      selectcolor="blue",
      variable=self.useScoreVar,
      )

    self.mbar.mbar.addmenuitem("Mailbox","separator")
    self.mbar.mbar.addcascademenu("Mailbox","Reply",font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addmenuitem("Reply","command",label="Reply",command=self.replyMenuCB,font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addmenuitem("Reply","command",label="Reply to all",command=self.replyAllMenuCB,font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addmenuitem("Reply","command",label="Reply to To",command=self.replyListMenuCB,font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addcascademenu("Mailbox","Forward",font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addmenuitem("Forward","command",label="Forward",command=self.mimeForwardCB,font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addmenuitem("Forward","command",label="Forward verbatim",command=self.verbatimForwardCB,font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addmenuitem("Mailbox","separator")
    self.mbar.mbar.addmenuitem("Mailbox","command",label="Save",command=self.saveMenuCB,font=myFont)
    self.mbar.mbar.addmenuitem("Mailbox","command",label="Trash",command=self.trashMenuCB,font=myFont)
    self.mbar.mbar.addcascademenu("Mailbox","Move",font=myFont,activebackground=kTkDefaultBG)
    self.mbar.mbar.addcascademenu("Mailbox","Copy",font=myFont,activebackground=kTkDefaultBG)
    self.updateMailboxNamesInMenu()
    self.mbar.doneAddingMenu()
    self.mbar.pack(side=LEFT)
    spacerFrame=Frame(controlsFrame,height=4,width=4)
    spacerFrame.pack(side=RIGHT)
    countLabelBorder=Frame(controlsFrame,borderwidth=1,bg="black")
    self.countLabel=Label(countLabelBorder,font=gPrefs.get("fontILike"),width=11)
    self.countLabel.pack(side=LEFT)
    countLabelBorder.pack(side=RIGHT)
    controlsFrame.pack(side=TOP,fill=X)
    lineAbove=Frame(self.root,height=1,bg="black")
    lineAbove.pack(side=TOP,fill=X)
    if self.mailboxName=="outbox":
      myHeaderStrs=kOutboxHeaders
    else:
      myHeaderStrs=kMailboxHeaders

    self.scrollArea=gridWithHeaders(self,
      headerStrs=myHeaderStrs,
      defaultWidthsChars=kOverviewColumnDefaultWidths,
      textGetCB=self.getStringsForOverview,
      contentsLenGetCB=self.numOverviewRows,
      doubleClickCB=self.mailDoubleClick,
      rowConfigureCB=self.tableRowConfigure,
      font=myFont,
      tableLinesToShow=kMailOverviewLinesToShow,
      )
    self.scrollArea.pack(side=TOP,fill=BOTH,expand=1)
    self.scrollArea.showMessage("") ### Hide grid to start with so less blinky???
    return None

  def saveSortState(self):
    val=self.dateSortVar.get()
    if val==0:
      textVal="ascending"
    elif val==1:
      textVal="descending"
    else:
      assert 1==0,"Impossible value for dateSortVar"

    d={"date": textVal,
       "score": self.useScoreVar.get(),
       "thread": self.threadMessagesVar.get(),
        "whichDate": self.whichDateVar}
    gState.mailboxSortSettings[self.mailboxName]=d
    return None

  def threadToggled(self):
    self.saveSortState()
    self.redisplayFromScratch()
    return None

  def scoreToggled(self):
    self.saveSortState()
    self.redisplayFromScratch()
    return None

  def sortChanged(self):
    self.saveSortState()
    self.redisplayFromScratch()
    return None

  def loadingProgress(self,pctDone):
    self.scrollArea.showMessage("Loading... %i%%" % pctDone)
    return None

  def redisplayFromScratch(self):
    # If sorter thread is running, stop it
    if self.sorterThread<>None and self.sorterThread.isAlive():
      self.messagesToSorterThread.put("stop")
      self.sorterThread.join()

      # Force the grid widget to erase old contents, prevents
      # jumpy look
      self.mailObjs=None
      self.scrollArea.redisplay()

    if self.mailObjs==None:
      #if self.maildirObj.doSummary:
      self.maildirObj=gAllMailboxes.getMaildir(self.mailboxName,self.loadingProgress)
      #else:
      #  self.maildirObj=gAllMailboxes.getMaildir(self.mailboxName)
      #self.maildirObj=gAllMailboxes.getMaildir(self.mailboxName)
      #if self.maildirObj.doSummary:
      #  myMailObjs=self.maildirObj.getMailObjs(self.loadingProgress)
      #else:
      #  myMailObjs=self.maildirObj.getMailObjs()
      myMailObjs=self.maildirObj.getMailObjs()
    else:
      # Recycle old objects to take advantage of cached data
      myMailObjs=self.mailObjs.getAsList()

    ## If sorter thread is running, stop it
    #if self.sorterThread<>None and self.sorterThread.isAlive():
    #  self.messagesToSorterThread.put("stop")
    #  self.sorterThread.join()
    #
    #  # Force the grid widget to erase old contents, prevents
    #  # jumpy look
    #  self.mailObjs=None
    #  self.scrollArea.redisplay()

    if self.threadMessagesVar.get():
      sorterClass=messageThreader
    else:
      sorterClass=simpleMailSorter

    if len(myMailObjs)>=kEnoughMailForFancyMailboxOpening:
      self.sorterThread=threading.Thread(target=sorterClass,
        args=(myMailObjs,
        self.dateSortVar.get(),
        self.useScoreVar.get(),
        self.whichDateVar.get(),
        self.messagesToSorterThread,
        self.messagesFromSorterThread))
      self.sorterThread.start()

      self.sortThreadCheckerAfter=gTkRoot.after(0,self.checkForSortMessages)
    else:
      self.mailObjs=sorterClass(myMailObjs,
        self.dateSortVar.get(),
        self.useScoreVar.get(),
        self.whichDateVar.get())
      self.setCountLabel()
      self.scrollArea.hideMessage()
      self.scrollArea.redisplay()

    return None


  def checkForSortMessages(self):
    while 1:
      try:
        message=self.messagesFromSorterThread.get(0)
        if message.type=="progress":
          self.scrollArea.showMessage(message.message)
        elif message.type=="done":
          self.mailObjs=message.data
          self.setCountLabel()
          self.scrollArea.hideMessage()
          self.scrollArea.redisplay()
          return None
        else:
          assert 1==0,"Unknown message type %s" % message.type
      except Queue.Empty:
        break
    self.sortThreadCheckerAfter=gTkRoot.after(50,self.checkForSortMessages)
    return None

  def getStringsForOverview(self,index):
    if self.mailObjs==None:
      return ""
    else:
      return self.mailObjs[index].getTextForOverview()

  def mailDoubleClick(self,index):
    self.openMailByIndex(index)
    return None

  def tableRowConfigure(self,row):
    row.labels[4].configure(anchor=E)
    return None

  def openSelectedMail(self):
    ind=self.scrollArea.getHighlightIndex()
    if ind<>None:
      self.openMailByIndex(ind)
    return None

  def numOverviewRows(self):
    if self.mailObjs==None:
      return 0
    else:
      return len(self.mailObjs)

  def getWidgetParent(self):
    return self.root

  def updateMailboxNamesInMenu(self):
    self.mbar.mbar.deletemenuitems("Move",0,999)
    self.mbar.mbar.deletemenuitems("Copy",0,999)

    saveNames=["(Default per prefs)"]
    if gPrefs.get("getSaveFoldersInboundAlt"):
      saveNames+=["(Alternate per prefs)"]
    saveNames+=gAllMailboxes.getNames()

    myFont=gPrefs.get("fontILike")
    for name in saveNames:
      self.mbar.mbar.addmenuitem("Move","command",
        label=name,command=moveMailToOtherMailbox(name,"move",self),
        font=myFont,activebackground=kTkDefaultBG)
      self.mbar.mbar.addmenuitem("Copy","command",
        label=name,command=moveMailToOtherMailbox(name,"copy",self),
        font=myFont,activebackground=kTkDefaultBG)
    return None


  # Callback bound to ordinary <Key> event can't distinguish
  # keys with modifier keys down from same keys without
  # modifiers down so we bind control-arrow keys separately
  # to this routine.
  def controlArrowCB(self,event):
    if event.keysym=="Up":
      self.scrollArea.scrollLines(-1)
    elif event.keysym=="Down":
      self.scrollArea.scrollLines(1)
    else:
      assert 1==0,"Key %s received by controlArrowCB but it doesn't know how to deal with that character" % event.keysym
    return "break"


  def keyCB(self,event):
    # Hmmm. Wanna be sure that we don't get called again while we're still processing a previous
    # keystroke. That can happen when something we call calls update() because it wants to
    # update the screen to keep the user informed. That may let another keystroke through.
    gotIt=self.processingKeystrokeLock.acquire(0)
    if not gotIt:
      #gLogger.add("Keystroke %s ignored, still processing another" % event.keysym)
      return "break"
    try:
      if event.keysym=="Up":
        self.scrollArea.moveSelection(-1)
      elif event.keysym=="Down":
        self.scrollArea.moveSelection(1)
      elif event.keysym=="Left":
        self.scrollArea.scrollXJump(-1)
      elif event.keysym=="Right":
        self.scrollArea.scrollXJump(1)
      elif event.keysym=="Return":
        self.openSelectedMail()
      elif event.char==" ": # like slrn
        self.openSelectedMail()      
      #~ elif event.char=="s":
        #~ self.saveSelectedMail()
      #~ elif event.char=="S":
        #~ self.saveSelectedMail(evenIfNotSeen=1)
      #~ elif event.char=="a":
        #~ self.saveSelectedMail(useAltFunc=1)
      #~ elif event.char=="A":
        #~ self.saveSelectedMail(evenIfNotSeen=1,useAltFunc=1)
      #~ elif event.char=="d":
        #~ self.trashSelectedMail()
      #~ elif event.char=="D":
        #~ self.trashSelectedMail(evenIfNotSeen=1)
      #~ elif event.char=="f":
        #~ self.forwardSelectedMail(decodeStyle="mime")
      #~ elif event.char=="F":
        #~ self.forwardSelectedMail(decodeStyle="nodecode")
      #~ elif event.char=="r":
        #~ self.replyToSelectedMail()
      #~ elif event.char=="R":
        #~ self.replyToSelectedMail(replyAddr="all")
      #~ elif event.char=="l":
        #~ self.replyToSelectedMail(replyAddr="to") # List reply
      elif event.char=="m":
        gSwitchboard.newMail()
      elif event.char=="t":  ### ???
        gTkRoot.lift()       ### ???
      elif event.char=="w":
        self.closeMe()

      if self.mailboxName=="outbox":
        # If trashing outbund mail, shouldn't matter if it's
        # marked as draft
        if event.char=="d":
          self.trashSelectedMail(evenIfNotSeen=1)
        elif event.char=="D":
          self.trashSelectedMail(evenIfNotSeen=1)
      else:
        # Some actions don't make sense for mail user has written
        if event.char=="s":
          self.saveSelectedMail()
        elif event.char=="S":
          self.saveSelectedMail(evenIfNotSeen=1)
        elif event.char=="a":
          self.saveSelectedMail(useAltFunc=1)
        elif event.char=="A":
          self.saveSelectedMail(evenIfNotSeen=1,useAltFunc=1)
        elif event.char=="d":
          self.trashSelectedMail()
        elif event.char=="D":
          self.trashSelectedMail(evenIfNotSeen=1)
        elif event.char=="f":
          self.forwardSelectedMail(decodeStyle="mime")
        elif event.char=="F":
          self.forwardSelectedMail(decodeStyle="nodecode")
        elif event.char=="r":
          self.replyToSelectedMail()
        elif event.char=="R":
          self.replyToSelectedMail(replyAddr="all")
        elif event.char=="l":
          self.replyToSelectedMail(replyAddr="to") # List reply
        elif event.char=="M":
          self.markSelectedMailAsRead()          
        
      # Deliberately let events through so we can't get into a massive
      # typeahead situation.
      gTkRoot.update()
    finally:
      self.processingKeystrokeLock.release()
    # No further processing of this event by Tkinter
    return "break"

  def closeWinCB(self):
    self.closeMe()
    return None

  def closeMe(self):
    ###XXX
    try:
      gTkRoot.after_cancel(self.sortThreadCheckerAfter)
    except TclError:
      pass
    ###XXX
    if self.sorterThread<>None:
      self.messagesToSorterThread.put("stop")
      self.sorterThread.join()

    self.root.destroy()
    gSwitchboard.removeMailboxBrowser(self)
    return None

  def lift(self):
    self.root.lift()
    return None

  def setCountLabel(self):
    # Very unlikely situation: checker has collected mail
    # while we're still in __init__ but before we've gotten
    # the lock that prevents changing our data while we're
    # being drawn.
    if self.mailObjs<>None:
      n=len(self.mailObjs)
      t=str(n)+pluralize(n," item")
      self.countLabel.configure(text=t)
    return None

  def getPathForMove(self):
    if len(self.mailObjs)==0 or self.scrollArea.getHighlightIndex()==None:
      return None
    mailObj=self.mailObjs[self.scrollArea.getHighlightIndex()]
    return mailObj.getPath()

  def openMailByIndex(self,index):
    mailObj=self.mailObjs[index]
    result=gSwitchboard.getByPath(mailObj.getPath())
    if result<>None:
      result.root.lift()
      return None

    if self.mailboxName=="outbox":
      # Mail is coming from the outbox. If it has addresses, put the cursor
      # in the text, if it doesn't, put the cursor in the To field.
      if mailObj.hasAddresses():
        mailEditor(mailObj.getPath(),startInText=1,isNew=0)
      else:
        mailEditor(mailObj.getPath(),isNew=0)
    else:
      mailViewer(mailObj.getPath(),self) # mailObjs so that next and prev buttons can work
    return None

  def markSelectedMailAsRead(self):
    if self.scrollArea.getHighlightIndex()==None:
      gLogger.add("Can't mark read, nothing selected")
      return None
      
    mailObj=self.mailObjs[self.scrollArea.getHighlightIndex()]
    
    if mailObj.whichOfThreeDirs()<>"new":
      gLogger.add("Can't mark read, already read")
      return None

    oldPath=mailObj.path
    mailObj.moveToCur()
    gSwitchboard.subdirChanged(oldPath,mailObj.path)

    return None

  def saveSelectedMail(self,evenIfNotSeen=0,useAltFunc=0):
    # Called if user hit the "s" key or clicked save button
    if self.scrollArea.getHighlightIndex()==None:
      gLogger.add("Can't save, nothing selected")
      return None

    if useAltFunc:
      filterFunc=gPrefs.get("getSaveFoldersInboundAlt")
    else:
      filterFunc=gPrefs.get("getSaveFoldersInbound")

    if not filterFunc:
      gLogger.add("Can't save, function to determine where not set")
      return None

    mailObj=self.mailObjs[self.scrollArea.getHighlightIndex()]
    if getSubdirFromPath(mailObj.path)=="new" and not evenIfNotSeen:
      gLogger.add("To save unread mail use cap-S")
      return None

    newMBs=filterFunc(mailObj.getHeaders())
    if type(newMBs)==types.StringType:
      newMBs=[newMBs]

    goodResult=1
    for box in newMBs:
      if box<>self.mailboxName:
        newMaildir=gAllMailboxes.getMaildir(box)
        result=newMaildir.copyToThisMaildir(mailObj)
        gSwitchboard.add(box,[mailObj.path])
        if not result:
          goodResult=0
    if goodResult:
      gLogger.add("Message saved to %s" % ", ".join(newMBs))
      oldPath=mailObj.path
      gSwitchboard.delete(self.mailboxName,[oldPath])
      os.unlink(oldPath)
      gSwitchboard.counts(self.mailboxName)
    return None

  def replyMenuCB(self):
    self.replyToSelectedMail()
    return None

  def replyListMenuCB(self):
    self.replyToSelectedMail(replyAddr="to")
    return None

  def replyAllMenuCB(self):
    self.replyToSelectedMail(replyAddr="all")
    return None

  def verbatimForwardCB(self):
    self.forwardSelectedMail(decodeStyle="nodecode")
    return None

  def mimeForwardCB(self):
    self.forwardSelectedMail(decodeStyle="mime")
    return None

  def replyToSelectedMail(self,replyAddr=None):
    if len(self.mailObjs)==0 or self.scrollArea.getHighlightIndex()==None:
      gLogger.add("Can't reply, nothing selected")
      return None
    self.mailObjs[self.scrollArea.getHighlightIndex()].showInEditorNewReply(replyAddr)
    return None

  def forwardSelectedMail(self,decodeStyle):
    if len(self.mailObjs)==0 or self.scrollArea.getHighlightIndex()==None:
      gLogger.add("Can't forward, nothing selected")
      return None
    if self.scrollArea.getHighlightIndex==None:
      return None
    self.mailObjs[self.scrollArea.getHighlightIndex()].showInEditorNewForward(decodeStyle)
    return None

  def add(self,pathList):
    if self.mailObjs<>[]:
      self.mailObjs.add(pathList)
      self.scrollArea.redisplay()
    return None

  def trashMenuCB(self):
    self.trashSelectedMail(evenIfNotSeen=1)
    return None

  def trashSelectedMail(self,evenIfNotSeen=0):
    if self.scrollArea.getHighlightIndex()==None:
      gLogger.add("Can't delete, nothing selected")
      return None
    if self.mailboxName=="trash":
      gLogger.add("That mail is already in the trash")
      return None

    index=self.scrollArea.getHighlightIndex()
    mailObj=self.mailObjs[index]

    if self.mailboxName=="outbox" and gSwitchboard.getEditorByPath(mailObj.path)<>None:
      gLogger.add("Can't trash that message, it's open in an editor")
      return None

    if getSubdirFromPath(mailObj.path)=="new" and not evenIfNotSeen and not self.mailboxName=="outbox":
      gLogger.add("To trash unread mail use cap-D or Trash button")
      return None

    newMaildir=gAllMailboxes.getMaildir("trash")
    result=newMaildir.copyToThisMaildir(mailObj)
    if result<>None:
      gLogger.add("Mail moved to trash")
      gSwitchboard.delete(self.mailboxName,[mailObj.path])
      oldPath=mailObj.getPath()
      gSwitchboard.add("trash",[oldPath])
      os.unlink(oldPath)
      gSwitchboard.counts(self.mailboxName)
    return None

  def saveMenuCB(self):
    self.saveSelectedMail(evenIfNotSeen=1)
    return None

  def change(self,stuff):
    for path in stuff:
      ix=self.mailObjs.getIndexByPath(path)
      self.mailObjs[ix].killOverviewCache()
      self.scrollArea.updateDisplayOf(ix)
    return None

  def subdirChanged(self,oldPath,newPath):
    ind=self.mailObjs.newPath(oldPath,newPath)
    self.scrollArea.updateDisplayOf(ind)
    return None


    ix=self.mailObjs.getIndexByPath(oldPath)
    self.mailObjs[ix].newPath(newPath)
    ###self.mailObjs.newPath(oldPath,newPath) ###XXX###
    self.scrollArea.updateDisplayOf(ix)
    return None

  def delete(self,stuff):
    for path in stuff:
      whereItWas=self.mailObjs.deleteByPath(path)
      # If we deleted the bottom item in the overview and it was the topmost one showing,
      # scroll up so that the overview doesn't appear empty.
      if whereItWas==len(self.mailObjs)==self.scrollArea.getHighlightIndex==self.scrollArea.indexAtTop:
        self.scrollArea.scrollPagesV(-1)
      self.scrollArea.redisplay()
    return None

  # These last three are used by the viewer so that nextMail(), prevMail(), and
  # the x-of-y label work. They might more properly belong in the mailSortHelper
  # object but they can't go there because the viewer is also opened by the
  # search window: the search window has to implement them too and it doesn't
  # have a mailSortHelper. Sometime, it might make sense to redo the search window
  # so it's more like an overview and give it a mailSortHelper.
  def howmanyItems(self):
    return len(self.mailObjs)

  def getIndexByPath(self,path):
    return self.mailObjs.getIndexByPath(path)

  def getPathByIndex(self,index):
    return self.mailObjs[index].path


class mainWin:

  def __init__(self):
    gTkRoot.protocol("WM_DELETE_WINDOW",quit)
    self.root=gTkRoot
    self.searchWindow=None
    self.aboutBoxWin=None
    self.addressBookWin=None
    self.keysWin=None
    self.root.title(kMyName+" "+kVersion+" -- Mailboxes")
    self.processingKeystrokeLock=threading.Lock()
    self.getMailboxInfo()
    self.createWidgets()
    # Ordinary key event callback
    self.root.bind("<Control-Up>",self.controlArrowCB)
    self.root.bind("<Control-Down>",self.controlArrowCB)
    self.root.bind("<Key>",self.keyCB)
    gSwitchboard.setOverview(self)
    self.checkQueueAfterID=gTkRoot.after(kGeneralAfterDelayMS,self.checkQueue)
    return None

  def createWidgets(self):
    # Hack. Otherwise our grid starts too small. Don't know why.
    self.root.configure(height=400,width=200)
    myFont=gPrefs.get("fontILike")
    topFrame=Frame(self.root)
    self.mbar=mainMenubar(topFrame)
    self.mbar.pack(side=LEFT,fill=X)
    topFrame.pack(side=TOP,fill=X)
    nextFrame=Frame(self.root)
    llBorder=Frame(nextFrame,borderwidth=1,bg="black")
    spacer=Frame(nextFrame,width=5,height=5)
    spacer.pack(side=LEFT)
    lastLogLine=Text(llBorder,font=myFont,bg="white",width=kLastLogLineWidth,
      height=kLastLogLineHeight,selectbackground="white",selectborderwidth=0,
      relief=FLAT,state=DISABLED,wrap=WORD,cursor="") # Use normal cursor
    lastLogLine.pack(side=LEFT,fill=X,expand=1)
    lastLogLine.bind("<Button-1>",self.lastLogLineClickedCB)
    # Keep the last log line from running up against the right margin.
    spacer=Frame(nextFrame,width=4,height=4)
    spacer.pack(side=RIGHT)
    gLogger.setLastLogLineWidget(lastLogLine)
    llBorder.pack(side=LEFT,fill=X,expand=1)
    nextFrame.pack(side=TOP,fill=X)

    spacerFrame=Frame(self.root,height=5,width=5)
    spacerFrame.pack(side=TOP)
    # We add a line above the grid's headers and cleverly make it the
    # right width so that the window ends up being the right width for
    # the labels in the grid. Why bother? Since we have only two labels
    # here, they're likely to end up needing less width than the grid
    # gets from the size of the window. And it looks funny if they're
    # narrower than the clipper. It's possible to stretch them to fill
    # the clipper but not before the clipper is drawn because they
    # won't know how wide they ought to be until then. So unless we
    # want them to jump after the window has been drawn, it's easiest
    # to make the clipper match their width in the first place. Once
    # it's all drawn, they'll stretch just fine if the user widens the
    # window. Normally, that's not an issue because we use the grid for
    # things that are routinely wider than the clipper.
    f=tkFont.Font(font=myFont)
    oneCharWidth=f.measure("x")
    myWidth=oneCharWidth*reduce(operator.add,kMainWinColWidths)+kScrollbarWidth+1 # Plus one for divider
    lineAbove=Frame(self.root,width=myWidth,bg="black",height=1)
    lineAbove.pack(side=TOP,fill=X)
    self.mailboxList=gridWithHeaders(self,
      tableLinesToShow=15,
      headerStrs=("Name","Count"),
      defaultWidthsChars=kMainWinColWidths,
      textGetCB=self.getMailboxText,
      contentsLenGetCB=self.howmanyMailboxes,
      font=myFont,
      doubleClickCB=self.openSelected,
      rowConfigureCB=self.tableRowConfigure,
      )
    self.mailboxList.pack(side=TOP,fill=BOTH,expand=1)
    return None

  def checkQueue(self):
    # For now, process all items in the queue before
    # going on to do anything else. May want something
    # more sophisticated later.
    while 1:
      try:
        item=gSwitchboard.messagesFromCheckerThread.get(0)
        if not hasattr(item,"kwArgs"):
          item.kwArgs={}
        item.func(*item.args,**item.kwArgs)
      except Queue.Empty:
        break
    self.checkQueueAfterID=gTkRoot.after(kGeneralAfterDelayMS,self.checkQueue)


  def getMailboxInfo(self):
    mailboxNames=gAllMailboxes.getNames()
    items=[]
    for name in mailboxNames:
      item=gAllMailboxes.getMaildir(name).getFileCounts()
      item.name=name
      items.append(item)
    self.mailboxInfo=items
    return None

  def howmanyMailboxes(self):
    return len(self.mailboxInfo)

  def getMailboxText(self,index):
    item=self.mailboxInfo[index]
    if item.new==0 and item.cur==0:
      countStr="none"
    elif item.name=="inbox" and item.new>0:
      countStr="%i (%i new)" % (item.new+item.cur,item.new)
    elif item.name=="outbox" and item.new>0:
      countStr="%i (%i %s)" % (item.new+item.cur,item.new,pluralize(item.new,"draft"))
    else:
      countStr="%i" % (item.new+item.cur)
    return (item.name,countStr)

  def openSelected(self,index):
    boxName=gAllMailboxes.getNames()[index]
    result=gSwitchboard.getBrowserByBox(boxName)
    if result<>None:
      result.lift()
    else:
      mailboxBrowser(boxName)
    return None

  def tableRowConfigure(self,row):
    row.labels[1].pack(fill=X,expand=1)
    row.labels[1].configure(anchor=E)
    return None

  def redoCounts(self,boxName):
    for count in range(len(self.mailboxInfo)):
      if self.mailboxInfo[count].name==boxName:
        break
    else:
      assert 1==0,'Mailbox "%s" not found in mailboxInfo' % boxName
    item=gAllMailboxes.getMaildir(boxName).getFileCounts()
    item.name=boxName
    self.mailboxInfo[count]=item
    # Maybe add option to grid widget to specify row to redo
    self.mailboxList.redisplay()
    return None

  def newMailboxCreated(self,boxName):
    # Could be cleverer here but don't think this needs to be fast
    self.getMailboxInfo()
    self.mailboxList.redisplay()
    return None

  def getWidgetParent(self):
    return self.root

  def lastLogLineClickedCB(self,eventIgnored):
    gLogger.show()
    return None

  # Callback bound to ordinary <Key> event can't distinguish
  # keys with modifier keys down from same keys without
  # modifiers down so we bind control-arrow keys separately
  # to this routine.
  def controlArrowCB(self,event):
    if event.keysym=="Up":
      self.mailboxList.scrollLines(-1)
    elif event.keysym=="Down":
      self.mailboxList.scrollLines(1)
    else:
      assert 1==0,"Key %s received by controlArrowCB but it doesn't know how to deal with that character" % event.keysym
    return "break"

  def keyCB(self,event):
    # Hmmm. Wanna be sure that we don't get called again while we're still processing a previous
    # keystroke. That can happen when something we call calls update() because it wants to
    # update the screen to keep the user informed. That may let another keystroke through.
    gotIt=self.processingKeystrokeLock.acquire(0)
    if not gotIt:
      #gLogger.add("Keystroke %s ignored, still processing another" % event.keysym)
      return "break"
    try:
      if event.keysym=="Up":
        self.mailboxList.moveSelection(-1)
      elif event.keysym=="Down":
        self.mailboxList.moveSelection(1)
      elif event.keysym=="Return":
        self.openSelected(self.mailboxList.getHighlightIndex())
      elif event.char==" ":
        self.openSelected(self.mailboxList.getHighlightIndex())
      elif event.char=="m":
        gSwitchboard.newMail()
      elif event.char=="q":
        quit()
      elif event.char=="Q":
        quit(evenIfEditorsOpen=1)
      # Deliberately let events through so we can't get into a massive
      # typeahead situation.
      gTkRoot.update()
    finally:
      self.processingKeystrokeLock.release()
    # No further processing of this event by Tkinter
    return "break"

def checkUnixEnv():
  if os.getuid()==0:
    print "I think it's unwise to read mail as root. In fact, I think it's"
    print "so unwise that I'm going to refuse to run."
    sys.exit(0)

  if os.geteuid()==0:
    print "You're running a copy of Python that is SUID-root. That's too weird for me."
    sys.exit(0)

  if os.getuid()<>os.geteuid():
    print "You're running a copy of Python that is SUID (though not to root). That's"
    print "still too weird for me."
    sys.exit(0)

  # gid checks here?
  return None


def main():
  global gAddressBook
  global gAllMailboxes
  global gChecker
  global gLastSaveOrOpenDir
  global gLogger
  ###global gMakeDisplayableOnly
  global gMousePointerHider
  global gNewMaildirFilename
  global gPrefs
  global gSwitchboard
  ###global timeoutsocket # Used if timeoutsocket module not available
  global gTkRoot
  global gUnfetchedMailNotifier
  global gWin32


  ### XXX tuple(email.__version__.split(".")) XXX
  # Stil need recent version, but can't take float of 2.3.1
  #if float(email.__version__)<0.92:
  #
  #  print "email module 0.92 or later required"
  #  sys.exit(1)
  socket.setdefaulttimeout(kSocketTimeout)
  gLastSaveOrOpenDir=os.path.expanduser("~")
  gSwitchboard=switchboard()
  gLogger=logger()
  gLogger.add(kMyName+" "+kVersion)
  gChecker=checker()
  ###gMakeDisplayableOnly=makeDisplayableOnly()
  gTkRoot=Tk()
  gTkRoot.withdraw()
  Pmw.initialise(gTkRoot)
  Pmw.reporterrorstofile(callbackErrorHandler())

  # Are we running under Windows?
  gWin32 = 0
  if sys.platform == "win32":
    gWin32 = 1

  if len(sys.argv)>2 or (len(sys.argv)>1 and (sys.argv[1]=="?" or sys.argv[1]=="help")):
    print "Usage: %s [alternate preferences file]" % os.path.basename(sys.argv[0])
    sys.exit(0)

  if not gWin32:
    checkUnixEnv()
    
  os.umask(0177) # octal
  gPrefs=prefs()
  gLogger.readyToUseFile()
  #if sys.modules.has_key("timeoutsocket"):
  #  timeoutsocket.setDefaultSocketTimeout(kSocketTimeout)
  #else:
  #  # If we haven't imported the timeoutsocket module
  #  # hack something up so that the exception clauses
  #  # "except timeoutsocket.Timeout" don't throw
  #  # NameErrors. The value is None so that it won't
  #  # ever match an exception.
  #  timeoutsocket=nullClass()
  #  timeoutsocket.Timeout=None
  gNewMaildirFilename=newMaildirFileName()
  gAllMailboxes=allMailboxes()
  gAllMailboxes.cleanupTmps()
  checkTempDir()
  tempfile.tempdir=gPrefs.get("tempDir")
  tempfile.template=kMyName+"Temp"
  loadState()
  gState.forgetAboutOldMailboxes()
  if gPrefs.get("hideMousePointerWhileTyping"):
    gMousePointerHider=mousePointerHider()
  if gPrefs.get("swdbFile"):
    gAddressBook=swdbAddressBook()###gPrefs,gLogger,gSwitchboard)
  else:
    gAddressBook=addressBook()
  gUnfetchedMailNotifier=unfetchedMailNotifier()
  mainWin()
  gTkRoot.deiconify()
  if gPrefs.get("autoStartCheckRegularly"):
    gChecker.setCheckingRegularly(1)
    gSwitchboard.checkRegularlyChanged(1)

  gTkRoot.mainloop()
  return None # Never actually returns


if __name__=="__main__":
  main()
